//
// File:         Allocator.h
// Description:  Raw allocation of small objects.
// Created:      Mar 2, 2000, 17:45
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#ifndef ALLOCATOR_H
//====================================================
#define ALLOCATOR_H
#include <stdlib.h>
#include <limits.h> 
#include <iostream.h>
#include "jargon.h"
#include "RuntimeError.h"
#include "Exit.h"
#include "BLDebugFlags.h"
#include "Debug.h"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_ALLOCATOR
 #define DEBUG_NAMESPACE "Allocator<MaxSmallObjectSize,StandardChunkSize>"
#endif 
#include "debugMacros.h"
//=================================================

#ifdef ALLOCATE_WITH_MIRROR
 #define ALLOCATE_WITH_MIRROR_TOTAL_MEMORY_SIZE_IN_WORDS 33554432
                                                         // 128 Mb
                                                         // 16777216
                                                         // 64Mb
 #define ALLOCATE_WITH_MIRROR_TOTAL_MEMORY_SIZE_DECREMENT_IN_WORDS 262144
                                                                   // 1Mb
#endif

template <ulong MaxSmallObjectSize,ulong StandardChunkSize>
class Allocator
{
 public:
  Allocator() 
   : _freeChunkList((void**)0),
     _occupiedByObjects(0L), 
     _occupiedByChunks(0L),
     _hardMemoryLimit(LONG_MAX),
     _unbufferedAllocationLimit(LONG_MAX),
     _bufferIsBeingUsed(false),
     _requestMemoryRelease(0),
     _reportAllocationFailure(0)
     #ifndef NO_DEBUG
      , _allChunks((ChunkList*)0)
     #endif
  {
   CALL("constructor Allocator()");
   ASSERT(MaxSmallObjectSize <= StandardChunkSize);
   createNewChunk();
   for (ulong i = 0; i <= MaxSmallObjectSize; i++) 
    {
     _freeList[i] = (void**)0;
     _numberOfTakenFromBuffer[i] = 0L;
     _numberOfActiveTakenFromBuffer[i] = 0L;
     _numberOfFree[i] = 0L; 
    };
   _totalBufferDeficit = 0L;
   _activeBufferDeficit = 0L;

   #ifdef ALLOCATE_WITH_MIRROR
   _debugWithMirrorMemorySize = ALLOCATE_WITH_MIRROR_TOTAL_MEMORY_SIZE_IN_WORDS; 
   request_all_memory:   
    _debugWithMirrorFreeMemory = (void**)(malloc(_debugWithMirrorMemorySize*sizeof(void*)));
    if (!_debugWithMirrorFreeMemory) 
     {     
      if (_debugWithMirrorMemorySize < ALLOCATE_WITH_MIRROR_TOTAL_MEMORY_SIZE_DECREMENT_IN_WORDS)
       Debug::debugFeatureInternalFault("_debugWithMirrorMemorySize < ALLOCATE_WITH_MIRROR_TOTAL_MEMORY_SIZE_DECREMENT_IN_WORDS");

      _debugWithMirrorMemorySize -= ALLOCATE_WITH_MIRROR_TOTAL_MEMORY_SIZE_DECREMENT_IN_WORDS; 
      goto request_all_memory;
     };
    _debugWithMirrorHalfMemorySize = _debugWithMirrorMemorySize/2;
    for (ulong i = 0; i < _debugWithMirrorHalfMemorySize; i++)
     {
      _debugWithMirrorFreeMemory[i] = (void*)0;
     };
    _debugWithMirrorEndOfFreeMemory = _debugWithMirrorFreeMemory + _debugWithMirrorMemorySize;
    _debugWithMirrorFreeMemory += _debugWithMirrorHalfMemorySize; 
   #endif 
  };
 

  Allocator( bool (*requestMemoryRelease)(ulong size), void (*reportAllocationFailure)(ulong size)) 
   : _freeChunkList((void**)0),
     _occupiedByObjects(0L), 
     _occupiedByChunks(0L),
     _hardMemoryLimit(LONG_MAX),
     _unbufferedAllocationLimit(LONG_MAX),
     _bufferIsBeingUsed(false),
     _requestMemoryRelease(requestMemoryRelease),
     _reportAllocationFailure(reportAllocationFailure)
     #ifndef NO_DEBUG
      , _allChunks((ChunkList*)0) 
     #endif
  {
   CALL("constructor Allocator(bool (*requestMemoryRelease)(ulong size), void (*reportAllocationFailure)(ulong size))");


   ASSERT(MaxSmallObjectSize <= StandardChunkSize);
   createNewChunk();
   for (ulong i = 0; i <= MaxSmallObjectSize; i++) 
    {
     _freeList[i] = (void**)0;
     _numberOfTakenFromBuffer[i] = 0L;
     _numberOfActiveTakenFromBuffer[i] = 0L;
     _numberOfFree[i] = 0L; 
    };
   _totalBufferDeficit = 0L;
   _activeBufferDeficit = 0L; 

   #ifdef ALLOCATE_WITH_MIRROR
   _debugWithMirrorMemorySize = ALLOCATE_WITH_MIRROR_TOTAL_MEMORY_SIZE_IN_WORDS; 
   request_all_memory:   
    _debugWithMirrorFreeMemory = (void**)(malloc(_debugWithMirrorMemorySize*sizeof(void*)));
    if (!_debugWithMirrorFreeMemory) 
     {     
      if (_debugWithMirrorMemorySize < ALLOCATE_WITH_MIRROR_TOTAL_MEMORY_SIZE_DECREMENT_IN_WORDS)
       Debug::debugFeatureInternalFault("_debugWithMirrorMemorySize < ALLOCATE_WITH_MIRROR_TOTAL_MEMORY_SIZE_DECREMENT_IN_WORDS");
      _debugWithMirrorMemorySize -= ALLOCATE_WITH_MIRROR_TOTAL_MEMORY_SIZE_DECREMENT_IN_WORDS; 
      goto request_all_memory;
     };
    _debugWithMirrorHalfMemorySize = _debugWithMirrorMemorySize/2;
    for (ulong i = 0; i < _debugWithMirrorHalfMemorySize; i++)
     {
      _debugWithMirrorFreeMemory[i] = (void*)0;
     };
    _debugWithMirrorEndOfFreeMemory = _debugWithMirrorFreeMemory + _debugWithMirrorMemorySize;
    _debugWithMirrorFreeMemory += _debugWithMirrorHalfMemorySize; 
   #endif 
  };

  ~Allocator() { CALL("destructor ~Allocator()"); };
  void setHardMemoryLimit(long lim)
  {
   CALL("setHardMemoryLimit(long lim)");
   _hardMemoryLimit = lim;
  }; 
  void setBufferSize(long percentage) 
  {
   CALL("setBufferSize(long percentage)");
   if (percentage > 0)    
    {
     if (percentage < 100)
      {
       _unbufferedAllocationLimit = (long)(_hardMemoryLimit * (1 - (((float)percentage)/100)));       
      }
     else _unbufferedAllocationLimit = 0L;
    }
   else _unbufferedAllocationLimit = _hardMemoryLimit;
  };
  void* allocate(ulong size)
  {
   CALL("allocate(ulong size)");  
   ASSERT(size);
   ASSERT(size <= MaxSmallObjectSize);
   void** res;
    
   take_from_free_list: 
    #ifndef DEBUG_ALLOC_NO_REUSE
     res = _freeList[size];
     if (res) 
      { 
       _freeList[size] = (void**)(*res); 
       if (_bufferIsBeingUsed)
        {     
 	 _numberOfFree[size]--;
         if (_numberOfTakenFromBuffer[size] > _numberOfFree[size])
	  {
           _totalBufferDeficit += size;
           _activeBufferDeficit += size;
          };
        };
       _occupiedByObjects += size;
       return (void*)res;
      };
    #else
     res = (void**)0; 
    #endif 

   allocate_in_current_chunk:
    // try to allocate in the current chunk
    if (size <= _freeInCurrChunk)
     {
      _freeInCurrChunk -= size;
      res = _nextInCurrentChunk;
      _nextInCurrentChunk += size;
      if (_bufferIsBeingUsed)
       {
        _numberOfTakenFromBuffer[size]++;
        _numberOfActiveTakenFromBuffer[size]++;
        _totalBufferDeficit += size;
        _activeBufferDeficit += size;
       }
      else 
       if (_occupiedByChunks - ((long)_freeInCurrChunk) > _unbufferedAllocationLimit)
        {
  	 _bufferIsBeingUsed = true;
         ASSERT(_numberOfTakenFromBuffer[size] == 0L);
         ASSERT(_numberOfActiveTakenFromBuffer[size] == 0L);
         #ifndef DEBUG_ALLOC_NO_REUSE
          ASSERT(_numberOfFree[size] == 0L);
         #endif
         _numberOfTakenFromBuffer[size] = 1L;
         _numberOfActiveTakenFromBuffer[size] = 1L;
         ASSERT(!_totalBufferDeficit);
         ASSERT(!_activeBufferDeficit);
         _totalBufferDeficit = size;
         _activeBufferDeficit = size;
        };
      _occupiedByObjects += size; 
      return (void*)res;
     };       

    // try a free chunk
   
    if (useFreeChunk()) goto allocate_in_current_chunk;

    // request a new chunk from the system
   
    if (_occupiedByChunks + ((long)StandardChunkSize) > _hardMemoryLimit)    
     { 
      if (_requestMemoryRelease && (_requestMemoryRelease(size)))
       {
	goto take_from_free_list;
       };
      if (_reportAllocationFailure) _reportAllocationFailure(size); 
      RuntimeError::report("allocation failure");
      Exit::exit(1);  
     };
    
    if (!createNewChunk()) 
     { 
      if (_requestMemoryRelease && (_requestMemoryRelease(size)))
       {
	goto take_from_free_list;
       };
      if (_reportAllocationFailure) _reportAllocationFailure(size);  
      RuntimeError::report("allocation failure");
      Exit::exit(1);  
     };
    goto allocate_in_current_chunk;
  }; // void* allocate(ulong size)


  void deallocate(void* obj,ulong size)
  {
   CALL("deallocate(void* obj,ulong size)");
   ASSERT(size);
   ASSERT(size <= MaxSmallObjectSize);
   _occupiedByObjects -= size;
   *((void**)obj) = _freeList[size];
   _freeList[size] = (void**)obj;
   if (_bufferIsBeingUsed)
    {
     if (_numberOfTakenFromBuffer[size] > _numberOfFree[size]) 
      {
       _totalBufferDeficit -= size;
       if (_numberOfActiveTakenFromBuffer[size] > 0L)
        { 
         _activeBufferDeficit -= size;
        };
      };
     _numberOfFree[size]++;
    }; 
  }; // void deallocate(void* obj,ulong size)

  void* tryToAllocateLarge(ulong size)
  {
   CALL("tryToAllocateLarge(ulong size)");
   void** res;
   // try in the current chunk
   if (size <= _freeInCurrChunk)
    {
     _freeInCurrChunk -= size;
     res = _nextInCurrentChunk;
     _nextInCurrentChunk += size;
     _occupiedByObjects += size; 
     return (void*)res;
    };

   // size > _freeInCurrChunk, request a piece of memory from the system 
    
   if (_occupiedByChunks + ((long)size) > _hardMemoryLimit)  return (void*)0; 

   #ifdef ALLOCATE_WITH_MIRROR
    #ifndef NO_DEBUG
     res = (void**)(debugWithMirrorMalloc(size + 1));
    #else
     res = (void**)(debugWithMirrorMalloc(size + 1));
    #endif
   #else
    #ifndef NO_DEBUG
     res = (void**)(malloc((size + 1)*sizeof(void*)));
    #else 
     res = (void**)(malloc(size*sizeof(void*)));
    #endif
   #endif

   if (!res) return (void*)0;

   _occupiedByChunks += size;
   #ifndef NO_DEBUG
    _allChunks = new ChunkList(res,_allChunks); // res is a new chunk
    res[0] = (void*)(size + 1); 
    res++;
   #endif

   return res;
  }; // void* tryToAllocateLarge(ulong size)

  void deallocateLarge(void* obj,ulong size) 
  {
   CALL("deallocateLarge(void* obj,ulong size)");
   ASSERT(size >= 2); 
   _occupiedByObjects -= size;
   ((void**)obj)[0] = _freeChunkList;
   ((void**)obj)[1] = (void*)size;
   _freeChunkList = (void**)obj;
  }; // void deallocateLarge(void* obj,ulong size) 

  long occupiedByObjects() const { return _occupiedByObjects; };
  long occupiedByChunks() const { return _occupiedByChunks; };
  long freeInTheCurrentChunk() const { return (long)_freeInCurrChunk; };
  long totalBufferDeficit() const { return _totalBufferDeficit; };
  long activeBufferDeficit() const { return _activeBufferDeficit; }; 
  void annulBufferDeficit() 
  {
   CALL("annulBufferDeficit()");
   for (ulong i = 0; i <= MaxSmallObjectSize; i++) 
    if (_numberOfActiveTakenFromBuffer[i] > _numberOfFree[i])
     { 
       //DF; cout << "ANNUL " << (_numberOfActiveTakenFromBuffer[i] - _numberOfFree[i])
       //<< " x " << i*sizeof(void*) << " bytes\n";
      _numberOfActiveTakenFromBuffer[i] = _numberOfFree[i];      
     };
   _activeBufferDeficit = 0L;     
  };

  #ifndef NO_DEBUG
   void** chunk(void* obj) const
   {
    ChunkList* chunks = _allChunks;
    while (chunks)
     {
      if (inChunk(obj,chunks->chunk)) { return chunks->chunk; }
      else chunks = chunks->next;
     }; 

    /******
    {
    ChunkList* chunks = _allChunks;
DF; cout << "OBJ" << (ulong)obj << "\n";
    while (chunks)
     {
       DF; cout << "CHUNK " << (ulong)(chunks->chunk) 
                << " + " << (ulong)(((void**)(chunks->chunk))[0]) << "\n";
      if (inChunk(obj,chunks->chunk)) { return chunks->chunk; }
      else chunks = chunks->next;
     };
    };
    ******/
    return (void**)0;              
   };
  #endif

  #ifdef ALLOCATE_WITH_MIRROR
   void* mirror(void* ptr)
   {
    return (void*)(((void**)ptr) - _debugWithMirrorHalfMemorySize);
   }; 
  #endif
 private:
  #ifndef NO_DEBUG
   struct ChunkList
   {
    ChunkList(void** ch,ChunkList* nxt) : chunk(ch), next(nxt) {};
    void* operator new (size_t) 
    {
     void* res = malloc(sizeof(ChunkList));
     if (!res) Debug::debugFeatureInternalFault("!Allocator<..>::ChunkList::operator new (size_t)::res");
     return res;
    };
    void** chunk;
    ChunkList* next;
   };
  #endif
 private:
  #ifdef ALLOCATE_WITH_MIRROR
   void* debugWithMirrorMalloc(size_t size)
   {
    CALL("debugWithMirrorMalloc(size_t size)");   
    if (_debugWithMirrorFreeMemory + size > _debugWithMirrorEndOfFreeMemory) return (void*)0;
    void* res = _debugWithMirrorFreeMemory;
    _debugWithMirrorFreeMemory += size;
    return res; 
   };   
  #endif
 
  bool useFreeChunk()
  {
   if (_freeChunkList)
    {
     _currentChunk = _freeChunkList;
     _freeChunkList = (void**)(_freeChunkList[0]);
     _nextInCurrentChunk = _currentChunk;
     _freeInCurrChunk = (ulong)(_currentChunk[1]);
     _occupiedByChunks += (long)(_currentChunk[1]);
     #ifndef NO_DEBUG
      _allChunks = new ChunkList(_currentChunk,_allChunks);
      _currentChunk[0] = (void*)_freeInCurrChunk;
      _nextInCurrentChunk++;
      _freeInCurrChunk--;
     #endif
     return true;
    }
   else return false;
  };

  bool createNewChunk()
  {
   #ifdef ALLOCATE_WITH_MIRROR
    _currentChunk = (void**)(debugWithMirrorMalloc(StandardChunkSize));
   #else
     _currentChunk = (void**)(malloc(StandardChunkSize*sizeof(void*)));
   #endif
   if (!_currentChunk) return false;
   _nextInCurrentChunk = _currentChunk;  
   _freeInCurrChunk = StandardChunkSize;
   _occupiedByChunks += StandardChunkSize;
   #ifndef NO_DEBUG
    _allChunks = new ChunkList(_currentChunk,_allChunks);
    _currentChunk[0] = (void*)_freeInCurrChunk;
    _nextInCurrentChunk++;
    _freeInCurrChunk--;
   #endif
   return true;
  };

  #ifndef NO_DEBUG
   static bool inChunk(void* obj,void** chunk) 
   {
    return (((void**)obj) >= chunk) && ((ulong)(((void**)obj) - chunk) < (ulong)(chunk[0]));
   }; 
  #endif

 private:
  void** _freeChunkList;
  void** _freeList[MaxSmallObjectSize+1];
  void** _currentChunk; 
  void** _nextInCurrentChunk;  
  ulong _freeInCurrChunk;
  long _occupiedByObjects;
  long _occupiedByChunks;
  long _hardMemoryLimit;
  long _unbufferedAllocationLimit;
  bool _bufferIsBeingUsed;  

  long _numberOfTakenFromBuffer[MaxSmallObjectSize+1];
  long _numberOfActiveTakenFromBuffer[MaxSmallObjectSize+1];
  long _numberOfFree[MaxSmallObjectSize+1];
  long _totalBufferDeficit;
  long _activeBufferDeficit;

  bool (*_requestMemoryRelease)(ulong size);
  void (*_reportAllocationFailure)(ulong size);
  #ifndef NO_DEBUG
   ChunkList* _allChunks;
  #endif
 
  #ifdef ALLOCATE_WITH_MIRROR
   ulong _debugWithMirrorMemorySize;
   ulong _debugWithMirrorHalfMemorySize;
   void** _debugWithMirrorFreeMemory;
   void** _debugWithMirrorEndOfFreeMemory;        
  #endif
}; //class Allocator<ulong MaxSmallObjectSize,ulong StandardChunkSize> 

//================================================
#ifdef ALLOCATE_WITH_MIRROR
 #undef ALLOCATE_WITH_MIRROR_TOTAL_MEMORY_SIZE_IN_WORDS 
 #undef ALLOCATE_WITH_MIRROR_TOTAL_MEMORY_SIZE_DECREMENT_IN_WORDS 
#endif
#undef DEBUG_NAMESPACE
//====================================================
#endif







