//
// File:         CommandLine.h
// Description:  Parsing command line and printing usage.
// Created:      Jan 25, 2000, 15:22
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//

#ifndef COMMAND_LINE_H
//======================================================================================
#define COMMAND_LINE_H
#include <string.h>
#include <limits.h>
#include <float.h> 
#include <iostream.h>
#include "jargon.h"
#include "BLDebugFlags.h"
#include "Exit.h"
#include "GList.h"
#include "Malloc.h"
#include "FormatText.h"
#include "StrToX.h"
#include "ClassDesc.h"
//======================================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_COMMAND_LINE
 #define DEBUG_NAMESPACE "CommandLine"
#endif
#include "debugMacros.h"
//======================================================================================
class CommandLine
{
 typedef GList<Malloc,const char*,CommandLine> StrList;
 typedef StrList::Iter StrListIter; 

 public: 
  enum ParamType { String, Long, Double };
  class Values : private StrListIter
  {
   public:
    Values() : StrListIter() {};
    operator bool() const { return StrListIter::rest(); };
    bool next(const char*& str);
    bool next(long& n);
    bool next(double& n); 
   private:
    Values(StrList* lst) : StrListIter(lst) {};
    static bool readInf(const char* str,bool& positive)
    {   
     if (!strcmp("inf",str)) { positive = true; return true; };
     if (!strcmp("+inf",str)) { positive = true; return true; };
     positive = false;
     return !strcmp("-inf",str);
    }; 
   friend class CommandLine; 
  };
 public:  
  CommandLine() 
   : optDescs(0),
     nonoptPrefSize(0),
     nonoptPrefNames(0),
     nonoptPrefComments(0),
     nonoptSuffSize(0),
     nonoptSuffNames(0),
     nonoptSuffComments(0),
     nonoptMiddleName(0),
     nonoptMiddleComment(0),
     nonoptMidValues(0),
     hidingChar('+')
  {
  };
  ~CommandLine() 
  {
   CALL("destructor ~CommandLine()"); 
   if (!Exit::fast)
    {
     OptDescList::destroy(optDescs); 
     StrList::destroy(nonoptMidValues);
    };
  };
 
  void setHidingChar(char ch) { hidingChar = ch; };

  // describing arguments
  void option(const char* optName,const char* cmt) // option without argument  
  {
   optDescs = new OptDescList(OptDesc(optName,cmt),optDescs); 
  };
  void option(const char* optName,ParamType paramType,const char* paramName,const char* cmt)
  {
   optDescs = new OptDescList(OptDesc(optName,paramType,paramName,cmt),optDescs); 
  };
  void nonoptPrefix(int num,
                    const char* const * names,
                    const char* const * comments)
  {
   nonoptPrefSize = num;
   nonoptPrefNames = names;
   nonoptPrefComments = comments;
  };
  void nonoptSuffix(int num,
                    const char* const * names,
                    const char* const * comments)
  {
   nonoptSuffSize = num;
   nonoptSuffNames = names;
   nonoptSuffComments = comments;
  };
  void nonoptMiddle(const char* name,const char* comment)
  {
   nonoptMiddleName = name;  
   nonoptMiddleComment = comment;
  };

  // parsing arguments  

  void arguments(int argC,const char* const * argV)
  {
   argc = 1;
   int i;
   for (i = 1; i < argC; i++) if (argV[i][0] != hidingChar) argc++;
   argv = new const char* [argc]; 
   argv[0] = argV[0];
   int j = 1;
   for (i = 1; i < argC; i++)
    if (argV[i][0] != hidingChar)
     {
      argv[j] = argV[i];
      j++;
     };
  };

  bool parse() { return (optParsSuccess = parseOptions()) && parseNonopt(); };
 
  // getting values

  Values values(const char* optName) const
  {
   CALL("values(const char* optName) const");
   OptDescList* desc = findDesc(optName,(strlen(optName) == 1));
   if (desc) return Values(desc->hd().values);
   DOP(DMSG("Option ") << optName << " was not described.\n");
   ICP("ICP1"); 
   return Values((StrList*)0);
  };

  const char* nonoptVal(const char* name) const
  {
   int i = findStr(name,nonoptPrefSize,nonoptPrefNames);
   if (i >= 0) return argv[nonoptArg + i];  
   i = findStr(name,nonoptSuffSize,nonoptSuffNames);
   if (i >= 0)
    {     
     return argv[argc - (nonoptSuffSize - i)];      
    };
   return (const char*)0;
  };

  Values nonoptMiddleValues() const { return Values(nonoptMidValues); };

  // getting help

  ostream& printCommand(ostream& str) const;
  void printUsage(const char* command) const;
  void printHelp(const char* command) const;
  void printError() const;
   

 private:
  struct OptDesc 
  { 
   OptDesc(const char* optName,const char* cmt) 
     : name(optName), parName(0), values(0), comment(cmt)
   {
    shortOpt = (strlen(optName) == 1);    
   };
   OptDesc(const char* optName,ParamType paramType,const char* paramName,const char* cmt) 
    : name(optName), ptype(paramType), parName(paramName), values(0), comment(cmt)
   {
    shortOpt = (strlen(optName) == 1);    
   };   

   void addVal(const char* val)
   {
    values = new StrList(val,values);
   };
   bool shortOpt;
   const char* name; 
   ParamType ptype;
   const char* parName;
   StrList* values; 
   const char* comment;
  };

 public: // made public to please VC++
  typedef GList<Malloc,OptDesc,CommandLine> OptDescList;  
 private:
  bool parseOptions();
  bool parseNonopt()
  {
   int numOfNonopt = argc - nonoptArg;   
   if (nonoptMiddleName) 
    if (numOfNonopt >= nonoptPrefSize + nonoptSuffSize)
     {
      for (int i = (argc - nonoptSuffSize) - 1; i >= nonoptArg + nonoptPrefSize; i--) 
       nonoptMidValues = new StrList(argv[i],nonoptMidValues);  
      return true;
     }
    else return false;
   return (numOfNonopt == nonoptPrefSize + nonoptSuffSize);
  }; 
  OptDescList* findDesc(const char* optName,bool shortOpt) const;     
  static bool strpref(const char* s1,const char* s2)
  { 
   return (strlen(s1) <= strlen(s2)) && (!strncmp(s1,s2,strlen(s1)));
  };  
  static bool checkType(const char* str,ParamType tp);
  static int findStr(const char* str,int arrSize,const char* const *arr); 
 private:
  OptDescList* optDescs;
  int argc;
  const char** argv; 
  int nonoptArg;
  int nonoptPrefSize;
  const char* const * nonoptPrefNames;
  const char* const * nonoptPrefComments;
  int nonoptSuffSize;
  const char* const * nonoptSuffNames;
  const char* const * nonoptSuffComments;
  const char* nonoptMiddleName;
  const char* nonoptMiddleComment;
  StrList* nonoptMidValues; 
  bool optParsSuccess;
  char hidingChar;
}; //class CommandLine 

//======================================================================================
#undef DEBUG_NAMESPACE
//=======================================================================================
#endif
