//
// File:         GlobAlloc.h
// Description:  Allocation of small objects with debugging facilities.
// Created:      Mar 2, 2000, 19:00
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk  
//====================================================
#ifndef GLOB_ALLOC_H
//====================================================
#define GLOB_ALLOC_H  
#include <stdlib.h>
#include "jargon.h"
#include "BLDebugFlags.h"  
#include "ClassDesc.h"
#include "ObjDesc.h"
//================================================
 
#ifdef ALLOCATE_WITH_MIRROR 
 #error GlobAlloc.h: ALLOCATE_WITH_MIRROR defined  
#endif
 
#ifdef DEBUG_ALLOC
 #define ALLOCATE_WITH_MIRROR
 #include "Allocator.h"
 #undef ALLOCATE_WITH_MIRROR
#else
 #undef ALLOCATE_WITH_MIRROR
 #include "Allocator.h"
 #undef ALLOCATE_WITH_MIRROR 
#endif  
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_ALLOC
 #define DEBUG_NAMESPACE "GlobAlloc"
#endif 
#include "debugMacros.h"
//=================================================

#define GLOB_ALLOC_CHUNK_SIZE 16000UL
#define GLOB_ALLOC_MAX_OBJECT_SIZE 4096UL


class GlobAlloc 
{
 public:
  static void setHardMemoryLimit(long lim)
  { 
   CALL("setHardMemoryLimit(long lim)");
   _allocator.setHardMemoryLimit(lim/(sizeof(void*)));
  };  
  static void setAllocationBufferSize(long abs) 
  { 
   _allocator.setBufferSize(abs);  
  };
  static void setRequestMemoryRelease(bool (*fun)(ulong size))
  {
   _requestMemoryRelease = fun;
  };
  static void setReportAllocationFailure(void (*fun)(ulong size))
  { 
   _reportAllocationFailure = fun;
  };  
  static long occupiedByObjects() { return sizeof(void*)*_allocator.occupiedByObjects(); };
  static long occupiedByChunks() { return sizeof(void*)*_allocator.occupiedByChunks(); }; 
  static long freeInTheCurrentChunk() { return sizeof(void*)*_allocator.freeInTheCurrentChunk(); };
  static long totalBufferDeficit() { return sizeof(void*)*_allocator.totalBufferDeficit(); };
  static long activeBufferDeficit() { return sizeof(void*)*_allocator.activeBufferDeficit(); };
  static void annulBufferDeficit() { _allocator.annulBufferDeficit(); };
 
  static void* allocate(size_t size) 
  { 
   CALL("allocate(size_t size)"); 
   ulong nsize = size/sizeof(void*);
   if (size % sizeof(void*)) nsize++;
   #ifndef DEBUG_ALLOC
    return _allocator.allocate(nsize);
   #else
    void* res = _allocator.allocate(nsize);
    void** resMirror = (void**)(_allocator.mirror(res));
    ObjDesc* objDesc = (ObjDesc*)(*resMirror);
    if (!objDesc) // fresh memory
     {
      objDesc = new ObjDesc();
      for (ulong i = 0; i < nsize; i++)
       resMirror[i] = (void*)objDesc;  
      #ifdef DEBUG_ALLOC_OBJ_PTR
       objDesc->objPtr = res;
      #endif  
     }
    else // recycled piece of memory
     {
      #ifdef DEBUG_ALLOC_OBJ_PTR  
       ASSERT(objDesc->objPtr == res);       
      #endif
      #ifdef DEBUG_ALLOC_OBJ_STATUS
       ASSERT(objDesc->status = ObjDesc::Recycled);
      #endif
     };
    #ifdef DEBUG_ALLOC_OBJ_SIZE
     objDesc->size = size;
    #endif
    #ifdef DEBUG_ALLOC_OBJ_STATUS
     objDesc->status = ObjDesc::Persistent;       
    #endif
    #ifdef DEBUG_ALLOC_OBJ_TYPE
     objDesc->classDesc = ClassDesc::universalClassDesc();
    #endif
    #ifdef DEBUG_ALLOC_EVENT_ID
     objDesc->eventId = Debug::event();
    #endif
    return res;
   #endif
  }; // void* allocate(size_t size) 

  static void deallocate(void* obj,size_t size) 
  {
   CALL("deallocate(void* obj,size_t size)");
   ulong nsize = size/sizeof(void*);
   if (size % sizeof(void*)) nsize++;
   ASSERT(_allocator.chunk(obj)); 
   #ifdef DEBUG_ALLOC           
    ObjDesc* objDesc = *((ObjDesc**)(_allocator.mirror(obj))); 
    ASSERT(objDesc);    
    #ifdef DEBUG_ALLOC_OBJ_PTR
     ASSERT(objDesc->objPtr == obj); 
    #endif
    #ifdef DEBUG_ALLOC_OBJ_STATUS
     ASSERT(objDesc->status == ObjDesc::Persistent);   
     objDesc->status = ObjDesc::Recycled;    
    #endif
    #ifdef DEBUG_ALLOC_OBJ_TYPE
     ASSERT(objDesc->classDesc == ClassDesc::universalClassDesc());
    #endif    
    #ifdef DEBUG_ALLOC_OBJ_SIZE
     if (objDesc->size != size) 
      { 
       DMSG("size = ") << size << "\nobjDesc:\n" << *objDesc << '\n';
       ASSERT(objDesc->size == size);
      };
    #endif
    #ifdef DEBUG_ALLOC_EVENT_ID
     objDesc->eventId = Debug::event();
    #endif
   #endif    
   _allocator.deallocate(obj,nsize);
  }; //void deallocate(void* obj,size_t size)  

  #ifndef NO_DEBUG
   static void* allocate(size_t size,ClassDesc* classDesc) 
   { 
    CALL("allocate(size_t size,ClassDesc* classDesc)");
    ulong nsize = size/sizeof(void*);
    if (size % sizeof(void*)) nsize++;
    #ifndef DEBUG_ALLOC
     return _allocator.allocate(nsize);
    #else
     void* res = _allocator.allocate(nsize);
     void** resMirror = (void**)(_allocator.mirror(res));

     ObjDesc* objDesc = (ObjDesc*)(*resMirror);
     if (!objDesc) // fresh memory
      {
       objDesc = new ObjDesc();
       for (ulong i = 0; i < nsize; i++)
        resMirror[i] = (void*)objDesc;  
       #ifdef DEBUG_ALLOC_OBJ_PTR
        objDesc->objPtr = res;
       #endif  
      }
     else // recycled piece of memory
      {
       #ifdef DEBUG_ALLOC_OBJ_PTR  
        ASSERT(objDesc->objPtr == res);       
       #endif
       #ifdef DEBUG_ALLOC_OBJ_STATUS
        ASSERT(objDesc->status = ObjDesc::Recycled);
       #endif
      };
     #ifdef DEBUG_ALLOC_OBJ_SIZE
      objDesc->size = size;
     #endif
     #ifdef DEBUG_ALLOC_OBJ_STATUS
      objDesc->status = ObjDesc::Persistent;       
     #endif
     #ifdef DEBUG_ALLOC_OBJ_TYPE
      objDesc->classDesc = classDesc;
     #endif
     #ifdef DEBUG_ALLOC_EVENT_ID
      objDesc->eventId = Debug::event();
     #endif
     return res;  

    #endif
   }; // void* allocate(size_t size,ClassDesc* classDesc) 
 
   static void deallocate(void* obj,size_t size,ClassDesc* classDesc) 
   {
    CALL("deallocate(void* obj,size_t size,ClassDesc* classDesc)"); 
    ulong nsize = size/sizeof(void*);
    if (size % sizeof(void*)) nsize++;
    ASSERT(_allocator.chunk(obj)); 
    #ifdef DEBUG_ALLOC
     ObjDesc* objDesc = *((ObjDesc**)(_allocator.mirror(obj))); 
     ASSERT(objDesc);
     #ifdef DEBUG_ALLOC_OBJ_PTR
      ASSERT(objDesc->objPtr == obj); 
     #endif
     #ifdef DEBUG_ALLOC_OBJ_STATUS
      if (objDesc->status != ObjDesc::Persistent)
       { 
        DMSG("objDesc:\n") << *objDesc << '\n';
        ASSERT(objDesc->status == ObjDesc::Persistent); 
       }; 
      objDesc->status = ObjDesc::Recycled;     
     #endif
     #ifdef DEBUG_ALLOC_OBJ_TYPE
      ASSERT(objDesc->classDesc == classDesc);
     #endif
     #ifdef DEBUG_ALLOC_OBJ_SIZE
      if (objDesc->size != size) 
       { 
        DMSG("size = ") << size << "\nobjDesc:\n" << *objDesc << '\n';
        ASSERT(objDesc->size == size);
       }; 
     #endif
     #ifdef DEBUG_ALLOC_EVENT_ID
      objDesc->eventId = Debug::event();
     #endif
    #endif
    _allocator.deallocate(obj,nsize);
   }; // void deallocate(void* obj,size_t size,ClassDesc* classDesc) 
  #endif 


  static void* tryToAllocateLarge(size_t size) 
  { 
   CALL("tryToAllocateLarge(size_t size)"); 
   ulong nsize = size/sizeof(void*);
   if (size % sizeof(void*)) nsize++;
   #ifndef DEBUG_ALLOC
    return _allocator.tryToAllocateLarge(nsize);
   #else
    void* res = _allocator.tryToAllocateLarge(nsize);
    if (!res) return (void*)0;
    void** resMirror = (void**)(_allocator.mirror(res));
    ObjDesc* objDesc = (ObjDesc*)(*resMirror);
    if (!objDesc) // fresh memory
     {
      objDesc = new ObjDesc();
      for (ulong i = 0; i < nsize; i++)
       resMirror[i] = (void*)objDesc;    
      #ifdef DEBUG_ALLOC_OBJ_PTR
       objDesc->objPtr = res;
      #endif  
     }
    else // recycled piece of memory
     {
      #ifdef DEBUG_ALLOC_OBJ_PTR  
       ASSERT(objDesc->objPtr == res);       
      #endif
      #ifdef DEBUG_ALLOC_OBJ_STATUS
       ASSERT(objDesc->status = ObjDesc::Recycled);
      #endif
     };
    #ifdef DEBUG_ALLOC_OBJ_SIZE
     objDesc->size = size;
    #endif
    #ifdef DEBUG_ALLOC_OBJ_STATUS
     objDesc->status = ObjDesc::Persistent;       
    #endif
    #ifdef DEBUG_ALLOC_OBJ_TYPE
     objDesc->classDesc = ClassDesc::universalClassDesc();
    #endif
    #ifdef DEBUG_ALLOC_EVENT_ID
     objDesc->eventId = Debug::event();
    #endif
    return res;
   #endif
  }; // void* tryToAllocateLarge(size_t size) 

  static void deallocateLarge(void* obj,size_t size) 
  {
   CALL("deallocateLarge(void* obj,size_t size)");
   ulong nsize = size/sizeof(void*);
   if (size % sizeof(void*)) nsize++;
   ASSERT(_allocator.chunk(obj)); 
   #ifdef DEBUG_ALLOC           
    ObjDesc* objDesc = *((ObjDesc**)(_allocator.mirror(obj))); 
    ASSERT(objDesc);    
    #ifdef DEBUG_ALLOC_OBJ_PTR
     ASSERT(objDesc->objPtr == obj); 
    #endif
    #ifdef DEBUG_ALLOC_OBJ_STATUS
     ASSERT(objDesc->status == ObjDesc::Persistent);   
     objDesc->status = ObjDesc::Recycled;    
    #endif
    #ifdef DEBUG_ALLOC_OBJ_TYPE
     ASSERT(objDesc->classDesc == ClassDesc::universalClassDesc());
    #endif     
    #ifdef DEBUG_ALLOC_OBJ_SIZE
     if (objDesc->size != size) 
      {  
       DMSG("size = ") << size << "\nobjDesc:\n" << *objDesc << '\n';
       ASSERT(objDesc->size == size);
      };
    #endif 
    #ifdef DEBUG_ALLOC_EVENT_ID
     objDesc->eventId = Debug::event();
    #endif
    // make t fresh memory since it will be sliced 
    delete objDesc;
    for (ulong i = 0; i < nsize; i++) 
     ((void**)(_allocator.mirror(obj)))[i] = (void*)0;
   #endif    
   _allocator.deallocateLarge(obj,nsize);
  }; //void deallocateLarge(void* obj,size_t size)  

  #ifndef NO_DEBUG
   static void* tryToAllocateLarge(size_t size,ClassDesc* classDesc) 
   { 
    CALL("tryToAllocateLarge(size_t size)");
    ulong nsize = size/sizeof(void*);
    if (size % sizeof(void*)) nsize++;
    #ifndef DEBUG_ALLOC
     return _allocator.tryToAllocateLarge(nsize);
    #else
     void* res = _allocator.tryToAllocateLarge(nsize);
     if (!res) return (void*)0;
     void** resMirror = (void**)(_allocator.mirror(res));
     ObjDesc* objDesc = (ObjDesc*)(*resMirror);
     if (!objDesc) // fresh memory
      {
       objDesc = new ObjDesc();
       for (ulong i = 0; i < nsize; i++)
        resMirror[i] = (void*)objDesc;    
       #ifdef DEBUG_ALLOC_OBJ_PTR
        objDesc->objPtr = res;
       #endif  
      }
     else // recycled piece of memory
      {
       #ifdef DEBUG_ALLOC_OBJ_PTR  
        ASSERT(objDesc->objPtr == res);       
       #endif
       #ifdef DEBUG_ALLOC_OBJ_STATUS
        ASSERT(objDesc->status = ObjDesc::Recycled);
       #endif
      };
     #ifdef DEBUG_ALLOC_OBJ_SIZE
      objDesc->size = size;
     #endif
     #ifdef DEBUG_ALLOC_OBJ_STATUS
      objDesc->status = ObjDesc::Persistent;       
     #endif
     #ifdef DEBUG_ALLOC_OBJ_TYPE
      objDesc->classDesc = classDesc;
     #endif
     #ifdef DEBUG_ALLOC_EVENT_ID
      objDesc->eventId = Debug::event();
     #endif
     return res;
    #endif
   }; // void* tryToAllocateLarge(size_t size,ClassDesc* classDesc) 

   static void deallocateLarge(void* obj,size_t size,ClassDesc* classDesc) 
   {
    CALL("deallocateLarge(void* obj,size_t size,ClassDesc* classDesc)"); 
    ulong nsize = size/sizeof(void*);
    if (size % sizeof(void*)) nsize++;
    ASSERT(_allocator.chunk(obj)); 
    #ifdef DEBUG_ALLOC
     ObjDesc* objDesc = *((ObjDesc**)(_allocator.mirror(obj))); 
     ASSERT(objDesc);
     #ifdef DEBUG_ALLOC_OBJ_PTR
      ASSERT(objDesc->objPtr == obj); 
     #endif
     #ifdef DEBUG_ALLOC_OBJ_STATUS
      if (objDesc->status != ObjDesc::Persistent)
       { 
        DMSG("objDesc:\n") << *objDesc << '\n';
        ASSERT(objDesc->status == ObjDesc::Persistent); 
       }; 
      objDesc->status = ObjDesc::Recycled;     
     #endif
     #ifdef DEBUG_ALLOC_OBJ_TYPE
      ASSERT(objDesc->classDesc == classDesc);
     #endif
     #ifdef DEBUG_ALLOC_OBJ_SIZE
      if (objDesc->size != size) 
       { 
        DMSG("size = ") << size << "\nobjDesc:\n" << *objDesc << '\n';
        ASSERT(objDesc->size == size);
       }; 
     #endif
     #ifdef DEBUG_ALLOC_EVENT_ID
      objDesc->eventId = Debug::event();
     #endif
     // make t fresh memory since it will be sliced 
     delete objDesc;
     for (ulong i = 0; i < nsize; i++) 
      ((void**)(_allocator.mirror(obj)))[i] = (void*)0;
    #endif
    _allocator.deallocateLarge(obj,nsize);
   }; // void deallocateLarge(void* obj,size_t size,ClassDesc* classDesc) 
  #endif 

  #ifdef DEBUG_ALLOC
   static ObjDesc* hostObj(void* ptr) 
   { 
    return *((ObjDesc**)(_allocator.mirror(ptr)));
   };
  #endif

 private:
  static bool requestMemoryRelease(ulong size); 
  static void reportAllocationFailure(ulong size);
 

 //private:   
 public: // made public to please VC++
  typedef Allocator<GLOB_ALLOC_MAX_OBJECT_SIZE,GLOB_ALLOC_CHUNK_SIZE> Alloc;
 private:
  static Alloc _allocator;
  static bool (*_requestMemoryRelease)(ulong size);
  static void (*_reportAllocationFailure)(ulong size);
}; // class GlobAlloc

//====================================================
#undef DEBUG_NAMESPACE
#undef GLOB_ALLOC_CHUNK_SIZE
#undef GLOB_ALLOC_MAX_OBJECT_SIZE 
//====================================================
#endif
