//
// File:         LinComb.h
// Description:  Representation of shared linear combinations.
// Created:      May 5, 2000, 18:20
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//============================================================================
#ifndef LIN_COMB_H
//============================================================================
#define LIN_COMB_H 
#ifndef NO_DEBUG_VIS
 #include <iostream.h>
#endif
#include "jargon.h"
#include "BLDebugFlags.h"
#ifdef DEBUG_ALLOC_OBJ_TYPE
 #include "ClassDesc.h"       
#endif
#ifndef NO_DEBUG
 #include "ObjDesc.h"
#endif  
#include "GDiscTree.h"
#include "Stack.h"
#include "Exit.h"
//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_LIN_COMB
 #define DEBUG_NAMESPACE "LinComb<class Alloc,ulong MaxSize>"
#endif
#include "debugMacros.h"
//============================================================================

// Allignment of bits in LinComb::Member:
//  tag   coefficient variable/constant 
//  [ 2 ]   [ 14 ]     [16]

template <class Alloc,ulong MaxSize>
class LinComb
{
 public:
  enum Tag
  {
   PlusVar = 0x00000000,
   MinusVar = 0x00000001, 
   PlusConst = 0x00000002,
   MinusConst = 0x00000003
  };
  class Member
  {
   public:
    Member() {};
    Member(const Tag& t,const ulong coeff,const ulong& var) : _content(t)
    {
     CALL("constructor Member(const Tag& t,const ulong& coeff,const ulong& var)");  
     ASSERT(var < 65536UL);
     ASSERT(coeff < 16384UL);
     ((_content <<= 30) |= (coeff <<= 16)) |= var;
    };
    Member(const Tag& t,const ulong& num) : _content(t)
    {
     CALL("constructor Member(const Tag& t,const ulong& num)");
     // for constants
     ASSERT(num < 65536UL);
     (_content <<= 30) |= num;
    }; 
    void makePlusVar(ulong co,const ulong& v)
    {
     CALL("makePlusVar(const ulong& co,const ulong& v)");
     ASSERT(v < 65536UL);
     ASSERT(co < 16384UL);
     (((_content = PlusVar) <<= 30) |= (co <<= 16)) |= v;
    };
    void makeMinusVar(ulong co,const ulong& v) 
    {
     CALL("makeMinusVar(const ulong& co,const ulong& v)");
     ASSERT(v < 65536UL);
     ASSERT(co < 16384UL);
     (((_content = MinusVar) <<= 30) |= (co <<= 16)) |= v;
    };
    void makePlusConst(const ulong& num)
    {
     CALL("makePlusConst(const ulong& num)");
     ASSERT(num < 65536UL);
     ((_content = PlusConst) <<= 30) |= num;
    };
    void makeMinusConst(const ulong& num)
    {
     CALL("makeMinusConst(const ulong& num)");      
     ASSERT(num < 65536UL);
     ((_content = MinusConst) <<= 30) |= num;
    };  

    Tag tag() const { return (Tag)(_content >> 30); };
    ulong coeff() const 
    {
     CALL("Member::coeff() const");
     ASSERT(isVariable());
     return (_content << 2) >> 18; 
    };
    ulong var() const 
    {
     CALL("Member::var() const");
     ASSERT(isVariable());
     return (_content << 16) >> 16; 
    };
    ulong constant() const 
    {
     CALL("Member::constant() const");
     ASSERT(isConstant());
     return (_content << 16) >> 16; 
    };
    void decompose(Tag& t,ulong& c,ulong& v) const
    {
     (t = _content) >>= 30;      
     ((c = _content) <<= 2) >>= 18;
     ((v = _content) <<= 16) >>= 16;  
    }; 
    bool isVariable() const { return tag() < PlusConst; };
    bool isConstant() const { return tag() >= PlusConst; };   
    bool operator==(const Member& m) const { return _content == m._content; };
    bool operator!=(const Member& m) const { return _content != m._content; };
    bool operator<(const Member& m) const { return _content < m._content; }; 
    bool operator<=(const Member& m) const { return _content <= m._content; }; 
    bool operator>(const Member& m) const { return _content > m._content; }; 
    bool operator>=(const Member& m) const { return _content >= m._content; };
    #ifndef NO_DEBUG_VIS
     ostream& output(ostream& str) const 
     {
      switch (tag()) 
       {
        case PlusVar: return str << " + " << coeff() << 'X' << var();
        case MinusVar: return str << " - " << coeff() << 'X' << var();
        case PlusConst: return str << " + " << constant();
        case MinusConst: return str << " - " << constant();
       };
      return str << " ************* Error ************ ";
     }; 
    #endif
   private:
    ulong _content;
  }; // class Member 
 
	

 public:
  LinComb(const Member& m) : _member(m), _refCounter(1L), _tail((LinComb*)0) 
  {
   CALL("constructor LinComb(const Member& m)");
  };
  LinComb(const Member& m,LinComb* t) : _member(m), _refCounter(1L), _tail(t) 
  {
   CALL("LinComb(const Member& m,LinComb* t)");
  }; 
  static ulong minSize() { return sizeof(LinComb); };
  static ulong maxSize() { return sizeof(LinComb); };
  void* operator new(size_t) 
  {
   #ifdef DEBUG_ALLOC_OBJ_TYPE 
    ALWAYS(_classDesc.registerAllocated(sizeof(LinComb)));
    return Alloc::allocate(sizeof(LinComb),&_classDesc);
   #else
    return Alloc::allocate(sizeof(LinComb));
   #endif 
  };
  void  operator delete(void* obj) 
  { 
   #ifdef DEBUG_ALLOC_OBJ_TYPE 
    ALWAYS(_classDesc.registerDeallocated(sizeof(LinComb)));
    Alloc::deallocate(obj,sizeof(LinComb),&_classDesc); 
   #else
    Alloc::deallocate(obj,sizeof(LinComb)); 
   #endif 
  };
  void destroy() 
  {
   CALL("destroy()");
   ASSERT(checkObj());
   delete this;
  }; 
  const Member& hd() const 
  {
   CALL("hd() const");
   ASSERT(checkObj());
   return _member; 
  };  
  const LinComb* const & tl() const 
  {
   CALL("tl() const");
   ASSERT(checkObj()); 
   return _tail; 
  };
  LinComb*& tl() 
  { 
   CALL("tl()");
   ASSERT(checkObj()); 
   return _tail;
  };
  const long& refCounter() const 
  {
   CALL("refCounter() const");
   ASSERT(checkObj());  
   return _refCounter; 
  };
  long& refCounter() 
  {
   CALL("refCounter()");
   ASSERT(checkObj());   
   return _refCounter; 
  }; 
 
  #ifndef NO_DEBUG 
   bool checkObj() const
   {
    #ifndef DEBUG_ALLOC
     return true;
    #else
     ObjDesc* objDesc = Alloc::hostObj((void*)this);
     if (!objDesc) 
      { 
       DMSG("LinComb<..>::checkObj()> Null pointer to object descriptor.\n");  
       return false; 
      };  
     bool res = true;
     #ifdef DEBUG_ALLOC_OBJ_SIZE
      if (objDesc->size !=  sizeof(LinComb)) 
       {
        DMSG("LinComb<..>::checkObj()> objDesc->size !=  sizeof(LinComb<..>)\n");
        res = false;
       };
     #endif
     #ifdef DEBUG_ALLOC_OBJ_STATUS
      if (objDesc->status != ObjDesc::Persistent) 
       {
        DMSG("LinComb<..>::checkObj()> objDesc->status != ObjDesc::Persistent\n");
        res = false;
       };
     #endif
     #ifdef DEBUG_ALLOC_OBJ_TYPE
      if (objDesc->classDesc != &_classDesc) 
       {
        DMSG("LinComb<..>::checkObj()> objDesc->classDesc != &_classDesc\n");
        res = false;
       };
     #endif
     if (!res) DOP(DMSG("LinComb<..>::checkObj()> host object ") << *objDesc << "\n"); 
     return res;
    #endif
   };
  #endif


 private:
  Member _member;
  long _refCounter;
  LinComb* _tail;
  #ifdef DEBUG_ALLOC_OBJ_TYPE
   static ClassDesc _classDesc;
  #endif
}; // template <class Alloc> class LinComb


#ifndef NO_DEBUG_VIS
 template <class Alloc,ulong MaxSize>
 ostream& operator<<(ostream& str,const LinComb<Alloc,MaxSize>::Member& mem) 
 {
  return mem.output(str);
 };

 template <class Alloc,ulong MaxSize>
 ostream& operator<<(ostream& str,const LinComb<Alloc,MaxSize>* lc) 
 {
  if (!lc) return str << '0';
  while (lc)
   {
     //str << " [" << lc->refCounter() << "] ";
    lc->hd().output(str);
    lc = lc->tl();
   };
  return str;
 };
#endif

//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_LIN_COMB
 #define DEBUG_NAMESPACE "LinCombSharing<class Alloc,ulong MaxLinCombSize>"
#endif
#include "debugMacros.h"
//============================================================================

template <class Alloc,ulong MaxLinCombSize>
class LinCombSharing
{
 public:
  LinCombSharing() 
   : _discTree(), 
     _members("_members"), 
     _linCombs("_linCombs")
   {
    CALL("constructor LinCombSharing()");
   };
  ~LinCombSharing() 
  { 
   CALL("destructor ~LinCombSharing()"); 
   if (!Exit::fast) destroy(); 
  };
  
    void destroy() { _discTree.destroyWithObjects(); };
    void resetIntegration() 
    {
     _members.reset();
    };
    void integrPlusVar(const ulong& co,const ulong& v) 
    {
     _members.next().makePlusVar(co,v);
     _members.push();
    };
    void integrMinusVar(const ulong& co,const ulong& v) 
    {
     _members.next().makeMinusVar(co,v);
     _members.push();
    };
    void integrPlusConst(const ulong& num)
    {
     _members.next().makePlusConst(num);
     _members.push();
    };
    void integrMinusConst(const ulong& num)
    {
     _members.next().makeMinusConst(num);
     _members.push();
    };  

    LinComb<Alloc,MaxLinCombSize>* integrate() 
    {
     CALL("integrate()");
     if (!_members) return (LinComb<Alloc,MaxLinCombSize>*)0;
     LinComb<Alloc,MaxLinCombSize>** indObj;
     const LinComb<Alloc,MaxLinCombSize>::Member* first;
     LinComb<Alloc,MaxLinCombSize>* linComb;
     _linCombs.reset(); 
     for (first = _members.begin(); first != _members.end(); first++)
      {
       indObj = integrateIntoIndex(first);
       linComb = *indObj;     
       if (linComb) { linComb->refCounter()++; goto link_created; };
       linComb = new LinComb<Alloc,MaxLinCombSize>(*first);   
       *indObj = linComb;
       _linCombs.push(linComb);
      };     
  
     linComb = _linCombs.pop();

     link_created:
      LinComb<Alloc,MaxLinCombSize>* tmp = linComb;    
      while (_linCombs)
       {
        linComb = _linCombs.pop();
        linComb->tl() = tmp;
        tmp = linComb;
       };
      return linComb;
    }; // LinComb* integrate() 

    void remove(LinComb<Alloc,MaxLinCombSize>* linComb)
    {
     CALL("remove(LinComb<Alloc,MaxLinCombSize>* linComb)");
     next:
      linComb->refCounter()--;
      if (!linComb->refCounter())
       {  
        LinComb<Alloc,MaxLinCombSize>* tmp = linComb->tl();
        removeFromIndex(linComb);
        delete linComb;
        linComb = tmp;
        if (linComb) goto next;
       };
    }; // void remove(LinComb<Alloc,MaxLinCombSize>* linComb) 

   private:
    typedef GDiscTree<Alloc,LinComb<Alloc,MaxLinCombSize>::Member,LinComb<Alloc,MaxLinCombSize> > DiscTree;

   private: 
    LinComb<Alloc,MaxLinCombSize>** integrateIntoIndex(const LinComb<Alloc,MaxLinCombSize>::Member* first)
    {
     CALL("integrateIntoIndex(const LinComb<Alloc,MaxLinCombSize>::Member* first)");
     _discTree.integrator().reset();   
     while (first != _members.end())
      {
       _discTree.integrator().skipOrInsert(*first);
       first++;
      };
     return _discTree.integrator().indexedObj();
    }; // LinComb<Alloc,MaxLinCombSize>** integrateIntoIndex(const LinComb<Alloc,MaxLinCombSize>::Member* first)


    bool removeFromIndex(LinComb<Alloc,MaxLinCombSize>* const linComb)
    {
     CALL("removeFromIndex(LinComb<Alloc,MaxLinCombSize>* linComb)");
     _discTree.remover().reset();
     LinComb<Alloc,MaxLinCombSize>* tmp = linComb;
     while (tmp) 
      {
       if (!_discTree.remover().skip(tmp->hd())) return false;
       tmp = tmp->tl();
      };
     LinComb<Alloc,MaxLinCombSize>** indObj = _discTree.remover().indexedObj();
     if ((!indObj) || (*indObj != linComb)) return false;
     _discTree.remover().destroyBranch();
     return true;
    }; // bool removeFromIndex(LinComb<Alloc,MaxLinCombSize>* linComb)

   private:
    DiscTree _discTree;
    Stack<LinComb<Alloc,MaxLinCombSize>::Member,MaxLinCombSize> _members;
    Stack<LinComb<Alloc,MaxLinCombSize>*,MaxLinCombSize> _linCombs; 
  }; // class LinCombSharing






//============================================================================
#endif
