//
// File:         OrdList.h
// Description:  Generic ordered association lists.
// Created:      Feb 25, 2000, 11:45
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//============================================================================
#ifndef ORD_LIST_H
//=============================================================================
#define ORD_LIST_H
#include "jargon.h"
#include "BLDebugFlags.h"
//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_ORD_LIST
 #define DEBUG_NAMESPACE "OrdList<class Alloc,class ValType,class KeyType>"
#endif
#include "debugMacros.h"
//============================================================================
template <class Alloc,class ValType,class KeyType>
class OrdList
{
 public:
  class Surfer;
  class Destruction;
  class Node
  {
   public:
    Node(const KeyType& k) : _next((Node*)0), _key(k){};     
    Node(const KeyType& k,Node* nxt) : _next(nxt), _key(k) {};    
    Node(const KeyType& k,const ValType& val,Node* nxt) : _value(val), _next(nxt), _key(k) {};
    void* operator new(size_t) { return Alloc::allocate(sizeof(Node)); };
    void operator delete(void* obj) { Alloc::deallocate(obj,sizeof(Node)); };
    Node*& next() { return _next; };
    const Node * const & next() const { return _next; };
    const ValType& value() const { return _value; };
    ValType& value() { return _value; };
    const KeyType& key() const { return _key; };
    KeyType& key() { return _key; };
   private:
    Node() {};
   private:
    ValType _value;
    Node* _next;
    KeyType _key;
   friend class OrdList;
   friend class OrdList::Surfer;
  }; // class Node

  class Surfer
  {
   public:
    Surfer() {};
    Surfer(const OrdList& lst) { reset(lst); };    
    void reset(const OrdList& lst)
    {
     _currNode = lst.first();
    };
    const Node* currNode() const { return _currNode; };
    bool find(const KeyType& searchKey) { return find(searchKey,_currNode); };
    bool find(const KeyType& searchKey,const Node*& node)
    {     
     check_node:
      if (_currNode)
       {     
	if (_currNode->key() > searchKey) { _currNode = _currNode->next(); goto check_node; };
        // _currNode->key() <= searchKey      
        if (_currNode->key() == searchKey) { node = _currNode; return true; };
        // _currNode->key() < searchKey  
        node = _currNode;
        return false;
       };
      node = (Node*)0;
      return false;
    };
   private:
    const Node* _currNode;
  }; // class Surfer

  class Destruction
  {
   public:
    Destruction() {};
    Destruction(OrdList& ol) { reset(ol); }; 
    void reset(OrdList& ol) 
    { 
     _node = ol.first();
     ol._size = 0;
     ol._list = (Node*)0;
    };
    bool next(ValType& val) 
    {
     if (!_node) return false;
     val = _node->value();
     Node* tmp = _node;
     _node = _node->next();
     delete tmp;
     return true;
    };
   private: 
    Node* _node;
  };

 public:
  OrdList() : _list((Node*)0), _size(0) {};
  ~OrdList() { CALL("destructor ~OrdList()");  destroy(); };
  void* operator new(size_t) { return Alloc::allocate(sizeof(OrdList)); };
  void operator delete(void* obj) { Alloc::deallocate(obj,sizeof(OrdList)); };
  void destroy();
  void reset() 
  {
   destroy();
   _list = (Node*)0;
   _size = 0;
  };
  operator bool() const { return (bool)_list; }; 
  const Node* first() const { return (const Node*)_list; };
  Node* first() { return _list; };
  unsigned long size() const { return _size; }; 
  unsigned long count() const 
  {
   unsigned long res = 0; 
   for (const Node* n = first(); n; n = n->next()) res++;
   return res;
  };
  bool contains(const KeyType& searchKey) const
  {
   Surfer surfer(*this);
   const Node* dummy;
   return surfer.find(searchKey,dummy);
  }; // bool contains (const KeyType& searchKey) const

  Node* insert(const KeyType& searchKey,bool& newNode); 
  bool remove(const KeyType& searchKey,ValType& val);
  #ifndef NO_DEBUG
   bool checkIntegrity() const { return (count() == size()); };
  #endif
 private:
  Node* _list;
  unsigned long _size; 
 friend class Surfer;
 friend class Destruction;   
}; // template <class Alloc,class ValType,class KeyType> class OrdList

template <class Alloc,class ValType,class KeyType>
OrdList<Alloc,ValType,KeyType>::Node* OrdList<Alloc,ValType,KeyType>::insert(const KeyType& searchKey,bool& newNode)
{
 CALL("insert(const KeyType& searchKey,bool& newNode)");
 Node** addr = &_list;
 Node* node = _list;
 KeyType nodeKey;
 check_node:
  if (node)
   {
    nodeKey = node->key();       
    if (nodeKey > searchKey)
     {
      addr = &(node->next()); 
      node = node->next();
      goto check_node;
     };
    // nodeKey <= searchKey
    if (nodeKey == searchKey) { newNode = false; return node; };
   };
  ASSERT(!node || (node->key() < searchKey));
  // create new node

  Node* res = new Node(searchKey,node);
  *addr = res;
  newNode = true;
  _size++;
  return res; 
}; // OrdList<Alloc,ValType,KeyType>::Node* OrdList<Alloc,ValType,KeyType>::insert(const KeyType& searchKey,bool& newNode)
 
template <class Alloc,class ValType,class KeyType>
bool OrdList<Alloc,ValType,KeyType>::remove(const KeyType& searchKey,ValType& val)
{
 CALL("remove(const KeyType& searchKey,ValType& val)");
 Node** addr = &_list;
 Node* node = _list;
 KeyType nodeKey;
 check_node:
  if (node)
   {
    nodeKey = node->key();       
    if (nodeKey > searchKey)
     {
      addr = &(node->next()); 
      node = node->next();
      goto check_node;
     };
    // nodeKey <= searchKey
    if (nodeKey == searchKey) 
     {
      val = node->value(); 
      *addr = node->next();
      delete node;
      _size--;
      return true; 
     };    
   };
  ASSERT(!node || (node->key() < searchKey));
  return false;
}; // bool OrdList<Alloc,ValType,KeyType>::remove(const KeyType& searchKey,ValType& val) 

template <class Alloc,class ValType,class KeyType>
void OrdList<Alloc,ValType,KeyType>::destroy()
{
 Node* tmp;
 while (_list) 
  {
   tmp = _list;
   _list = _list->next();
   delete tmp; 
  }; 
}; // void OrdList<Alloc,ValType,KeyType>::destroy()


//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_ORD_LIST
 #define DEBUG_NAMESPACE "VoidOrdList<class Alloc,class KeyType>"
#endif
#include "debugMacros.h"
//============================================================================

template <class Alloc,class KeyType>
class VoidOrdList
{
 public:
  class Surfer;
  class Node
  {
   public:
    Node(const KeyType& k) : _next((Node*)0), _key(k){};     
    Node(const KeyType& k,Node* nxt) : _next(nxt), _key(k) {};  
    void* operator new(size_t) { return Alloc::allocate(sizeof(Node)); };
    void operator delete(void* obj) { Alloc::deallocate(obj,sizeof(Node)); };
    Node*& next() { return _next; };
    const Node * const & next() const { return _next; };
    const KeyType& key() const { return _key; };
    KeyType& key() { return _key; };
   private:
    Node() {};
   private:
    Node* _next;
    KeyType _key;
   friend class VoidOrdList;
   friend class VoidOrdList::Surfer;
  }; // class Node

  class Surfer
  {
   public:
    Surfer() {};
    Surfer(const VoidOrdList& lst) { reset(lst); };    
    void reset(const VoidOrdList& lst)
    {
     _currNode = lst.first();
    };
    const Node* currNode() const { return _currNode; };
    bool find(const KeyType& searchKey) { return find(searchKey,_currNode); };
    bool find(const KeyType& searchKey,const Node*& node)
    {     
     check_node:
      if (_currNode)
       {     
	if (_currNode->key() > searchKey) { _currNode = _currNode->next(); goto check_node; };
        // _currNode->key() <= searchKey      
        if (_currNode->key() == searchKey) { node = _currNode; return true; };
        // _currNode->key() < searchKey  
        node = _currNode;
        return false;
       };
      node = (Node*)0;
      return false;
    };
   private:
    const Node* _currNode;
  }; // class Surfer

 public:
  VoidOrdList() : _list((Node*)0), _size(0) {};
  ~VoidOrdList() { CALL("destructor ~VoidOrdList()"); destroy(); };
  void* operator new(size_t) { return Alloc::allocate(sizeof(VoidOrdList)); };
  void operator delete(void* obj) { Alloc::deallocate(obj,sizeof(VoidOrdList)); };
  void destroy();
  void reset() 
  {
   destroy();
   _list = (Node*)0;
   _size = 0;
  };
  operator bool() const { return (bool)_list; }; 
  const Node* first() const { return (const Node*)_list; };
  Node* first() { return _list; };
  unsigned long size() const { return _size; }; 
  unsigned long count() const 
  {
   unsigned long res = 0; 
   for (const Node* n = first(); n; n = n->next()) res++;
   return res;
  };
  bool contains(const KeyType& searchKey) const
  {
   Surfer surfer(*this);
   const Node* dummy;
   return surfer.find(searchKey,dummy);
  }; // bool contains (const KeyType& searchKey) const

  bool insert(const KeyType& searchKey); 
  bool remove(const KeyType& searchKey);
  #ifndef NO_DEBUG
   bool checkIntegrity() const { return (count() == size()); };
  #endif
 private:
  Node* _list;
  unsigned long _size; 
 friend class Surfer; 
}; // template <class Alloc,class KeyType> class VoidOrdList

template <class Alloc,class KeyType>
bool VoidOrdList<Alloc,KeyType>::insert(const KeyType& searchKey)
{
 CALL("insert(const KeyType& searchKey)");
 Node** addr = &_list;
 Node* node = _list;
 KeyType nodeKey;
 check_node:
  if (node)
   {
    nodeKey = node->key();       
    if (nodeKey > searchKey)
     {
      addr = &(node->next()); 
      node = node->next();
      goto check_node;
     };
    // nodeKey <= searchKey
    if (nodeKey == searchKey) return false;
   };
  ASSERT(!node || (node->key() < searchKey));
  // create new node

  Node* res = new Node(searchKey,node);
  *addr = res;
  _size++;
  return true; 
}; // bool VoidOrdList<Alloc,KeyType>::insert(const KeyType& searchKey)
 
template <class Alloc,class KeyType>
bool VoidOrdList<Alloc,KeyType>::remove(const KeyType& searchKey)
{
 CALL("remove(const KeyType& searchKey)");
 Node** addr = &_list;
 Node* node = _list;
 KeyType nodeKey;
 check_node:
  if (node)
   {
    nodeKey = node->key();       
    if (nodeKey > searchKey)
     {
      addr = &(node->next()); 
      node = node->next();
      goto check_node;
     };
    // nodeKey <= searchKey
    if (nodeKey == searchKey) 
     {
      *addr = node->next();
      delete node;
      _size--;
      return true; 
     };    
   };
  ASSERT(!node || (node->key() < searchKey));
  return false;
}; // bool VoidOrdList<Alloc,KeyType>::remove(const KeyType& searchKey)

template <class Alloc,class KeyType>
void VoidOrdList<Alloc,KeyType>::destroy()
{
 Node* tmp;
 while (_list) 
  {
   tmp = _list;
   _list = _list->next();
   delete tmp; 
  }; 
}; // void VoidOrdList<Alloc,KeyType>::destroy()





//======================================================================
#undef DEBUG_NAMESPACE
//======================================================================
#endif
