//
// File:         SmallHeaderSkipList.h
// Description:  Generic skip lists with small (flexible size) headers.
// Created:      Apr 23, 2001, 18:10
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
// Note:         More memory efficient than skip lists with fixed size header,  
//               although some operations are slightly slower.  
// Note:         A specialised version for ValType = void is to be implemented.        
//============================================================================
#ifndef SMALL_HEADER_SKIP_LIST_H
//=============================================================================
#define SMALL_HEADER_SKIP_LIST_H 
#include <limits.h>                                    // for ULONG_MAX
#ifndef NO_DEBUG_VIS 
 #include <iostream.h>                                 // for debugging output
#endif
#include "jargon.h"
#include "BLDebugFlags.h"
#ifdef DEBUG_ALLOC_OBJ_TYPE
 #include "ClassDesc.h"       
#endif
#include "Label.h"                                    // for debugging output
#include "Tab.h"                                      // for debugging output

//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_SKIP_LIST
 #define DEBUG_NAMESPACE "SmallHeaderSkipList<class Alloc,class ValType,class KeyType,ulong Base,class InstanceId>"
#endif
#include "debugMacros.h"
//============================================================================

template <class Alloc,class ValType,class KeyType,ulong Base,class InstanceId>
class SmallHeaderSkipList
{
 public:
  class Surfer;
  class Destruction;
  class Node
  {
   public:
    KeyType key() const { return _key; };
    const ValType& value() const { return _value; };
    ValType& value() { return _value; };
    Node* next() { return *(jump()); };
    const Node* next() const { return *(jump()); };    
    static ulong minSize() { return sizeof(Node) + sizeof(Node*); }; 
    static ulong maxSize() { return ULONG_MAX; }; 


    #ifndef NO_DEBUG_VIS
     ostream& output(ostream& ostr,int maxLevel) const
     {
      ostr << '[' << Label((void*)this) << "] " << Tab<Alloc>::ULong(key()) << " ";
      for (int n = 0; n <= level(); n++)
       {
        ostr << "<" << Label((void*)(jump()[n])) << "> ";
       };
      for (int n = level() + 1; n <= maxLevel; n++)
       {
        ostr << (Tab<Alloc>::String(Label::Length + 3,"|   "));
       };
      return ostr;
     };
    #endif

   private:
    int level() const { return _level; };

   private: 
	// the following functions are public because Surfer and SmallHeaderSkipList use them,
	// "friend" can't be used due to VC++
    Node** jump() { return (Node**)(this + 1); }; 
    const Node* const * jump() const { return (const Node* const *)(this + 1); };
    Node** jump(int lev) { return ((Node**)(this + 1)) + lev; }; 
    const Node* const * jump(int lev) const { return ((const Node* const *)(this + 1)) + lev; };
   
    Node(int l,const KeyType& k) : _level(l), _key(k) {};
    Node* next(int lev) { return ((Node**)(this + 1))[lev]; }; 
    const Node* next(int lev) const { return ((const Node* const *)(this + 1))[lev]; };
      
    static const Node* node(const Node * const * jmp)  // This is a hack! Must be removed.
    {
     return ((Node*)jmp) - 1;
    };

    static Node* node(Node ** jmp)           // This is a hack! Must be removed.
    {
     return ((Node*)jmp) - 1;
    };
  
    void* operator new(size_t size,int lv) 
    {    
     CALL("operator new(size_t size,int lv)");  
     #ifdef DEBUG_ALLOC_OBJ_TYPE
      ALWAYS(_classDesc.registerAllocated(size + (lv+1)*sizeof(Node*)));
      return Alloc::allocate(size + (lv+1)*sizeof(Node*),&_classDesc);
     #else
      return Alloc::allocate(size + (lv+1)*sizeof(Node*));
     #endif
    };
    void operator delete(void* obj)
    {
     CALL("operator delete(void* obj)");
     #ifdef DEBUG_ALLOC_OBJ_TYPE
      ALWAYS(_classDesc.registerDeallocated(sizeof(Node) + (((Node*)obj)->level() + 1)*sizeof(Node*)));
      Alloc::deallocate(obj,sizeof(Node) + (((Node*)obj)->level() + 1)*sizeof(Node*),&_classDesc);
     #else
      Alloc::deallocate(obj,sizeof(Node) + (((Node*)obj)->level() + 1)*sizeof(Node*));
     #endif
    };   
   private: 
    int _level;
    ValType _value;
    KeyType _key;    
    #ifdef DEBUG_ALLOC_OBJ_TYPE
     static ClassDesc _classDesc;
    #endif   
   friend class SmallHeaderSkipList<Alloc,ValType,KeyType,Base,InstanceId>;
   friend class Surfer;
   friend class Destruction;
  }; // class Node

  class Surfer
  {
   public:
    Surfer() {};
    Surfer(const SmallHeaderSkipList& sl) { reset(sl); };
    void reset(const SmallHeaderSkipList& sl) 
    {
     CALL("Surfer::reset(const SmallHeaderSkipList& sl)");
     ASSERT(sl.checkIntegrity());
     _currMaxLevel = sl._currMaxLevel;
     _jump = sl._topIndex;
     _currNode = _jump[0];
    };
    const Node* currNode() const { return _currNode; };
    bool find(const KeyType& searchKey) { return find(searchKey,_currNode); };
    bool find(const KeyType& searchKey,const Node*& node)
    {
     CALL("Surfer::find(const KeyType& searchKey,const Node*& node)");
     ml_next_node:
      node = _jump[_currMaxLevel];   
      if (node) 
       {
    	ASSERT(node->level() == _currMaxLevel); 
	    if (node->key() < searchKey)
		{
          if (!_currMaxLevel) return false;
          int currLevel = _currMaxLevel - 1;
          const Node* const * jump_ = _jump;        
	     next_node:
	      node = jump_[currLevel];
          ASSERT(node);
          ASSERT(node->level() >= currLevel);
          if (node->key() == searchKey) return true;
          if (node->key() < searchKey)
		  { // try smaller level  
	        if (!currLevel) return false;   
            currLevel--;  
           }
          else // node->key() > searchKey, proceed with the same level            
           jump_ = node->jump();
           goto next_node;
         };
        // node->key() >= searchKey 
        _jump = node->jump();
        if (node->key() == searchKey) return true;
        goto ml_next_node;
       }
      else // !node         
       {
	if (!_currMaxLevel) return false;
        _currMaxLevel--;
        goto ml_next_node;
       };      
    }; // bool find(const KeyType& searchKey,const Node*& node) 
   private:
    int _currMaxLevel;
    const Node* const * _jump;
    const Node* _currNode;
  }; // class Surfer  

  class Destruction
  {
   public:
    Destruction() {};
    Destruction(SmallHeaderSkipList& sl) { reset(sl); }; 
    void reset(SmallHeaderSkipList& sl) 
    { 
     _node = sl.first();
     sl._insertCount = 0;
     sl._size = 0;

     #ifdef DEBUG_ALLOC_OBJ_TYPE
      ALWAYS(_headerClassDesc.registerDeallocated((_currMaxLevel + 1) * sizeof(Node*)));
      Alloc::deallocate(_topIndex,(_currMaxLevel + 1) * sizeof(Node*),&_headerClassDesc);
     #else
      Alloc::deallocate(_topIndex,(_currMaxLevel + 1) * sizeof(Node*));
     #endif
     sl._currMaxLevel = 0;

     #ifdef DEBUG_ALLOC_OBJ_TYPE
      ALWAYS(_headerClassDesc.registerAllocated((_currMaxLevel + 1) * sizeof(Node*),&_headerClassDesc)); 
      _topIndex = Alloc::allocate((_currMaxLevel + 1) * sizeof(Node*),&_headerClassDesc); 
     #else
      _topIndex = Alloc::allocate((_currMaxLevel + 1) * sizeof(Node*)); 
     #endif
     sl._topIndex[0] = (Node*)0;
    };
    bool next(ValType& val) 
    {
     if (!_node) return false;
     val = _node->value();
     Node* tmp = _node;
     _node = _node->next();
     delete tmp;
     return true;
    };
   private: 
    Node* _node;
  };

 public:
  SmallHeaderSkipList() : _currMaxLevel(0), _insertCount(0), _size(0) 
   {
    CALL("constructor SmallHeaderSkipList()"); 
    #ifdef DEBUG_ALLOC_OBJ_TYPE
     ALWAYS(_headerClassDesc.registerAllocated((_currMaxLevel + 1) * sizeof(Node*))); 
     _topIndex = (Node**)Alloc::allocate((_currMaxLevel + 1) * sizeof(Node*),&_headerClassDesc); 
    #else
     _topIndex = (Node**)Alloc::allocate((_currMaxLevel + 1) * sizeof(Node*)); 
    #endif
    _topIndex[0] = (Node*)0; 
   };
  ~SmallHeaderSkipList() 
   {
    CALL("destructor ~SmallHeaderSkipList()");
    destroy(); 
   };
  void* operator new(size_t) 
  {
   CALL("operator new(size_t)"); 
   #ifdef DEBUG_ALLOC_OBJ_TYPE
    ALWAYS(_classDesc.registerAllocated(sizeof(SmallHeaderSkipList)));
    return Alloc::allocate(sizeof(SmallHeaderSkipList),&_classDesc);
   #else
    return Alloc::allocate(sizeof(SmallHeaderSkipList));
   #endif 
  };
  void operator delete(void* obj) 
  {
   CALL("operator delete(void* obj)"); 
   #ifdef DEBUG_ALLOC_OBJ_TYPE
    ALWAYS(_classDesc.registerDeallocated(sizeof(SmallHeaderSkipList)));
    Alloc::deallocate(obj,sizeof(SmallHeaderSkipList),&_classDesc);
   #else
    Alloc::deallocate(obj,sizeof(SmallHeaderSkipList));
   #endif 
  };
  static ulong minSize() { return sizeof(SmallHeaderSkipList); }; // has nothing to do with size()
  static ulong maxSize() { return sizeof(SmallHeaderSkipList); }; // has nothing to do with size()
  static ulong headerMinSize() { return sizeof(Node*); }; // has nothing to do with size()
  static ulong headerMaxSize() { return ULONG_MAX; }; 
  void destroy();
  void fastReset() 
  {
   CALL("fastReset()");
   _currMaxLevel = 0;
   _insertCount = 0;
   _size = 0;
   _topIndex[0] = (Node*)0;
  };
  void reset() 
  {
   CALL("reset()");
   destroy();
   _currMaxLevel = 0;
   _insertCount = 0;
   _size = 0;
   _topIndex[0] = (Node*)0;
  };
  operator bool() const { return _topIndex[0]; }; 
  const Node* first() const { return _topIndex[0]; };
  Node* first() { return _topIndex[0]; };
  ulong size() const { return _size; }; 

  ulong count() const 
  {
   ulong res = 0; 
   for (const Node* n = first(); n; n = n->next()) res++;
   return res;
  };

  bool contains(const KeyType& searchKey) const
  {
   Surfer surfer(*this);
   const Node* dummy;
   return surfer.find(searchKey,dummy);
  }; // bool contains (const KeyType& searchKey) const

  Node* insert(const KeyType& searchKey,bool& newNode)  
  {
   CALL("insert(const KeyType& searchKey,bool& newNode)");
   ASSERT(checkIntegrity());

   Node* update[32];

   Node* currNode = Node::node(_topIndex);

   int currLevel = _currMaxLevel;
   Node* nextNode = currNode->next(currLevel); 

   KeyType nodeKey;
   check_next_node:
    if (nextNode)
     {
      ASSERT(nextNode->level() <= currLevel);
      nodeKey = nextNode->key();
      if (nodeKey > searchKey) 
       {
	currNode = nextNode;
	nextNode = nextNode->next(currLevel);        
        goto check_next_node;
       };
      // nodeKey <= searchKey
      if (nodeKey == searchKey) { newNode = false; ASSERT(checkIntegrity()); return nextNode; };
      // nodeKey < searchKey, "if (nextNode)" is not needed anymore
      update[currLevel] = currNode;
      while (currLevel) 
       { 
        currLevel--;
        nextNode = currNode->next(currLevel);
        nodeKey = nextNode->key();
        while (nodeKey > searchKey)
	 {
	  currNode = nextNode;
          nextNode = nextNode->next(currLevel); 
          nodeKey = nextNode->key();
         };
        // nodeKey <= searchKey
        if (nodeKey == searchKey) { newNode = false; ASSERT(checkIntegrity()); return nextNode; };
        // nodeKey < searchKey, "if (nextNode)" is not needed anymore
        update[currLevel] = currNode;
       };
      ASSERT(nextNode);
      ASSERT(nextNode->key() < searchKey);
     }
    else // !next_node
     {
      update[currLevel] = currNode;
      if (currLevel) 
       {
	currLevel--;
        nextNode = currNode->next(currLevel);
        goto check_next_node;
       };
     };
 
    ASSERT(!currLevel);    

    // insert new node
    newNode = true;
    ASSERT(_insertCount < INT_MAX);
    _insertCount++;  
    _size++;
    currLevel = randomLevel();

    currNode = new (currLevel) Node(currLevel,searchKey); 
    Node** nodeJump = currNode->jump();  
    Node** jump; 

    if (currLevel > _currMaxLevel)
     {
      // _topIndex must be expanded 

      #ifdef DEBUG_ALLOC_OBJ_TYPE
       ALWAYS(_headerClassDesc.registerAllocated((currLevel + 1) * sizeof(Node*)));
       Node** newTopIndex = (Node**)Alloc::allocate((currLevel + 1) * sizeof(Node*),&_headerClassDesc);
      #else
       Node** newTopIndex = (Node**)Alloc::allocate((currLevel + 1) * sizeof(Node*));
      #endif

      for (int lev = _currMaxLevel; lev>=0; lev--)
       newTopIndex[lev] = _topIndex[lev];

      #ifdef DEBUG_ALLOC_OBJ_TYPE
       ALWAYS(_headerClassDesc.registerDeallocated((_currMaxLevel + 1) * sizeof(Node*)));
       Alloc::deallocate(_topIndex,(_currMaxLevel + 1) * sizeof(Node*),&_headerClassDesc);
      #else
       Alloc::deallocate(_topIndex,(_currMaxLevel + 1) * sizeof(Node*));   
      #endif

      Node* pseudonode = Node::node(_topIndex);      
      
      for (int lev = _currMaxLevel; lev>=0; lev--)
       { 
        if (update[lev] == pseudonode)      
         {
	  jump = newTopIndex + lev;
	 }
	else jump = update[lev]->jump() + lev;  
        ASSERT(jump);
        ASSERT((*jump) ? ((*jump)->key() < searchKey) : true);
        nodeJump[lev] = *jump;       
        *jump = currNode;
       };

      _topIndex = newTopIndex;

      do
       {
        _currMaxLevel++;
        _topIndex[_currMaxLevel] = currNode;
        nodeJump[_currMaxLevel] = (Node*)0;
       }
      while (_currMaxLevel < currLevel);
     }
    else // currLevel <= _currMaxLevel
     {       
      for (int lev = currLevel; lev>=0; lev--)
       { 
        jump = update[lev]->jump() + lev;  
        ASSERT(jump);
        ASSERT((*jump) ? ((*jump)->key() < searchKey) : true); 
        nodeJump[lev] = *jump;       
        *jump = currNode;
       };
     }; 
    ASSERT(checkIntegrity());
    return currNode;
  }; // Node* insert(const KeyType& searchKey,bool& newNode) 

  bool remove(const KeyType& searchKey,ValType& val) 
  {
   CALL("remove(const KeyType& searchKey,ValType& val)");
   ASSERT(checkIntegrity());
   Node** jump = _topIndex;  
   int currLevel = _currMaxLevel; 
   Node* node;
   next_node:   
    node = jump[currLevel]; 
    if (node)
     {
      ASSERT(node->level() >= currLevel);  
      if (node->key() == searchKey)
       { 
        jump[currLevel] = node->jump()[currLevel]; 
        if (currLevel) { currLevel--; goto next_node; };
        val = node->value();
        if (((!(_topIndex[_currMaxLevel])) && _currMaxLevel))
	 {
          // _topIndex must shrink 
	  int oldMaxLevel = _currMaxLevel;
	  do 
           { 
            _currMaxLevel--; 
           }
          while ((!(_topIndex[_currMaxLevel])) && _currMaxLevel);

          #ifdef DEBUG_ALLOC_OBJ_TYPE
	   ALWAYS(_headerClassDesc.registerAllocated((_currMaxLevel + 1) * sizeof(Node*)));
           Node** newTopIndex = (Node**)Alloc::allocate((_currMaxLevel + 1) * sizeof(Node*),&_headerClassDesc);
          #else
           Node** newTopIndex = (Node**)Alloc::allocate((_currMaxLevel + 1) * sizeof(Node*));
          #endif

          for (int lev = _currMaxLevel; lev>=0; lev--)
           newTopIndex[lev] = _topIndex[lev];

          #ifdef DEBUG_ALLOC_OBJ_TYPE
	   ALWAYS(_headerClassDesc.registerDeallocated((oldMaxLevel + 1) * sizeof(Node*)));
           Alloc::deallocate(_topIndex,(oldMaxLevel + 1) * sizeof(Node*),&_headerClassDesc);
          #else
           Alloc::deallocate(_topIndex,(oldMaxLevel + 1) * sizeof(Node*));
          #endif
          _topIndex = newTopIndex;                 
         };
        delete node;
        _size--;
        ASSERT(checkIntegrity());
        ASSERT(!contains(searchKey));
        return true;
       };
      if (node->key() > searchKey) 
       {
	jump = node->jump();        
        goto next_node;
       };
      // node->key() < searchKey, try smaller level
      if (currLevel) { currLevel--; goto next_node; }; 
      ASSERT(checkIntegrity());    
      ASSERT(!contains(searchKey)); 
      return false;
     };
    
    // !node
    if (currLevel) { currLevel--; goto next_node; };
    ASSERT(checkIntegrity());
    ASSERT(!contains(searchKey));
    return false;
  }; // bool remove(const KeyType& searchKey,ValType& val)  

  #ifndef NO_DEBUG
   bool checkIntegrity() const
   {
    if (_currMaxLevel)
     for (int lev = 0; lev <= _currMaxLevel; lev++)
      {
       if (!(_topIndex[lev])) return false;
      };
    return (count() == size()); 
   }; // bool checkIntegrity() const
  #endif

  ostream& output(ostream& ostr) const
  {
   for (const Node* node = first(); node; node = node->next())
    {
     node->output(ostr,_currMaxLevel) << "\n";
    };
   return ostr;
  };
  ostream& outputKeys(ostream& ostr) const
  {
   for (const Node* node = first(); node; node = node->next())
    {
     ostr << node->key() << "\n";
    };
   return ostr;
  };

 private:
  int randomLevel() const
  { 
   int res = 0;
   ulong tmp = _insertCount;
   while (!(tmp % Base))
    { 
     res++;
     tmp = tmp/Base;
    };
   return res;
  }; // randomLevel() 
 private:
  Node** _topIndex;
  int _currMaxLevel;
  ulong _insertCount;
  ulong _size;
  #ifdef DEBUG_ALLOC_OBJ_TYPE
   static ClassDesc _classDesc;
   static ClassDesc _headerClassDesc;
  #endif
 friend class Surfer; 
 friend class Destruction;   
}; // template <class Alloc,class ValType,class KeyType,ulong Base,class InstanceId> class SmallHeaderSkipList

#ifndef NO_DEBUG_VIS
 template <class Alloc,class ValType,class KeyType,ulong Base,class InstanceId>
 ostream& operator << (ostream& ostr,const SmallHeaderSkipList<Alloc,ValType,KeyType,Base,InstanceId>& lst)
 {
  return lst.output(ostr);
 };
#endif

template <class Alloc,class ValType,class KeyType,ulong Base,class InstanceId>
void SmallHeaderSkipList<Alloc,ValType,KeyType,Base,InstanceId>::destroy()
{
 Node* node = _topIndex[0];
 Node* tmp;
 while (node) 
  { 
   tmp = node;   
   node = node->next(); 
   delete tmp;
  };  
 #ifdef DEBUG_ALLOC_OBJ_TYPE
  ALWAYS(_headerClassDesc.registerDeallocated((_currMaxLevel + 1) * sizeof(Node*)));
  Alloc::deallocate(_topIndex,(_currMaxLevel + 1) * sizeof(Node*),&_headerClassDesc);
 #else 
  Alloc::deallocate(_topIndex,(_currMaxLevel + 1) * sizeof(Node*));
 #endif
}; // void SmallHeaderSkipList<Alloc,ValType,KeyType,Base,InstanceId>::destroy()

//======================================================================
#undef DEBUG_NAMESPACE
//======================================================================
#endif
