//
// File:         ForwardMatchingIndex.h
// Description:  Interface for the code trees for matching.
// Created:      ???
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
// Note:         This is a part of the Gematogen library.
//====================================================
#ifndef FORWARD_MATCHING_INDEX_H
//==================================================================
#define FORWARD_MATCHING_INDEX_H
#ifndef NO_DEBUG_VIS
 #include <iostream.h>
#endif
#include "jargon.h"
#include "GematogenDebugFlags.h" 
#include "MultisetOfVariables.h"
#include "ForwardMatchingIndexTree.h"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_MATCHING_INDEX
 #define DEBUG_NAMESPACE "ForwardMatchingIndex<MaxTermSize,MaxNumOfVariables,MaxNumOfFunctors,Alloc,Symbol,Flatterm,LeafInfo,InstanceId>"
#endif 
#include "debugMacros.h"
//=================================================
template <ulong MaxTermSize,ulong MaxNumOfVariables,ulong MaxNumOfFunctors,class Alloc,class Symbol,class Flatterm,class LeafInfo,class InstanceId>
class ForwardMatchingIndex
 { 
  public: 
   typedef ForwardMatchingIndexTreeInstruction<MaxTermSize,MaxNumOfVariables,Alloc,Symbol,Flatterm,LeafInfo,InstanceId> Instruction;
   typedef ForwardMatchingIndexTreeLeaf<MaxTermSize,MaxNumOfVariables,Alloc,Symbol,Flatterm,LeafInfo,InstanceId> LeafInstr;
  public:
   ForwardMatchingIndex() : integrator(subst), removal(subst), retrieval(), _output()
    {
     _output.SetSubst(subst);
     for (ulong f = 0;f<MaxNumOfFunctors;f++) tree[f] = (Instruction*)0;     
    };
   ~ForwardMatchingIndex() { CALL("destructor ~ForwardMatchingIndex()"); };

   //LeafInstr* Integrate(TERM* t);    
   //LeafInstr* Integrate(Flatterm* t);     
   
   inline void integrationReset(const Symbol& fun)
   {
    CALL("integrationReset(const Symbol& fun)");
    ASSERT(fun.isComplex());
    integrator.Reset(&(Tree(fun)));
   };    
   inline void integrationVar(const Symbol& var) 
   {
    CALL("integrationVar(const Symbol& var)");
    ASSERT(var.isVariable());
    integrator.PushVar(var);
   }; 
   inline void integrationFunc(const Symbol& fun)
   {
    CALL("integrationFunc(const Symbol& fun)");
    ASSERT(fun.isComplex());
    integrator.PushFunc(fun);
   };
   LeafInstr* integrationGetLeaf(bool& newLeaf)  
   {
    CALL("integrationGetLeaf()");
    return integrator.Leaf(newLeaf);
   };   
  

   MultisetOfVariables<MaxNumOfVariables>& integrationGetVariables() { return integrator.Vars(); };
   const Flatterm** integrationMapVariable(ulong v) { return integrator.MapVar(v); };

   bool removalReset(const Symbol& fun)
   {
    CALL("removalReset(const Symbol& fun)");
    ASSERT(fun.isComplex());
    Instruction** tree_addr = &(Tree(fun));  
    if (!(*tree_addr)) return false;
    removal.Reset(tree_addr);
    return true;
   };
   
   bool removalVar(const Symbol& var) 
   {
    CALL("removalVar(const Symbol& var)");
    ASSERT(var.isVariable());
    return removal.SkipVar(var);
   };
   
   bool removalFunc(const Symbol& fun) 
   {
    CALL("removalFunc(const Symbol& fun)");
    ASSERT(fun.isComplex());
    return removal.SkipFunc(fun);
   };
 
   bool removalFinish(LeafInstr*& leaf,Instruction**& branch_to_destroy)
   {
    CALL("removalGetLeaf(LeafInstr*& leaf,Instruction**& branch_to_destroy)");
    if (!removal.SkipComps()) return false;
    leaf = removal.Leaf();
    ASSERT(leaf->isLeaf());   
    branch_to_destroy = removal.BranchToDestroy();
    return true;
   };

   /***************
   bool Remove(TERM* t,LeafInstr*& leaf,Instruction**& branch_to_destroy); 
   bool Remove(Flatterm* t,LeafInstr*& leaf,Instruction**& branch_to_destroy);
   ****************/
#
   LeafInstr* Match(const Flatterm* query);
   LeafInstr* MatchAgain() 
   { 
    CALL("MatchAgain()"); 
    return retrieval.MatchAgain(); 
   };

   void Destroy()       
   {
    for (ulong f = 0;f<MaxNumOfFunctors;f++)  
     if (tree[f])
      { 
       ForwardMatchingIndexTree<MaxTermSize,MaxNumOfVariables,Alloc,Symbol,Flatterm,LeafInfo,InstanceId>::Destroy(tree[f]);
       tree[f] = (Instruction*)0;     
      };
   };
   #ifndef NO_DEBUG
    bool clean() const
    {
     for (ulong f = 0;f<MaxNumOfFunctors;f++)  
      if (tree[f]) return false;
     return true; 
    };
    const ForwardMatchingIndexTreeIntegrator<MaxTermSize,MaxNumOfVariables,Alloc,Symbol,Flatterm,LeafInfo,InstanceId>& debugIntegrator() { return integrator; };
   #endif
   #ifndef NO_DEBUG_VIS
    ostream& output(ostream& str) const;
   #endif

  private:
   Instruction*& Tree(const Symbol& func) 
   {
    return tree[func.functor()]; 
   };

  private:
   Instruction* tree[MaxNumOfFunctors];
   const Flatterm* subst[MaxTermSize];
   ForwardMatchingIndexTreeIntegrator<MaxTermSize,MaxNumOfVariables,Alloc,Symbol,Flatterm,LeafInfo,InstanceId> integrator;
   ForwardMatchingIndexTreeRemoval<MaxTermSize,MaxNumOfVariables,Alloc,Symbol,Flatterm,LeafInfo,InstanceId> removal;
   ForwardMatchingIndexTreeRetrieval<MaxTermSize,MaxNumOfVariables,Alloc,Symbol,Flatterm,LeafInfo,InstanceId> retrieval;
   ForwardMatchingIndexTreeOutput<MaxTermSize,MaxNumOfVariables,Alloc,Symbol,Flatterm,LeafInfo,InstanceId> _output;

 }; // template <ulong MaxTermSize,ulong MaxNumOfVariables,ulong MaxNumOfFunctors,class Alloc,class Symbol,class Flatterm,class LeafInfo,class InstanceId> class ForwardMatchingIndex

template <ulong MaxTermSize,ulong MaxNumOfVariables,ulong MaxNumOfFunctors,class Alloc,class Symbol,class Flatterm,class LeafInfo,class InstanceId>
inline ForwardMatchingIndexTreeLeaf<MaxTermSize,MaxNumOfVariables,Alloc,Symbol,Flatterm,LeafInfo,InstanceId>* ForwardMatchingIndex<MaxTermSize,MaxNumOfVariables,MaxNumOfFunctors,Alloc,Symbol,Flatterm,LeafInfo,InstanceId>::Match(const Flatterm* query)
 {
  CALL("Match(const Flatterm* query)");
  ASSERT(query->symbolRef().isComplex());
  return retrieval.Match(query + 1,Tree(query->symbolRef()));    
 }; // LeafInstr* ForwardMatchingIndex<MaxTermSize,MaxNumOfVariables,MaxNumOfFunctors,Alloc,Symbol,Flatterm,LeafInfo,InstanceId>::Match(const Flatterm* query) 


#ifndef NO_DEBUG_VIS
 template <ulong MaxTermSize,ulong MaxNumOfVariables,ulong MaxNumOfFunctors,class Alloc,class Symbol,class Flatterm,class LeafInfo,class InstanceId>
 ostream& ForwardMatchingIndex<MaxTermSize,MaxNumOfVariables,MaxNumOfFunctors,Alloc,Symbol,Flatterm,LeafInfo,InstanceId>::output(ostream& str) const
 {
  for (ulong f = 0; f < MaxNumOfFunctors; f++)
   {
    const Instruction* t = tree[f];
    if (t)
     {
      Symbol functor(0UL,f,Symbol::PositivePolarity);
      str << "HEADER " << functor << '\n';
      _output.Tree(str,t) << "\nEND OF TREE FOR " << functor << "\n\n";
     };  
   };  
  return str;
 }; // ostream& ForwardMatchingIndex<MaxTermSize,MaxNumOfVariables,MaxNumOfFunctors,Alloc,Symbol,Flatterm,LeafInfo,InstanceId>::output(ostream& str) const
#endif

//================================================
#undef DEBUG_NAMESPACE  
//==================================================================
#endif





















