//
// File:         AutoFunctorPrecedence.h
// Description:  Automatic assignment of precedences to the functors and predicates of the signature.
// Created:      Apr 24, 2001, 16:50
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#ifndef AUTO_FUNCTOR_PRECEDENCE_H
//=================================================
#define AUTO_FUNCTOR_PRECEDENCE_H
#ifndef NO_DEBUG_VIS
 #include <iostream.h>
#endif
#include "jargon.h"
#include "HellDebugFlags.h"
#include "HellConst.h"
#include "Array.h"
#include "Malloc.h"
#include "GlobAlloc.h"
#include "Tuple.h"
#include "GSortTree.h"
//=================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_AUTO_FUNCTOR_PRECEDENCE
 #define DEBUG_NAMESPACE "AutoFunctorPrecedence"
#endif 
#include "debugMacros.h"
//=================================================  

class AutoFunctorPrecedence
{
 public:
  AutoFunctorPrecedence() 
    : _dependenceOnArity(0L),
      _somePrecedenceSet(false),
      _minimalPrecedence(0L),
      _maximalPrecedence(0L),
      _minFunctorNumber(0UL),
      _maxFunctorNumber(0UL),
      #ifdef DEBUG_NAMESPACE
       _precedence(0L,"AutoFunctorPrecedence::_precedence"), 
       _precedenceIsFixed(false,"AutoFunctorPrecedence::_precedenceIsFixed"),
       _arity(0UL,"AutoFunctorPrecedence::_arity"),
       _registeredFunctor(false,"AutoFunctorPrecedence::_registeredFunctor")
      #else
       _precedence(0L,0), 
       _precedenceIsFixed(false,0),
       _arity(0UL,0),
       _registeredFunctor(false,0)
      #endif
  {
   CALL("constructor AutoFunctorPrecedence()");   
  };
  
  ~AutoFunctorPrecedence()
  {
   CALL("destructor ~AutoFunctorPrecedence()");
  };
    
  void setDependenceOnArity(long typeOfDependence) { _dependenceOnArity = typeOfDependence; };

  bool precedenceIsFixed(ulong funNum) { return _precedenceIsFixed[funNum]; };  

  void assignFixed(ulong funNum,long precedence) 
  { 
   CALL("assignFixed(ulong funNum,long precedence)");
   _registeredFunctor[funNum] = true;
   long normPrec = normalisedPrecedence(precedence);  
   if (!_somePrecedenceSet) 
    { 
     _somePrecedenceSet = true;
     _minimalPrecedence = normPrec;
     _maximalPrecedence = normPrec;
    }
   else 
    if (normPrec > _maximalPrecedence) { _maximalPrecedence = normPrec; }
    else 
     if (normPrec < _minimalPrecedence) { _minimalPrecedence = normPrec; };
   
   _precedence[funNum] = normPrec;    
   _precedenceIsFixed[funNum] = true;   
  };

  void assignArity(ulong funNum,ulong arity) 
  { 
   CALL("assignArity(ulong funNum,ulong arity)");
   _registeredFunctor[funNum] = true; 
   _arity[funNum] = arity;
  };
  
  void assignAllPrecedences() 
  { 
   CALL("assignAllPrecedences()");
   if (_dependenceOnArity)
    { 
     ArityFuncSortTree sortTree;   

     for (ulong fn = 0; fn < _registeredFunctor.size(); fn++)
      if (!_precedenceIsFixed[fn]) sortTree.insert(ArityFuncPair(_arity[fn],fn));

     if (_dependenceOnArity > 0L)     
      {
       // greater arity => greater precedence 
       ArityFuncSortTree::RightLeftIter sortTreeIter(sortTree);      
       while (sortTreeIter)
	{ 
	 assignMinimalPrecedence(sortTreeIter.value().el2); 
         sortTreeIter.next();     
        };
      }
     else // _dependenceOnArity < 0L 
      {
       // greater arity => smaller precedence   
       ArityFuncSortTree::LeftRightIter sortTreeIter(sortTree);
       while (sortTreeIter)
	{ 
	 assignMinimalPrecedence(sortTreeIter.value().el2); 
         sortTreeIter.next();     
        };
      };
    }  
   else // !_dependenceOnArity
    {
     for (ulong fn = 0; fn < _registeredFunctor.size(); fn++)
      if (!_precedenceIsFixed[fn]) assignMinimalPrecedence(fn);
    };
  };

  bool symbolPrecedence(ulong funNum,long& precedence)
  { 
   if (!_registeredFunctor[funNum]) return false;
   precedence = _precedence[funNum];
   return true; 
  };
  
  bool headerPrecedence(bool negative,ulong predNum,long& precedence) 
  {
   CALL("headerPrecedence(bool negative,ulong predNum,long& precedence)"); 
   if (!_registeredFunctor[predNum]) return false;
   if (negative)
    {
     long prec;
     ALWAYS(symbolPrecedence(predNum,prec));
     precedence = prec + 1;
     return true;
    }
   else return symbolPrecedence(predNum,precedence);
  };
 private: 
  typedef Array<Malloc,bool,32,AutoFunctorPrecedence> BoolArray;
  typedef Array<Malloc,long,32,AutoFunctorPrecedence> LongArray;
  typedef Array<Malloc,ulong,32,AutoFunctorPrecedence> ULongArray;
  typedef Tuple2<ulong,ulong> ArityFuncPair; 
  typedef GSortTree<GlobAlloc,ArityFuncPair,AutoFunctorPrecedence> ArityFuncSortTree;  
 private: 
  long normalisedPrecedence(long inputPrecedence) { return inputPrecedence*2; };
  long previousNormalisedPrecedence(long normalisedPrecedence) { return normalisedPrecedence - 2; };
  void assignMinimalPrecedence(ulong funNum) 
  { 
   CALL("assignMinimalPrecedence(ulong funNum)");
   _minimalPrecedence = previousNormalisedPrecedence(_minimalPrecedence);
   _precedence[funNum] = _minimalPrecedence;
  };
 private: 
  long _dependenceOnArity;
  long _somePrecedenceSet;
  long _minimalPrecedence;
  long _maximalPrecedence;
  ulong _minFunctorNumber;
  ulong _maxFunctorNumber;
  LongArray _precedence; 
  BoolArray _precedenceIsFixed;
  ULongArray _arity;
  BoolArray _registeredFunctor;
}; // class AutoFunctorPrecedence




//================================================
#undef DEBUG_NAMESPACE   
//==================================================
#endif
