//
// File:         BackDemodulation.cpp
// Description:  Back simplification by ordered unit equalities.
// Created:      Apr 16, 2000, 16:20
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#include "BackDemodulation.h"
//====================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BACK_DEMODULATION
 #define DEBUG_NAMESPACE "BackDemodulation"
#endif
#include "debugMacros.h"        
//====================================================

void BackDemodulation::integrate(Clause* cl)
{
 CALL("integrate(Clause* cl)");   
 ASSERT(_clauseNumIndex->find(cl->Number()) == cl);
 ulong clauseNum = cl->Number(); 
 TERM* lit;
 TERM::Iterator iter;
 TERM* term;
 ulong ln = 0;
 for(LiteralList::Iterator litIter(cl->LitList());litIter.notEnd();litIter.next())
  {
   lit = litIter.currentLiteral();
   if (lit->arity())
    {
     iter.Reset(lit);
     iter.Next();
     do
      {
       term = iter.CurrPos();
       if (term->isComplex()) term->RegisterOccurence(clauseNum,ln);
      }
     while (iter.Next());
    };         
   ln++;
  };
}; // void BackDemodulation::integrate(Clause* cl)
 
void BackDemodulation::remove(Clause* cl) 
{
 CALL("remove(Clause* cl)");
 ASSERT(_clauseNumIndex->find(cl->Number()) == cl);
 ulong clauseNum = cl->Number(); 
 TERM* lit;
 TERM::Iterator iter;
 TERM* term;

 for(LiteralList::Iterator litIter(cl->LitList());litIter.notEnd();litIter.next())
  {
   lit = litIter.currentLiteral();
   if (lit->arity())
    {
     iter.Reset(lit);
     iter.Next();
     do
      {
       term = iter.CurrPos();
       if (term->isComplex()) term->RemoveOccurences(clauseNum);
      }
     while (iter.Next());
    };   
  };
}; // void BackDemodulation::remove(Clause* cl)  


bool BackDemodulation::query(const Clause* cl) 
{ 
 CALL("query(const Clause* cl)");  
  
 if (!cl->Active()) return false;
 if (cl->Unit() || (_bDOnSplittingBranchesFlag && (cl->LitList()->numOfSelectedLits() == 1)))
  {   
   const TERM* lit = cl->LitList()->firstLit(); // the selected one
   if (lit->Positive() && lit->IsEquality())
    {
     if (_bDOnSplittingBranchesFlag)
      {
       _splittingBranches.reset();
       LiteralList::Iterator iter(cl->LitList());
       iter.next();
       while (iter.notEnd())
        {
	 if (iter.currentLiteral()->IsSplitting()) 
	  {
	   ASSERT(!iter.currentLiteral()->arity());
	   _splittingBranches.add(iter.currentLiteral()->HeaderNum()); 
          }
	 else return false;
         iter.next();
        };
      }
     else 
      if (!cl->Unit()) return false;

     _query = cl;
     _queryNum = cl->Number();
     _candidates = (const TERM::OccList::Node*)0;
     if (lit->IsOrderedEq())
      {
       _queryLHS = lit->Args()->First();
       ASSERT(_queryLHS->isComplex());
       //RARE(!(_queryLHS->arity()));
       _matching.SetQuery(_queryLHS);
       _queryRHS = *(lit->Args()->Next());
       _swapQuerySides = false;   
       _matching.Constraint().makeAlwaysValid(true);   
       return true;
      }
     else // not preordered
      {
       if (_degree <= 0) return false;
       _swapQuerySides = true;
       if (query(lit->Arg1(),lit->Arg2())) 
        {
         _matching.Constraint().makeAlwaysValid(false);
         return true;
        };
       _swapQuerySides = false;
       _matching.Constraint().makeAlwaysValid(false);
       return query(lit->Arg2(),lit->Arg1());
      };
    };
  };
 return false; 
}; // bool BackDemodulation::query(const Clause* cl) 

bool BackDemodulation::nextSimplified() 
{ 
 CALL("nextSimplified()");
 if (!_query->Active()) return false;  
 
 COP("X10");
   
 check_candidates:
  if (_candidates) 
   { 
    ulong clauseNum = _candidates->key();
    if (clauseNum == _queryNum) 
     { 
      _candidates = _candidates->next();
      goto check_candidates; 
     }; 

    // DF; cout << "CLNUM " << clauseNum << "\n";
 
   _recentlySimplified = (Clause*)_clauseNumIndex->find(clauseNum);

   if (!_recentlySimplified)
    {
     _candidates = _candidates->next();
     goto check_candidates; 
    };

    ASSERT(_recentlySimplified);
    if (!_recentlySimplified->Active())
     { 
      _candidates = _candidates->next();
      goto check_candidates; 
     }; 

    if (_bDOnSplittingBranchesFlag)
     {
      ulong numOfBranchesToCover = _splittingBranches.numOfDiffElements();
      if (!numOfBranchesToCover) return true;
      if (_recentlySimplified->Length() <= numOfBranchesToCover)
       {
        _candidates = _candidates->next();
        goto check_candidates; 
       };     
      LiteralList::Iterator iter(_recentlySimplified->LitList());
      do
       {
        if ((iter.currentLiteral()->IsSplitting()) 
            && (_splittingBranches.scoreOf(iter.currentLiteral()->HeaderNum())))
	 {
	  ASSERT(!iter.currentLiteral()->arity());          
          ASSERT(numOfBranchesToCover);
          numOfBranchesToCover--;
          if (!numOfBranchesToCover) { _numOfBDOnSplittingBranches++; return true; };
         };          
	iter.next();
       }
      while (iter.notEnd());
      ASSERT(numOfBranchesToCover);
      _candidates = _candidates->next();
      goto check_candidates; 
     };

    return true;
   };
  try_next_match:
   if (!_matching.NextMatch()) 
    { 
     if (_swapQuerySides)
      {
       _swapQuerySides = false;
       const TERM* lit = _query->LitList()->firstLit();
       if (!query(lit->Arg2(),lit->Arg1())) return false;
       _matching.Constraint().makeAlwaysValid(false);
       goto try_next_match;  
      }
     else return false;
    };

   _redex = _matching.Match();

   const TERM::OccList* occ = _redex->Occurences();   

   if (occ)
    {
     _candidates = occ->first();
     goto check_candidates;
    }
   else goto try_next_match; 
}; // nextSimplified() 
 
bool BackDemodulation::collectInference() 
{
 CALL("collectInference()");
 ASSERT(_candidates);
 if (rewrite(_candidates)) { _candidates = _candidates->next(); return true; };
 _candidates = _candidates->next();   
 return false; 
}; // bool collectInference() 



//====================================================
