//
// File:         BackDemodulation.h
// Description:  Back simplification by ordered unit equalities.
// Created:      Apr 16, 2000, 16:20
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#ifndef BACK_DEMODULATION_H
//====================================================
#define BACK_DEMODULATION_H
#include "jargon.h"
#include "HellDebugFlags.h"
#include "ClauseNumIndex.h"
#include "term.h"
#include "clause.h"
#include "NewClause.h"    
#include "sharing.h"
#include "MultisetOfVariables.h"
#include "b_matching.h"
#include "LiteralList.h"
#include "RuntimeError.h"
//====================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BACK_DEMODULATION
 #define DEBUG_NAMESPACE "BackDemodulation"
#endif
#include "debugMacros.h"
//====================================================

class BackDemodulation
{
 public:
  typedef Multiset<ulong,HellConst::MaxNumOfDifferentHeaders> SetOfSplittingBranches;
 public:
  BackDemodulation(NewClause* freshClause,SHARING_INDEX* sharingIndex,ClauseNumIndex* clauseNumIndex) 
   : _bDOnSplittingBranchesFlag(false),
     _resolvent(freshClause),
     _matching(sharingIndex,freshClause),
     _clauseNumIndex(clauseNumIndex),
     _termIter(),
     _queryLHSVars(),
     _queryRHSVars(),
     _rewrittenLitIter(),
     _RHSIter(),
     _splittingBranches(),
     _numOfBDOnSplittingBranches(0UL)
  {
   CALL("constructor BackDemodulation(NewClause* freshClause,SHARING_INDEX* sharingIndex,ClauseNumIndex* clauseNumIndex)");
  };
  ~BackDemodulation() 
  { 
   CALL("destructor ~BackDemodulation()");  
   if (!Exit::fast) destroy(); 
  };
  void destroy() {}; // to be implemented
  void setDegree(long d) { _degree = d; };
  void setBDOnSplittingBranchesFlag(bool fl) 
  {
   if (_bDOnSplittingBranchesFlag && (!fl)) RuntimeError::report("Switching _bDOnSplittingBranchesFlag off.");
   _bDOnSplittingBranchesFlag = fl; 
  };
  bool query(const Clause* cl);
  bool nextSimplified();
  bool collectInference();
  
  Clause* simplified() { return _recentlySimplified; }; 
  void integrate(Clause* cl);
  void remove(Clause* cl);
  ulong statBDOnSplittingBranches() const { return _numOfBDOnSplittingBranches; };
  void resetStat() { _numOfBDOnSplittingBranches = 0UL; };
 private:
  BackDemodulation() : _matching((SHARING_INDEX*)0,(NewClause*)0) {};

  bool query(const TERM* lhsref,const TERM* rhsref) 
  {
   CALL("query(const TERM* lhsref,const TERM* rhsref)");
   if (lhsref->isVariable() || rhsref->isVariable()) return false;
   ASSERT(lhsref->IsReference());
   ASSERT(rhsref->IsReference());
   _queryLHS = lhsref->First();
   ASSERT(_queryLHS->isComplex()); 
   _queryLHSVars.Reset();
   _queryLHSFuncs = 0;
   collectVarsAndFuncs(_queryLHS,_queryLHSVars,_queryLHSFuncs);

   _queryRHS = *rhsref;
   _queryRHSVars.Reset();
   _queryRHSFuncs = 0;
   _termIter.Reset(_queryRHS.First());
   ulong rhsvar;
   do
    {
     if (_termIter.CurrentSymIsVar())
      {
       rhsvar = _termIter.symbolRef().var();
       if (!_queryLHSVars.Belongs(rhsvar)) return false;
       _queryRHSVars.Register(rhsvar);
      }
     else _queryRHSFuncs++;
    }
   while (_termIter.Next());

   //cout << "CONSTR FOR " << _query << "\n"; DF;      

   _matching.Constraint().assumeGreater(_queryLHS,_queryLHSVars,_queryLHSFuncs,
                                        _queryRHS.First(),_queryRHSVars,_queryRHSFuncs);

   //DF; cout << "BDQ " << _query << "\n";
   //DF; _matching.Constraint().output(cout) << "\n";    
   
   _matching.Constraint().compile();

   //DF; _matching.Constraint().outputCode(cout) << "\n\n\n"; 

   _matching.SetQuery(_queryLHS);
   return appropriateDegree(_matching.Constraint().category());
   //return false; DF;
  };
   
  bool appropriateDegree(const KBConstraint::Category& cat) const 
  {
   switch (cat) 
    {
     case KBConstraint::NearlyTautology: return (_degree >= 2);
     case KBConstraint::Volatile: return (_degree >= 3);
     case KBConstraint::NearlyContradiction: return (_degree >= 4);
     case KBConstraint::PurelyLexicographic: return (_degree >= 4);
    };
   return false;
  };

  void collectVarsAndFuncs(const TERM* t,MultisetOfVariables<HellConst::MaxNumOfVariables>& vms,ulong& funcs)
  {
   CALL("collectVarsAndFuncs(const TERM* t,MultisetOfVariables<HellConst::MaxNumOfVariables>& vms,ulong& funcs)");  
   ASSERT(t->isComplex());
   _termIter.Reset(t); 
   do 
    {
     if (_termIter.CurrentSymIsVar())
      {
       vms.Register(_termIter.symbolRef().var());
      } 
     else funcs++;
    }
   while (_termIter.Next());          
  }; // void collectVarsAndFuncs(const TERM* t,MultisetOfVariables<HellConst::MaxNumOfVariables>& vms,ulong& funcs)


  bool rewrite(const TERM::OccList::Node* occ)
  {
   CALL("rewrite(const TERM::OccList::Node* occ)");

   _resolvent->Reset();
   TERM::SubtermMask litMask = occ->value();
   ASSERT((bool)litMask);
   
   ulong litNum = 0;
   for (LiteralList::Iterator iter(_recentlySimplified->LitList());iter.notEnd();iter.next())
   {
    if (litMask.bit(litNum)) // this literal contains rewritten redex       
     {
      if (!rewriteLit(iter.currentLiteral(),iter.currentElement()->isInherentlySelected())) return false;
     }
    else // literal remains unchanhed 
     if (!_resolvent->PushLiteral(iter.currentLiteral(),iter.currentElement()->isInherentlySelected())) return false;      
    litNum++;
   };

   _resolvent->PushAncestor((Clause*)_query);
   _resolvent->PushAncestor((Clause*)_recentlySimplified);                
   _resolvent->UsedRule(ClauseBackground::BackDemod);  
   _resolvent->MakeVIP();                               
   return true;  
  }; // bool rewrite(TERM::OccList::Node* occ)

  bool rewriteLit(TERM* lit,bool inhSel)             
  {
   CALL("rewriteLit(TERM* lit,bool inhSel)");
   ASSERT(_redex->isComplex());
   TERM header = *lit;
   if (header.IsEquality() && header.IsOrderedEq()) header.MakeUnordered(); 
 
   if (!_resolvent->OpenNonpropLiteral(header,inhSel)) return false; 
   _rewrittenLitIter.Reset(lit);
   ALWAYS(_rewrittenLitIter.Next());
   check_position:    
    if (_rewrittenLitIter.CurrPos() == _redex)
     {
      if (_queryRHS.isVariable()) 
       {
	if (!_matching.CollectSubst(_queryRHS.var())) return false;
       }
      else
       {
	ASSERT(_queryRHS.IsReference());
        _RHSIter.Reset(_queryRHS.First());
        if (!_resolvent->PushFunc(_RHSIter.symbolRef())) return false;
        while (_RHSIter.Next())
	 {
	  if (_RHSIter.CurrentSymIsVar())
	   {
	    if (!_matching.CollectSubst(_RHSIter.symbolRef().var())) return false;
           }
          else
	   if (!_resolvent->PushFunc(_RHSIter.symbolRef())) return false;
	 }; 
       };
      if (_rewrittenLitIter.SkipTerm()) goto check_position;
     }
    else
     {
      if (!_resolvent->PushSymbol(_rewrittenLitIter.symbolRef())) return false;
      if (_rewrittenLitIter.Next()) goto check_position;
     };
   return _resolvent->CloseLiteral();
  }; // bool rewriteLit(TERM* lit,bool inhSel)  

 private:
  bool _swapQuerySides;
  long _degree;
  bool _bDOnSplittingBranchesFlag;
  NewClause* _resolvent;
  B_MATCHING _matching;
  ClauseNumIndex* _clauseNumIndex;
  const Clause* _query;
  ulong _queryNum;
  const TERM* _queryLHS;
  TERM _queryRHS;
  TERM::Iterator _termIter;
  MultisetOfVariables<HellConst::MaxNumOfVariables> _queryLHSVars;
  MultisetOfVariables<HellConst::MaxNumOfVariables> _queryRHSVars;
  ulong _queryLHSFuncs;
  ulong _queryRHSFuncs;
  const TERM::OccList::Node* _candidates; 
  Clause* _recentlySimplified;
  TERM* _redex;
  TERM::Iterator _rewrittenLitIter;
  TERM::Iterator _RHSIter;
  SetOfSplittingBranches _splittingBranches;
  ulong _numOfBDOnSplittingBranches;
}; // class BackDemodulation


//====================================================
#undef DEBUG_NAMESPACE
//====================================================
#endif
