//
// File:         BackwardSubsumption.h
// Description:  Interface for several BS related modules.
// Created:      Feb 26, 2000, 20:20
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//============================================================================
#ifndef BACKWARD_SUBSUMPTION_H
//=============================================================================
#define BACKWARD_SUBSUMPTION_H 
#include "jargon.h"
#include "HellDebugFlags.h"
#include "HellInterceptMacros.h"
#include "HellConst.h"
#include "BSIndex.h"
#include "BSMachine.h"
#include "ClauseNumIndex.h"
#include "clause.h"
#include "term.h"
#include "SubsumptionCheck.h" 

//===========================================================================

// This is for experimental comparison with the old discrimination tree based implementation

#include "OldBackwardSubsumption.h" 

//#define OLD_BACKWARD_SUBSUMPTION_IMPLEMENTATION
#undef OLD_BACKWARD_SUBSUMPTION_IMPLEMENTATION

#ifdef OLD_BACKWARD_SUBSUMPTION_IMPLEMENTATION
 #define OLD_IMPL(act) act
#else
 #define OLD_IMPL(act)
#endif

//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BACKWARD_SUBSUMPTION
 #define DEBUG_NAMESPACE "BackwardSubsumption"
#endif
#include "debugMacros.h"
//============================================================================

class BackwardSubsumption
{
 public:
  typedef BSIndex::LTPair LTPair;  
  typedef BSIndex::LAPair LAPair;
  typedef BSIndex::LATTriple LATTriple;
  typedef BSIndex::LList LList;
  typedef BSIndex::LTList LTList;
  typedef BSIndex::LAList LAList;
  typedef BSIndex::LATList LATList;
  typedef BSIndex::CSkipList CSkipList;
  typedef BSIndex::CLSkipList CLSkipList;
  typedef BSIndex::CLTSkipList CLTSkipList;
  typedef BSIndex::CLASkipList CLASkipList;
  typedef BSIndex::CLATSkipList CLATSkipList;
 public:
  BackwardSubsumption(ClauseNumIndex* clauseNumIndex) 
   : _index(),
     _integrator(&_index),
     _removal(&_index),
     _retrieval(&_index),
     _machine(),        
     _termTraversal(DOP("BackwardSubsumption::_termTraversal")),
     _clauseNumIndex(clauseNumIndex),
     _setMode(false),
     _optimisedMultisetMode(false)
     #ifdef DEBUG_NAMESPACE
      , _debugSubsumptionCheck()
     #endif 
     #ifdef OLD_BACKWARD_SUBSUMPTION_IMPLEMENTATION
      , _oldImplementation(clauseNumIndex)
     #endif
  {
   CALL("constructor BackwardSubsumption(ClauseNumIndex* clauseNumIndex)");

   //DF; cout << "&_termTraversal = " << (long)(&_termTraversal) << "\n";
   //DF; if (!_termTraversal.checkIntegrity()) cout << "BRRRRRRRRR\n";
   //DF; cout << _termTraversal.wrapper[0].Content() << "\n";
   //DF; cout <<  TERM::dummyPositiveHeader1.Content() << "\n"; 
  };
  ~BackwardSubsumption() 
  {
   CALL("destructor ~BackwardSubsumption()");      
  };
  bool subsumptionAllowedInSetMode(const Clause* subsumingCl,const Clause* subsumedCl)
  {
   return subsumingCl->NumOfPosEq() <= subsumedCl->NumOfPosEq(); 
  };
  bool subsumptionAllowedInMultisetMode(const Clause* subsumingCl,const Clause* subsumedCl)
  {
   return true; 
  };
  void destroyIndex() { CALL("destroyIndex()"); _index.destroy(); };
  #ifndef NO_DEBUG
   bool clean() const { return _index.clean(); };
  #endif        
  void integrate(const Clause* cl);
  void remove(const Clause* cl);
  bool query(const Clause* cl);
  bool subsumeNext() 
  {
   CALL("subsumeNext()");
   bool res;
   if (_setMode) 
    {
     res = subsumeNextSetMode();
    }
   else 
    if (_optimisedMultisetMode && _queryClause->allMaximalSelected())
     {
      res = subsumeNextSetMode();
     }
    else 
     res = subsumeNextMultisetMode();
  
   #ifdef INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING
    if (!res) { INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING_RETRIEVAL_TOTAL; };
   #endif 

   return res;
  };
  bool subsumeNextSetMode();
  bool subsumeNextMultisetMode();
  const Clause* subsumed() const { return _recentlySubsumedCl; };
  void setSetMode(bool fl) { _setMode = fl; }; 
  void setOptimisedMultisetMode(bool fl) { _optimisedMultisetMode = fl; }; 
 private:
  bool preproQuery(const Clause* cl);
  bool preproQueryLit(const TERM* lit);
  bool preproQuerySymLit(const TERM* lit);
  bool preproQueryComplexTerm(const TERM* t);
  bool preproQueryComplexTermArguments(const TERM* t);
  bool preproQuerySymLitComplexTerm(const TERM* t);
  bool preproQuerySymLitComplexTermArguments(const TERM* t);
 private:
  BSIndex _index;
  BSIndex::Integrator _integrator;
  BSIndex::Removal _removal;
  BSIndex::Retrieval _retrieval;
  BSMachine _machine; 
 public: 
  TERM::Traversal _termTraversal;
 private: 
  ClauseNumIndex* _clauseNumIndex;
  const Clause* _recentlySubsumedCl;
  const Clause* _queryClause; 
  bool _setMode;
  bool _optimisedMultisetMode;
  #ifdef DEBUG_NAMESPACE
   SubsumptionCheck _debugSubsumptionCheck;
  #endif
  #ifdef OLD_BACKWARD_SUBSUMPTION_IMPLEMENTATION
   OldBackwardSubsumption _oldImplementation;
  #endif
}; // class BackwardSubsumption

inline bool BackwardSubsumption::preproQuery(const Clause* cl)
{
 CALL("preproQuery(const Clause* cl)");
 _machine.resetQuery(cl); 

 for (LiteralList::Iterator iter(cl->LitList()); iter.notEnd(); iter.next())
  if (!preproQueryLit(iter.currentLiteral())) return false;

 return _machine.endOfQuery();
}; // bool BackwardSubsumption::preproQuery(const Clause* cl)


inline bool BackwardSubsumption::preproQueryLit(const TERM* lit)
{
 CALL("preproQueryLit(const TERM* lit)");
 if (lit->arity())
  { 
   if (lit->IsSymLitHeader()) 
    { 
     return (preproQuerySymLit(lit) && _machine.queryEndOfSymLit()); 
    }
   else // ordinary literal
    {
     if (!_retrieval.litHeader(*lit)) return false;
     _machine.queryNextLit();     
     return  (preproQueryComplexTermArguments(lit) && _machine.queryEndOfLit());  
    };
  }
 else // propositional
  {
   const CSkipList* plcands;
   return ((_retrieval.propLit(*lit,plcands)) && (_machine.queryPropLit(plcands)));
  };
}; // bool BackwardSubsumption::preproQueryLit(const TERM* lit) 


inline bool BackwardSubsumption::preproQueryComplexTermArguments(const TERM* t) 
{
 CALL("preproQueryComplexTermArguments(const TERM* t)");
 ASSERT(t->isComplex());
 ASSERT(t->arity());
 const CLTSkipList* vcands;
 bool allVars = true;
 const TERM* argRef;
 for (argRef = t->Args(); ((!argRef->IsNull()) && allVars); argRef = argRef->Next())
  allVars = argRef->isVariable();
 
 // first argument
 argRef = t->Args();
 if (argRef->isVariable())
  {
   // primary only if all arguments are variables  
   if ((!_retrieval.variable(vcands)) || (!_machine.queryVariable(*argRef,vcands,allVars)))
    return false;  
  }
 else 
  if (!preproQueryComplexTerm(argRef->First())) return false;

 // other arguments
 argRef = argRef->Next();
 while (!argRef->IsNull())
  {
   if (argRef->isVariable())
    {
     // always secondary  
     if ((!_retrieval.variable(vcands)) || (!_machine.queryVariable(*argRef,vcands,false)))
      return false;  
    }
   else 
    if (!preproQueryComplexTerm(argRef->First())) return false; 
   argRef = argRef->Next();    
  };   
 return true;
}; // bool BackwardSubsumption::preproQueryComplexTermArguments(const TERM* t)  


inline bool BackwardSubsumption::preproQueryComplexTerm(const TERM* t)
{
 CALL("preproQueryComplexTerm(const TERM* t)");
 ASSERT(t->isComplex());
 const CLSkipList* ccands;
 if (t->arity())
  {
   return _retrieval.nonconstFunc(*t) && preproQueryComplexTermArguments(t);
  }
 else // constant
  return _retrieval.constant(*t,ccands) && _machine.queryConstant(*t,ccands);
}; // bool BackwardSubsumption::preproQueryComplexTerm(const TERM* t) 

inline bool BackwardSubsumption::preproQuerySymLit(const TERM* lit)
{
 CALL("preproQuerySymLit(const TERM* lit)");
 ASSERT(lit->IsSymLitHeader());
 if (!_retrieval.symLitHeader(*lit)) return false;
 _machine.queryNextSymLit();
 const TERM* argRef1 = lit->Args();
 const TERM* argRef2 = argRef1->Next(); 
 const CLATSkipList* slvcands;

 // first argument
 if (argRef1->isVariable()) 
  {   
    // primary only if the second argument is also a variable
   if ((!_retrieval.symLitVariable(slvcands)) 
       || (!_machine.querySymLitVariable(*argRef1,slvcands,argRef2->isVariable())))
    return false;      
  }
 else 
  if (!preproQuerySymLitComplexTerm(argRef1->First())) return false;

 // second argument
 if (!_retrieval.symLitSecondArg()) return false;
 _machine.querySymLitSecondArg();

 if (argRef2->isVariable()) 
  {   
   return (_retrieval.symLitVariable(slvcands)) 
           && (_machine.querySymLitVariable(*argRef2,slvcands,false)); // always secondary position
  }
 else return preproQuerySymLitComplexTerm(argRef2->First()); 
}; // bool BackwardSubsumption::preproQuerySymLit(const TERM* lit)

inline bool BackwardSubsumption::preproQuerySymLitComplexTerm(const TERM* t)
{
 CALL("preproQuerySymLitComplexTerm(const TERM* t)");
 ASSERT(t->isComplex());
 const CLASkipList* slccands;

 if (t->arity())
  {
   return _retrieval.symLitNonconstFunc(*t) && preproQuerySymLitComplexTermArguments(t);
  }
 else // constant
  return _retrieval.symLitConstant(*t,slccands) && _machine.querySymLitConstant(*t,slccands);
}; // bool BackwardSubsumption::preproQuerySymLitComplexTerm(const TERM* t) 


inline bool BackwardSubsumption::preproQuerySymLitComplexTermArguments(const TERM* t) 
{
 CALL("preproQuerySymLitComplexTermArguments(const TERM* t)");
 ASSERT(t->isComplex());
 ASSERT(t->arity());
 const CLATSkipList* slvcands;

 bool allVars = true;
 const TERM* argRef;
 for (argRef = t->Args(); ((!argRef->IsNull()) && allVars); argRef = argRef->Next())
  allVars = argRef->isVariable();
 
 // first argument
 argRef = t->Args();          
 if (argRef->isVariable())
  {
   // primary only if all arguments are variables  
   if ((!_retrieval.symLitVariable(slvcands)) || (!_machine.querySymLitVariable(*argRef,slvcands,allVars)))
    return false;  
  }
 else 
  if (!preproQuerySymLitComplexTerm(argRef->First())) return false;

 // other arguments
 argRef = argRef->Next();
 while (!argRef->IsNull())
  {
   if (argRef->isVariable())
    {
     // always secondary  
     if ((!_retrieval.symLitVariable(slvcands)) || (!_machine.querySymLitVariable(*argRef,slvcands,false)))
      return false;  
    }
   else 
    if (!preproQuerySymLitComplexTerm(argRef->First())) return false; 
   argRef = argRef->Next();    
  };   
 return true;
}; // bool BackwardSubsumption::preproQuerySymLitComplexTermArguments(const TERM* t)  

 
//======================================================================
#undef DEBUG_NAMESPACE
//======================================================================
#endif
