//
// File:         KBCommand.h
// Description:  Commands of an abstract machine for checking KB constraints.
// Created:      Apr 23, 2000, 19:00
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#ifndef KB_COMMAND_H
//====================================================
#define KB_COMMAND_H
#ifndef NO_DEBUG_VIS
 #include <iostream.h>
#endif
#include "jargon.h"
#include "HellDebugFlags.h"
#include "HellConst.h"
#include "term.h"   
#include "variables.h"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_KB_COMMAND
 #define DEBUG_NAMESPACE "KBCommand"
#endif 
#include "debugMacros.h"
//================================================= 

class KBCommand
{
 public:
  enum Tag 
   {
    Success,
    Failure,
    SuccIfWeightGreater,  // must fail if less
    FailIfWeightGreater, // does not mean success if less!
    FailIfNonground, 
    SuccIfWeightLess, // succes if less, continue if equal, fail otherwise
    InitWR,
    AddWR,
    SuccIfWRGreater,
    FailIfWRLess,
    FailIfWRGreater,
    SuccIfWRLess, // succes if less, continue if equal, fail otherwise 
    InitPosSubst,
    InitNegSubst,
    PosSubst,
    NegSubst,
    CompareSubst,  // success if greater, continue if can be equal, fail otherwise
    LexCompareComplexComplex // success if term1 is greater than term2, 
                             // continue if equal and fail otherwise
   };
  
 public:
  KBCommand() {};
  const Tag& tag() const { return _tag; };
  void mkSuccess()
  {
   _tag = Success;
  };

  void mkFailure()
  {
   _tag = Failure;
  };

  void mkSuccIfWeightGreater(const ulong& var,const ulong& coeff,const ulong& weight) 
  {
   _tag = SuccIfWeightGreater;
   _var = var;
   _coeff = coeff;
   _weightConst = weight; 
  };

  void mkFailIfWeightGreater(const ulong& var,const ulong& coeff,const ulong& weight)
  {
   _tag = FailIfWeightGreater;
   _var = var;
   _coeff = coeff;
   _weightConst = weight; 
  };

  void mkFailIfNonground(const ulong& var)
  {
   _tag = FailIfNonground;
   _var = var;
  };

  void mkSuccIfWeightLess(const ulong& var,const ulong& coeff,const ulong& weight)
  {
   _tag = SuccIfWeightLess;
   _var = var;
   _coeff = coeff;
   _weightConst = weight; 
  };

  void mkInitWR()
  {
   _tag = InitWR;
  };

  void mkAddWR(const ulong& var,const ulong& coeff)
  {
   _tag = AddWR;
   _var = var;
   _coeff = coeff;
  };

  void mkSuccIfWRGreater(const ulong& weight)
  {
   _tag = SuccIfWRGreater;
   _weightConst = weight; 
  };

  void mkFailIfWRLess(const ulong& weight)
  {
   _tag = FailIfWRLess;
   _weightConst = weight; 
  };

  void mkFailIfWRGreater(const ulong& weight)
  {
   _tag = FailIfWRGreater;
   _weightConst = weight; 
  };

  void mkSuccIfWRLess(const ulong& weight)
  {
   _tag = SuccIfWRLess;
   _weightConst = weight; 
  };

  void mkInitPosSubst(const ulong& coeff)
  {
   _tag = InitPosSubst;
   _coeff = coeff;
  };

  void mkInitNegSubst(const ulong& coeff)
  {
   _tag = InitNegSubst; 
   _coeff = coeff;
  };

  void mkPosSubst(const ulong& var,const ulong& coeff)
  {
   _tag = PosSubst;
   _var = var;
   _coeff = coeff;
  };

  void mkNegSubst(const ulong& var,const ulong& coeff)
  {
   _tag = NegSubst;
   _var = var;
   _coeff = coeff;
  }; 

  void mkCompareSubst()
  {
   _tag = CompareSubst;
  };

  void mkLexCompareComplexComplex(const TERM* t1,const TERM* t2)
  {
   _tag = LexCompareComplexComplex;
   _term1 = t1;
   _term2 = t2;
  };

  const ulong& var() const { return _var; };
  const ulong& coeff() const { return _coeff; };
  const ulong& weightConst() const { return _weightConst; };
  const TERM* const & term1() const 
  {
   CALL("term1() const");
   ASSERT(_tag == LexCompareComplexComplex);
   return _term1; 
  }; 
  const TERM* const & term2() const 
  {
   CALL("term2() const");
   ASSERT(_tag == LexCompareComplexComplex);
   return _term2; 
  };  

  #ifndef NO_DEBUG_VIS
   ostream& output(ostream& str) const;
   static ostream& outputCode(ostream& str,const KBCommand* begin,const KBCommand* end);
  #endif
    
 private: 
  Tag _tag;
  union
  {
   ulong _var;
   const TERM* _term1;
  };
  union
  {
   ulong _coeff;
   const TERM* _term2;
  };
  ulong _weightConst;
}; // class KBCommand

#ifndef NO_DEBUG_VIS
 inline ostream& KBCommand::output(ostream& str) const
 {
  CALL("output(ostream& str) const");
  switch (_tag)
   {
    case Success: return str << "Success";
    case Failure: return str << "Failure";
    case SuccIfWeightGreater: 
     return str << "SuccIfWeightGreater " << "[var " << var() << ",coeff " << coeff() << ",wc " << weightConst() << ']'; 
    case FailIfWeightGreater: 
     return str << "FailIfWeightGreater " << "[var " << var() << ",coeff " << coeff() << ",wc " << weightConst() << ']';
    case FailIfNonground: return str << "FailIfNonground " <<  "[var " << var() << ']';
    case SuccIfWeightLess: 
     return str << "SuccIfWeightLess " << "[var " << var() << ",coeff " << coeff() << ",wc " << weightConst() << ']';
    case InitWR: return str << "InitWR";
    case AddWR: return str << "AddWR " << "[var " << var() << ",coeff " << coeff() << ']';
    case SuccIfWRGreater: return str << "SuccIfWRGreater " << "[wc " << weightConst() << ']';
    case FailIfWRLess: return str << "FailIfWRLess " << "[wc " << weightConst() << ']';
    case FailIfWRGreater: return str << "FailIfWRGreater " << "[wc " << weightConst() << ']';
    case SuccIfWRLess: return str << "SuccIfWRLess " << "[wc " << weightConst() << ']';
    case InitPosSubst: return str << "InitPosSubst";
    case InitNegSubst: return str << "InitNegSubst";
    case PosSubst: return str << "PosSubst " << "[var " << var() << ",coeff " << coeff() << ']';
    case NegSubst: return str << "NegSubst " << "[var " << var() << ",coeff " << coeff() << ']';
    case CompareSubst: return str << "CompareSubst ";
    case LexCompareComplexComplex: return str << "LexCompareComplexComplex " << "[term1 " << term1() << ",term 2 " << term2() << ']'; 
    default: ICP("ICP0"); return str << "Unknown";
   };
 }; // ostream& KBCommand::output(ostream& str) const

 inline ostream& operator<<(ostream& str,const KBCommand& com) 
 {
  return com.output(str);
 };
 inline ostream& operator<<(ostream& str,const KBCommand* com) 
 {
  return com->output(str);
 };
 inline ostream& KBCommand::outputCode(ostream& str,const KBCommand* begin,const KBCommand* end)
 {
  CALL("outputCode(ostream& str,const KBCommand* begin,const KBCommand* end)");
  for (;begin != end; begin++)
   str << begin << '\n';
  return str;     
 }; // ostream& KBCommand::outputCode(ostream& str,const KBCommand* begin,const KBCommand* end) 
#endif
//================================================
#undef DEBUG_NAMESPACE
//================================================
#endif
