//
// File:         KBConstraint.h
// Description:  Representation of simple KB ordering constrains.
// Created:      Apr 23, 2000, 17:00
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#ifndef KB_CONSTRAINT_H
//====================================================
#define KB_CONSTRAINT_H
#ifndef NO_DEBUG_VIS
 #include <iostream.h>
#endif
#include "jargon.h"
#include "HellDebugFlags.h"
#include "HellConst.h"
#include "term.h"   
#include "MultisetOfVariables.h"
#include "KBCommand.h"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_KB_CONSTRAINT
 #define DEBUG_NAMESPACE "KBConstraint"
#endif 
#include "debugMacros.h"
//================================================= 

class KBConstraint
{
 public:
  typedef MultisetOfVariables<HellConst::MaxNumOfVariables> VMS;  
  enum Category { NearlyTautology, NearlyContradiction, Volatile, PurelyLexicographic };
 public:
  KBConstraint() : _posVars(), _negVars() {};
  void assumeGreater(const TERM* lhs,
                     VMS& lhsVars,
                     ulong lhsFuncs,
                     const TERM* rhs,
                     VMS& rhsVars,
                     ulong rhsFuncs)
  {
   CALL("assumeGreater(const TERM* lhs,VMS& lhsVars,ulong lhsFuncs,const TERM* rhs,VMS& rhsVars,ulong rhsFuncs)");
   ASSERT(lhs->isComplex());
   ASSERT(rhs->isComplex());  
   ASSERT(rhsVars.SubsetOf(lhsVars));
   _lhsTerm = lhs;
   _rhsTerm = rhs;   
   lhsVars.ResetIterator();
   ulong lhsOcc;
   ulong rhsOcc;
   _posVars.Reset();
   _negVars.Reset();
   _activeVars.Reset();
   for (const ulong* var = lhsVars.NextVar(); var; var = lhsVars.NextVar())
    {
     lhsOcc = lhsVars.Occurences(*var);
     rhsOcc = rhsVars.Occurences(*var);
     if (lhsOcc != rhsOcc)
      {     
       _activeVars.Register(*var);
       if (lhsOcc > rhsOcc)
        {
  	 _posVars.Register(*var,lhsOcc - rhsOcc);  
        }
       else // lhsOcc < rhsOcc
        _negVars.Register(*var,rhsOcc - lhsOcc); 
      }; 
    };
   if (lhsFuncs > rhsFuncs)
    {      
     _posFuncs = lhsFuncs - rhsFuncs;
     _negFuncs = 0;
    }
   else // lhsFuncs <= rhsFuncs
    {
     _posFuncs = 0;
     _negFuncs = rhsFuncs - lhsFuncs;
    };
   _lexComp = lhs->CompareLex(rhs);
   classify();
   //DF; output(cout); 
  }; //  void assumeGreater(const TERM* lhs,...

  void compile(KBCommand*& code) 
  {
   CALL("compile(KBCommand*& code)");
   switch (_category)
    {
     case NearlyTautology: compileNearlyTautology(code); break; 
     case NearlyContradiction: compileNearlyContradiction(code); break; 
     case Volatile: compileVolatile(code); break; 
     case PurelyLexicographic: compilePurelyLexicographic(code); break;
     #ifdef DEBUG_NAMESPACE
      default: ICP("ICP0"); break;
     #else
      #ifdef _SUPPRESS_WARNINGS_
       default: 
        code->mkFailure();
        code++;
        break;
      #endif        
     #endif 
    }; 
  }; // void compile(KBCommand*& code)  

 
  const Category& category() const { return _category; };
  //bool checkable() const { return _category != PurelyLexicographic; };
  const Comparison& lexComp() const { return _lexComp; };
  const ulong& posFuncs() const { return _posFuncs; };
  const ulong& negFuncs() const { return _negFuncs; };
  VMS& posVars() { return _posVars; };
  VMS& negVars() { return _negVars; };

  ulong posSubstWeight() const { return _posSubstFuncs + _posSubstVars.Size(); };
  ulong negSubstWeight() const { return _negSubstFuncs + _negSubstVars.Size(); };

  void initPosSubst(const ulong& num)
  {
   _posSubstVars.Reset();
   _posSubstFuncs = num;
  };
  
  void initNegSubst(const ulong& num)
  {
   _negSubstVars.Reset();
   _negSubstFuncs = num;
   evalPosSubstWeight();
  };
 
  void posSubstVar(const ulong& svar,const ulong& coeff)
  {
   _posSubstVars.Register(svar,coeff);
  };

  bool negSubstVar(const ulong& svar,const ulong& coeff)
  {
   CALL("negSubstVar(const ulong& svar,const ulong& coeff)");
   _negSubstVars.Register(svar,coeff);
   ASSERT(_posSubstWeight == posSubstWeight());
   return (negSubstWeight() <= _posSubstWeight) 
           && (_posSubstVars.Occurences(svar) >= _negSubstVars.Occurences(svar));  
  };
  
  void posSubstFunc(const ulong& coeff)
  {
   _posSubstFuncs += coeff;
  };

  bool negSubstFunc(const ulong& coeff)
  {
   CALL("negSubstFunc(const ulong& coeff)");
   _negSubstFuncs += coeff;
   ASSERT(_posSubstWeight == posSubstWeight());
   return negSubstWeight() <= _posSubstWeight; 
  };
  
  void evalPosSubstWeight() { _posSubstWeight = posSubstWeight(); };

  bool substGreaterOrEq(bool& canBeEqual)
  {
   CALL("substGreaterOrEq(bool& canBeEqual)");
   if (!_posSubstVars.GreaterOrEqual(_negSubstVars,canBeEqual)) return false;
   if (canBeEqual)
    { 
     if (_posSubstFuncs == _negSubstFuncs) return true;
     canBeEqual = false;
     return (_posSubstFuncs > _negSubstFuncs);
    }
   else // _posSubstVars is strictly greater than _negSubstVars
    {
     ASSERT(_posSubstVars.Size() > _negSubstVars.Size());
     canBeEqual = ((_posSubstFuncs + _posSubstVars.Size()) == (_negSubstFuncs + _negSubstVars.Size()));
     return (_posSubstFuncs + _posSubstVars.Size()) >= (_negSubstFuncs + _negSubstVars.Size());     
    };
  }; // bool substGreaterOrEq(bool& canBeEqual) 

  MultisetOfVariables<HellConst::MaxNumOfVariables>& activeVars() { return _activeVars; };

  bool active(const ulong& v) const { return _activeVars.Occurences(v); };
  bool activePositive(const ulong& v) const { return _posVars.Occurences(v); };
  bool activeNegative(const ulong& v) const { return _negVars.Occurences(v); };

 #ifndef NO_DEBUG_VIS
  static const char* spell(const Category& cat);
  ostream& output(ostream& str);
  ostream& outputInstance(ostream& str); 
 #endif

 protected: 
  void classify()
  {
   CALL("classify()");
   if (_posVars.NumOfDifferent())
    {
     if (_negVars.NumOfDifferent())  
      {
       _category = Volatile;
      }
     else
      {
       ASSERT((!_posFuncs) && _negFuncs);
       _category = NearlyTautology;
      };
    }   
   else // !_posVars.NumOfDifferent()
    if (_negVars.NumOfDifferent())  
     {
      ASSERT(_posFuncs && (!_negFuncs));
      _category = NearlyContradiction;
     }
    else // _posVars and _negVars are empty
     {
      ASSERT((!_posFuncs) && (!_negFuncs) && (_lexComp == Incomparable));
      _category = PurelyLexicographic;
     }; 
  }; // void classify()

  void compileNearlyTautology(KBCommand*& code)
  {
   CALL("compileNearlyTautology(KBCommand*& code)");
   _posVars.ResetIterator();  
   const ulong* var = _posVars.NextVar();
   if (_posVars.NumOfDifferent() == 1) 
    {
     code->mkSuccIfWeightGreater(*var,_posVars.Occurences(*var),_negFuncs);    
     code++;
     if (_lexComp == Greater)
      { 
       code->mkSuccess();
      }
     else
      {
       if (_lexComp == Incomparable) 
	{
         code->mkLexCompareComplexComplex(_lhsTerm,_rhsTerm);
         code++;
        };
       code->mkFailure();
      };
     code++;
     return;
    };

   // more than 1 variable in _posVars  
   code->mkInitWR();
   code++;
   do
    {
     code->mkAddWR(*var,_posVars.Occurences(*var));
     code++;
     code->mkSuccIfWRGreater(_negFuncs);      
     code++;
    }
   while ((var = _posVars.NextVar()));

   if (_lexComp == Greater)
    {
     code->mkFailIfWRLess(_negFuncs);
     code++;
     code->mkSuccess();
    }
   else
    {
     if (_lexComp == Incomparable)
      {
       code->mkFailIfWRLess(_negFuncs);
       code++;
       code->mkLexCompareComplexComplex(_lhsTerm,_rhsTerm);
       code++;
      };
     code->mkFailure();
    };
   code++;
  }; // void compileNearlyTautology(KBCommand*& code)

  void compileNearlyContradiction(KBCommand*& code)
  {
   CALL("compileNearlyContradiction(KBCommand*& code)");
   _negVars.ResetIterator();  
   const ulong* var = _negVars.NextVar();
   if (_negVars.NumOfDifferent() == 1) 
    {
     code->mkFailIfWeightGreater(*var,_negVars.Occurences(*var),_posFuncs);    
     code++;
     code->mkFailIfNonground(*var);
     code++;
     if (_lexComp == Greater)
      {
       code->mkSuccess();
      }
     else
      {
       // can be less or equal only
       code->mkSuccIfWeightLess(*var,_negVars.Occurences(*var),_posFuncs);
       code++;
       if (_lexComp == Incomparable)
	{
         code->mkLexCompareComplexComplex(_lhsTerm,_rhsTerm);
         code++;
        };      
       code->mkFailure();
      };
     code++;
     return;
    };
   
   // more than 1 variable in _negVars   
   code->mkInitWR();
   code++;
   do
    {
     code->mkAddWR(*var,_negVars.Occurences(*var));
     code++;
     code->mkFailIfWRGreater(_posFuncs);
     code++;
     code->mkFailIfNonground(*var);
     code++;
    }
   while ((var = _negVars.NextVar()));

   if (_lexComp == Greater)
    {
     code->mkSuccess();
    }
   else
    {
     code->mkSuccIfWRLess(_posFuncs);
     code++;
     if (_lexComp == Incomparable)
      {
       code->mkLexCompareComplexComplex(_lhsTerm,_rhsTerm);
       code++;
      };      
     code->mkFailure();
    };
   code++;   
  }; // void compileNearlyContradiction(KBCommand*& code)

  void compileVolatile(KBCommand*& code)
  {
   CALL("compileVolatile(KBCommand*& code)");

   code->mkInitPosSubst(_posFuncs);
   code++;
   _posVars.ResetIterator();  
   for (const ulong* var = _posVars.NextVar(); var; var = _posVars.NextVar())
    {
     code->mkPosSubst(*var,_posVars.Occurences(*var));
     code++;
    };

   code->mkInitNegSubst(_negFuncs);
   code++; 
   _negVars.ResetIterator();  
   for (const ulong* var_ = _negVars.NextVar(); var_; var_ = _negVars.NextVar())
    {
     code->mkNegSubst(*var_,_negVars.Occurences(*var_));
     code++;
    };

   code->mkCompareSubst();
   code++; 
   if (_lexComp == Greater)
    {
     code->mkSuccess();
    }
   else
    {
     if (_lexComp == Incomparable) 
      {
       code->mkLexCompareComplexComplex(_lhsTerm,_rhsTerm);
       code++;
      };
     code->mkFailure();
    };
   code++;
  }; // void compileVolatile(KBCommand*& code) 

  void compilePurelyLexicographic(KBCommand*& code)
  {
   CALL("compilePurelyLexicographic(KBCommand*& code)");
   code->mkLexCompareComplexComplex(_lhsTerm,_rhsTerm);
   code++;
   code->mkFailure();
   code++; 
  }; // void compilePurelyLexicographic(KBCommand*& code)

 protected:
  const TERM* _lhsTerm;
  const TERM* _rhsTerm;
  VMS _posVars;
  VMS _negVars; 
  ulong _posFuncs;
  ulong _negFuncs;
  Comparison _lexComp;
  Category _category; 
 
  VMS _activeVars;
 
  VMS _posSubstVars;
  VMS _negSubstVars;
  ulong _posSubstFuncs;
  ulong _negSubstFuncs;
  ulong _posSubstWeight;
}; // class KBConstraint

    
#ifndef NO_DEBUG_VIS
 inline const char* KBConstraint::spell(const Category& cat)
 {
  switch (cat) 
   {
    case NearlyTautology: return "NearlyTautology";
    case NearlyContradiction: return "NearlyContradiction";
    case Volatile: return "Volatile";
    case PurelyLexicographic: return "PurelyLexicographic";
    default: return "Unknown"; 
   };  
 };

 inline ostream& KBConstraint::output(ostream& str)
 {
  str << spell(_category) << ' ';
  str << _posFuncs;
  if (_posVars.NumOfDifferent()) _posVars.output(str << " + ");
  str << " VERSUS ";
  str << _negFuncs;
  if (_negVars.NumOfDifferent()) _negVars.output(str << " + ");
  str << " (lex " << _lexComp << ')';
  return str;
 };

 inline ostream& KBConstraint::outputInstance(ostream& str)
 {
  str << _posSubstFuncs;
  if (_posSubstVars.NumOfDifferent()) _posSubstVars.output(str << " + ");
  str << " VERSUS "; 
  str << _negSubstFuncs;
  if (_negSubstVars.NumOfDifferent()) _negSubstVars.output(str << " + ");
  return str;
 };

#endif




//================================================
#undef DEBUG_NAMESPACE
//================================================= 
#endif

