//
// File:         PagedPassiveClauseStorage.h
// Description:  Various signature related operations.  
// Created:      Apr 27, 2001, 17:10
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#ifndef PAGED_PASSIVE_CLAUSE_STORAGE_H
//=================================================
#define PAGED_PASSIVE_CLAUSE_STORAGE_H
#include <iostream.h>
#include "jargon.h"
#include "HellDebugFlags.h"
#ifdef DEBUG_ALLOC_OBJ_TYPE 
 #include "ClassDesc.h"
#endif
#include "HellConst.h"
#include "GlobAlloc.h"
#include "BitWord.h"
#include "PageQueueDisciplineStorage.h"
#include "term.h"
#include "ClauseBackground.h"
#include "clause.h"
//=================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PAGED_PASSIVE_CLAUSE_STORAGE
 #define DEBUG_NAMESPACE "PagedPassiveClauseStorage"
#endif 
#include "debugMacros.h"
//=================================================  

class PagedPassiveClauseStorage
{
 public: 
  class ClauseProfile
  {
   public:
    ClauseProfile() {};
    long clauseNumber() const { return _clauseNumber; };
    long clauseSize() const { return _clauseSize; };
    const BitWord& rules() const { return _rules; };
    ulong numOfLiterals() const { return _numOfLiterals; };
    ulong numOfAncestors() const { return _numOfAncestors; };
    ostream& output(ostream& str) const;
   private:
    long _clauseNumber;
    long _clauseSize;
    BitWord _rules;
    ulong _numOfLiterals;
    ulong _numOfAncestors;
   friend class PagedPassiveClauseStorage;
  }; // class ClauseProfile
  
  class ClauseLiterals
  {
   public:
    ClauseLiterals() {};
    bool nextLiteral() 
    {
     CALL("ClauseLiterals::nextLiteral()");
     ASSERT(!_numOfRemainingSymbols);
     if (_numOfRemainingLiterals)
      {
       _numOfRemainingLiterals--;
       _numOfRemainingSymbols = _cursor->Content();
       _cursor++;
       _currentLiteralIsInherentlySelected = (bool)_cursor->Content();
       _cursor++;
       return true;
      }
     else return false;   
    };
    bool nextSymbol(TERM& sym) 
    {
     CALL("ClauseLiterals::nextSymbol(TERM& sym)");
     if (_numOfRemainingSymbols) 
      {
       _numOfRemainingSymbols--;
       sym = *_cursor;
       _cursor++;
       return true;
      }
     else return false;
    };
    bool currentLiteralIsInherentlySelected() const
    {
     CALL("currentLiteralIsInherentlySelected() const");
     return _currentLiteralIsInherentlySelected;
    };
    void reset() 
    {
     CALL("ClauseLiterals::reset()");    
     _cursor = _beginning;
     _numOfRemainingLiterals = _numberOfLiterals;
     _numOfRemainingSymbols = 0;
    };
    
   private:
    const TERM* _beginning;
    ulong _numberOfLiterals;
    const TERM* _cursor;
    ulong _numOfRemainingLiterals;
    ulong _numOfRemainingSymbols;
    bool _currentLiteralIsInherentlySelected;
   friend class PagedPassiveClauseStorage;
  }; // class ClauseLiterals

  class ClauseAncestors
  {
   public:
    ClauseAncestors() {};
    bool nextAncestor(Clause*& anc) 
    {
     CALL("ClauseAncestors::nextAncestor(Clause*& anc)");
     if (_numOfRemainingAncestors)
      {
       _numOfRemainingAncestors--;
       anc = *_cursor;
       _cursor++;
       return true;
      }
     else return false;
    };
    void reset()
    {
     CALL("ClauseAncestors::reset()");    
     _cursor = _beginning;
     _numOfRemainingAncestors = _numberOfAncestors;
    };
   private:
    Clause** _beginning;
    ulong _numberOfAncestors;
    Clause** _cursor;
    ulong _numOfRemainingAncestors;
   friend class PagedPassiveClauseStorage;
  }; // class ClauseAncestors

  class StoredClause
  {
   public:   
    StoredClause() : _profile(), _literals(), _ancestors() {};
    ClauseProfile& profile() { return _profile; };
    ClauseLiterals& literals() { return _literals; };
    ClauseAncestors& ancestors() { return _ancestors; };

    long clauseNumber() const { return _profile.clauseNumber();  };
    long clauseSize() const { return _profile.clauseSize();  };
    const BitWord& rules() const { return _profile.rules();  };
    ulong numOfLiterals() const { return  _profile.numOfLiterals(); };
    ulong numOfAncestors() const { return  _profile.numOfAncestors(); };   

    void resetLiterals() { _literals.reset(); };
    bool nextLiteral() { return _literals.nextLiteral(); };
    bool nextSymbol(TERM& sym) { return _literals.nextSymbol(sym); };

    void resetAncestors() { _ancestors.reset(); };
    bool nextAncestor(Clause*& anc) { return _ancestors.nextAncestor(anc); };

    bool isVIP() const { return rules().bit(ClauseBackground::VIP); };    
    bool isOrphan() 
    {
     CALL("StoredClause::isOrphan()");
     resetAncestors();
     Clause* ancestor;
     while (nextAncestor(ancestor)) 
       if (!ancestor->Active()) return true;
     return false;  
    };
    ostream& output(ostream& str);    
    
   private: 
    ClauseProfile _profile;
    ClauseLiterals _literals;
    ClauseAncestors _ancestors;
   friend class PagedPassiveClauseStorage;
  }; // class StoredClause

  class StoredClauseHandler
  {
   public:
    StoredClauseHandler() {};
    operator bool() const { return (bool)_memory; };
   private:
    void* _memory;
   friend class PagedPassiveClauseStorage;
  }; // class StoredClauseHandler 

  class Iterator
  {
   public:
    Iterator() :_currentClause() {};
    void reset(const PagedPassiveClauseStorage& storage)
    {
     CALL("Iterator::reset(const PagedPassiveClauseStorage& storage)");
     _storage = (PagedPassiveClauseStorage*)(&storage);
     _nextCategoryNumber = _storage->_minNonemptyCatNum + 1;
     _nextObject = _storage->_category[_storage->_minNonemptyCatNum].head();
    };  
 
    bool nextClause() 
    {
     CALL("Iterator::nextClause()");
     while (!_nextObject)
      {
       // try next category
       if (_nextCategoryNumber > _storage->_maxNonemptyCatNum) return false;      
       _nextObject = _storage->_category[_nextCategoryNumber].head();
       _nextCategoryNumber++; 
      };    

     PagedPassiveClauseStorage::readFromMemory(PagedPassiveClauseStorage::Storage::objectContent(_nextObject),
                                               _currentClause.profile(),
                                               _currentClause.literals(),
                                               _currentClause.ancestors());
     _nextObject = Storage::nextObject(_nextObject);  
     return true;
    };
    StoredClause& currentClause() { return _currentClause; };   
   private:
    PagedPassiveClauseStorage* _storage;
    StoredClause _currentClause;
    long _nextCategoryNumber; 
    void* _nextObject; 
  }; // class Iterator  
 

 public:
  PagedPassiveClauseStorage() 
  {
   CALL("constructor PagedPassiveClauseStorage()");
   init();
  };  
  ~PagedPassiveClauseStorage() 
  {
   CALL("destructor ~PagedPassiveClauseStorage()");
   if (!Exit::fast) destroy(); 
  };     
  
  static ulong pageSize() { return PageSize*sizeof(void*); }; 
  void init()
  {
   CALL("init()");
   _ageWeightRatio = 0;
   _maximalPriorityFlag = false;
   _minNonemptyCatNum = MaximalCategoryNumber + 1; 
   _maxNonemptyCatNum = -1;
   _toBeSelectedByWeight = 0;
   _totalNumberOfClauses = 0;
   _numberOfSelectedClauses = 0;  
   _numberOfCancelledClauses = 0; 

   _freePages.init();
   for (ulong cat = 0; cat <= MaximalCategoryNumber; cat++)
    {
     _category[cat].init(&_freePages);
    };
  
   #ifdef DEBUG_NAMESPACE
    _debugClauseOpenForWriting = false; 
   #endif  
  }; // void init()

  void destroy() 
  {
   CALL("destroy()");
   for (ulong cat = 0; cat <= MaximalCategoryNumber; cat++)
    {
     _category[cat].destroy();
    };
   _freePages.destroy();
  }; // void destroy() 

  void reset() { CALL("reset()"); destroy(); init(); }; 
  
  void setAgeWeightRatio(long awr) 
  { 
   CALL("setAgeWeightRatio(long awr)");
   ASSERT(awr >= 0);
   _ageWeightRatio = awr;
   _toBeSelectedByWeight = awr; 
  };
  void maximalPriorityOn() { _maximalPriorityFlag = true; };
  void maximalPriorityOff() { _maximalPriorityFlag = false; };

  // Properties 
 
  bool empty() { return _maxNonemptyCatNum < _minNonemptyCatNum; };
  bool nonempty() { return _maxNonemptyCatNum >= _minNonemptyCatNum; };
  long totalNumberOfClauses() const { return _totalNumberOfClauses; };
  long numberOfSelectedClauses() const { return _numberOfSelectedClauses; };
  long numberOfCancelledClauses() const { return _numberOfCancelledClauses; };
  long numberOfCurrentlyKeptClauses() const 
  { 
   return _totalNumberOfClauses - _numberOfSelectedClauses - _numberOfCancelledClauses;
  };

  // Writing new clauses

  void  openClause(StoredClauseHandler& handler,long clNum,long clSize,const BitWord rules,ulong numOfLits,ulong numOfAncestors); 

  bool openClause(long clNum,long clSize,const BitWord rules,ulong numOfLits,ulong numOfAncestors)
  {
   CALL("openClause(long clNum,long clSize,const BitWord rules,ulong numOfLits,ulong numOfAncestors)");
   StoredClauseHandler handler;
   openClause(handler,clNum,clSize,rules,numOfLits,numOfAncestors);
   return (bool)handler;
  }; // bool openClause(long clNum,long clSize,const BitWord rules,ulong numOfLits,ulong numOfAncestors)
 
  void openLiteral(bool inherentlySelected)
  {
   CALL("openLiteral(bool inherentlySelected)");
   ASSERT(_debugClauseOpenForWriting); 
   ASSERT(_cursor + sizeInWords(sizeof(TERM)) <= _debugEndOfCurrentClause);
   ASSERT(_debugNumberOfLiteralsToWrite);
   _currentLiteral = (TERM*)_cursor;
   _currentLiteralSize = 0UL;
   _cursor += sizeInWords(sizeof(TERM));
   ((TERM*)_cursor)->Make((ulong)inherentlySelected);
   _cursor += sizeInWords(sizeof(TERM));
  }; // void openLiteral(bool inherentlySelected)

  void writeSymbol(const TERM& sym) 
  {
   CALL("writeSymbol(const TERM& sym)");
   ASSERT(_debugNumberOfLiteralsToWrite);
   ASSERT(_cursor + sizeInWords(sizeof(TERM)) <= _debugEndOfCurrentClause);
   *((TERM*)_cursor) = sym;
   _cursor += sizeInWords(sizeof(TERM));
   _currentLiteralSize++;
  }; // void writeSymbol(const TERM& sym) 
  
  void closeLiteral()
  { 
   CALL("closeLiteral()");
   ASSERT(_debugNumberOfLiteralsToWrite);
   DOP(_debugNumberOfLiteralsToWrite--);
   _currentLiteral->Make(_currentLiteralSize);       
  }; // void closeLiteral()

  void writeAncestor(const Clause* ancestor)
  {
   CALL("writeAncestor(const Clause* ancestor)");
   ASSERT(_debugClauseOpenForWriting); 
   ASSERT(!_debugNumberOfLiteralsToWrite);
   ASSERT(_debugNumberOfAncestorsToWrite);
   ASSERT(_cursor + 1 <= _debugEndOfCurrentClause);
   *_cursor = (void*)ancestor;     
   _cursor++;
   DOP(_debugNumberOfAncestorsToWrite--);
  }; // void writeAncestor(const Clause* ancestor)

  void endOfClause()
  {
   CALL("endOfClause()");
   ASSERT(_debugClauseOpenForWriting);  
   DOP(_debugClauseOpenForWriting = false);
   ASSERT(!_debugNumberOfLiteralsToWrite);
   ASSERT(!_debugNumberOfAncestorsToWrite);
   ASSERT(!empty()); 
   ASSERT(_category[_minNonemptyCatNum].nonempty());
   ASSERT(_category[_maxNonemptyCatNum].nonempty());
  }; // void endOfClause()


  // Selection and popping the selected clause. 
  bool select(StoredClauseHandler& selectedClauseHandler);  

  bool select(StoredClause& selectedClause);

  void popSelected();
       
  // Cancelling clauses at the end of the queue and recycling pages. 

  bool worstClause(ClauseProfile& prof);
  bool cancelWorstClause();

  bool recyclePage()
  {
   CALL("recyclePage()");
   return _freePages.recyclePage();
  }; // bool recyclePage()

  static void loadClause(const StoredClauseHandler& handler,StoredClause& clause)
  {
   CALL("loadClause(const StoredClauseHandler& handler,StoredClause& clause)");
   ASSERT((bool)handler);
   readFromMemory(handler._memory,clause.profile(),clause.literals(),clause.ancestors());
  }; // void loadClause(const StoredClauseHandler& handler,StoredClause& clause)



 private: 
  static const ulong PageSize = 1024; 
  typedef PageQueueDisciplineStorage<GlobAlloc,PageSize,PagedPassiveClauseStorage> Storage;
 private:
  long categoryNum(long clauseSize) { return clauseSize; };
  static ulong sizeInWords(ulong sizeInBytes) 
  {
   ulong res = sizeInBytes/sizeof(void*);
   if (sizeInBytes % sizeof(void*)) res++;
   return res;
  }; // ulong sizeInWords(ulong sizeInBytes)  

  static void selectFrom(Storage* cat,StoredClauseHandler& selectedClauseHandler)
  {
   CALL("selectFrom(Storage* cat,StoredClauseHandler& selectedClauseHandler)");
   ASSERT(cat);
   ASSERT(cat->nonempty());      
   selectedClauseHandler._memory = Storage::objectContent(cat->head());
  }; // void selectFrom(Storage* cat,StoredClauseHandler& selectedClauseHandler)
 
  static void readFromMemory(void* memory,ClauseProfile& prof,ClauseLiterals& lits,ClauseAncestors& ancestors);

 private:
  static const ulong MaximalCategoryNumber = HellConst::MaxClauseSize;
 private:
  // parameters 
  long _ageWeightRatio;
  bool _maximalPriorityFlag; 

  // statistics
  long _minNonemptyCatNum; 
  long _maxNonemptyCatNum;
  long _toBeSelectedByWeight;

  long _totalNumberOfClauses;
  long _numberOfSelectedClauses;  
  long _numberOfCancelledClauses; 

  // storage
  Storage::FreePages _freePages;
  Storage _category[MaximalCategoryNumber + 1];
  Storage* _currentCategory;
  long _selectedCategoryNumber;
  bool _lastSelectedByWeight;
  void** _cursor;
  TERM* _currentLiteral;
  ulong _currentLiteralSize;
  #ifdef DEBUG_NAMESPACE
   bool _debugClauseOpenForWriting; 
   void** _debugEndOfCurrentClause;
   ulong _debugNumberOfLiteralsToWrite;
   ulong _debugNumberOfAncestorsToWrite;
  #endif  
 friend Iterator;   
}; // class PagedPassiveClauseStorage

//====================================================
#undef DEBUG_NAMESPACE
//====================================================
#endif
