//
// File:         PathIndexTree.h
// Description:  Trees for path indexing.
// Created:      Feb 25, 2000, 15:00
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//============================================================================
#ifndef PATH_INDEX_TREE_H
//=============================================================================
#define PATH_INDEX_TREE_H 
#include "jargon.h"
#include "HellDebugFlags.h"
#include "term.h"
#include "Stack.h"
#ifdef DEBUG_ALLOC_OBJ_TYPE
 #include "ClassDesc.h"
#endif
//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PATH_INDEX_TREE
 #define DEBUG_NAMESPACE "PathIndexTree<Alloc,IndObj,ConstIndObj,MaxTermDepth,MaxTermSize>"
#endif
#include "debugMacros.h"
//============================================================================

template <class Alloc,class IndObj,class ConstIndObj,ulong MaxTermDepth,ulong MaxTermSize>
// IndObj and ConstIndObj must have operator bool() const
class PathIndexTree
{
 public:
  class OccNode;
  class FuncNode
  {
   public:
    FuncNode(const TERM& f) : _func(f), _altern((FuncNode*)0) // for nonconst. functors only
    {
     CALL("FuncNode::(constructor FuncNode(const TERM& f))");
     ASSERT(_func.isComplex());
     ulong arity = f.arity();
     ASSERT(arity);
     OccNode** deeperPos = deeper();
     for (ulong argn = 0; argn < arity; argn++) deeperPos[argn] = (OccNode*)0;  
    }; 
    FuncNode(const TERM& c,const ConstIndObj& initIndObj) 
     : _func(c), _altern((FuncNode*)0) // for constants only
    {
     CALL("FuncNode::(constructor FuncNode(const TERM& c,const ConstIndObj& initIndObj))");
     ASSERT(_func.isComplex());
     ASSERT(!(_func.arity()));
     indObj() = initIndObj;
    };

    void* operator new(size_t,ulong arity) // for nonconst. functors only
    { 
     CALL("operator new(size_t,ulong arity)");
     ASSERT(arity);
     #ifdef DEBUG_ALLOC_OBJ_TYPE
      ALWAYS(_funcClassDesc.registerAllocated(sizeof(FuncNode) + arity*sizeof(OccNode*)));
      return Alloc::allocate(sizeof(FuncNode) + arity*sizeof(OccNode*),&_funcClassDesc); 
     #else
      return Alloc::allocate(sizeof(FuncNode) + arity*sizeof(OccNode*)); 
     #endif
    };

    void* operator new(size_t) // for constants only 
    { 
     CALL("FuncNode::operator new(size_t)");
     #ifdef DEBUG_ALLOC_OBJ_TYPE
      ALWAYS(_constClassDesc.registerAllocated(sizeof(FuncNode) + sizeof(ConstIndObj)));
      return Alloc::allocate(sizeof(FuncNode) + sizeof(ConstIndObj),&_constClassDesc);  
     #else 
      return Alloc::allocate(sizeof(FuncNode) + sizeof(ConstIndObj));  
     #endif
    };
    void operator delete(void* obj) 
    {
     CALL("FuncNode::operator delete(void* obj)");
     ulong arity = ((FuncNode*)obj)->func().arity();
     if (arity)
      {
       #ifdef DEBUG_ALLOC_OBJ_TYPE
        ALWAYS(_funcClassDesc.registerDeallocated(sizeof(FuncNode) + arity*sizeof(OccNode*)));
        Alloc::deallocate(obj,sizeof(FuncNode) + arity*sizeof(OccNode*),&_funcClassDesc);
       #else
        Alloc::deallocate(obj,sizeof(FuncNode) + arity*sizeof(OccNode*));
       #endif
      }
     else
      {
       #ifdef DEBUG_ALLOC_OBJ_TYPE
        ALWAYS(_constClassDesc.registerDeallocated(sizeof(FuncNode) + sizeof(ConstIndObj)));
        Alloc::deallocate(obj,sizeof(FuncNode) + sizeof(ConstIndObj),&_constClassDesc);
       #else
        Alloc::deallocate(obj,sizeof(FuncNode) + sizeof(ConstIndObj));
       #endif
      };
    };     
    static ulong funcMinSize() { return sizeof(FuncNode) + sizeof(OccNode*); };
    static ulong funcMaxSize() { return sizeof(FuncNode) + HellConst::MaxNumOfTermArguments*sizeof(OccNode*); };
    static ulong constMinSize() { return sizeof(FuncNode) + sizeof(ConstIndObj); };
    static ulong constMaxSize() { return sizeof(FuncNode) + sizeof(ConstIndObj); };

    const TERM& func() const { return _func; }; 
    FuncNode*& altern() { return _altern; };
    const FuncNode* const & altern() const { return _altern; };
    ConstIndObj& indObj() // for constants only
    {
     CALL("FuncNode::indObj()");
     ASSERT(isConstant());
     return *((ConstIndObj*)(this + 1));
    };
    const ConstIndObj& indObj() const // for constants only
    {
     CALL("FuncNode::indObj() const");
     ASSERT(isConstant());
     return *((const ConstIndObj*)(this + 1));
    };
    OccNode** deeper() // for nonconst. functors only
    {
     CALL("FuncNode::deeper()");
     ASSERT(isNonconst());
     return (OccNode**)(this + 1);
    };   
    const OccNode* const * deeper() const // for nonconst. functors only
    {
     CALL("FuncNode::deeper() const");
     ASSERT(isNonconst());
     return (const OccNode* const *)(this + 1);
    };
    ulong arity() const { return func().arity(); };
    bool isConstant() const { return !(func().arity()); };
    bool isNonconst() const { return func().arity(); };
    bool obsolete() const 
    {
     if (isConstant()) return !(indObj());
     for (ulong a = 0; a < arity(); a++) if (deeper()[a]) return false; 
     return true;  
    };   
   private:
    TERM _func;
    FuncNode* _altern;
    #ifdef DEBUG_ALLOC_OBJ_TYPE
     static ClassDesc _funcClassDesc;
     static ClassDesc _constClassDesc;
    #endif 
   }; // class FuncNode

  class OccNode
  { 
   public:
    OccNode(const IndObj& initIndObj) : _indObj(initIndObj), _complex((FuncNode*)0) {};
    void* operator new(size_t) 
    {
     CALL("OccNode::operator new(size_t)");
     #ifdef DEBUG_ALLOC_OBJ_TYPE
      ALWAYS(_classDesc.registerAllocated(sizeof(OccNode)));
      return Alloc::allocate(sizeof(OccNode),&_classDesc); 
     #else 
      return Alloc::allocate(sizeof(OccNode)); 
     #endif
    };
    void operator delete(void* obj) 
    { 
     CALL("OccNode::operator delete(void* obj)");
     #ifdef DEBUG_ALLOC_OBJ_TYPE
      ALWAYS(_classDesc.registerDeallocated(sizeof(OccNode)));
      Alloc::deallocate(obj,sizeof(OccNode),&_classDesc); 
     #else 
      Alloc::deallocate(obj,sizeof(OccNode)); 
     #endif
    };
    static ulong minSize() { return sizeof(OccNode); };
    static ulong maxSize() { return sizeof(OccNode); };   
    IndObj& indObj() { return _indObj; };    
    const IndObj& indObj() const { return _indObj; };
    FuncNode*& complex() { return _complex; }; 
    const FuncNode* const & complex() const { return _complex; }; 
   private:
    IndObj _indObj;
    FuncNode* _complex;
    #ifdef DEBUG_ALLOC_OBJ_TYPE
     static ClassDesc _classDesc;
    #endif  
  }; // class OccNode

  class Integrator
  {
   //public:
    //typedef PathIndexTree<Alloc,IndObj,ConstIndObj,MaxTermDepth,MaxTermSize>::OccNode OccNode;
    //typedef PathIndexTree<Alloc,IndObj,ConstIndObj,MaxTermDepth,MaxTermSize>::FuncNode FuncNode; 

   public:
    Integrator(const IndObj& initIndObj,const ConstIndObj& initConstIndObj)
      : _initIndObj(initIndObj),
       _initConstIndObj(initConstIndObj),
       _backtrackFuncNodes(),
       _backtrackArgNums()
    {
    };
    void reset(OccNode** tree) 
    { 
     _occAddr = tree;
     createOccNodeIfNeeded();
     _backtrackFuncNodes.reset();
     _backtrackArgNums.reset();
    };
  
    void nonconstFunc(const TERM& f)
    {
     CALL("Integrator::nonconstFunc(const TERM& f)");
     ulong arity = f.arity();
     ASSERT(arity);
     ASSERT(*_occAddr);
     FuncNode** nodeAddr = &((*_occAddr)->complex());
     FuncNode* node = *nodeAddr;
     while (node)
      {
       if (node->func() == f) 
        { 
         _occAddr = node->deeper(); 
         createOccNodeIfNeeded();      
         if (arity > 1)
  	  { 
	   _backtrackFuncNodes.push(node);
           _backtrackArgNums.push(1);
          };
         return;
        };  
       //node->func() != f
       nodeAddr = &(node->altern());
       node = *nodeAddr;
      };
     // !node, there is no node with this functor, create one
     node = new (arity) FuncNode(f);
     *nodeAddr = node;
     _occAddr = node->deeper();
     *_occAddr = new OccNode(_initIndObj);        
     if (arity > 1)
      { 
       _backtrackFuncNodes.push(node);
       _backtrackArgNums.push(1);
      };
    }; // void nonconstFunc(const TERM& f) 

    void constant(const TERM& c)
    {
     CALL("Integrator::constant(const TERM& c)");    
     ASSERT(!c.arity());
     ASSERT(*_occAddr);
     FuncNode** nodeAddr = &((*_occAddr)->complex());
     _lastConstNode = *nodeAddr;
     while (_lastConstNode)
      {
       if (_lastConstNode->func() == c) 
        { 
         backtrack();
         return;
        };  
       //_lastConstNode->func() != c
       nodeAddr = &(_lastConstNode->altern());
       _lastConstNode = *nodeAddr;
      };
     // !node, there is no node with this constant, create one
     _lastConstNode = new FuncNode(c,_initConstIndObj);
     *nodeAddr = _lastConstNode;
     backtrack();
   }; // void constant(const TERM& c)

   void variable()
   {
    CALL("Integrator::variable()");
    ASSERT(*_occAddr);
    backtrack();  
   };
  
   IndObj& indexedObj() const { return (*_occAddr)->indObj(); };
   ConstIndObj& constIndObj() const { return _lastConstNode->indObj(); };  

  private:
   void createOccNodeIfNeeded() const
   {
    if (!(*_occAddr)) 
     {
      *_occAddr = new OccNode(_initIndObj);
     };   
   };
   void backtrack()
   {
    CALL("Integrator::backtrack()");
    if (_backtrackFuncNodes.nonempty()) 
     { 
      FuncNode* node = _backtrackFuncNodes.pop();    
      ulong argn = _backtrackArgNums.pop();
      _occAddr = node->deeper() + argn;
      createOccNodeIfNeeded();
      argn++;
      ASSERT(argn <= node->arity());
      if (argn < node->arity())
       {
        _backtrackFuncNodes.push(node);    
        _backtrackArgNums.push(argn);
       };  
     };
   }; // void backtrack()
  private:
   const IndObj _initIndObj;
   const ConstIndObj _initConstIndObj;
   OccNode** _occAddr;
   FuncNode* _lastConstNode;
   Stack<FuncNode*,MaxTermDepth> _backtrackFuncNodes;
   Stack<ulong,MaxTermDepth> _backtrackArgNums;
  }; // class Integrator


 class Retrieval
 {
  //public:
   //typedef PathIndexTree<Alloc,IndObj,ConstIndObj,MaxTermDepth,MaxTermSize>::OccNode OccNode;
   //typedef PathIndexTree<Alloc,IndObj,ConstIndObj,MaxTermDepth,MaxTermSize>::FuncNode FuncNode; 

  public:
   Retrieval() : _backtrackFuncNodes(), _backtrackArgNums() {};
   bool reset(OccNode** tree) 
   { 
    _occAddr = tree;
    _backtrackFuncNodes.reset();
    _backtrackArgNums.reset();
    DOP(debugEndOfTerm = false);
   return *_occAddr;
   }; 
   bool nonconstFunc(const TERM& f)
   {
    CALL("Retrieval::nonconstFunc(const TERM& f)");
    ASSERT(!debugEndOfTerm);
    if (!(*_occAddr)) return false;
    ulong arity = f.arity();
    ASSERT(arity);
    
    FuncNode* node = (*_occAddr)->complex();
    while (node)
     {
      if (node->func() == f) 
       { 
        _occAddr = node->deeper();     
        if (arity > 1)
 	 { 
	  _backtrackFuncNodes.push(node);
          _backtrackArgNums.push(1);
         };
        return true;
       };  
      //node->func() != f
      node = node->altern();
     };
    // !node
    return false;
   }; // bool nonconstFunc(const TERM& f)

   bool constant(const TERM& c)
   {
    CALL("Retrieval::constant(const TERM& c)"); 
    ASSERT(!debugEndOfTerm);   
    ASSERT(!c.arity());
    if (!(*_occAddr)) return false;
    _lastConstNode = (*_occAddr)->complex();
    while (_lastConstNode)
     {
      if (_lastConstNode->func() == c) 
       { 
        backtrack();
        return true;
       }  
      else _lastConstNode = _lastConstNode->altern();
     };
    // !_lastConstNode, there is no node with this constant
    return false;
   }; // bool constant(const TERM& c)  

   void after() { CALL("Retrieval::after()"); ASSERT(!debugEndOfTerm); backtrack(); };

   IndObj& indexedObj() const { return (*_occAddr)->indObj(); };
   ConstIndObj& constIndObj() const { return _lastConstNode->indObj(); };  
  private:
   void backtrack()
   {
    CALL("Retrieval::backtrack()");
    if (_backtrackFuncNodes.nonempty()) 
     { 
      FuncNode* node = _backtrackFuncNodes.pop();    
      ulong argn = _backtrackArgNums.pop();
      _occAddr = node->deeper() + argn;
      argn++;
      ASSERT(argn <= node->arity());
      if (argn < node->arity())
       {
        _backtrackFuncNodes.push(node);    
        _backtrackArgNums.push(argn);
       };  
     }
    else 
     {
      DOP(debugEndOfTerm = true);
     };
   }; // void backtrack()

  private:
   OccNode** _occAddr;
   FuncNode* _lastConstNode;
   Stack<FuncNode*,MaxTermDepth> _backtrackFuncNodes;
   Stack<ulong,MaxTermDepth> _backtrackArgNums;
   #ifdef DEBUG_NAMESPACE
    bool debugEndOfTerm;  
   #endif
 }; // class Retrieval

 class Removal
 {
  //public:
   //typedef PathIndexTree<Alloc,IndObj,ConstIndObj,MaxTermDepth,MaxTermSize>::OccNode OccNode;
   //typedef PathIndexTree<Alloc,IndObj,ConstIndObj,MaxTermDepth,MaxTermSize>::FuncNode FuncNode; 

  public:
   Removal()
    : _backtrackFuncNodes(),
      _backtrackArgNums(),
      _candidatesForRecycling()
   {
   };
 
   void reset() 
   { 
    _candidatesForRecycling.reset();
   };

   void reset(OccNode** tree) 
   { 
    _occAddr = tree;
    _backtrackFuncNodes.reset();
    _backtrackArgNums.reset();
    DOP(debugEndOfTerm = false);
   };

   bool nonconstFunc(const TERM& f)
   {
    CALL("Removal::nonconstFunc(const TERM& f)");
    ASSERT(!debugEndOfTerm);
    if (!(*_occAddr)) return false;
    ulong arity = f.arity();
    ASSERT(arity);
   
    _candForRecycling = &((*_occAddr)->complex());
    FuncNode* node = (*_occAddr)->complex();

    while (node)
     {
      if (node->func() == f) 
       { 
        _occAddr = node->deeper();     
        if (arity > 1)
 	 { 
 	  _backtrackFuncNodes.push(node);
          _backtrackArgNums.push(1);
         };
        ASSERT(*_candForRecycling);
        _candidatesForRecycling.push(_candForRecycling);
        return true;
       };  
      //node->func() != f
      _candForRecycling = &(node->altern());
      node = node->altern();
     };
    // !node
    return false;
   }; // bool nonconstFunc(const TERM& f)

   bool constant(const TERM& c)
   {
    CALL("Removal::constant(const TERM& c)"); 
    ASSERT(!debugEndOfTerm);   
    ASSERT(!c.arity());
    if (!(*_occAddr)) return false;
    _candForRecycling = &((*_occAddr)->complex());
    _lastConstNode = (*_occAddr)->complex();
    while (_lastConstNode)
     {
      if (_lastConstNode->func() == c) 
       { 
        backtrack();
        ASSERT(*_candForRecycling);
        _candidatesForRecycling.push(_candForRecycling);
        return true;
       }  
      else 
       {
        _candForRecycling = &(_lastConstNode->altern());
        _lastConstNode = _lastConstNode->altern();
       };
     };
    // !_lastConstNode, there is no node with this constant
    return false;
   }; // bool constant(const TERM& c)  

   bool occurences() { return (bool)(*_occAddr); }; 

   void after() { CALL("Removal::after()"); ASSERT(!debugEndOfTerm); backtrack(); }; 

   IndObj& indexedObj() const { CALL("Removal::indexedObj() const"); ASSERT(*_occAddr);  return (*_occAddr)->indObj(); };
   ConstIndObj& constIndObj() const { CALL("Removal::constIndObj() const"); ASSERT(_lastConstNode); return _lastConstNode->indObj(); };  

   void recycleObsoleteNodes()
   {
    CALL("Removal::recycleObsoleteNodes()"); 
    while (_candidatesForRecycling.nonempty()) recycleObsolete(*(_candidatesForRecycling.pop()));
   };

  private:
   void backtrack()
   {
    CALL("Removal::backtrack()");
    if (_backtrackFuncNodes.nonempty()) 
     { 
      FuncNode* node = _backtrackFuncNodes.pop();    
      ulong argn = _backtrackArgNums.pop();
      _occAddr = node->deeper() + argn;
      argn++;
      ASSERT(argn <= node->arity());
      if (argn < node->arity())
       {
        _backtrackFuncNodes.push(node);    
        _backtrackArgNums.push(argn);
       };  
     }
    else 
     {
      DOP(debugEndOfTerm = true);
     };
   }; // void backtrack()
 
  private:
   OccNode** _occAddr;
   FuncNode* _lastConstNode;
   FuncNode** _candForRecycling;
   Stack<FuncNode*,MaxTermDepth> _backtrackFuncNodes;
   Stack<ulong,MaxTermDepth> _backtrackArgNums;
   Stack<FuncNode**,MaxTermSize> _candidatesForRecycling;
   #ifdef DEBUG_NAMESPACE
    bool debugEndOfTerm;  
   #endif
 }; // class Removal

 class Destruction
 {
  //public:
   //typedef PathIndexTree<Alloc,IndObj,ConstIndObj,MaxTermDepth,MaxTermSize>::OccNode OccNode;
   //typedef PathIndexTree<Alloc,IndObj,ConstIndObj,MaxTermDepth,MaxTermSize>::FuncNode FuncNode; 

  public:
   Destruction()
    : _backtrackFuncNodes(),
      _backtrackArgNums()
   {
   };

   bool reset(OccNode** tree,bool& constant) 
   { 
    _occAddr = tree;
    _tree = tree;
    constant = false;
    _lastConstNode = (FuncNode*)0;
    return *_occAddr;
   };
 
   void recycle() { destroyTree(*_tree); };

   bool next(bool& constant)
   {
    CALL("Removal::next(bool& constant)");  

    ASSERT(((_lastConstNode) ? (!constIndObj()) : (!indexedObj()))); // very strong assumption     

    FuncNode* nextFunc;
    if (_lastConstNode)  // was constant
     {
      ASSERT(_lastConstNode->isConstant());
      ASSERT(!_lastConstNode->indObj()); 
      nextFunc = _lastConstNode->altern();
      check_func_node:
       if (nextFunc)
        { 
         if (nextFunc->isConstant()) 
          {
           _lastConstNode = nextFunc;
           constant = true;
           return true;
          }
         else
          {
           for (ulong a = 0; a < nextFunc->arity(); a++)
  	    if (nextFunc->deeper()[a])
	     {
	      if (a + 1 < nextFunc->arity())  
	       { 
	        _backtrackFuncNodes.push(nextFunc);
                _backtrackArgNums.push(a + 1);                
               }
              else
	       if (nextFunc->altern()) 
                {
	 	 _backtrackFuncNodes.push(nextFunc->altern());
                 _backtrackArgNums.push(0);
                };
              _occAddr = nextFunc->deeper() + a;
              _lastConstNode = (FuncNode*)0;
              constant = false;
              return true; 
             };
           nextFunc = nextFunc->altern();
           goto check_func_node;   
          };
       }      
      else // try to backtrack
       {
        if (_backtrackFuncNodes.empty()) return false;     
        nextFunc = _backtrackFuncNodes.pop();  
        if (nextFunc->isConstant())
         {
          _backtrackArgNums.pop();
          _lastConstNode = nextFunc;
          constant = true;
          return true;
         };
        for (ulong a = _backtrackArgNums.pop(); a < nextFunc->arity(); a++) 
         if (nextFunc->deeper()[a])
 	  {
           if (a + 1 < nextFunc->arity())  
	    { 
	     _backtrackFuncNodes.push(nextFunc);
             _backtrackArgNums.push(a + 1);                
            }
           else
	    if (nextFunc->altern()) 
             {
	      _backtrackFuncNodes.push(nextFunc->altern());
              _backtrackArgNums.push(0);
             };
           _occAddr = nextFunc->deeper() + a;
	   _lastConstNode = (FuncNode*)0;
           constant = false;
           return true;
          };
        nextFunc = nextFunc->altern();
        goto check_func_node;
       };       
     }
    else // was not a constant 
     {
      nextFunc = (*_occAddr)->complex();
      goto check_func_node;
     };
   }; // bool next(bool& constant) 

   IndObj& indexedObj() const { CALL("Removal::indexedObj() const"); ASSERT(*_occAddr);  return (*_occAddr)->indObj(); };
   ConstIndObj& constIndObj() const { CALL("Removal::constIndObj() const"); ASSERT(_lastConstNode); return _lastConstNode->indObj(); }; 
 
  private:
   OccNode** _occAddr;
   OccNode** _tree;
   FuncNode* _lastConstNode;
   Stack<FuncNode*,MaxTermDepth> _backtrackFuncNodes;
   Stack<ulong,MaxTermDepth> _backtrackArgNums;
 }; // class Destruction

 public: // for class PathIndexTree
  static void destroyTree(FuncNode*& tree)
  {   
   CALL("destroyTree(FuncNode*& tree)");
   if (tree)
    {
     destroyTree(tree->altern());
     if (tree->isNonconst())
      {             
       for (ulong a = 0; a < tree->arity(); a++) 
        {
	 if (tree->deeper()[a])
	  {
	   ASSERT(!((tree->deeper()[a])->indObj()));
           destroyTree(tree->deeper()[a]);
          }; 
        };
      }
     else
      {       
       ASSERT(!(tree->indObj()));
      };
     delete tree;
     tree = (FuncNode*)0;
    };
  };

  static void destroyTree(OccNode*& tree)   
  {
   CALL("destroyTree(OccNode*& tree)");
   if (tree)   
    {
     ASSERT(!(tree->indObj()));
     destroyTree(tree->complex());     
     delete tree;
     tree = (OccNode*)0;
    };
  };
  
  static void recycleObsolete(FuncNode*& tree)
  {
   CALL("recycleObsolete(FuncNode*& tree)");
   if (tree) 
    {
     bool obsolete = true;
     if (tree->isConstant()) { obsolete = !tree->indObj(); }
     else
      {
       for (ulong a = 0; a < tree->arity(); a++) 
	{ 
	 recycleObsolete(tree->deeper()[a]);
         if (tree->deeper()[a]) obsolete = false; 
        };
      };
     if (obsolete) 
      { 
       FuncNode* tmp = tree;
       tree = tmp->altern();
       delete tmp;
      }; 
    };   
  }; // void recycleObsolete(FuncNode*& tree) 

  static void recycleObsolete(OccNode*& tree)
  {
   CALL("recycleObsolete(OccNode*& tree)");
   if (tree && (!tree->indObj()) && (!tree->complex()))
    {
     delete tree; 
     tree = (OccNode*)0;
    };                  
  }; // void recycleObsolete(OccNode*& tree)
}; // template <class Alloc,class IndObj,class ConstIndObj,ulong MaxTermDepth> class PathIndexTree



//======================================================================
#undef DEBUG_NAMESPACE
//======================================================================
#endif
