//
// File:         SharedKBConstr.h
// Description:  Representation of simple shared KB ordering constrains.
// Created:      May 8, 2000, 10:20
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#ifndef SHARED_KB_CONSTR_H
//====================================================
#define SHARED_KB_CONSTR_H
#ifndef NO_DEBUG_VIS
 #include <iostream.h>
#endif
#include "jargon.h"
#include "HellDebugFlags.h"
#ifdef DEBUG_ALLOC_OBJ_TYPE
 #include "ClassDesc.h"       
#endif
#ifndef NO_DEBUG
 #include "ObjDesc.h"
#endif
#include "HellConst.h"
#include "GlobAlloc.h"
#include "LinComb.h"
#include "term.h"   
#include "MultisetOfVariables.h"
#include "KBConstraint.h"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_SHARED_KB_CONSTR
 #define DEBUG_NAMESPACE "SharedKBConstr"
#endif 
#include "debugMacros.h"
//================================================= 

class SharedKBConstr 
{
 public: 
  typedef LinComb<GlobAlloc,HellConst::MaxNumOfVariables> WeightForm;
  typedef LinCombSharing<GlobAlloc,HellConst::MaxNumOfVariables> WeightFormSharing;
  typedef KBConstraint::Category Category; 
 public:
  void  operator delete(void* obj) 
  { 
   #ifdef DEBUG_ALLOC_OBJ_TYPE 
    ALWAYS(_classDesc.registerDeallocated(sizeof(SharedKBConstr)));
    GlobAlloc::deallocate(obj,sizeof(SharedKBConstr),&_classDesc);  
   #else
    GlobAlloc::deallocate(obj,sizeof(SharedKBConstr));   
   #endif
  };
  static ulong minSize() { return sizeof(SharedKBConstr); };
  static ulong maxSize() { return sizeof(SharedKBConstr); };
  const Category& category() const 
  {
   CALL("category() const");
   ASSERT(checkObj());
   return _category; 
  };
  const WeightForm* posWeightForm() const { return _posWeightForm; }; 
  const WeightForm* negWeightForm() const { return _negWeightForm; }; 
  const ulong& posFuncs() const { return _posConst; };
  const ulong& negFuncs() const { return _negConst; };
  const TERM* lhsTerm() const { return _lhsTerm; };
  const TERM* rhsTerm() const { return _rhsTerm; };
  const Comparison& lexComp() const { return _lexComp; };
  long refCounter() const { return _refCounter; };
  long& refCounter() { return _refCounter; }; 

  static void assumeGreater(const TERM* lhs,
                            MultisetOfVariables<HellConst::MaxNumOfVariables>& lhsVars,
                            ulong lhsFuncs,
                            const TERM* rhs,
                            MultisetOfVariables<HellConst::MaxNumOfVariables>& rhsVars,
                            ulong rhsFuncs);

  static const Category& collectedCategory() { return _raw.category(); };

  static SharedKBConstr* getShared();

  void destroy()
  {
   CALL("destroy()");
   switch (_category)
    {      
     case KBConstraint::NearlyTautology: 
      _weightFormSharing.remove(_posWeightForm);
      break; 
     case KBConstraint::NearlyContradiction: 
      _weightFormSharing.remove(_negWeightForm);
      break; 
     case KBConstraint::Volatile: 
      _weightFormSharing.remove(_posWeightForm);
      _weightFormSharing.remove(_negWeightForm);
      break; 
     case KBConstraint::PurelyLexicographic: return;
    };
  }; // void destroy()
 
  ~SharedKBConstr() { CALL("destructor ~SharedKBConstr()"); destroy(); };

  #ifndef NO_DEBUG_VIS
   static ostream& outputAssumed(ostream& str) { return _raw.output(str); };
   ostream& output(ostream& str) const
   {
    if (!this) return str << "Null";
    str << KBConstraint::spell(_category) << ' ';
    switch (_category)
     {    
      case KBConstraint::NearlyTautology: 
       str << _posWeightForm << " vs " << _negConst;
       break;
      case KBConstraint::NearlyContradiction:
       str << _posConst << " vs " <<  _negWeightForm;
       break;
      case KBConstraint::Volatile: 
       str << _posWeightForm << " vs " << _negWeightForm;
       break;
      case KBConstraint::PurelyLexicographic: 
       break;
     };
    str << " lex.comp.: ";
    if (_lexComp == Incomparable)
     {
      str << _lhsTerm << " vs " << _rhsTerm;
     }
    else str << _lexComp; 
    return str;
   };
  #endif  


  #ifndef NO_DEBUG 
   bool checkObj() const
   {
    #ifndef DEBUG_ALLOC
     return true;
    #else
     ObjDesc* objDesc = GlobAlloc::hostObj((void*)this);
     if (!objDesc) 
      { 
       DMSG("SharedKBConstr::checkObj()> Null pointer to object descriptor.\n");  
       return false; 
      };  
     bool res = true;
     #ifdef DEBUG_ALLOC_OBJ_SIZE
      if (objDesc->size !=  sizeof(SharedKBConstr)) 
       {
        DMSG("SharedKBConstr::checkObj()> objDesc->size !=  sizeof(SharedKBConstr)\n");
        res = false;
       };
     #endif
     #ifdef DEBUG_ALLOC_OBJ_STATUS
      if (objDesc->status != ObjDesc::Persistent) 
       {
        DMSG("SharedKBConstr::checkObj()> objDesc->status != ObjDesc::Persistent\n");
        res = false;
       };
     #endif
     #ifdef DEBUG_ALLOC_OBJ_TYPE
      if (objDesc->classDesc != &_classDesc) 
       {
        DMSG("SharedKBConstr::checkObj()> objDesc->classDesc != &_classDesc\n");
        res = false;
       };
     #endif
     if (!res) DOP(DMSG("SharedKBConstr::checkObj()> host object ") << *objDesc << "\n"); 
     return res;
    #endif
   };
  #endif

  static void destroySharingIndex();

 private:
  SharedKBConstr(const Category& cat,
                 WeightForm* posWF,
                 const ulong& negConst,
                 const TERM* lhs, 
                 const TERM* rhs,
                 const Comparison& lexComp) 
   : _category(cat),
     _lexComp(lexComp),
     _lhsTerm(lhs),
     _rhsTerm(rhs),
     _refCounter(0)  
  {
   CALL("SharedKBConstr(const Category& cat,WeightForm* posWF,const ulong& negConst,const Comparison& lexComp)");
   ASSERT(cat == KBConstraint::NearlyTautology); 
   _posWeightForm = posWF;
   _negConst = negConst;  
  };

  SharedKBConstr(const Category& cat,
                 WeightForm* posWF,
                 WeightForm* negWF,
                 const TERM* lhs, 
                 const TERM* rhs,
                 const Comparison& lexComp) 
   : _category(cat),
     _lexComp(lexComp),
     _lhsTerm(lhs),
     _rhsTerm(rhs),
     _refCounter(0)     
  {
   CALL("SharedKBConstr(const Category& cat,WeightForm* posWF,WeightForm* negWF,const Comparison& lexComp)");
   ASSERT(cat == KBConstraint::Volatile); 
   _posWeightForm = posWF;
   _negWeightForm = negWF;  
  };

  SharedKBConstr(const Category& cat,
                 const ulong& posConst,
                 WeightForm* negWF,
                 const TERM* lhs, 
                 const TERM* rhs,
                 const Comparison& lexComp) 
   : _category(cat),
     _lexComp(lexComp),
     _lhsTerm(lhs),
     _rhsTerm(rhs),
     _refCounter(0)     
  {
   CALL("SharedKBConstr(const Category& cat,const ulong& posConst,WeightForm* negWF,const Comparison& lexComp)");
   ASSERT(cat == KBConstraint::NearlyContradiction); 
   _posConst = posConst;
   _negWeightForm = negWF;  
  };
    
  SharedKBConstr(const Category& cat,
                 const TERM* lhs, 
                 const TERM* rhs) 
   : _category(cat),
     _lexComp(Incomparable),
     _lhsTerm(lhs),
     _rhsTerm(rhs),
     _refCounter(0)     
  {
   CALL("SharedKBConstr(const Category& cat,const ulong& posConst,WeightForm* negWF,const Comparison& lexComp)");
   ASSERT(cat == KBConstraint::PurelyLexicographic); 
  };    


  void* operator new(size_t) 
  {
   #ifdef DEBUG_ALLOC_OBJ_TYPE  
    ALWAYS(_classDesc.registerAllocated(sizeof(SharedKBConstr)));
    return GlobAlloc::allocate(sizeof(SharedKBConstr),&_classDesc); 
   #else
    return GlobAlloc::allocate(sizeof(SharedKBConstr)); 
   #endif
  };
  
  static SharedKBConstr* getSharedNearlyTautology()
  {
   CALL("getSharedNearlyTautology()");
   ASSERT(_raw.category() == KBConstraint::NearlyTautology);

   _weightFormSharing.resetIntegration();
   _raw.posVars().ResetIterator();
   for (const ulong* var = _raw.posVars().NextVar(); var; var = _raw.posVars().NextVar())
    _weightFormSharing.integrPlusVar(_raw.posVars().Occurences(*var),*var);             
   WeightForm* posWF = _weightFormSharing.integrate();
   return new SharedKBConstr(KBConstraint::NearlyTautology,posWF,_raw.negFuncs(),_lhs,_rhs,_raw.lexComp());
  };

  static SharedKBConstr* getSharedNearlyContradiction()
  {
   CALL("getSharedNearlyContradiction()");
   ASSERT(_raw.category() == KBConstraint::NearlyContradiction); 

   _weightFormSharing.resetIntegration();
   _raw.negVars().ResetIterator();
   for (const ulong* var = _raw.negVars().NextVar(); var; var = _raw.negVars().NextVar())
    _weightFormSharing.integrPlusVar(_raw.negVars().Occurences(*var),*var);             
   WeightForm* negWF = _weightFormSharing.integrate();
   return new SharedKBConstr(KBConstraint::NearlyContradiction,_raw.posFuncs(),negWF,_lhs,_rhs,_raw.lexComp());
  };
 
  static SharedKBConstr* getSharedVolatile()
  {
   CALL("getSharedNearlyVolatile()");
   ASSERT(_raw.category() == KBConstraint::Volatile);

   _weightFormSharing.resetIntegration();   
   if (_raw.posFuncs()) _weightFormSharing.integrPlusConst(_raw.posFuncs()); 
   _raw.posVars().ResetIterator(); 
   const ulong* var;
   for (var = _raw.posVars().NextVar(); var; var = _raw.posVars().NextVar())
    _weightFormSharing.integrPlusVar(_raw.posVars().Occurences(*var),*var);             
   WeightForm* posWF = _weightFormSharing.integrate();

   _weightFormSharing.resetIntegration();
   if (_raw.negFuncs()) _weightFormSharing.integrPlusConst(_raw.negFuncs()); 
   _raw.negVars().ResetIterator();
   for (var = _raw.negVars().NextVar(); var; var = _raw.negVars().NextVar())
    _weightFormSharing.integrPlusVar(_raw.negVars().Occurences(*var),*var);             
   WeightForm* negWF = _weightFormSharing.integrate();

   return new SharedKBConstr(KBConstraint::Volatile,posWF,negWF,_lhs,_rhs,_raw.lexComp());
  };

  static SharedKBConstr* getSharedPurelyLexicographic()
  {
   CALL("getSharedPurelyLexicographic()");
   ASSERT(_raw.category() == KBConstraint::PurelyLexicographic);
   return new SharedKBConstr(KBConstraint::PurelyLexicographic,_lhs,_rhs); 
  }; // SharedKBConstr* getSharedPurelyLexicographic()

 private:
  Category _category; 
  union
  {
   WeightForm* _posWeightForm;
   ulong _posConst;
  };

  union
  {
   WeightForm* _negWeightForm;
   ulong _negConst;
  };
  Comparison _lexComp;
  const TERM* _lhsTerm;
  const TERM* _rhsTerm;
  long _refCounter;
  #ifdef DEBUG_ALLOC_OBJ_TYPE
   static ClassDesc _classDesc;
  #endif
  static const TERM* _lhs;
  static const TERM* _rhs;
  static KBConstraint _raw;
  static WeightFormSharing _weightFormSharing;
}; // class SharedKBConst

#ifndef NO_DEBUG_VIS
inline ostream& operator<<(ostream& str,const SharedKBConstr& con)
{
 return con.output(str);
};
inline ostream& operator<<(ostream& str,const SharedKBConstr* con)
{
 return con->output(str);
};
#endif


//================================================= 
#endif
