//
// File:         Signature.h
// Description:  Various signature related operations.  
// Created:      Apr 20, 2001, 13:50
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#ifndef SIGNATURE_H
//=================================================
#define SIGNATURE_H
#include <stddef.h>
#include <iostream.h>
#include <string.h>
#include "jargon.h"
#include "HellDebugFlags.h"
#ifdef DEBUG_ALLOC_OBJ_TYPE 
 #include "ClassDesc.h"
#endif
#include "HellConst.h"
#include "Array.h"
#include "Malloc.h"
#include "Comparison.h"
//=================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_SIGNATURE
 #define DEBUG_NAMESPACE "Signature"
#endif 
#include "debugMacros.h"
//=================================================  
class Clause;
class Signature
{
 public:
  Signature() 
   : _inputVariableName(DOP("Signature::_inputVariableName")),
     _nextInputVariableNumber(0UL),
     _functorName(DOP("Signature::_functorName")),
     _isInputFunctor(DOP("Signature::_isInputFunctor")),
     _isPredicate(DOP("Signature::_isPredicate")),
     _isCommutative(DOP("Signature::_isCommutative")),
     _arity(DOP("Signature::_arity")),
     _nextFunctorNumber(0UL),
     _nextSkolemConstantSuffix(0UL),
     _nextSkolemFunctorSuffix(0UL),
     _nextSkolemPredicateSuffix(0UL),
     #ifdef DEBUG_NAMESPACE
      _positiveHeaderPrecedence(DefaultHeaderPrecedence,"Signature::_positiveHeaderPrecedence"),  
      _negativeHeaderPrecedence(DefaultHeaderPrecedence,"Signature::_negativeHeaderPrecedence"),
     #else
      _positiveHeaderPrecedence(DefaultHeaderPrecedence,0),  
      _negativeHeaderPrecedence(DefaultHeaderPrecedence,0),
     #endif
     _minimalHeaderPrecedence(DefaultHeaderPrecedence),
     _maximalHeaderPrecedence(DefaultHeaderPrecedence)
  {
   CALL("constructor Signature()");
   ASSERT(PositivePolarity == 0UL);
   ASSERT(NegativePolarity == 1UL);
   ASSERT(DefaultHeaderPrecedence == 0L);
   ASSERT(MinimalPossibleHeaderPrecedence == (LONG_MIN/4)); 
   ASSERT(MaximalPosiibleHeaderPrecedence == (LONG_MAX/4));

   _functorName[HellConst::UnordEqNum] = "equal"; // equality predicate in the TPTP format 
   _isInputFunctor[HellConst::UnordEqNum] = true;
   _isPredicate[HellConst::UnordEqNum] = true;
   _isCommutative[HellConst::UnordEqNum] = true;           
   _arity[HellConst::UnordEqNum] = 2UL;
   if (HellConst::UnordEqNum >= _nextFunctorNumber) _nextFunctorNumber = HellConst::UnordEqNum + 1;
  
   _functorName[HellConst::OrdEqNum] = "Equal"; // can not be an input functor in the TPTP format 
   _isInputFunctor[HellConst::OrdEqNum] = true;
   _isPredicate[HellConst::OrdEqNum] = true; 
   _isCommutative[HellConst::OrdEqNum] = true; 
   _arity[HellConst::OrdEqNum] = 2UL; 
   if (HellConst::OrdEqNum >= _nextFunctorNumber) _nextFunctorNumber = HellConst::OrdEqNum + 1; 

   _functorName[HellConst::ReservedPred0] = "ReservedPred0"; // can not be an input functor in the TPTP format 
   _isPredicate[HellConst::ReservedPred0] = true; 
   _arity[HellConst::ReservedPred0] = 2UL; 
   if (HellConst::ReservedPred0 >= _nextFunctorNumber) _nextFunctorNumber = HellConst::ReservedPred0 + 1; 

   _functorName[HellConst::ReservedPred1] = "ReservedPred1"; // can not be an input functor in the TPTP format 
   _isPredicate[HellConst::ReservedPred1] = true; 
   _arity[HellConst::ReservedPred1] = 2UL; 
   if (HellConst::ReservedPred1 >= _nextFunctorNumber) _nextFunctorNumber = HellConst::ReservedPred1 + 1; 

   _functorName[HellConst::ReservedPred2] = "ReservedPred2"; // can not be an input functor in the TPTP format
   _isPredicate[HellConst::ReservedPred2] = true;
   _arity[HellConst::ReservedPred2] = 2UL; 
   if (HellConst::ReservedPred2 >= _nextFunctorNumber) _nextFunctorNumber = HellConst::ReservedPred2 + 1; 

   // Fill the "holes" between the reserved functors
   
   for (ulong fn = 0; fn < _nextFunctorNumber; fn++) 
    if (!reservedFunctorNumber(fn)) _functorName[fn] = "NoFunctor"; // can not be an input functor in the TPTP format  

   // Create canonical names for all variables
   for (ulong vn = 0; vn < HellConst::MaxNumOfVariables; vn++)
    {
     _canonicalVariableName[vn] = createCanonicalVariableName(vn);
    };
  };
  
  ~Signature() 
  {
   CALL("destructor ~Signature()");
  };
 
  // Registering new symbols: 

  void resetInputVariables() 
  { 
   CALL("resetInputVariables()");
   while (_nextInputVariableNumber)  
    {
     _nextInputVariableNumber--; 
     delete [] (_inputVariableName[_nextInputVariableNumber]);
    };
  };
  ulong inputVariableNumber(const char* varName)
  {
   CALL("inputVariableNumber(const char* varName)");
   for (ulong vn = 0; vn < _nextInputVariableNumber; vn++)
    if (!(strcmp(varName,_inputVariableName[vn]))) return vn;
   _inputVariableName[_nextInputVariableNumber] = new char[strlen(varName) + 1];
   strcpy(_inputVariableName[_nextInputVariableNumber],varName);
   _nextInputVariableNumber++;
   return _nextInputVariableNumber - 1; 
  };
  ulong inputFunctorNumber(const char* funName)
  {
   CALL("inputFunctorNumber(const char* funName)");
   for (ulong fn = 0; fn < _nextFunctorNumber; fn++)
    if (!(strcmp(funName,_functorName[fn]))) return fn;
   _functorName[_nextFunctorNumber] = new char[strlen(funName) + 1];
   strcpy(_functorName[_nextFunctorNumber],funName);
   _isInputFunctor[_nextFunctorNumber] = true;
   _isPredicate[_nextFunctorNumber] = false;   
   _isCommutative[_nextFunctorNumber] = false; 
   _arity[_nextFunctorNumber] = 0UL;
   _nextFunctorNumber++;
   return _nextFunctorNumber - 1;  
  };

  ulong inputFunctorNumber(const char* funName,ulong arity)
  {
   CALL("inputFunctorNumber(const char* funName,ulong arity)");
   ulong res = inputFunctorNumber(funName); 
   _arity[res] = arity;
   return res;
  };

  // Getting symbol properties: 

  bool isPredicate(ulong funNum) { return _isPredicate[funNum]; };     
  ulong arity(ulong funNum) { return _arity[funNum]; };  
  bool isSkolem(ulong funNum) { return !_isInputFunctor[funNum]; };   
  bool isCommutative(ulong funNum) { return _isCommutative[funNum]; };
  const char* functorName(ulong funNum) { return _functorName[funNum]; };  

  // Setting/changing symbol properties:
   
  void setFunctorProperties(ulong funNum,ulong arity,bool isPredicate)
  {
   CALL("setFunctorProperties(ulong funNum,ulong arity,bool isPredicate)");
   _arity[funNum] = arity;
   _isPredicate[funNum] = isPredicate;
  };
  void makePredicate(ulong funNum) 
  {
   CALL("makePredicate(ulong funNum)"); 
   _isPredicate[funNum] = true;
  };
  void makeNonpredicate(ulong funNum)
  {
   CALL("makeNonpredicate(ulong funNum)");
   _isPredicate[funNum] = false;
  };
  void makeCommutative(ulong funNum)
  {
   CALL("makeCommutative(ulong funNum)");
   _isCommutative[funNum] = true;
  }; 
  void makeNoncommutative(ulong funNum)
  {
   CALL("makeNoncommutative(ulong funNum)");
   _isCommutative[funNum] = false;
  };
  void setCommutativity(const Clause* const *commutativityTable,const Clause* const *symmetryTable)
  {
   CALL("setCommutativity(const Clause* const *commutativityTable,const Clause* const *symmetryTable)");
   for (ulong fn = 0; fn < _nextFunctorNumber; fn++)     
    _isCommutative[fn] = (commutativityTable && commutativityTable[fn]) || (symmetryTable && symmetryTable[fn]);
  };
  
  // Operations on the signature as a whole:

  ulong nextFunctorNumber() const { return _nextFunctorNumber; };
  ulong newSkolemConstantNumber()
  {
   CALL("newSkolemConstantNumber()");
   static char tmpString[256]; 
   strcpy(tmpString,HellConst::SkolemConstantNamePrefix);
   char* endOfPrefix = tmpString + strlen(HellConst::SkolemConstantNamePrefix);
   char* suffix;
   try_suffix:
    suffix = endOfPrefix; 
    print(_nextSkolemConstantSuffix,suffix);
    suffix[0] = '\0';
    if (isRegisteredFunctorName(tmpString)) 
     {
      // a function with this name already exists, try bigger suffix  
      _nextSkolemConstantSuffix++;
      goto try_suffix;
     };
    // this candidate is OK, register it
   
   _functorName[_nextFunctorNumber] = new char[strlen(tmpString) + 1];
   strcpy(_functorName[_nextFunctorNumber],tmpString);
   _isInputFunctor[_nextFunctorNumber] = false;
   _isPredicate[_nextFunctorNumber] = false;   
   _isCommutative[_nextFunctorNumber] = false; 
   _arity[_nextFunctorNumber] = 0UL;
   _nextFunctorNumber++;
   return _nextFunctorNumber - 1;  
  }; // ulong newSkolemConstantNumber()
  
  ulong newSkolemFunctorNumber(ulong arity)
  {
   CALL("newSkolemFunctorNumber(arity)");
   static char tmpString[256]; 
   strcpy(tmpString,HellConst::SkolemFunctorNamePrefix);
   char* endOfPrefix = tmpString + strlen(HellConst::SkolemFunctorNamePrefix);
   char* suffix;
   try_suffix:
    suffix = endOfPrefix; 
    print(_nextSkolemFunctorSuffix,suffix);
    suffix[0] = '\0';
    if (isRegisteredFunctorName(tmpString)) 
     {
      // a function with this name already exists, try bigger suffix  
      _nextSkolemFunctorSuffix++;
      goto try_suffix;
     };
    // this candidate is OK, register it
   
   _functorName[_nextFunctorNumber] = new char[strlen(tmpString) + 1];
   strcpy(_functorName[_nextFunctorNumber],tmpString);
   _isInputFunctor[_nextFunctorNumber] = false;
   _isPredicate[_nextFunctorNumber] = false;   
   _isCommutative[_nextFunctorNumber] = false; 
   _arity[_nextFunctorNumber] = arity;
   _nextFunctorNumber++;
   return _nextFunctorNumber - 1;  
  }; // ulong newSkolemFunctorNumber(ulong arity)

  ulong newSkolemPredicateNumber(ulong arity)
  {
   CALL("newSkolemPredicateNumber(arity)");
   static char tmpString[256]; 
   strcpy(tmpString,HellConst::SkolemPredicateNamePrefix);
   char* endOfPrefix = tmpString + strlen(HellConst::SkolemPredicateNamePrefix);
   char* suffix;
   try_suffix:
    suffix = endOfPrefix; 
    print(_nextSkolemPredicateSuffix,suffix);
    suffix[0] = '\0';
    if (isRegisteredFunctorName(tmpString)) 
     {
      // a function with this name already exists, try bigger suffix  
      _nextSkolemPredicateSuffix++;
      goto try_suffix;
     }; 
    // this candidate is OK, register it
   
   _functorName[_nextFunctorNumber] = new char[strlen(tmpString) + 1];
   strcpy(_functorName[_nextFunctorNumber],tmpString);
   _isInputFunctor[_nextFunctorNumber] = false;
   _isPredicate[_nextFunctorNumber] = true;   
   _isCommutative[_nextFunctorNumber] = false; 
   _arity[_nextFunctorNumber] = arity;
   _nextFunctorNumber++;
   return _nextFunctorNumber - 1;  
  }; // ulong newSkolemPredicateNumber(ulong arity)
 
  const char* canonicalVariableName(ulong varNum)
  {
   CALL("canonicalVariableName(ulong varNum)");
   return _canonicalVariableName[varNum];
  };

  // Weights and precedences of functors and headers

  void setHeaderPrecedence(ulong polarity,ulong predNum,long precedence)
  {
   CALL("setHeaderPrecedence(ulong polarity,ulong predNum,long precedence)");   
   ASSERT((polarity == PositivePolarity) || (polarity == NegativePolarity)); 
   if (polarity) { _negativeHeaderPrecedence[predNum] = precedence; }
   else _positiveHeaderPrecedence[predNum] = precedence;
  };
  
  void setMinimalPrecedence(ulong polarity,ulong predNum)
  {
   CALL("setMinimalPrecedence(ulong polarity,ulong predNum)");
   ASSERT((polarity == PositivePolarity) || (polarity == NegativePolarity));     
   _minimalHeaderPrecedence--;
   setHeaderPrecedence(polarity,predNum,_minimalHeaderPrecedence);
  };

  void setMaximalPrecedence(ulong polarity,ulong predNum)
  {
   CALL("setMaximalPrecedence(ulong polarity,ulong predNum)");
   ASSERT((polarity == PositivePolarity) || (polarity == NegativePolarity));     
   _maximalHeaderPrecedence++;
   setHeaderPrecedence(polarity,predNum,_maximalHeaderPrecedence);
  };

  long headerPrecedence(ulong polarity,ulong predNum)
  {
   CALL("headerPrecedence(ulong polarity,ulong predNum)");
   ASSERT((polarity == PositivePolarity) || (polarity == NegativePolarity)); 
   return (polarity) ? _negativeHeaderPrecedence[predNum] : _positiveHeaderPrecedence[predNum];
  };
  
  Comparison compareHeaderPrecedences(ulong polarity1,ulong predNum1,ulong polarity2,ulong predNum2)
  {
   CALL("compareHeaderPrecedences(ulong polarity1,ulong predNum1,ulong polarity2,ulong predNum2)");
   ASSERT((polarity1 == PositivePolarity) || (polarity1 == NegativePolarity));
   ASSERT((polarity2 == PositivePolarity) || (polarity2 == NegativePolarity));  
   long precedenceDifference = headerPrecedence(polarity1,predNum1) - headerPrecedence(polarity2,predNum2);
   if (precedenceDifference)
    { 
     return (precedenceDifference > 0) ? Greater : Less; 
    }
   else return Equal;
  };      

  // Output for debugging
  
  #ifndef NO_DEBUG
   ostream& output(ostream& str)
   { 
    str << "%=================== Signature: =====================\n";
    for (ulong fn = 0; fn < _nextFunctorNumber; fn++)
     {
      str << fn << ": " << _functorName[fn] << "/" << _arity[fn] << " ";
      if (_isInputFunctor[fn]) { str << "input  "; } else str << "skolem ";
      if (_isPredicate[fn]) 
       { 
        str << "predicate    ";
        if (_isCommutative[fn]) { str << "symmetric      "; } else str << "nonsymmetric  "; 
       } 
      else
       {
        str << "nonpredicate ";
        if (_isCommutative[fn]) { str << "commutative    "; } else str << "noncommutative "; 
       };
      str << "\n";
     };
    for (ulong fn = 0; fn < _nextFunctorNumber; fn++)
     if (_isPredicate[fn]) 
      { 
       str << "precedence(++" << _functorName[fn] << ") = " << headerPrecedence(PositivePolarity,fn) << ",     "
           << "precedence(--" << _functorName[fn] << ") = " << headerPrecedence(NegativePolarity,fn) << "\n";
      } 
     else
      { 
       str << "precedence(" << _functorName[fn] << ") = " << headerPrecedence(PositivePolarity,fn) << "\n";
      };
    return str << "%============ End of signature. =================\n";
   };
  #endif

 public:
  static const ulong PositivePolarity = 0UL;
  static const ulong NegativePolarity = 1UL;
  static const long DefaultHeaderPrecedence = 0L;
  static const long MinimalPossibleHeaderPrecedence = LONG_MIN/4; 
  static const long MaximalPosiibleHeaderPrecedence = LONG_MAX/4;  

 private:
  typedef Array<Malloc,char*,32,Signature> StrArray;
  typedef Array<Malloc,bool,32,Signature> BoolArray;
  typedef Array<Malloc,ulong,32,Signature> ULongArray;
  typedef Array<Malloc,long,32,Signature> LongArray;
 private: 
  bool reservedFunctorNumber(ulong funNum) 
  {
   return (funNum == HellConst::UnordEqNum) 
          || (funNum == HellConst::OrdEqNum) 
          || (funNum == HellConst::ReservedPred0)  
          || (funNum == HellConst::ReservedPred1)  
          || (funNum == HellConst::ReservedPred2) 
          ;                   
  };
  static void print(ulong num,char*& target)
  {
   if (!num) { target[0] = '0'; target++; return; }; 
   char lastDigit;
   switch (num % 10) 
    {   
     case 0UL: lastDigit = '0'; break;
     case 1UL: lastDigit = '1'; break;
     case 2UL: lastDigit = '2'; break;   
     case 3UL: lastDigit = '3'; break;
     case 4UL: lastDigit = '4'; break;
     case 5UL: lastDigit = '5'; break;   
     case 6UL: lastDigit = '6'; break;
     case 7UL: lastDigit = '7'; break;
     case 8UL: lastDigit = '8'; break;
     case 9UL: lastDigit = '9'; break;
     #ifdef DEBUG_NAMESPACE
      default: ICP("ICP0"); return; 
     #else
      #ifdef _SUPPRESS_WARNINGS_
       default: return;  
      #endif
     #endif 
    };
   num = num / 10;     
   if (num) print(num,target);
   target[0] = lastDigit;
   target++;
  }; // void print(ulong num,char*& target) 
 
  bool isRegisteredFunctorName(const char* str)
  {
   CALL("isRegisteredFunctorName(const char* str)");
   for (ulong fn = 0; fn < _nextFunctorNumber; fn++)
    if (!(strcmp(str,_functorName[fn]))) return true;
   return false;  
  }; // bool isRegisteredFunctorName(const char* str)

  static const char* createCanonicalVariableName(ulong varNum)
  {
   CALL("createCanonicalVariableName(ulong varNum)");
   static char tmpString[256];
   strcpy(tmpString,HellConst::CanonicalVariableNamePrefix);
   char* cursor = tmpString + strlen(HellConst::CanonicalVariableNamePrefix);
   print(varNum,cursor);
   cursor[0] = '\0';
   char* res = new char[strlen(tmpString) + 1]; 
   strcpy(res,tmpString);
   return res;
  };

 private:
  StrArray _inputVariableName;
  ulong _nextInputVariableNumber; 
  
  StrArray _functorName;
  BoolArray _isInputFunctor;
  BoolArray _isPredicate;
  BoolArray _isCommutative;
  ULongArray _arity;
  ulong _nextFunctorNumber;
 
  ulong _nextSkolemConstantSuffix;
  ulong _nextSkolemFunctorSuffix;
  ulong _nextSkolemPredicateSuffix;  

  const char* _canonicalVariableName[HellConst::MaxNumOfVariables];

  LongArray _positiveHeaderPrecedence;  
  LongArray _negativeHeaderPrecedence;
  long _minimalHeaderPrecedence;
  long _maximalHeaderPrecedence;
}; 


//================================================
#undef DEBUG_NAMESPACE   
//==================================================
#endif
