//
// File:         UnifKBConstr.h
// Description:  Checking KB ordering constraints.
// Created:      May 19, 2000, 19:00
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#ifndef UNIF_KB_CONSTR_H
//====================================================
#define UNIF_KB_CONSTR_H
#include "jargon.h"
#include "HellDebugFlags.h"
#include "HellConst.h"
#include "KBConstraint.h"
#include "SharedKBConstr.h"
#include "prefix.h"
#include "term.h"
#include "variables.h"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_UNIF_KB_CONSTR
 #define DEBUG_NAMESPACE "UnifKBConstr"
#endif
#include "debugMacros.h"
//=================================================

class UnifKBConstr
{
 public:
  UnifKBConstr() : _constraintInstance(), _constrIntVarRenaming() {};
  bool checkConstr(const SharedKBConstr* constr,ulong index);
  bool checkNearlyTautology(const SharedKBConstr* constr,ulong index);
  bool checkNearlyContradiction(const SharedKBConstr* constr,ulong index);
  bool checkVolatile(const SharedKBConstr* constr,ulong index); 
  bool checkLexicographically(const TERM* lhs,const TERM* rhs,ulong index);  
 private:
  KBConstraint _constraintInstance;
  VAR_TO_INT_RENAMING _constrIntVarRenaming;
}; // class UnifKBConstr

inline bool UnifKBConstr::checkLexicographically(const TERM* lhs,const TERM* rhs,ulong index)
{
 CALL("checkLexicographically(const TERM* lhs,constr TERM* rhs,ulong index)");
 static ST_TERM_WITH_SUBST lhsIter;
 static ST_TERM_WITH_SUBST rhsIter;
 lhsIter.Reset(index,(TERM*)lhs);
 lhsIter.Reset(index,(TERM*)rhs);
 PREF_SYM lhsSym; 
 PREF_SYM rhsSym; 
 check_iter:
  if (lhsIter.Empty()) return false; // equal terms
  lhsSym = lhsIter.CurrSym();
  rhsSym = rhsIter.CurrSym();
  if (lhsSym.isVariable())
   {
    if (rhsSym.isVariable())
     {
      if (lhsSym == rhsSym)
       { 
	lhsIter.SkipSym();
        rhsIter.SkipSym();
        goto check_iter;
       };       
     };
    return false; // incomparable
   }
  else
   if (rhsSym.isVariable())
    {
     return false; // incomparable
    }
   else // both are functors
    {
     if (lhsSym == rhsSym)
      { 
       lhsIter.SkipSym();
       rhsIter.SkipSym();
       goto check_iter;
      };      
     bool res = lhsSym.Func().greaterPrecedence(rhsSym.Func()); 
     return res;
    };
}; // bool UnifKBConstr::checkLexicographically(const TERM* lhs,constr TERM* rhs,ulong index)

inline bool UnifKBConstr::checkConstr(const SharedKBConstr* constr,ulong index)
{
 CALL("checkConstr(const SharedKBConstr* constr,ulong index)");

 ASSERT(constr);
 switch (constr->category())
  {
   case KBConstraint::NearlyTautology: return checkNearlyTautology(constr,index);            
   case KBConstraint::NearlyContradiction: return checkNearlyContradiction(constr,index);
   case KBConstraint::Volatile: return checkVolatile(constr,index);
   case KBConstraint::PurelyLexicographic: return checkLexicographically(constr->lhsTerm(),constr->rhsTerm(),index);
  };
 #ifdef DEBUG_NAMESPACE
  ICP("ICP0"); return false;
 #else
  #ifdef _SUPPRESS_WARNINGS_
   return false;
  #endif
 #endif 
}; // bool UnifKBConstr::checkConstr(const SharedKBConstr* constr,ulong index)

inline bool UnifKBConstr::checkNearlyTautology(const SharedKBConstr* constr,ulong index)
{
 CALL("checkNearlyTautology(const SharedKBConstr* constr,ulong index)");

 const SharedKBConstr::WeightForm* posWF = constr->posWeightForm();
 ASSERT(posWF);
 ulong negFuncs = constr->negFuncs(); 
 ulong posWeight = 0;
 SharedKBConstr::WeightForm::Member mem;
 do 
  {
   mem = posWF->hd();
   ASSERT(mem.tag() == SharedKBConstr::WeightForm::PlusVar);

   VAR* var = VAR::Var(mem.var(),index);
   bool free_var;
   void* bind = var->Unref(free_var);
   if (free_var) 
    {
     posWeight += mem.coeff();
    } 
   else posWeight += mem.coeff() * PREFIX::WeightWithSubst((const PREFIX_SYM*)bind);

   if (posWeight > negFuncs) 
    {
      //DF; cout << "SUCC FAST\n";
     return true;
    };
   posWF = posWF->tl();
  } 
 while (posWF);

 if (posWeight == negFuncs)
  {
   // weights are equal, check lexocographically
   if (constr->lexComp() == Greater) return true;
   if (constr->lexComp() == Less) return false;
   return checkLexicographically(constr->lhsTerm(),constr->rhsTerm(),index); 
  }; 

 return false;
}; // bool UnifKBConstr::checkNearlyTautology(const SharedKBConstr* constr,ulong index)

inline bool UnifKBConstr::checkNearlyContradiction(const SharedKBConstr* constr,ulong index)
{
 CALL("checkNearlyContradiction(const SharedKBConstr* constr,ulong index)");
 const SharedKBConstr::WeightForm* negWF = constr->negWeightForm();
 ASSERT(negWF);
 ulong posFuncs = constr->posFuncs(); 
 ulong negWeight = 0;
 SharedKBConstr::WeightForm::Member mem;
 do 
  {
   mem = negWF->hd();
   ASSERT(mem.tag() == SharedKBConstr::WeightForm::PlusVar);

   VAR* var = VAR::Var(mem.var(),index);
   bool free_var;
   void* bind = var->Unref(free_var);
   if (free_var) 
    {
      //DF; cout << "FAIL FREE VAR\n";
     return false;
    };
   ulong weight;
   if (!PREFIX::GroundnessAndWeight((const PREFIX_SYM*)bind,weight)) 
    {
      //DF; cout << "FAIL NONGROUNDNESS\n";
     return false;     
    };

   negWeight += mem.coeff() * weight;
   if (negWeight > posFuncs)
    { 
      //DF; cout << "FAIL WEIGHT\n";
     return false;
    };
   negWF = negWF->tl();
  }
 while (negWF);

 if (negWeight < posFuncs) 
  {
    //DF; cout << "        SUCC WEIGHT\n";
   return true;
  };

 ASSERT(negWeight == posFuncs);
 // check lexicorgaphically
 if (constr->lexComp() == Greater)
  {
    //DF; cout << "        SUCC LEX\n";  
   return true;
  }
 else
  if (constr->lexComp() == Less) { return false; }
  else 
   {
    return checkLexicographically(constr->lhsTerm(),constr->rhsTerm(),index);
   };
}; // bool UnifKBConstr::checkNearlyContradiction(const SharedKBConstr* constr,ulong index)

inline bool UnifKBConstr::checkVolatile(const SharedKBConstr* constr,ulong index)   
{
 CALL("checkVolatile(const SharedKBConstr* constr,ulong index)");
 const SharedKBConstr::WeightForm* posWF = constr->posWeightForm();
 const SharedKBConstr::WeightForm* negWF = constr->negWeightForm();
 ASSERT(posWF);
 ASSERT(negWF);
 SharedKBConstr::WeightForm::Member mem;
 _constrIntVarRenaming.Reset();
 ulong substCoeff;
 PREFIX::ITER iter;

 // collecting instance of the positive part
 _constraintInstance.initPosSubst(0UL); 
 do 
  {
   mem = posWF->hd();
   if (mem.tag() == SharedKBConstr::WeightForm::PlusConst)
    {
     _constraintInstance.posSubstFunc(mem.constant());
    }
   else
    {
     ASSERT(mem.tag() == SharedKBConstr::WeightForm::PlusVar);
     substCoeff = mem.coeff();

     VAR* var = VAR::Var(mem.var(),index);
     bool free_var;
     void* bind = var->Unref(free_var);
     if (free_var) 
      {
       _constraintInstance.posSubstVar(_constrIntVarRenaming.Rename((VAR*)bind),substCoeff);
      }
     else
      {
       iter.Reset((PREFIX_SYM*)bind);
       do
	{
	 if (iter.CurrVar())
	  {
           _constraintInstance.posSubstVar(_constrIntVarRenaming.Rename(iter.CurrVar()),substCoeff); 
          }
         else
	  {
           _constraintInstance.posSubstFunc(substCoeff); 
          };
         iter.SkipSym();
        }
       while (iter.NotEmpty());
      };

    };
   posWF = posWF->tl();
  }
 while (posWF); 

 // collecting instance of the negative part

 _constraintInstance.initNegSubst(0UL);
 do 
  {
   mem = negWF->hd();
   if (mem.tag() == SharedKBConstr::WeightForm::PlusConst)
    {
     if (!_constraintInstance.negSubstFunc(mem.constant()))
      {
	//DF; cout << "FAIL WEIGHT 1\n";   
       return false;
      };
    }
   else
    {
     ASSERT(mem.tag() == SharedKBConstr::WeightForm::PlusVar);
     substCoeff = mem.coeff();

     VAR* var = VAR::Var(mem.var(),index);
     bool free_var;
     void* bind = var->Unref(free_var);

     if (free_var) 
      {
       if (!_constraintInstance.negSubstVar(_constrIntVarRenaming.Rename((VAR*)bind),substCoeff))
	{
	  //DF; cout << "FAIL VAR 1\n";   
         return false;
        };
      }
     else
      {
       iter.Reset((PREFIX_SYM*)bind);
       do
	{
	 if (iter.CurrVar())
	  {
	   if (!_constraintInstance.negSubstVar(_constrIntVarRenaming.Rename(iter.CurrVar()),substCoeff))
            {
	      //DF; cout << "FAIL VAR 2\n";   
             return false;
            };
          }
         else
	  {
	   if (!_constraintInstance.negSubstFunc(substCoeff))
	    {
	      //DF; cout << "FAIL WEIGHT 2\n";   
             return false;             
            };
          };
         iter.SkipSym();
        }
       while (iter.NotEmpty());
      };
    };
   negWF = negWF->tl();
  }
 while (negWF);   

 // both positive and negative parts are collected
 //DF; _constraintInstance.outputInstance(cout << "                        CONSTR INST ") << "\n";   
 ASSERT(_constraintInstance.posSubstWeight() >= _constraintInstance.negSubstWeight());
 if (_constraintInstance.posSubstWeight() == _constraintInstance.negSubstWeight())
  {
    //DF; cout << "FAIL LEX\n";     
   if (constr->lexComp() == Greater) return true;
   if (constr->lexComp() == Less) return false;
   return checkLexicographically(constr->lhsTerm(),constr->rhsTerm(),index);   
  };

 //DF; cout << "                    SUCC\n"; 
 return true;  

}; //bool UnifKBConstr::checkVolatile(const SharedKBConstr* constr,ulong index)  



//================================================
#undef DEBUG_NAMESPACE
//=================================================
#endif
