//
// File:         clause.h
// Description:  Representation of stored clauses.
// Created:      Feb 13, 2000, 20:30
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#ifndef CLAUSE_H
//===========================================================================
#define CLAUSE_H
#include <iostream.h>
#include <stddef.h>
#include "jargon.h"
#include "HellDebugFlags.h"
#ifdef DEBUG_ALLOC_OBJ_TYPE
 #include "ClassDesc.h"       
#endif
#include "GlobAlloc.h"
#include "GList.h"
#include "BitWord.h"
#include "term.h"
#include "ClauseBackground.h"
#include "PrologTerm.h"
#include "LiteralList.h"
//===========================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CLAUSE
 #define DEBUG_NAMESPACE "Clause"
#endif 
#include "debugMacros.h"
//=================================================                            
class Clause     
{
 public:
  class PrintRep;
  class Traversal;
  class OpenUnshared; 
  class InitStatic
  {
   public: 
    InitStatic() 
    {
     if (!_count) { Clause::initStatic(); _count++; };    
    };
    ~InitStatic() {};
   private:
    long _count;
  };
 private:
  class ClauseList {};  
 public: 
  typedef GList<GlobAlloc,Clause*,ClauseList> List;
  enum MainSet
   {
    MainSetFree = 0UL, // fixed 
    MainSetNew = 1UL,
    MainSetPassive = 2UL,
    MainSetReservedPassive = 3UL, 
    MainSetActive = 4UL,
    MainSetHyperRules = 5UL,
    MainSetDefinitions = 6UL,

    // end of active sets

    MainSetTrash = 7UL,
    MainSetHistory = 8UL
   };
  enum FlagNumber
   {
    MainSetPieceBegin = 0UL,
    MainSetPieceEnd = 3UL,

    FlagInClauseIndex = 4UL,
    FlagLiteralsShared = 5UL,
    FlagInForwardSubsumptionIndex = 6UL,
    FlagInBackwardSubsumptionIndex = 7UL,    
    FlagInForwardDemodulationIndex = 8UL,
    FlagInBackDemodulationIndex = 9UL, 
    FlagInInferenceDispatcherAsActive = 10UL,
    FlagInInferenceDispatcherAsDefinition = 11UL,
    FlagInClauseNumIndex = 12UL,
    FlagInQueueForBackSimplification = 13UL,

    FlagHasChildren = 30UL,
    FlagIsBlocked = 31UL          
   };
 public:
  static void initStatic()
  {
   _selectedPosEqPenaltyCoefficient = 1.0;
  };
  static void setSelectedPosEqPenaltyCoefficient(float coeff) 
  {
   _selectedPosEqPenaltyCoefficient = coeff;
   if (_selectedPosEqPenaltyCoefficient < 1.0) _selectedPosEqPenaltyCoefficient = 1.0;
   _useSelectedPosEqPenaltyCoefficient = (_selectedPosEqPenaltyCoefficient != 1.0); 
  };

  Clause(LiteralList* lits) : litList (lits), _flags(), num_of_pos_eq(0) {};

   // for making arrays of dummy clauses in KEPT 
  Clause() // : next(this), previous(this) 
  {
   next = this;
   previous = this;
  };

  void* Clause::operator new(size_t)
  {
   CALL("operator new(size_t)");
   #ifdef DEBUG_ALLOC_OBJ_TYPE 
    ALWAYS(_classDesc.registerAllocated(sizeof(Clause)));
    return GlobAlloc::allocate(sizeof(Clause),&_classDesc);
   #else
    return GlobAlloc::allocate(sizeof(Clause));
   #endif
  };
  void Clause::operator delete(void* obj)
  { 
   CALL("operator delete(void* obj)");
   #ifdef DEBUG_ALLOC_OBJ_TYPE
    ALWAYS(_classDesc.registerDeallocated(sizeof(Clause)));
    GlobAlloc::deallocate(obj,sizeof(Clause),&_classDesc);
   #else
    GlobAlloc::deallocate(obj,sizeof(Clause));
   #endif
  };

  void clearFlags()
  {
   _flags.clear();
  }; 
  void setMainSet(const MainSet& ms) 
  {
   CALL("setMainSet(const MainSet& ms)");
   ASSERT(!(ms >> 4));   
   _flags.setPiece(MainSetPieceBegin,MainSetPieceEnd,ms);      
   ASSERT(mainSet() == ms);
  };
  MainSet mainSet() const { return (MainSet)_flags.piece(MainSetPieceBegin,MainSetPieceEnd); };
 
  bool flag(const FlagNumber& fn) const { return _flags.bit(fn); };
  
  void setFlag(const FlagNumber& fn) { _flags.set(fn); };
  void clearFlag(const FlagNumber& fn) { _flags.clear(fn); };

  #ifndef NO_DEBUG 
   bool checkObj() const
   {
    #ifndef DEBUG_ALLOC
     return true;
    #else 
     ObjDesc* objDesc = GlobAlloc::hostObj((void*)this);
     if (!objDesc) 
      { 
       DMSG("Clause::checkObj()> Null pointer to object descriptor.\n");  
       return false; 
      }; 
     bool res = true;   
     #ifdef DEBUG_ALLOC_OBJ_SIZE
      if (objDesc->size !=  sizeof(Clause)) 
       {
        DMSG("Clause::checkObj()> objDesc->size !=  sizeof(Clause)\n");
        res = false;
       };
     #endif
     #ifdef DEBUG_ALLOC_OBJ_STATUS
      if (objDesc->status != ObjDesc::Persistent) 
       {
        DMSG("Clause::checkObj()> objDesc->status != ObjDesc::Persistent\n");
        DMSG("Clause::checkObj()> clause # "); DMSG(number); DMSG('\n');  
         res = false;
      };
     #endif
     #ifdef DEBUG_ALLOC_OBJ_TYPE
      if (objDesc->classDesc != &_classDesc) 
       {
        DMSG("Clause::checkObj()> objDesc->classDesc != &_classDesc\n");
        res = false;
       };
     #endif
     if (!res) DOP(DMSG("Clause::checkObj()> host object ") << *objDesc << "\n"); 
     return res;
    #endif
   }; // bool checkObj() const
  #endif  

  Clause* Next() const { return next; };
  void SetNext(Clause* c) { next = c; };
  Clause* Previous() const { return previous; };
  void SetPrevious(Clause* p) { previous = p; };
  LiteralList* LitList() const 
  {
   CALL("LitList() const");
   ASSERT(checkObj()); 
   return litList; 
  };
  
  ClauseBackground* Ancestors() const 
  {
   CALL("Ancestors()"); 
   ASSERT(checkObj()); 
   return ancestors; 
  };  
  void SetAncestors(ClauseBackground* a) 
  { 
   ancestors = a; 
  };
  ulong Number() const 
  {
   CALL("Number() const");
   ASSERT(checkObj()); 
   return number;
  };
  void SetNumber(long n) 
  {
   number = n; 
  };

  long weight() const 
  {
   CALL("weight() const"); 
   ASSERT(checkObj());
   long wt = size(); // not _size!
   if ((_useSelectedPosEqPenaltyCoefficient) 
       && (posEqSelected()))
    {
     wt = (long)(wt*_selectedPosEqPenaltyCoefficient);
    };
   if (wt > (long)HellConst::MaxClauseWeight) 
    {
     wt = HellConst::MaxClauseWeight;
    }
   else 
    if (wt < 1)
     {
      wt = 1;
     };
   return wt;
  };

  long size() const 
  {
   CALL("size() const"); 
   ASSERT(checkObj());
   if (_size < 0) return ((-1)*_size);
   return _size;
  };

    
  void AssignMaxPriority() 
  { 
   if (_size > 0) _size = (-1)*_size; 
  };
  bool MaxPriority() const 
  {
   CALL("MaxPriority() const"); 
   ASSERT(checkObj()); 
   return (_size < 0); 
  };    
  long NumOfPosEq() const 
  {
   CALL("NumOfPosEq() const"); 
   ASSERT(checkObj()); 
   return num_of_pos_eq; 
  };

  ulong NumOfEqLits() const 
  {
   CALL("NumOfEqLits() const");   
   ASSERT(checkObj()); 
   ulong res = 0;

   for(LiteralList::Iterator iter(LitList());iter.notEnd();iter.next()) 
    if (iter.currentElement()->isEquality()) res++;

   return res;       
  };

  void SetNumOfPosEq(long n) 
  { 
   num_of_pos_eq = n; 
   };
  long MinResWeight() const 
  {
   CALL("MinResWeight() const"); 
   ASSERT(checkObj());
   return _minChildSize; 
  };
  //void SetMinResWeight(long mrw) {  _minChildSize = mrw; };
  void setMinChildSize(long mcs) {  _minChildSize = mcs; };
  ulong Length() const 
  { 
   CALL("Length() const");
   ASSERT(checkObj()); 
   return LitList()->length();
  };
  void setSize(long s) { _size = s; };
  //void SetWeight(long w) { _size = w; };

  bool Active() const
  {
   CALL("Active() const");
   return  mainSet() < MainSetTrash; 
   //return Store() < CS_TRASH_USED;
  };
 
  bool Blocked() const 
  { 
   CALL("Blocked() const"); 
   ASSERT(checkObj());
   return flag(FlagIsBlocked); 
  };
  
  void Block() 
  { 
   CALL("Block()"); 
   ASSERT(checkObj()); 
   setFlag(FlagIsBlocked); 
   // if (number > 0) number = (-1)*number; 
  };
  void Unblock() 
  {
   CALL("Unblock()"); 
   ASSERT(checkObj()); 
   clearFlag(FlagIsBlocked); 
   // if (number < 0) number = (-1)*number; 
  };

  void MakeVIP() { Ancestors()->MakeVIP(); };
  bool IsVIP() const { return Ancestors()->IsVIP(); };  
  

   void MarkAsName()
   {
    CALL("MarkAsName()");
    ASSERT(checkObj());
    ancestors->MarkAsName();
   };

   bool IsSplitName() const 
   {
    CALL("IsSplitName() const");
    ASSERT(checkObj());
    return (ancestors->TheRules().bit(ClauseBackground::Split) 
            && ancestors->TheRules().bit(ClauseBackground::Name));
   };

   ulong NumberOfLiteral(const TERM* lit) const
   {
    CALL("NumberOfLiteral(TERM* lit) const");
    ASSERT(checkObj());
    ulong i = 0;
    
    for (LiteralList::Iterator iter(LitList());;iter.next())
     {
      ASSERT(i < LitList()->length());
      if (iter.currentLiteral() == lit) return i;
      i++; 
     };
    ICP("0");
   }; // ulong NumberOfLiteral(const TERM* lit)

   bool Empty() const 
   { 
    CALL("Empty() const"); 
    return (!(Length())); 
    //return !(weight());
   };
   bool Unit() const { CALL("Unit() const"); return (Length() == 1); };
   bool PositiveUnitEquality() const
   { 
    if (!Unit()) return false;    
    const TERM* lit = LitList()->firstElem()->literal();
    return (lit->Positive()) && (lit->IsEquality());
   };
   bool PreorderedPositiveUnitEquality() const
   { 
    if (!Unit()) return false;    
    const TERM* lit = LitList()->firstElem()->literal();
    return (lit->Positive()) && (lit->IsEquality()) && (lit->IsOrderedEq());
   };

   bool posEqSelected() const 
   {
    CALL("posEqSelected() const");
    ulong numOfSel = LitList()->numOfSelectedLits();
    for (LiteralList::Iterator iter(LitList()); numOfSel; iter.next())
     {
      if (iter.currentElement()->isEquality() && iter.currentElement()->positive())
       return true; 
      numOfSel--;
     };
    return false;
   };

   bool MakeDefinitionOf(const TERM& definedHeader);

   bool DefinitionUnfoldingSelected(const bool* definedHeaderTable) const;
 
   bool Definition(const bool* definedHeaderTable,TERM& definedHeader) const;

   bool HornClause() const;

   bool IsRule() const; // <=> contains both positive and negative nonsplitting literals 
   
   bool SymmetryLaw() const;

   bool TotalityLaw() const;

   bool CommutativityLaw(ulong& func) const;

   bool ContainsEliminated(const bool* eliminated_pred) const
   {
    for (LiteralList::Iterator iter(LitList()); iter.notEnd(); iter.next())
     if (eliminated_pred[iter.currentElement()->functor()]) return true;
    return false;
   }; 

   void Destroy();
   void DestroyUnshared();
   static void InsertAfter(Clause* c,Clause* db);
   static void InsertBefore(Clause* c,Clause* db);
   static void Remove(Clause* c);
   static Clause* MakeEmptyClause();
   
   Clause* unsharedCopy() const;
  
   bool IsOrphan() const 
   {
    CALL("IsOrphan() const");
    ASSERT(checkObj());
    for (ClauseBackground::ClauseList* cl = Ancestors()->ancestors(); cl; cl = cl->tl())
     if (!cl->hd()->Active()) return true;
    return false;
   };

   bool allMaximalSelected() const
   {
    CALL("allMaximalSelected() const");
    return !Ancestors()->NegativeSelection(); 
   };

   //=========== output: ======================
   ostream& outputAncestors(ostream& str) const
   {
    if (IsVIP()) str << "vip ";
    str << '['; 
    BitWord rules = Ancestors()->TheRules();

    ClauseBackground::outputRules(str,rules);
    for (ClauseBackground::ClauseList* cl = Ancestors()->ancestors(); cl; cl = cl->tl())
     {
      str << cl->hd()->Number();
      if (cl->tl()) str << ',';
     };

    return str << ']';
   }; // ostream& outputAncestors(ostream& str) const

   ostream& output(ostream& str) const  
   {
    if (mainSet() == MainSetActive) str << "* ";
    //if (Store() == CS_USED) str << "* ";  
    if (mainSet() == MainSetDefinitions) str << "DEF ";
    //if (Store() == CS_DEFINITION) str << "DEF ";
    str << Number() << ". ";
    if (Empty()) { str << '#'; }
    else LitList()->output(str);

    str << " /" << weight() << '/' << size() << '/' << MinResWeight() << "/ "
        <<  NumOfPosEq() << "pe ";
    return outputAncestors(str);   
   }; // ostream& output(ostream& str) const

   ostream& outputAsPrologTerm(ostream& str) const;
      
   ostream& outputInTPTPFormat(ostream& str) const;
  

  private:  // structure
   Clause* next;
   Clause* previous;
   LiteralList* litList; 
   ClauseBackground* ancestors;
   long number;
   long _size;
   long _minChildSize;
   //CLAUSE_STORE store;
   BitWord _flags;   
   ulong num_of_pos_eq;
   #ifdef DEBUG_ALLOC_OBJ_TYPE
    static ClassDesc _classDesc;
   #endif
   static float _selectedPosEqPenaltyCoefficient; 
   static bool _useSelectedPosEqPenaltyCoefficient;
  friend InitStatic;
}; // class Clause

inline ostream& operator<<(ostream& str,const Clause& cl)
{
 return cl.output(str);
}; 

inline ostream& operator<<(ostream& str,const Clause* cl)
{
 return cl->output(str);
}; 


//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CLAUSE
 #define DEBUG_NAMESPACE "Clause::Traversal"
#endif
#include "debugMacros.h"
//=================================================
class Clause::Traversal
{
 public:
  enum State 
   { 
    EndOfClause,
    LitHeader,
    Func,
    Var               
   };
 public: 
  Traversal() : termTrav(), iter() { CALL("constructor Traversal()"); };
  Traversal(const Clause* c) : termTrav(), iter()
  {
   CALL("constructor Traversal(const Clause* c)");
   reset(c);  
  };   
  ~Traversal() { CALL("destructor ~Traversal()"); };
  void reset(const Clause* c)
  {
   CALL("reset(const Clause* c)");
   if (c->Empty()) { currState = EndOfClause; }
   else 
    {
     currState = LitHeader;
                 
     nextLit = c->LitList()->firstLit(); 
 
     termTrav.reset(nextLit);

     iter.reset(c->LitList());   
     iter.next();
    };
  };  
  const State& state() const { return currState; };
  void next()
  {
   CALL("next()");
   termTrav.next();   
   if (termTrav.state() == TERM::Traversal::End) // end of literal
    if (iter.end()) { currState = EndOfClause; }
    else 
     {
      nextLit = iter.currentLiteral();
      currState = LitHeader;
      termTrav.reset(nextLit);
      iter.next();
     }
   else currState = (termTrav.state() == TERM::Traversal::Func) ? Func : Var;
  };

  void after()
  {
   CALL("after()");
   termTrav.after();
   if (termTrav.state() == TERM::Traversal::End) // end of literal
    if (iter.end()) { currState = EndOfClause; }
    else 
     {
      nextLit = iter.currentElement()->literal();
      currState = LitHeader;
      termTrav.reset(nextLit);
      iter.next();
     }
   else currState = (termTrav.state() == TERM::Traversal::Func) ? Func : Var;   
  };

  const TERM& symbol() const { return termTrav.symbol(); };
  const TERM& term() const { return termTrav.term(); };
 private:
  State currState;
  TERM::Traversal termTrav;
  const TERM* nextLit;
  LiteralList::Iterator iter;
}; // class Clause::Traversal 




//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CLAUSE
 #define DEBUG_NAMESPACE "Clause::OpenUnshared"
#endif
#include "debugMacros.h"
//=================================================

class Clause::OpenUnshared
{
 public:
  OpenUnshared() : _currLit(), _literals(), _ancestors() 
  {
   CALL("constructor OpenUnshared()");
   reset(); 
  };
  ~OpenUnshared() { CALL("destructor ~OpenUnshared()"); };
  void reset()
  {
   _length = 0UL;
   _size = 0UL;
   _numOfPosEq = 0UL;
   _currLit.reset();
   _literals.reset();
   _ancestors.Reset();
  };
  bool litHeader(const TERM& hd)
  {
   CALL("litHeader(const TERM& hd)");
   if (_length >= HellConst::MaxNumOfLiterals) { recycle(); return false; };   
   if (_length) _literals.push(_currLit.result().First());  
   _length++;
   _size += (1 + hd.arity());
   if (hd.Positive() && hd.IsEquality()) _numOfPosEq++;
   _currLit.reset();
   if (!_currLit.litHeader(hd)) { recycle(); return false; }; 
   return true;
  };
  bool functor(const TERM& f)
  {
   CALL("functor(const TERM& f)");
   _size += f.arity();
   if (!_currLit.functor(f)) { recycle(); return false; }; 
   return true;
  };
  void variable(const TERM& v) { _currLit.variable(v); };
  
  void ancestor(Clause* cl) { _ancestors.Push(cl); };
  void usedRule(ClauseBackground::Rule r) { _ancestors.UsedRule(r); }; 

  Clause* result()  
  {
   CALL("result()");
   LiteralList* lits = new(_length) LiteralList(_length,_length);
   if (_length) // nonempty clause
    {
     LiteralList::Iterator iter(lits);
     while (_literals)
      {
       ASSERT(iter.notEnd()); 
       iter.currentElement()->setLiteral(_literals.pop());
       iter.next();
      };
     ASSERT(iter.last());
     iter.currentElement()->setLiteral(_currLit.result().First());
    };
   Clause* res = new Clause(lits);
   res->SetAncestors(_ancestors.NewAncestors()); 
   res->setSize(_size);
   res->SetNumOfPosEq(_numOfPosEq);
   res->setMinChildSize(0);   
   return res;
  };

  void recycle() 
  {
   CALL("recycle()");
   while (_literals) _literals.pop()->DestroyLit();
   _currLit.recycle();    
  };
 private:
  ulong _length;
  ulong _size;
  ulong _numOfPosEq;
  TERM::OpenUnshared _currLit;
  Stack<TERM*,HellConst::MaxNumOfLiterals> _literals;
  OpenClauseBackground _ancestors;
}; // class Clause::OpenUnshared


//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CLAUSE
 #define DEBUG_NAMESPACE "Clause"
#endif 
#include "debugMacros.h"
//=================================================


inline bool Clause::DefinitionUnfoldingSelected(const bool* definedHeaderTable) const
{
 CALL("DefinitionUnfoldingSelected(const bool* definedHeaderTable) const");
 LiteralList::Iterator iter(LitList());
 ulong sel = LitList()->numOfSelectedLits();
 while (sel) 
  {
   ASSERT(iter.notEnd());
   TERM header = *(iter.currentLiteral());
   header.Inversepolarity();
   if (definedHeaderTable[header.HeaderNum()]) return true; 
   iter.next();
   sel--;
  };
 return false;
}; // bool Clause::DefinitionUnfoldingSelected(const bool* definedHeaderTable) const 

inline Clause* Clause::unsharedCopy() const
{
 CALL("unsharedCopy() const");
 LiteralList::Iterator iter(LitList());
 LiteralList* targetLitList = new (LitList()->length()) LiteralList(LitList()->length(),LitList()->numOfSelectedLits());
 LiteralList::Iterator targetIter(targetLitList);

 while (iter.notEnd()) 
  {
   ASSERT(targetIter.notEnd());
   targetIter.currentElement()->setLiteral(iter.currentLiteral()->unsharedLitCopy(),iter.currentElement()->mark());
   iter.next();
   targetIter.next();
  };
 ASSERT(iter.end());
 ASSERT(targetIter.end());
 
 Clause* res = new Clause(targetLitList);
 res->ancestors = Ancestors()->copy();
 res->number = number;
 res->_size = _size;
 res->_minChildSize = _minChildSize;

 //res->_flags = _flags; 
 ASSERT(!res->_flags);
 res->setMainSet(MainSetFree);
 if (flag(FlagHasChildren)) res->setFlag(FlagHasChildren);

 res->num_of_pos_eq = num_of_pos_eq;
 return res;
}; // Clause* Clause::unsharedCopy() const

inline void Clause::InsertAfter(Clause* c,Clause* db)
 {
  Clause* next_cl = db->next;
  c->next = next_cl;
  next_cl->previous = c;
  db->next = c;
  c->previous = db;
 }  // Clause::InsertAfter()


inline void Clause::InsertBefore(Clause* c,Clause* db)
 {
  Clause* previous_cl = db->previous;
  previous_cl->next = c;
  c->previous = previous_cl;
  c->next = db;
  db->previous = c;
 }  // Clause::InsertBefore()

inline void Clause::Destroy()         
 {
  CALL("Destroy()");
  ASSERT(checkObj());
  delete litList;
  delete ancestors;
  delete this;
 };  // Clause::Destroy()

inline void Clause::DestroyUnshared()         
 {
  CALL("DestroyUnshared()");
  ASSERT(checkObj());
  for (LiteralList::Iterator iter(LitList());iter.notEnd();iter.next()) 
   iter.currentLiteral()->destroyUnsharedLit();
   
  delete litList;
  delete ancestors;
  delete this;
 };  // Clause::DestroyUnshared()



inline void Clause::Remove(Clause* c)
 {
  Clause* previous_cl = c->previous;
  Clause* next_cl = c->next;

  previous_cl->next = next_cl;
  next_cl->previous = previous_cl;
 };  // Clause::Remove()

inline Clause* Clause::MakeEmptyClause()   
 {
  return new Clause(new (0UL) LiteralList(0UL,0UL));
 };

inline bool Clause::IsRule() const
{
 CALL("IsRule() const");
 ASSERT(!Empty());
 if (Unit()) return false;
 LiteralList::Iterator iter(LitList());
 while (iter.notEnd() && iter.currentLiteral()->IsSplitting()) iter.next();
 if (iter.end()) return false; // only splitting literals 
 
 TERM::Polarity firstLitPol = iter.currentLiteral()->polarity();
 iter.next();
 while (iter.notEnd())
  {
   if ((!iter.currentLiteral()->IsSplitting()) 
       && (iter.currentLiteral()->polarity() != firstLitPol)) 
    return true; 
   iter.next();   
  };
 return false; // same polarity everywhere 
}; // bool Clause::IsRule() const

static Clause::InitStatic clauseInitStatic;

//===========================================================================
#undef DEBUG_NAMESPACE
//===========================================================================
#endif
