//
// File:         forward_superposition.h
// Description:  Representation of stored terms.
// Created:      Feb 14, 2000, 21:15
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#ifndef FORWARD_SUPERPOSITION_H
//============================================================
#define FORWARD_SUPERPOSITION_H
#include "jargon.h"
#include "HellDebugFlags.h"
#include "HellConst.h"
#include "Stack.h"
#include "para_machine.h"
#include "NewClause.h"
#include "weight_formula.h"
#include "or_index.h"
#include "UnifKBConstr.h"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_SUPERPOSITION
 #define DEBUG_NAMESPACE "FPARA_MACH"
#endif
#include "debugMacros.h"
//=================================================
class FPARA_MACH 
 {
  private:
   enum STATE { INIT,  
                PARA_FROM_COMPLEX, PARA_FROM_VAR };
  private:   
   STATE state;
   PARA_MACH para_mach;
   PREFIX_SYM* goal;
   TERM pseudo_header; 
   ORD_CONSTR* ordering_constraint;
   OR_INDEX* _unifIndex;
  public: 
   FPARA_MACH(ORD_CONSTR* constr,OR_INDEX* unifIndex) 
    : para_mach(),
      pseudo_header((TERM::Arity)1,(TERM::Functor)0,TERM::PositivePolarity),
      ordering_constraint(constr),
      _unifIndex(unifIndex)
    {};
   ~FPARA_MACH() { CALL("destructor ~FPARA_MACH()"); };
   void SetIndex(ulong ind) { para_mach.SetIndex(ind); };
   ulong GetIndex() const { return para_mach.GetIndex(); };
   PREFIX_SYM* Goal() const { return goal; };
   void LoadTermToRewrite(PREFIX_SYM* t)
    {
     CALL("LoadTermToRewrite(PREFIX_SYM* t)");
     // t must be a plain complex term with all brackets set
     goal = t;
     para_mach.CompileComplexTerm(goal);
     state = INIT;
    };
    
   RES_PAIR_LIST* NextCandidate();
 }; // class FPARA_MACH 

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_SUPERPOSITION
 #define DEBUG_NAMESPACE "FPARA_MACHINE"
#endif
#include "debugMacros.h"
//================================================= 
 
class FPARA_MACHINE : public FPARA_MACH
 {
  private:
   Stack<PREFIX_SYM*,HellConst::MaxTermSize> complex_positions;
   PREFIX_SYM* literal;
   bool inherentlySelected;
   RES_PAIR_LIST* candidate;
   ulong index;
   ORD_CONSTR eq_ord_constr; // experimental
   PREFIX_SYM* eq_second_arg;
   NewClause* _resolvent;
   bool _noRulesFlag;
  public:
   FPARA_MACHINE(NewClause* res,OR_INDEX* unifIndex) 
    : FPARA_MACH(&eq_ord_constr,unifIndex), 
      complex_positions("complex_positions"),
      eq_ord_constr(),
      _resolvent(res)
    {};
   ~FPARA_MACHINE() { CALL("destructor ~FPARA_MACHINE()"); };
   void setNoRulesFlag(bool fl) { _noRulesFlag = fl; };
   void SetIndex(ulong ind) { index = ind; FPARA_MACH::SetIndex(ind); };
   const ulong& GetIndex() const { return index; };
   RES_PAIR_LIST* Candidate() const { return candidate; };
   bool LoadLiteral(PREFIX_SYM* lit,bool inhSel);
   bool LoadRestOfLit();
   bool Paramodulate();

   bool CollectResult()
    {
     CALL("CollectResult()");
     return CollectRewrittenLiteral() && CollectRestOfTheRewritingClause();
    };
  private:
   bool CollectRewrittenLiteral()
    {
     CALL("CollectRewrittenLiteral()");
     return _resolvent->PushRewrittenLit(literal,
                                         FPARA_MACH::Goal(),
                                         candidate->ReplacingTerm(),
                                         index,
                                         inherentlySelected);
    };  
    
   bool CollectRestOfTheRewritingClause();
 }; // class FPARA_MACHINE : public FPARA_MACH
 
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_SUPERPOSITION
 #define DEBUG_NAMESPACE "FPARA_PREMISE"
#endif
#include "debugMacros.h"
//=================================================
class FPARA_PREMISE
 {
  private:
   Clause* clause;
   ulong index;
   PREFIX_SYM max_literal[HellConst::MaxTermSize*HellConst::MaxNumOfLiterals];
   bool _inherentlySelected[HellConst::MaxNumOfLiterals];
   ulong num_of_max_literals;
   
   LiteralList* _allLiterals;
   LiteralList::Element* _firstNonselectedLit;

   ulong next_max_lit;
   bool eq_res_applicable;
   WEIGHT_FORMULA weight_form;
   NewClause* _resolvent;
  public:
   FPARA_PREMISE(NewClause* res) : weight_form(), _resolvent(res) {};
   ~FPARA_PREMISE() { CALL("destructor ~FPARA_PREMISE()"); };
   void SetIndex(ulong ind) { index = ind; weight_form.SetIndex(ind); };
   Clause* Query() const { return clause; };
   void ResetIter() { next_max_lit = num_of_max_literals; };
   bool EqResApplicable() const { return eq_res_applicable; };
   void CreateWeightFormula()
    {
     CALL("CreateWeightFormula()");
     weight_form.Reset();

     for(LiteralList::Iterator iter(_allLiterals,_firstNonselectedLit);iter.notEnd();iter.next())
      weight_form.Compile(iter.currentLiteral());


     
     ResetIter();
     for (PREFIX_SYM* m_lit = NextMaxLitMem(); m_lit; m_lit = NextMaxLitMem())
      {
       TERM header = m_lit->Head().Func();
       if (header.IsEquality() && header.IsOrderedEq()) weight_form.Compile(m_lit->Arg2()); 
      };
      
    };
   
   void Load(Clause* cl);

   PREFIX_SYM* NextMaxLit(bool& inherentlySelected)
    {
     if (next_max_lit)
      {
       next_max_lit--;
       return MaxLit(next_max_lit,inherentlySelected);
      };
     return (PREFIX_SYM*)0;
    }; 

   PREFIX_SYM* NextMaxLitMem()
    {
     if (next_max_lit)
      {
       next_max_lit--;
       return MaxLitMem(next_max_lit);
      };
     return (PREFIX_SYM*)0;
    }; 

   long MinResultWeight()
    {
     return weight_form.Compute();
    };
    
   bool CollectLiterals();
  private:
   PREFIX_SYM* MaxLit(const ulong& n,bool& inherentlySelected) 
   { 
    inherentlySelected = _inherentlySelected[n];
    return max_literal + (n*HellConst::MaxTermSize); 
   };
   PREFIX_SYM* MaxLitMem(const ulong& n) 
   { 
    return max_literal + (n*HellConst::MaxTermSize); 
   };
 }; // class FPARA_PREMISE

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_SUPERPOSITION
 #define DEBUG_NAMESPACE "FORWARD_SUPERPOSITION"
#endif
#include "debugMacros.h"
//=================================================
class FORWARD_SUPERPOSITION
 {
  public:
   FORWARD_SUPERPOSITION(NewClause* res,OR_INDEX* unifIndex)
    : fpara_machine(res,unifIndex),
      fpara_premise(res),
      _resolvent(res),
      _unifConstr(),
      _noInfBetweenRulesFlag(false)
    {
     fpara_premise.SetIndex(0);
     fpara_machine.SetIndex(1);
    }; 
   ~FORWARD_SUPERPOSITION() { CALL("destructor ~FORWARD_SUPERPOSITION()"); };
   void setNoInfBetweenRulesFlag(bool fl)
   {
    _noInfBetweenRulesFlag = fl;     
   };
   bool Load(Clause* cl);
   bool NextInference()
   { 
    CALL("NextInference()");
    switch (_continuation)
     {
      case ContFSup:
       while (NextFSupInference())
	if (CollectResult()) return true;
       // no more superpositions, start equality resolution 
       if (!fpara_premise.EqResApplicable()) return false;
       fpara_premise.ResetIter();
       _continuation = ContEqRes;
       // no break here, proceed with the next case
      case ContEqRes:
       while (NextEqRes())
        {
         if (CollectEqResResult())
          {
           UndoEqRes();         
           return true;
          };
         UndoEqRes(); 
        }; 
       // no more inferences
       return false;
     };
    ICP("ICP1");
    return false;
   };
  private:
   bool NextFSupInference();
   void UndoEqRes()
    {
     GlobalSubst.UnbindLastSection();
    }; 
   bool NextEqRes();
   bool CollectResult();
   bool CollectEqResResult();
  
   bool checkConstr(const SharedKBConstr* constr);
  private: 
   enum Continuation { ContFSup, ContEqRes };
  private:
   FPARA_MACHINE fpara_machine;
   FPARA_PREMISE fpara_premise;
   NewClause* _resolvent;
   Continuation _continuation;
   UnifKBConstr _unifConstr;
   bool _noInfBetweenRulesFlag;
 }; // class FORWARD_SUPERPOSITION


//================================================
#undef DEBUG_NAMESPACE 
//=================================================================
#endif
