//
// File:         hyper_rule.h
// Description:  Clauses-rules in hyperresolution.
// Created:      Feb 15, 2000, 13:10
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#ifndef HYPER_RULE_H
//======================================================================
#define HYPER_RULE_H
#ifndef NO_DEBUG_VIS
 #include <iostream.h>
#endif
#include "jargon.h"
#include "HellDebugFlags.h"
#include "HellConst.h"
#include "clause.h"
#include "prefix.h"
#include "variables.h"
#include "premise_factoring.h"
#include "iu_command.h"
#include "iu_translator.h"
#include "unification.h"
#include "NewClause.h"
#include "Stack.h"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_HYPER_RULE
 #define DEBUG_NAMESPACE "HYPER_RULE"
#endif
#include "debugMacros.h"
//=================================================
class HYPER_RULE 
 {
  // A rule has form: LHS -> RHS
  private: // structure
   static bool LHS_Negative; // determines orientation of rules in hyperresolution
   Clause* clause; // the input clause itself
   bool compiled;
   PREFIX_SYM** RightHandSide; // array of pointers to terms in word representation
   PREFIX_SYM** LeftHandSide;  // array of pointers to terms in word representation
   IU_COMMAND** LeftHandSideCode; 
   PREFIX_SYM** endOfLHS;
   PREFIX_SYM** endOfRHS;
   ulong sizeOfLHS;
   ulong sizeOfRHS;
   ulong number;
   static long lastRule;
   enum { ordinaryRuleIndex = 0 };
   enum { unitConflictRuleIndex = HellConst::MaxNumOfStandardIndices + 0 };
   static bool unit_conflict;
   ulong ruleIndex;
   PREFIX_SYM* wordRepMemory;
   static IU_TRANSLATOR Compiler;
   static PREFIX_SYM tmpWordMemory[HellConst::MaxTermSize];
   IU_COMMAND* codeMemory;
   ulong memorySize;

  public: // methods
   HYPER_RULE() 
    {
    };
   ~HYPER_RULE() 
    {
     CALL("destructor ~HYPER_RULE()");
     delete [] wordRepMemory;
     
     if (compiled) 
      { 
       delete [] LeftHandSideCode;
       delete [] codeMemory;
      };
     
     delete [] RightHandSide;
     delete [] LeftHandSide;
     clause->Destroy();
    }; 
   static void PositiveHyper(bool flag) { LHS_Negative = flag; };
   static bool isPositiveHyper() { return LHS_Negative; };
   static void SetUnitConflict(bool flag) { unit_conflict = flag; };
   ulong Number() const { return number; };
   static long LastRule() { return lastRule; };
   ulong RuleIndex() const { return ruleIndex; };

   static bool IsARule(const Clause* c); 
   void MakeRule(Clause* c,bool compile);

   Clause* TheClause() const { return clause; };
   bool Compiled() const { return compiled; };
   PREFIX_SYM** RHS() { return RightHandSide; };
   PREFIX_SYM** LHS() { return LeftHandSide; };
   PREFIX_SYM** EndOfLHS() { return endOfLHS; };
   PREFIX_SYM** EndOfRHS() { return endOfRHS; };
   IU_COMMAND** LHSCode() { return LeftHandSideCode; };
   ulong SizeOfLHS() const { return sizeOfLHS; };
   ulong SizeOfRHS() const { return sizeOfRHS; };
   bool LongLHS() const { return (SizeOfLHS() > 1); };
   bool Unit() const { return ((SizeOfLHS() == 1) && (!(SizeOfRHS()))); };
   #ifndef NO_DEBUG_VIS
    ostream& output(ostream& str) const;
   #endif
 }; // class HYPER_RULE

#ifndef NO_DEBUG_VIS
 inline ostream& operator<<(ostream& str,const HYPER_RULE& rule) 
 {
  return rule.output(str);
 };

 inline ostream& operator<<(ostream& str,const HYPER_RULE* rule) 
 {
  return rule->output(str);
 };
#endif

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_HYPER_RULE
 #define DEBUG_NAMESPACE "LHS_LIT_COMPLEXITY"
#endif
#include "debugMacros.h"
//=================================================
class LHS_LIT_COMPLEXITY
 {
  private:
   ulong numOfFuncs;
   ulong numOfVarOccs;
   ulong numOfFirstVarOccs;
   LARGE_VAR_COUNTER var_counter;
  public:
   void Estimate(PREFIX_SYM* lit)
    {
     PREFIX::ITER iter;
     iter.Reset(lit);
     iter.SkipSym();
     ulong num_of_funcs = 0;
     ulong num_of_var_occs = 0;
     ulong num_of_first_var_occs = 0;
     while (!(iter.Empty()))
      {
       if (iter.CurrVar())
        {
         num_of_var_occs++;
         if (!(var_counter.Register(iter.CurrVar()))) 
          {
           num_of_first_var_occs++;
          };
        }
       else
        {
         num_of_funcs++;
        };
       iter.SkipSym();
      };
     numOfFuncs = num_of_funcs;
     numOfVarOccs = num_of_var_occs;
     numOfFirstVarOccs = num_of_first_var_occs;
    }; // void Estimate(PREFIX_SYM* lit)
    
   Comparison Compare(LHS_LIT_COMPLEXITY& est)
    {
     if (numOfVarOccs - numOfFirstVarOccs > est.numOfVarOccs - est.numOfFirstVarOccs)
      {
       return Greater;
      }
     else
      {
       if (numOfVarOccs - numOfFirstVarOccs < est.numOfVarOccs - est.numOfFirstVarOccs)
        {
         return Less;
        }
       else
        {
         if (numOfFuncs >= est.numOfFuncs)
          {
           return Greater;
          }
         else
          {
           return Less;
          };
        };
      };
    };
 }; // class LHS_LIT_COMPLEXITY

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_HYPER_RULE
 #define DEBUG_NAMESPACE "HYPER_ACTIVE_RULE"
#endif
#include "debugMacros.h"
//=================================================
class HYPER_ACTIVE_RULE : public HYPER_RULE
 {
  private:
   enum { currentFactIndex = 1 };
   enum { unitConflictSuspectIndex = HellConst::MaxNumOfStandardIndices + 1 };
  private:
   static OR_PREMISE_WITH_FACTORING currentFact;
   static OR_PREMISE unitConflictSuspect;
   static ulong cutLHSLiteral;
   static ulong currentLHSLiteral;
   static bool used[HellConst::MaxNumOfLiterals];
   static bool active[HellConst::MaxNumOfLiterals];
   static Stack<ulong,HellConst::MaxNumOfLiterals> inactive;
   static ulong numOfActive;
  private: // dinamic ordering of lhs literals
   static LHS_LIT_COMPLEXITY est1,est2;
   static LHS_LIT_COMPLEXITY* maxEst;
   static LHS_LIT_COMPLEXITY* currEst;
   static bool dinamicLHSOrdering;
  private: // restriction on application of a rule
   static bool empty_RHS_only; 
    // "empty_RHS_only == true" means that only rules with empty RHS are applicable 
  public:   
   static void Initialise(NewClause* resolvent)
    {
     dinamicLHSOrdering = true; // default
     empty_RHS_only = false;
     currentFact.assignResolvent(resolvent);
     unitConflictSuspect.setResolvent(resolvent); 
    }; 
   static void SetDinamicLHSOrdering(bool flag) 
    { dinamicLHSOrdering = flag; };
   static void EmptyRHSOnly() { empty_RHS_only = true; };
   static void SetCurrentFact(Clause* fact)
    {
     currentFact.Load(currentFactIndex,fact);
    };
   static void SetUnitConflictSuspect(Clause* fact)
    {
     unitConflictSuspect.Load(unitConflictSuspectIndex,fact);
    };
   static Clause* FactClause() { return (Clause*)currentFact.TheClause(); };
   static Clause* UnitConflictSuspectClause() 
    { return (Clause*)unitConflictSuspect.TheClause(); };
   static void UnloadCurrentFact() { currentFact.Unload(); }; 
   static void UnloadUnitConflictSuspect() { unitConflictSuspect.Unload(); };
   bool TryToApply()
    {
     if ((empty_RHS_only) && (SizeOfRHS())) { return false; };
     
     GlobalSubst.UnbindAll();
     
     for (ulong l = 0; l < SizeOfLHS(); l++) { active[l] = true; };
     
     PREFIX_SYM** last_lit = currentFact.MaxLiterals() + currentFact.LastMax();
     
     for (PREFIX_SYM** lhs_lit = LHS(); lhs_lit < EndOfLHS(); lhs_lit++)
      {
       for (PREFIX_SYM** lit = currentFact.MaxLiterals(); lit <= last_lit; lit++)
        {
         if (Compiled() ? Unify(*(LHSCode()+(lhs_lit - LHS())),*lit)
                        : UNIFICATION::UnifyComplex(*lit,*lhs_lit))
          {
           currentFact.UnmarkCut();
           currentFact.Cut(lit - currentFact.MaxLiterals());
           cutLHSLiteral = lhs_lit - LHS();
           used[cutLHSLiteral] = true;
           active[cutLHSLiteral] = false;
           numOfActive = SizeOfLHS() - 1;
           for (ulong l = cutLHSLiteral + 1; l < SizeOfLHS(); l++)
            {
             used[l] = false;
            };
           return true;
          };
        };
      };
     return false; 
    }; // bool TryToApply()
    
   bool NextApplication()
    {
     GlobalSubst.UnbindAll();
     PREFIX_SYM** last_lit = currentFact.MaxLiterals() + currentFact.LastMax();
     // first try to resolve the current LHS literal with another literal from the fact
     PREFIX_SYM** first_lit = currentFact.MaxLiterals() + currentFact.CutLit() + 1;
     PREFIX_SYM** lhs_lit = LHS() + cutLHSLiteral;
     
     for (PREFIX_SYM** lit = first_lit; lit <= last_lit; lit++)
      {
       if (Compiled() ? Unify(*(LHSCode()+(lhs_lit - LHS())),*lit)
                      : UNIFICATION::UnifyComplex(*lit,*lhs_lit))
	{
         currentFact.UnmarkCut();
         currentFact.Cut(lit - currentFact.MaxLiterals());
         return true;
        };
      };
     
     // try other LHS literals
     
     active[cutLHSLiteral] = true;
     lhs_lit++;
     while (lhs_lit < EndOfLHS())
      {
       for (PREFIX_SYM** lit = currentFact.MaxLiterals(); lit <= last_lit; lit++)
        {
         if (UNIFICATION::UnifyComplex(*lit,*lhs_lit))
          {
           currentFact.UnmarkCut();
           currentFact.Cut(lit - currentFact.MaxLiterals());
           cutLHSLiteral = lhs_lit - LHS();
           used[cutLHSLiteral] = true;
           active[cutLHSLiteral] = false;
           return true;
          };
        };
       lhs_lit++;
      };
     return false; 
    }; // bool NextApplication()
   
   bool Conflict()
    {
     // the rule and the current suspect must be unit here!
     // index of the rule must be the corresponding one
     
     if (Compiled())
      {
       return Unify(*(LHSCode()),*(unitConflictSuspect.MaxLiterals()));
      }
     else
      {
       return UNIFICATION::UnifyComplex(*(unitConflictSuspect.MaxLiterals()),
                                        *(LHS()));  
      };  
                          
    };
   
  private:
   bool CollectRHSLiterals(NewClause* resolvent)
    {
     CALL("CollectRHSLiterals(NewClause* resolvent)");
     PREFIX_SYM** end_of_rhs = EndOfRHS();
     for (PREFIX_SYM** rhs_lit = RHS(); rhs_lit < end_of_rhs; rhs_lit++)
      {
       if (!(resolvent->PushPrefixLit(*rhs_lit,false))) { return false; };
      };
     return true;
    }; 
  public:
   static void ResetFactoring() { currentFact.ResetFactoring(); };
   static bool NextFactor() 
    {
     if (currentFact.WasFactored())
      {
       return currentFact.GetNextFactor(); 
      }
     else
      {
       return currentFact.GetFirstFactor(0);
      };
    };
   bool CollectLiterals(NewClause* resolvent)
    {
     CALL("CollectLiterals(NewClause* resolvent)");
     if (!currentFact.TheClause()->Active()) return false;      
     currentFact.setResolvent(resolvent);
     return ((currentFact.CollectNonmarkedLiterals()) && (CollectRHSLiterals(resolvent)));
    };
  public: // iteration over LHS with respect to the ordering
   static void ResetIterOverLHS()
    {
    };
   PREFIX_SYM* NextLHSLiteral(bool& wasUsed) 
    {
     if (!numOfActive) { return (PREFIX_SYM*)0; };
     currentLHSLiteral = 0;
     while (!(active[currentLHSLiteral])) { currentLHSLiteral++; }; 
     if ((numOfActive > 1) && (dinamicLHSOrdering))
      // more then one active literal, maximal must be found
      {
       maxEst = &est1;
       currEst = &est2;
       LHS_LIT_COMPLEXITY* tmp;
       
       maxEst->Estimate((LHS())[currentLHSLiteral]);
       ulong restOfActive = numOfActive - 1;
       ulong lhs_lit = currentLHSLiteral + 1;
       
       try_next_active:
        // first find next active literal
        while (!(active[lhs_lit])) { lhs_lit++; };
        restOfActive--;
        currEst->Estimate((LHS())[lhs_lit]);
        if (maxEst->Compare(*currEst) == Less)
         {
          tmp = maxEst;
          maxEst = currEst;
          currEst = tmp;
          currentLHSLiteral = lhs_lit;
         };
        if (restOfActive) 
         {
          lhs_lit++;
          goto try_next_active; 
         };
      }; 
      
     PREFIX_SYM* res = (LHS())[currentLHSLiteral];
     wasUsed = used[currentLHSLiteral];
     active[currentLHSLiteral] = false;
     inactive.push(currentLHSLiteral);
     numOfActive--;
     return res;
    };
   static void Rollback()
    {
     numOfActive++;
     currentLHSLiteral = inactive.pop();
     active[currentLHSLiteral] = true; 
    };
   static void Rollback(ulong n) { while (n) { Rollback(); n--; }; };
 };

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_HYPER_RULE
 #define DEBUG_NAMESPACE "HYPER_RULE"
#endif
#include "debugMacros.h"
//=================================================

inline bool HYPER_RULE::IsARule(const Clause* c)
{
 CALL("HYPER_RULE::IsARule(Clause* c)");
 // count sizes of LHS and RHS
 ulong sizeOfLHS_ = 0;
  
 for(LiteralList::Iterator iter(c->LitList());iter.notEnd();iter.next())
  if (iter.currentElement()->negative() == LHS_Negative) sizeOfLHS_++;

 return (bool)sizeOfLHS_; 
}; // bool HYPER_RULE::IsARule(const Clause* c)

inline void HYPER_RULE::MakeRule(Clause* c,bool compile)
 {
  CALL("MakeRule(Clause* c,bool compile)");
  ASSERT(IsARule(c)); 
  compiled = compile;
  // first simply count sizes of LHS and RHS
  sizeOfLHS = 0;
  sizeOfRHS = 0;
  
  for (LiteralList::Iterator iter(c->LitList());iter.notEnd();iter.next())
   if (iter.currentElement()->negative() == LHS_Negative)
    {
     sizeOfLHS++;
    } 
   else sizeOfRHS++;
   
  ASSERT(sizeOfLHS);

  clause = c;

  RightHandSide = new PREFIX_SYM* [sizeOfRHS];
  LeftHandSide = new PREFIX_SYM* [sizeOfLHS];
  if ((unit_conflict) && (Unit()))
   {
    ruleIndex = unitConflictRuleIndex;
   }
  else
   {
    ruleIndex = ordinaryRuleIndex;
   };

  ulong lhs = 0;
  ulong rhs = 0;
  memorySize = (c->size()); 
  wordRepMemory = new PREFIX_SYM[memorySize];

  PREFIX_SYM* memory = wordRepMemory; 
  
  if (compiled)
   {
    LeftHandSideCode = new IU_COMMAND* [sizeOfLHS];
     
    codeMemory = new IU_COMMAND[((HellConst::MaxTermSize + 2)*sizeOfLHS)]; // the size can be optimized
   }
  else
   {
    LeftHandSideCode = (IU_COMMAND**)0; // just in case
    codeMemory = (IU_COMMAND*)0; // just in case
   };
   
  IU_COMMAND* code_memory = codeMemory;
  PREFIX_SYM* tmp_memory; 
  for (LiteralList::Iterator iter(c->LitList());iter.notEnd();iter.next())
   if (iter.currentElement()->negative() == LHS_Negative)
    {
     // LHS
     LeftHandSide[lhs] = memory;
        
     if (compiled)
      {
       LeftHandSideCode[lhs] = code_memory;
       tmp_memory = tmpWordMemory;
       StandardTermToPrefix(ruleIndex,iter.currentLiteral(),tmp_memory);
       Compiler.CompileLiteral(tmpWordMemory,code_memory,memory,HellConst::MaxTermSize);
          
       LeftHandSideCode[lhs]->Inversepolarity();
          
      }
     else
      {
       StandardTermToPrefix(ruleIndex,iter.currentLiteral(),memory);
      };

     LeftHandSide[lhs]->Head().Inversepolarity();
     lhs++;    
    } 
   else
    {
     // RHS
     RightHandSide[rhs] = memory;
     rhs++;
     StandardTermToPrefix(ruleIndex,iter.currentLiteral(),memory);
    };
  
  endOfLHS = LeftHandSide + sizeOfLHS;
  endOfRHS = RightHandSide + sizeOfRHS;
  
  lastRule++;
  number = lastRule;
 }; // void HYPER_RULE::MakeRule(Clause* c,bool compile) 

#ifndef NO_DEBUG_VIS
 inline ostream& HYPER_RULE::output(ostream& str) const
 {
  return str << "RULE #" << Number() << ' ' << TheClause() << '\n';
 }; // ostream& HYPER_RULE::output(ostream& str) const
#endif
 
//================================================
#undef DEBUG_NAMESPACE
//======================================================================
#endif
