#ifndef KEPT_H
//======================================================================
#define KEPT_H
#include <limits.h>
#include "jargon.h"
#include "HellDebugFlags.h"
#include "HellConst.h"
#include "clause.h"
#include "clause_set.h"
//====================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_KEPT
 #define DEBUG_NAMESPACE "KEPT"
#endif
#include "debugMacros.h"
//====================================================
const int NUMBER_OF_KEPT_CATEGORIES = HellConst::MaxClauseWeight + 1;

class KEPT
{
 public:
  KEPT() : pick_given_ratio(0), magic_coefficient(0.0) { Reset(); };
  ~KEPT() { CALL("destructor ~KEPT()"); }; 
  void Reset()
  {
   max_priority_flag = false;
   num_of_currently_kept = 0L;
   num_of_used = 0L;
   num_of_removed = 0L;
   num_of_cancelled = 0L;
   numOfReachable = LONG_MAX;
   for (int cat = 0; cat < NUMBER_OF_KEPT_CATEGORIES; cat++) StatReset(cat);
   minCat = 0;
   maxCat = 0;
   max_accepted_cat = NUMBER_OF_KEPT_CATEGORIES-1;
   picked_by_weight = 0;
  };
  void DestroyAllClauses();
  void SetPickGivenRatio(long pgr) 
  {
   pick_given_ratio = pgr; 
   if (pgr > 0) magic_coefficient = 1.0 + ((1.0)/((float)pgr));
  };
  void MaxPriorityOn() { max_priority_flag = true; };
  void MaxPriorityOff() { max_priority_flag = false; }; 
  long Size() const { return num_of_currently_kept; };
  long Total() const { return num_of_currently_kept + num_of_used + num_of_removed + num_of_cancelled; };
  long CurrentlyKept() const { return num_of_currently_kept; };
  long NumOfReachable() const { return numOfReachable; };
  long UsedAndRemoved() const { return (num_of_used + num_of_removed); };
  bool MaxPriorityEmpty() { return EmptyCat(0); };
  long UnreachableKept() const
  {
   return (num_of_currently_kept > numOfReachable)? (num_of_currently_kept - numOfReachable) : 0; 
  };
  void SetNumOfReachable(long n) 
  {
   CALL("SetNumOfReachable(const long n)");
   ASSERT(CheckStatistics());    
   numOfReachable = n; 
   CorrectMaxAcceptedCat();
  };
  long MaxAcceptedWeight() const
  {
   if (num_of_currently_kept < numOfReachable) return HellConst::MaxClauseWeight; 
   if (pick_given_ratio > 0) { return Weight(max_accepted_cat); }
   else return 0L; // ??
  };
  long MaxPresentWeight() const { return Weight(maxCat); };
  long MinPresentWeight() const { return Weight(minCat); };
  //void InsertInputClause(Clause* const c) { InsertAs(c,0); };

  void Cancel(Clause* c) { Clause::Remove(c); StatCancelled(Category(c)); };
  void Remove(Clause* c) 
  { 
   Clause::Remove(c); 
   StatRemoved(Category(c)); 
  };
  void Insert(Clause* c) 
  { 
   if (max_priority_flag) 
    { 
     c->AssignMaxPriority();
    };
   InsertAs(c,Category(c)); 
  };
  Clause* Next();
  Clause* TheWorstClause()
  {
   CALL("TheWorstClause()");
   for (int cat = maxCat; cat >= minCat; cat--) 
    if (!(EmptyCat(cat))) return category[cat].Previous();
   return (Clause*)0;
  };

  Clause* TheWorstUnblockedClause()
  {
   CALL("TheWorstUnblockedClause()");
   for (int cat = maxCat; cat >= minCat; cat--) 
    if (!(EmptyCat(cat))) 
     for (Clause* cand = category[cat].Previous(); cand != &(category[cat]); cand = cand->Previous())
      if (!cand->Blocked()) return cand;
   return (Clause*)0;   
  };

  void ResetIterator()
  {
   if (EmptyCat(minCat)) { next_clause = (Clause*)0; return; };
   iter_cat = minCat;
   next_clause = category[iter_cat].Next();
  };
  Clause* IterNext() 
  {
   Clause* res = next_clause;     
   if (res)
    {
     next_clause = res->Next();
     if (next_clause == &(category[iter_cat]))
      {
       // try another category
       iter_cat++;       
       while ((iter_cat <= maxCat) && (EmptyCat(iter_cat))) { iter_cat++; };
       if (iter_cat > maxCat) 
        { 
         next_clause = (Clause*)0; 
        }
       else { next_clause = category[iter_cat].Next(); };
      };
    };
   return res;
  }; // Clause* IterNext() 
 
 private:   
  void NormalizeMinMaxCat(int lb,int rb)
  {
   minCat = lb;
   maxCat = rb;
   while ((EmptyCat(minCat)) && (minCat < maxCat)) minCat++;
   while ((EmptyCat(maxCat)) && (minCat < maxCat)) maxCat--;
  };
  
  void StatReset(int cat) 
  {
   currently_kept[cat] = 0L;
   used[cat] = 0L;
   removed[cat] = 0L;
  };
  
  void StatInserted(int cat)
  {
   currently_kept[cat]++;
   num_of_currently_kept++;    
   if (cat > maxCat) { NormalizeMinMaxCat(minCat,cat); }
   else
    if (cat < minCat) NormalizeMinMaxCat(cat,maxCat);
  };
   
  void StatUsed(int cat)
  {
   used[cat]++;
   currently_kept[cat]--;
   num_of_used++;     
   num_of_currently_kept--;
   numOfReachable--; // experimental
   if (EmptyCat(cat) && ((cat == minCat) || (cat == maxCat)))
    NormalizeMinMaxCat(minCat,maxCat);
  };
   
  void StatRemoved(int cat)
  {
   removed[cat]++;
   currently_kept[cat]--;
   num_of_removed++;
   num_of_currently_kept--;
   numOfReachable--; // experimental
   if (EmptyCat(cat) && ((cat == minCat) || (cat == maxCat)))
    NormalizeMinMaxCat(minCat,maxCat);
  };
  
  void StatCancelled(int cat)
  {
   num_of_cancelled++;
   currently_kept[cat]--;
   num_of_currently_kept--;
   if (EmptyCat(cat) && ((cat == minCat) || (cat == maxCat)))
    NormalizeMinMaxCat(minCat,maxCat);     
  };

  long Total(int cat) const { return currently_kept[cat] + used[cat] + removed[cat]; };
 
  long NumOfLighter(int cat)
  {
   long res = 0L;
   for (int c = 0; c < cat; c++) res += currently_kept[c]; 
   return res;
  };
   
  long NumOfHeavier(int cat)
  {
   long res = 0L;
   for (int c = cat + 1; c <= maxCat; c++) res += currently_kept[c];
   return res;
  };


  /********** old version, used in CASC-16 *********/
  bool Reachable(int cat)
  {
   CALL("Reachable(int cat)");
   if (numOfReachable > num_of_currently_kept) return true;   
   long num_of_lighter_or_same_weight = NumOfLighter(cat) + currently_kept[cat];
   if (!num_of_lighter_or_same_weight) return true; 
   if (numOfReachable <= num_of_lighter_or_same_weight) return false;
   long magic_number = (long)(num_of_lighter_or_same_weight * magic_coefficient);
   return (numOfReachable - (num_of_currently_kept - numOfReachable) > magic_number);
  };
  /***********************/


  /********************
  bool Reachable(int cat)
  {
   CALL("Reachable(int cat)");   
    if (numOfReachable > num_of_currently_kept) return true; 
   long num_of_lighter_or_same_weight = NumOfLighter(cat) + currently_kept[cat];
   if (!num_of_lighter_or_same_weight) return true; 
   if (numOfReachable <= num_of_lighter_or_same_weight) return false;
   long magic_number = (long)(num_of_lighter_or_same_weight * magic_coefficient);
   return (magic_number < num_of_currently_kept); 
  };
  **********************/

 
  void CorrectMaxAcceptedCat();

  int Category(int weight) 
  { 
   if (pick_given_ratio > 0) return weight;
   return 0;
  };

  long Weight(int cat) const { return cat; };
   
  int Category(Clause* c) 
  {
   if (c->MaxPriority()) return 0; 
   return Category(c->weight());
  };
  void InsertAs(Clause* c,int cat) { Clause::InsertBefore(c,&(category[cat])); StatInserted(cat); };
  bool EmptyCat(int cat) { return (category[cat].Next() == &(category[cat])); };
         
 public: // for debugging
  #ifndef NO_DEBUG
   bool clean() const;
   long CountCurrentlyKept(int cat);
   bool CheckStatistics();
  #endif 

 private: // main data
  bool max_priority_flag;
  long pick_given_ratio; // <=0 if no weighting is used (default)
  float magic_coefficient; // must be = 1 + (1/pick_given_ratio);
  int picked_by_weight;
  Clause category[NUMBER_OF_KEPT_CATEGORIES];
  int minCat;
  int maxCat;
  int max_accepted_cat;
 private: // statistics
  long currently_kept[NUMBER_OF_KEPT_CATEGORIES];
  long used[NUMBER_OF_KEPT_CATEGORIES];
  long removed[NUMBER_OF_KEPT_CATEGORIES];
  long num_of_currently_kept;
  long num_of_used;
  long num_of_removed;
  long num_of_cancelled;
  long numOfReachable;

 private: // iterator
  Clause* next_clause;
  int iter_cat; 
}; // class KEPT

inline void KEPT::CorrectMaxAcceptedCat()
{
 if (num_of_currently_kept < numOfReachable) 
  {
   max_accepted_cat = NUMBER_OF_KEPT_CATEGORIES-1;
   return;
  };
 if (Reachable(max_accepted_cat))
  {
   while (Reachable(max_accepted_cat + 1))
    {
     max_accepted_cat++;
     if (max_accepted_cat > maxCat) 
      {
       max_accepted_cat = NUMBER_OF_KEPT_CATEGORIES-1;
       return;
      };
    };
  }
 else
  {
   do
    {
     max_accepted_cat--;
    }
   while ((max_accepted_cat > 0) && (!(Reachable(max_accepted_cat))));
  };     
}; // void KEPT::CorrectMaxAcceptedCat()

inline Clause* KEPT::Next() 
{
 if (!(EmptyCat(0)))
  {
   Clause* res = category[0].Next();
   Clause::Remove(res);

   /*******
DF; cout << "INPUT\n";
DF; cout << minCat << " <= " << 0 << " <= " << maxCat << "\n";
DF; cout << "NUM " << res->Number() << "\n"; 
   *********/
   StatUsed(0); 
   return res;
  };
  
 if (pick_given_ratio > 0)
  {
   if (EmptyCat(minCat))
    {
     return (Clause*)0; 
    };
   if (picked_by_weight < pick_given_ratio)
    // pick by weight
    {
     picked_by_weight++;
     Clause* res = category[minCat].Next();
     Clause::Remove(res);

     /***********
DF; cout << "WEIGHT\n";
DF; cout << minCat << " <= " << minCat << " <= " << maxCat << "\n";
DF; cout << "NUM " << res->Number() << "\n"; 
     ************/

     StatUsed(minCat); 
     return res;
    }
   else // pick by history
    {
     unsigned long oldestNum = category[minCat].Next()->Number();
     int oldestCat = minCat;
     for (int cat = minCat + 1; cat<=maxCat; cat++)
      {
       if ((!(EmptyCat(cat))) && (category[cat].Next()->Number() < oldestNum))
        {
         oldestNum = category[cat].Next()->Number();
         oldestCat = cat;
        };
      };
     picked_by_weight = 0;
     Clause* res = category[oldestCat].Next();
     Clause::Remove(res);

     /*************
DF; cout << "AGE\n";
DF; cout << minCat << " <= " << oldestCat << " <= " << maxCat << "\n";
DF; cout << "NUM " << res->Number() << "\n"; 
     ************/

     StatUsed(oldestCat);
     return res;
    };
  }
 else
  {
   if (EmptyCat(0)) { return (Clause*)0; };
   Clause* res = category[0].Next();
   Clause::Remove(res);
   /************
DF; cout << "QUEUE\n";
DF; cout << minCat << " <= " << 0 << " <= " << maxCat << "\n";
DF; cout << "NUM " << res->Number() << "\n"; 
   *************/
   StatUsed(0);
   return res;
  };
}; //  Clause* KEPT::Next() 

inline void KEPT::DestroyAllClauses()
{
 CALL("DestroyAllClauses()");
 for (int cat = minCat; cat <= maxCat; cat++)
  {
   while (!(EmptyCat(cat)))
    {
     Clause* c = category[cat].Next();
     Clause::Remove(c);
     c->Destroy();
    };
  };
 Reset();  
}; // void KEPT::DestroyAllClauses()


#ifndef NO_DEBUG
 inline bool KEPT::clean() const
 {
  if (max_priority_flag 
      || num_of_currently_kept
      || num_of_used
      || num_of_removed) return false;
  if (numOfReachable != LONG_MAX) return false;
  for (int cat = 0; cat < NUMBER_OF_KEPT_CATEGORIES; cat++)
   {
    if (currently_kept[cat] || used[cat] || removed[cat]) return false;
    if (category[cat].Next() != (category + cat)) return false;
    if (category[cat].Previous() != (category + cat)) return false;  
   };         
  
 if (minCat || maxCat || picked_by_weight) return false;
  if (max_accepted_cat != NUMBER_OF_KEPT_CATEGORIES-1) return false;
   return true;
 }; // bool KEPT::clean() const

 inline long KEPT::CountCurrentlyKept(int cat)
  {
   Clause* cl = category[cat].Next();
   long res = 0L;
   while (cl != &(category[cat]))
    {
     res++;
     cl = cl->Next();
    };
   return res;
  };
   
 inline bool KEPT::CheckStatistics()
  {
   long sum_used = 0L;
   long sum_removed = 0L;
   long sum_currently_kept = 0L;
   for (int c = 0; c < NUMBER_OF_KEPT_CATEGORIES; c++)
    {
     sum_used += used[c];
     sum_removed += removed[c];
     sum_currently_kept += currently_kept[c];
        
     long real_curr_kept = CountCurrentlyKept(c);
     if (currently_kept[c] != real_curr_kept)
      {
       DOP(DMSG("KEPT::CheckStatistics() > ") << "currently_kept[" << c << "] CORRUPT!\n");
       DOP(DMSG("KEPT::CheckStatistics() > stat = ") << currently_kept[c] << ", real = " << real_curr_kept << "\n");
       return false;
      };
    };  
   if (sum_used != num_of_used) { return false; };
   if (sum_removed != num_of_removed) { return false; };
   if (sum_currently_kept != num_of_currently_kept) { return false; };
   if (minCat > maxCat) { return false; };
   if (minCat < maxCat) 
    {
     if (EmptyCat(minCat)) 
      {
       DOP(DMSG("KEPT::CheckStatistics() > ") 
            << "minCat(" << minCat << ") < maxCat("   
            << maxCat << ") but EMPTY!\n");
       return false;
      };
     if (EmptyCat(maxCat))
      {
       DOP(DMSG("KEPT::CheckStatistics() > ") 
           << "maxCat(" << maxCat << ") > minCat(" << minCat << ") but EMPTY!\n");
       return false;
      };
    };
   return true; 
  }; // bool KEPT::CheckStatistics() 
#endif 


//====================================================
#undef DEBUG_NAMESPACE
//====================================================
#endif
