//
// File:         term.h
// Description:  Representation of stored terms.
// Created:      Oct 15, 1999, 15:00
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#ifndef TERM_H
//=================================================
#define TERM_H
#include <stddef.h>
#include <iostream.h>
#include "jargon.h"
#include "HellDebugFlags.h"
#include "HellConst.h"
#include "Signature.h"
#include "GlobAlloc.h"
#include "Comparison.h"
#include "Stack.h"
#include "BitWord.h"
#include "SmallHeaderSkipList.h"
#ifdef DEBUG_ALLOC_OBJ_TYPE 
 #include "ClassDesc.h"
#endif
//=================================================

// Allignment of bits in the representation of symbols:

//       number    polarity splitting       arity        flag
//  [     16     ]   [1]       [1]       [    12     ]   [ 2 ] 

//=================================================

const ulong TermArityMask         = 0x00003FFC;
const ulong TermArityInvMask     = 0xFFFF9003;
const ulong TermSplittingMask     = 0x00004000;
const ulong TermSplittingInvMask = 0xFFFFBFFF; 
const ulong TermPolarityMask      = 0x00008000;
const ulong TermPolarityInvMask  = 0xFFFF7FFF;
const ulong TermVarMask           = 0xFFFF0000;
const ulong TermVarInvMask       = 0x0000FFFF;   
const ulong TermFunctorMask       = 0xFFFF0000;
const ulong TermFunctorInvMask   = 0x0000FFFF;
const ulong TermFlagMask          = 0x00000003;
const ulong TermFlagInvMask      = 0xFFFFFFFC;

enum TermFlag 
 { 
  TermRef      = 0x00000000, 
  TermVar      = 0x00000001, 
  TermComplex  = 0x00000002 
 };

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TERM
 #define DEBUG_NAMESPACE "TERM"
#endif
#include "debugMacros.h"
//=================================================

class TERM
{
 public:
  typedef ulong Arity;
  typedef ulong Functor;
  typedef ulong Polarity;
  static const Polarity PositivePolarity = 0UL;
  static const Polarity NegativePolarity = 1UL; 
  inline Polarity inverse(const Polarity& pol) { return !pol; };
  class OpenCompact;
  class Iterator;
  class IteratorWithDepth;
  class Traversal;
  class OpenUnshared;
  class FastOpenUnshared;
  class SubtermMask                              
  {
   // ATTENTION! THIS WILL NOT WORK IF FUNCTORS CAN HAVE ARITY Greater THAN 32 
   public:
    SubtermMask() : _content(0UL) {};
    operator bool() const { return (bool)_content; };
    bool bit(const ulong& n) const { return _content.bit(n); };
    void set(const ulong& n) 
    {
     CALL("SubtermMask::set(const ulong& n)");
     ASSERT(n < HellConst::MaxNumOfTermArguments);
     _content.set(n);
    };
    void clear() { _content.init(); };
    #ifndef NO_DEBUG_VIS
     ostream& output(ostream& str) const
     {
      for (ulong n = 0; n < HellConst::MaxNumOfTermArguments; n++)
       {
	if (bit(n)) str << n << ' ';
       };
      return str;
     };
    #endif
   private:
    BitWord _content;
  }; // class SubtermMask
  //typedef SkipList<GlobAlloc,SubtermMask,ulong,2,28,TERM> OccList;
  typedef SmallHeaderSkipList<GlobAlloc,SubtermMask,ulong,2,TERM> OccList;

  class InitStatic
  {
   public: 
    InitStatic() 
    {
     if (!_count) { TERM::initStatic(); _count++; };    
    };
    ~InitStatic() {};
   private:
    long _count;
  };

 public: 
 
  //======= construction/initialisation/destruction

  static void setSignature(Signature* sig) 
  {
   _signature = sig; 
  };  
  TERM() {};
  TERM(const ulong& cont) : content(cont) {};
  TERM(TERM* const & r) : ref(r) {};
  TERM(const TERM& t) : content(t.content) {};
  TERM(TermFlag,ulong var) // variable
   : content(var) 
  {
   (content <<= 16) |= TermVar;
  };
 
  TERM(Arity a,Functor f,Polarity p) : content(f)
  { 
   CALL("costructor TERM(Arity a,Functor f,Polarity p)");

   ASSERT(a <= HellConst::MaxNumOfTermArguments);
   ASSERT(((ulong)p) <= 1);
   (((((content <<= 1) |= p) <<= 13) |= a) <<= 2) |= TermComplex;  
  }; 

  void Make(const ulong& c) { content = c; };
  void MakeVar(ulong var)
  {   
   ((content = var) <<= 16) |= TermVar;   
  };
  void MakeComplex(Arity a,Functor f,Polarity p)
  {
   CALL("MakeComplex(Arity a,Functor f,Polarity p)");
   ASSERT(a <= HellConst::MaxNumOfTermArguments);
   ((((((content = f) <<= 1) |= p) <<= 13) |= a) <<= 2) |= TermComplex;  
  }; 
  void MakeReference(TERM* r) { ref = r; };
  void MakeHeader(ulong hdNum) // without arity!
  {
   ulong fun = hdNum/2;
   ulong pol = hdNum%2;
   MakeComplex((TERM::Arity)0,(TERM::Functor)fun,(Polarity)pol); 
  };

  void MakeSplitting() 
  {
   content |= TermSplittingMask;
  }; 

  //========== getting properties: ================

  TermFlag Flag() const { return TermFlag((content << 30) >> 30); };

  Arity arity() const { return (content << 18) >> 20; };

  Polarity polarity() const { return (content << 16) >> 31; };
  bool Positive() const { return !polarity(); };
  bool Negative() const { return (bool)polarity(); };
  bool IsSplitting() const 
  {
   return (content << 17) >> 31;
  };
  Functor functor() const { return content >> 16; }; 
  ulong HeaderNum() const 
  {
   return functor()*2 + polarity();
  };
  static Polarity HeaderPolarity(ulong headerNum) 
  {
   return (headerNum % 2);
  };
  static bool HeaderPositive(ulong headerNum) 
  {
   return !(headerNum % 2);
  };
  static bool HeaderNegative(ulong headerNum) 
  {
   return (bool)(headerNum % 2);
  };
  ulong var() const { return content >> 16; }; 
  const ulong& Content() const { return content; };

  //=========== changing properties: ==================  
 
  void SetArity(ulong a)
  {
   CALL("SetArity(ulong a)");
   ASSERT(a <= HellConst::MaxNumOfTermArguments);
   (content &= TermArityInvMask) |= (a <<= 2);    
  };
  void SetPolarity(Polarity p)
  {
   (content &= TermPolarityInvMask) |= (p <<= 15);   
  };   
  void SetVar(ulong v) // doesn't set the flag
  {
   (content &= TermVarInvMask) |= (v << 16);    
  };
  void makePositive() { SetPolarity(PositivePolarity); };
  void makeNegative() { SetPolarity(NegativePolarity); };
  void Inversepolarity() 
  { 
   if (Positive()) { SetPolarity(NegativePolarity); } else SetPolarity(PositivePolarity);
  };
  void SetFunctor(ulong f)
  {
   (content &= TermFunctorInvMask) |= (f <<= 16);
  };

  void SetFlag(TermFlag fl) 
  {
   CALL("SetFlag(TermFlag fl)"); 
   ASSERT(fl <= TermComplex);
   ((content >>= 2) <<= 2) |= fl;
  };   
  void MakeUnordered() { SetFunctor(HellConst::UnordEqNum); };
  void MakeOrdered() { SetFunctor(HellConst::OrdEqNum); };  
  
  //====== allocation: ===============================

  void* operator new(size_t,unsigned arity) 
  {    
   CALL("operator new(size_t,unsigned arity)");
   // arity can be 0 when literals are allocated for an empty clause
   return GlobAlloc::allocate((arity + 1)*sizeof(TERM) + sizeof(void*));
  };
  void operator delete(void* obj) 
  {
   CALL("operator delete(void* obj)");
   ASSERT(((TERM*)obj)->arity()); // constants are always shared in some way  
   GlobAlloc::deallocate(obj,(((TERM*)obj)->arity() + 1)*sizeof(TERM) + sizeof(void*));
  }; 

  static TERM* AllocLit(const TERM& header)
  {
   TERM* res = (TERM*)GlobAlloc::allocate((header.arity() + 1)*sizeof(TERM) + sizeof(void*));
   res++;
   *res = header;
   return res;
  };  
  void DeleteLit() 
  { 
   GlobAlloc::deallocate(this - 1,(arity() + 1)*sizeof(TERM) + sizeof(void*));
  };

  //===================== copying: ====================

  TERM& operator=(const TERM& t) { content = t.content; return *this; };
  TERM* CopyTop() const { return new(arity()) TERM(*this); };
  TERM* CopyLitTop() const { return AllocLit(*this); };

  //=============== comparison: =================================

  bool operator==(const TERM& t) const { return content == t.content; };
  bool operator!=(const TERM& t) const { return content != t.content; };
  bool operator<(const TERM& t) const { return content < t.content; };
  bool operator>(const TERM& t) const { return content > t.content; };
  bool greaterPrecedence(ulong f) const
  {
   return (functor() < f);
  };
  bool greaterPrecedence(const TERM& f) const 
  {
   return greaterPrecedence((ulong)(f.functor())); 
  };
  Comparison CompareFunctor(ulong f) const
  {
   if (functor() < f) return Greater;     
   if (functor() > f) return Less; 
   return Equal;
  };
  Comparison CompareFunctor(const TERM& f) const 
  {
   CALL("CompareFunctor(const TERM& f) const");
   return CompareFunctor((ulong)(f.functor()));  
  };   
  bool ComplementaryHeader(const TERM& hd) const 
  {
   return (functor() == hd.functor()) && (polarity() != hd.polarity());
  };
  static bool TERM::EqualComplexTerms(const TERM* t1,const TERM* t2);
    
  Comparison CompareLex(const TERM* t) const;

    
  //======= recognising: ========================

  bool IsReference() const { return !Flag(); };
  bool isVariable() const { return Flag() == TermVar; };
  bool isNotVariable() const { return Flag() != TermVar; };  
  bool isComplex() const { return Flag() == TermComplex; };
  bool IsConstant() const { return isComplex() && (!arity()); };
  bool InNonconstFunc() const { return isComplex() && arity(); };
  bool IsNull() const { return Flag() == TermComplex; };
  bool IsSymLitHeader() const 
  {
    //return false; 
   return IsEquality(); 
  };
  bool IsEquality() const { return HellConst::isEquality(functor()); };
  bool IsOrderedEq() const 
  {
   CALL("IsOrderedEq()");
   ASSERT(IsEquality());
   return HellConst::isOrderedEq(functor());
  };
  bool IsUnorderedEq() const 
  {
   CALL("IsUnorderedEq()");
   ASSERT(IsEquality());  
   return HellConst::isUnorderedEq(functor());
  };

  bool IsReservedHeader() const { return HellConst::isReservedPred(functor()); };

  bool IsDefHeader() const // all arguments are different variables
  {
   CALL("IsDefHeader()");
   ASSERT(isComplex());
   for (const TERM* ar1 = Args(); !ar1->IsNull(); ar1 = ar1->Next())
    {
     if (!ar1->isVariable()) return false; 
     for (const TERM* ar2 = Args(); ar2 != ar1; ar2 = ar2->Next())     
      if (*ar2 == *ar1) return false;
    };
   return true;
  };
  
  static bool Symmetry(const TERM* t1,const TERM* t2)
  {
   CALL("Symmetry(const TERM* t1,const TERM* t2)");
   ASSERT(t1->isComplex() && t2->isComplex());
   if (t1->functor() != t2->functor()) return false;    
   if (t1->arity() != 2) return false;
   if ((!t1->IsDefHeader()) || (!t2->IsDefHeader())) return false;
   if ((*(t1->Arg1()) != *(t2->Arg2()))
       || (*(t1->Arg2()) != *(t2->Arg1()))) return false; 
   return true;
  };

  //========== surfing term structure: =================

  TERM* First() const { return ref; };   
  TERM* Args() { return this + arity(); };
  const TERM* Args() const { return this + arity(); };
  TERM* Next() { return this - 1; };
  const TERM* Next() const { return this - 1; };
  TERM* Previous() { return this + 1; };
  const TERM* Previous() const { return this + 1; };
  TERM* NthArg(ulong n) { return Args() - n; };   
  const TERM* NthArg(ulong n) const { return Args() - n; };
  TERM* Arg1() { return Args(); };
  const TERM* Arg1() const { return Args(); };
  TERM* Arg2() { return Args()->Next(); };
  const TERM* Arg2() const { return Args()->Next(); };

  //============= creating term structure: =========

  void SetFirst(TERM* t) { ref = t; };      

  //================ miscelanious: =============

  ulong size() const;
  ulong depth() const; 

  TERM unsharedCopy() const
  {
   CALL("unsharedCopy()"); 
   if (isVariable()) return (*this);
   ASSERT(IsReference());
   const TERM* t = First();
   ulong arity = t->arity();
   TERM* nt;
   if (arity) 
    { 
     nt = new (arity) TERM((TERM::Arity)arity,(TERM::Functor)t->functor(),t->polarity()); 
     for (ulong i = 0; i < arity; i++)
      *(nt->NthArg(i)) = t->NthArg(i)->unsharedCopy();  
    }
   else nt = sharedConst(t->functor());
   return TERM(nt);
  };
 
  TERM* unsharedLitCopy() const
  {
   CALL("unsharedLitCopy() const");
   TERM* res = AllocLit(*this);   
   for (ulong i = 0; i < arity(); i++)
    *(res->NthArg(i)) = NthArg(i)->unsharedCopy();
   ASSERT(EqualComplexTerms(res,this));
   return res;
  };      

  void destroyUnsharedComplex()
  {
   CALL("destroyUnsharedComplex()");
   ASSERT(isComplex());
   if (arity())
    {
     Stack<TERM*,HellConst::MaxTermSize> subterms;
     for (TERM* argRef = Args(); !argRef->IsNull(); argRef = argRef->Next())
      if (argRef->IsReference()) subterms.push(argRef->First());
     delete this;

     TERM* currSubterm;
     while (subterms.nonempty())
      {
       currSubterm = subterms.pop();
       if (currSubterm->arity())
	{
         ASSERT(currSubterm->isComplex());
         for (TERM* argRef = currSubterm->Args(); !argRef->IsNull(); argRef = argRef->Next())
          if (argRef->IsReference()) subterms.push(argRef->First());
         delete currSubterm;
        };
      };
    };
  }; // destroyUnsharedComplex()

  void destroyUnsharedLit()
  {
   CALL("destroyUnsharedLit()");
   ASSERT(isComplex());      
   for (TERM* argRef = Args(); !argRef->IsNull(); argRef = argRef->Next())
    if (argRef->IsReference()) argRef->First()->destroyUnsharedComplex();
   DeleteLit();
  };
  
  void Destroy() { Destroy(this); };
  void DestroyLit() { DestroyLit(this); };
  static void Destroy(TERM* t)
  {
   if (t->arity())
    {
     for (TERM* arg = t->Args(); !t->IsNull(); t = t->Next())
      if (arg->IsReference()) Destroy(arg->First());
     delete t;
    };
  };
  static void DestroyLit(TERM* t)
  {
   for (TERM* arg = t->Args(); !t->IsNull(); t = t->Next())
    if (arg->IsReference()) Destroy(arg->First());
   t->DeleteLit();
  }; 


  //TERM* Info() { return this + arity() + 1; }; // for nonliteral terms
 
  OccList* Occurences() 
  {
   return (OccList*)((this + arity() + 1)->First()); 
  }; 
  const OccList* Occurences() const
  {
   return (const OccList*)((this + arity() + 1)->First()); 
  }; 
  void InitOccurences() { (this + arity() + 1)->SetFirst((TERM*)0); };
  void DestroyOccurences() 
  {
   CALL("DestroyOccurences()");               
   ASSERT(isComplex());
   OccList* occList = Occurences();
   if (occList) 
    {
     occList->destroy();
     InitOccurences();
    };
  };
  void RegisterOccurence(ulong clauseNum,ulong litNum) 
  {
   CALL("RegisterOccurence(ulong clauseNum,ulong litNum)");
   ASSERT(isComplex());
   OccList* occList = Occurences();
   if (!occList) { occList = new OccList(); (this + arity() + 1)->SetFirst((TERM*)occList); };
   bool newNode;
   OccList::Node* node = occList->insert(clauseNum,newNode);
   if (newNode) node->value().clear();
   node->value().set(litNum);
  };

  void RemoveOccurences(ulong clauseNum)
  {
   CALL("RemoveOccurences(ulong clauseNum)");
   ASSERT(isComplex());
   OccList* occList = Occurences();
   if (!occList) return;
   SubtermMask dummy;
   occList->remove(clauseNum,dummy);
   if (!(*occList)) { delete occList; InitOccurences(); };
  };


  void** InfoField() const { return (void**)(this - 1); }; // for literals
  inline bool Plain() const; // "all arguments are pairwise distinct variables"  
  static ulong UnordHeaderNum(ulong header_num)
   {
    CALL("UnordHeaderNum(ulong header_num)");
    ASSERT((header_num/2 == HellConst::UnordEqNum) || (header_num/2 == HellConst::OrdEqNum));
    return HellConst::UnordEqNum*2 + header_num%2; 
   };
  
  static TERM* sharedConst(ulong num) { return _sharedConst + (2*num); };

  //================ output: ==================
  static ostream& output(ostream& str,const TERM& t) 
  { 
   CALL("output(ostream& str,const TERM& t)");
   COP("COP0");       
   if (t.isVariable()) { COP("VAR"); return str << 'X' << t.var(); }
   else
    if (t.isComplex())
     {
      COP("COMPLEX");
      if (t.IsSplitting()) str << "[";
      if (t.Negative()) str << ((t.IsEquality()) ? '!' : '~');
      if (t.IsEquality()) 
	if (t.IsOrderedEq()) { return str << "=="; }    
        else return str << '=';
      // not equality      
      COP("COP10");    
      str << _signature->functorName(t.functor());
      if (t.IsSplitting()) str << "]";
      return str;
     }
    else // t.IsReference()
     { 
      COP("REFERENCE");
      return output(str,t.First());
     };         
  }; // ostream& output(ostream& str,const TERM& t) 

  static ostream& output(ostream& str,const TERM* t) 
  {   
   if (t->IsEquality())
    { 
     output(str,*(t->Args()));
     output(str,*t);
     return output(str,*(t->Args()->Next()));   
    };
   // !t->IsEquality()
   output(str,*t);
   if (!t->arity()) return str; 
   str << '(';
   for (const TERM* arg = t->Args(); !arg->IsNull(); arg = arg->Next())
    { 
     output(str,*arg);
     if (!arg->Next()->IsNull()) str << ',';
    };
   return str << ')';
  }; //ostream& output(ostream& str,const TERM* t)  
 
  ostream& output(ostream& str) const { return output(str,this); };
  
  ostream& outputInTPTPFormat(ostream& str) const
  {
   CALL("outputInTPTPFormat(ostream& str) const");
   if (isVariable()) { return str << 'X' << var(); }
   else
    if (isComplex())
     {
      if (IsEquality()) 
       {
	str << "equal";
       }
      else str << _signature->functorName(functor());
      if (!arity()) return str;
      str << '(';
      for (const TERM* arg = Args(); !arg->IsNull(); arg = arg->Next())
       { 
        arg->outputInTPTPFormat(str);
        if (!arg->Next()->IsNull()) str << ',';
       };
      return str << ')';
     }
    else // IsReference() 
     return First()->outputInTPTPFormat(str);      
  }; // ostream& outputInTPTPFormat(ostream& str) const

  ostream& outputAsPrologTerm(ostream& str) const
  {
   CALL("TERM::outputAsPrologTerm(ostream& str) const");
   if (isVariable()) { return str << "\'X" << var() << '\''; }
   else 
    if (isComplex())
     {
      if (IsEquality()) 
       { 
	str << '('; 
	Arg1()->outputAsPrologTerm(str);
	if (IsOrderedEq())
	 {
	  str << " => ";
         }
        else str << " = "; 
        Arg2()->outputAsPrologTerm(str);
        str << ')';
        return str;
       }
      else str << _signature->functorName(functor());
      if (!arity()) return str;
      str << '(';
      for (const TERM* arg = Args(); !arg->IsNull(); arg = arg->Next())
       { 
        arg->outputAsPrologTerm(str);
        if (!arg->Next()->IsNull()) str << ',';
       };
      return str << ')';
     }
    else // IsReference() 
     return First()->outputAsPrologTerm(str);  
  }; 

  static const TERM& dummyPositiveHeader0() { return _dummyPositiveHeader0; };
  static const TERM& dummyPositiveHeader1() { return _dummyPositiveHeader1; };
  static const TERM& positiveOrdEqHeader() { return _positiveOrdEqHeader; };
  static const TERM& negativeOrdEqHeader() { return _negativeOrdEqHeader; };
  static const TERM& positiveUnordEqHeader() { return _positiveUnordEqHeader; };
  static const TERM& negativeUnordEqHeader() { return _negativeUnordEqHeader; };

 private:
  static void initStatic()
  {    
   for (ulong c = 0; c < HellConst::MaxNumOfFunctors; c++)
    {
     sharedConst(c)->MakeComplex((TERM::Arity)0,(TERM::Functor)c,TERM::PositivePolarity);
     sharedConst(c)->InitOccurences();
    };
   _dummyPositiveHeader0.MakeComplex((TERM::Arity)0,(TERM::Functor)HellConst::ReservedPred0,TERM::PositivePolarity);
   _dummyPositiveHeader1.MakeComplex((TERM::Arity)1,(TERM::Functor)HellConst::ReservedPred1,TERM::PositivePolarity);
   _positiveOrdEqHeader.MakeComplex((TERM::Arity)2,(TERM::Functor)HellConst::OrdEqNum,TERM::PositivePolarity);
   _negativeOrdEqHeader.MakeComplex((TERM::Arity)2,(TERM::Functor)HellConst::OrdEqNum,TERM::NegativePolarity);
   _positiveUnordEqHeader.MakeComplex((TERM::Arity)2,(TERM::Functor)HellConst::UnordEqNum,TERM::PositivePolarity);
   _negativeUnordEqHeader.MakeComplex((TERM::Arity)2,(TERM::Functor)HellConst::UnordEqNum,TERM::NegativePolarity);
  };

 private:
  union 
   {
    ulong content;
    TERM* ref;
   };
  static TERM _sharedConst[2*HellConst::MaxNumOfFunctors];
  static Signature* _signature;
  static TERM _dummyPositiveHeader0;
  static TERM _dummyPositiveHeader1;
  static TERM _positiveOrdEqHeader;
  static TERM _negativeOrdEqHeader;
  static TERM _positiveUnordEqHeader;
  static TERM _negativeUnordEqHeader;  
 friend InitStatic;
};




//=================================================================
#undef DEBUG_NAMESPACE   
#ifdef DEBUG_TERM
 #define DEBUG_NAMESPACE "TERM::Traversal"
#endif 
#include "debugMacros.h"
//=================================================

class TERM::Traversal 
{
 public:
  enum State { Func, Var, End };
 public:
  Traversal() : subterms()
  {
   CALLM("?","constructor Traversal()");
   DOP(_debugName = "??");
   wrapper[0] = TERM::dummyPositiveHeader1();   
   //wrapper[0].SetFlag(TermComplex);
   ASSERT_IN(wrapper->IsNull(),"end");
  }; 

  Traversal(const char* name) : subterms()
  {
   CALLM(name,"constructor Traversal()");
   DOP(_debugName = name);
   wrapper[0] = TERM::dummyPositiveHeader1();  
   //wrapper[0].SetFlag(TermComplex);  
   ASSERT_IN(wrapper->IsNull(),"end");
  }; 
  
  Traversal(const TERM& t) : subterms()
  {
   CALLM("?","constructor Traversal(const TERM& t)");
   DOP(_debugName = "???");
   wrapper[0] = TERM::dummyPositiveHeader1();  
   //wrapper[0].SetFlag(TermComplex);

   wrapper[1] = t;
   trm = wrapper + 1;
   if (t.isVariable()) 
    {
     currState = Var;
     sym = trm;
    }
   else // t is reference to a complex term
    {
     currState = Func;
     sym = trm->First();
    }; 
   ASSERT_IN(wrapper->IsNull(),"end");
  };

  Traversal(const char* name,const TERM& t) : subterms()
  {
   CALLM(name,"constructor Traversal(const char* name,const TERM& t)");
   DOP(_debugName = name);
   wrapper[0] = TERM::dummyPositiveHeader1();  
   //wrapper[0].SetFlag(TermComplex);

   wrapper[1] = t;
   trm = wrapper + 1;
   if (t.isVariable()) 
    {
     currState = Var;
     sym = trm;
    }
   else // t is reference to a complex term
    {
     currState = Func;
     sym = trm->First();
    }; 
   ASSERT_IN(wrapper->IsNull(),"end");
  };

  Traversal(const TERM* t) : subterms()
  {
   CALLM("?","constructor Traversal(const TERM* t)");
   DOP(_debugName = "???");
   ASSERT(t->isComplex());
   wrapper[0] = TERM::dummyPositiveHeader1();  
   //wrapper[0].SetFlag(TermComplex);

   wrapper[1].SetFirst((TERM*)t);
   trm = wrapper + 1;
   currState = Func;
   sym = trm->First();
   ASSERT_IN(wrapper->IsNull(),"end");
  };
  Traversal(const char* name,const TERM* t) : subterms()
  {
   CALLM(name,"constructor Traversal(const char* name,const TERM* t)");
   DOP(_debugName = name);
   ASSERT(t->isComplex());
   wrapper[0] = TERM::dummyPositiveHeader1();  
   //wrapper[0].SetFlag(TermComplex);

   wrapper[1].SetFirst((TERM*)t);
   trm = wrapper + 1;
   currState = Func;
   sym = trm->First();
   ASSERT_IN(wrapper->IsNull(),"end");
  };


  void reset(const TERM& t)
  {
   CALLM(_debugName,"reset(const TERM& t)");
   ASSERT(wrapper->IsNull());
   subterms.reset();
   wrapper[1] = t;
   trm = wrapper + 1;
   if (t.isVariable()) 
    {
     currState = Var;
     sym = trm;
    }
   else
    {
     currState = Func;
     sym = trm->First();
    }; 
   ASSERT_IN(wrapper->IsNull(),"end");
  };

  void reset(const TERM* t)
  {
   CALLM(_debugName,"reset(const TERM& t)");
   ASSERT(wrapper->IsNull());
   ASSERT(t->isComplex());
   subterms.reset();
   wrapper[1].SetFirst((TERM*)t);
   trm = wrapper + 1;
   currState = Func;
   sym = trm->First();
   ASSERT_IN(wrapper->IsNull(),"end");
  };

  const State& state() const { return currState; };
  operator bool() const { return currState != End; };

  void next()
  {
   CALLM(_debugName,"next()");
   ASSERT(wrapper->IsNull());
   const TERM* nextTrm = trm->Next();
   if ((currState == Func) && (sym->arity()))
    {
     if (!nextTrm->IsNull()) subterms.push(nextTrm);
     trm = sym->Args();
    }
   else // variable or constant
    {
     if (nextTrm->IsNull())
      { 
       if (subterms.empty()) { currState = End; ASSERT(wrapper->IsNull()); return; }
       else 
        {          
         trm = subterms.pop();
        };
      }
     else 
      {
       trm = nextTrm;
      };
    };
   if (trm->isVariable())
    {
     currState = Var;
     sym = trm;
    }
   else // reference to a nonvariable term
    {
     currState = Func;
     sym = trm->First();
    };
   ASSERT(wrapper->IsNull());
  };
  
  void after()
  {
   CALLM(_debugName,"after()");
   ASSERT(wrapper->IsNull());
   trm = trm->Next();
   if (trm->IsNull()) 
    if (subterms.empty()) { currState = End; ASSERT(wrapper->IsNull()); return; } 
    else trm = subterms.pop();
   if (trm->isVariable())
    {
     currState = Var;
     sym = trm;
    }
   else // reference to a nonvariable term
    {
     currState = Func;
     sym = trm->First();
    };   
   ASSERT(wrapper->IsNull());
  };

  const TERM& symbol() const { return *sym; }; 
  const TERM& term() const { return *trm; };

  #ifndef NO_DEBUG
   bool checkIntegrity() 
   {
    return wrapper->IsNull();
   };
  #endif

 private:
  TERM wrapper[2];
  State currState;
  Stack<const TERM*,HellConst::MaxTermSize> subterms;
  const TERM* trm;
  const TERM* sym;
  #ifndef NO_DEBUG
   const char* _debugName;
  #endif
}; // class TERM::Traversal

//=================================================
#undef DEBUG_NAMESPACE  
#ifdef DEBUG_TERM
 #define DEBUG_NAMESPACE "TERM::Iterator"
#endif
#include "debugMacros.h"
//=================================================
class TERM::Iterator // for complex terms only!
{
 private:
  Stack<TERM*,HellConst::MaxTermSize> subterms;   
  TERM* sym;
  bool isVar;  
 public:
  Iterator() : subterms() {};   
  Iterator(const TERM* t) 
   : subterms() 
   { 
    sym = (TERM*)t;
    isVar = false;
   };     
  void Reset(const TERM* t) 
  { 
   subterms.reset(); 
   sym = (TERM*)t;
   isVar = false;
  };                   
  TERM Symbol() const { return *sym; };
  TERM& symbolRef() { return *sym; };
  const TERM& symbolRef() const { return *sym; };
  TERM* CurrPos() const { return sym; };
  bool CurrentSymIsVar() const { return isVar; };
 public:
  bool Next()
  {
   if (isVar)
    {
     sym = sym->Next();
    }
   else
    {
     sym = sym->Args();  
    };
   if (sym->IsNull()) 
    {
     if (subterms.empty()) { return false; };
     sym = subterms.pop();
    };
   if (!((isVar = sym->isVariable()))) // reference
    {
     if (!(sym->Next()->IsNull())) { subterms.push(sym->Next()); };
     sym = sym->First(); 
    };
   return true;
  };
  bool NextVar()
  {
   while (Next())
    {
     if (isVar) { return true; };
    };
   return false;
  };
  bool SkipTerm()
  {
   if (isVar)
    {
     sym = sym->Next();
     if (sym->IsNull()) 
      {
       if (subterms.empty()) { return false; };
       sym = subterms.pop();
      };
    }
   else
    {
     if (subterms.empty()) { return false; };
     sym = subterms.pop();
    }; 
   if (!((isVar = sym->isVariable()))) // reference
    {
     if (!(sym->Next()->IsNull())) { subterms.push(sym->Next()); };
     sym = sym->First(); 
    };
   return true; 
  }; 
}; // class TERM::Iterator 
//=================================================================
#undef DEBUG_NAMESPACE  
#ifdef DEBUG_TERM
 #define DEBUG_NAMESPACE "BracketCounter"
#endif
#include "debugMacros.h"
//=================================================
template <ulong MaxDepth>
class BracketCounter
{
 private:
  int holes[MaxDepth];
  int* next_holes;
  int curr_holes;  
 public:
  BracketCounter() : curr_holes(0) {  next_holes = holes; };
  void Reset() { curr_holes = 0; next_holes = holes; };
  bool IsOpen() const { return curr_holes; };
  bool AllClosed() const { return !(curr_holes); };
  void Open(int arity) 
  {
   CALL("Open(int arity)");  
   ASSERT(arity > 0); 
   ASSERT(next_holes < (holes + MaxDepth));
   *next_holes = curr_holes-1;
   next_holes++;
   curr_holes = arity;
  };
  unsigned Close()
  {
   curr_holes--;
   if (curr_holes) return 0; // curr_holes can be -1 ! 
   unsigned closed = 1;
   pop_holes:
    if (next_holes == (holes + 1)) return closed;
    next_holes--;
    ASSERT(next_holes >= holes); 
    curr_holes = (*next_holes);
    if (curr_holes) return closed;
    closed++;
    goto pop_holes;
  };
}; // class BracketCounter<ulong MaxDepth>
//=================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TERM
 #define DEBUG_NAMESPACE "DepthCounter"
#endif
#include "debugMacros.h"
//=================================================
class DepthCounter
{
 private:
  unsigned depth;
 public: 
  DepthCounter() {};
  void Reset(unsigned d) { depth = d; };
  unsigned Depth() const { return depth; };
  void Down() { depth++; };
  void Up(unsigned n) { depth -= n; };
};
//=================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TERM
 #define DEBUG_NAMESPACE "TERM::IteratorWithDepth"
#endif
#include "debugMacros.h"
//=================================================
class TERM::IteratorWithDepth  // for complex terms only!
 : private TERM::Iterator,
   private BracketCounter<HellConst::MaxTermDepth>,
   private DepthCounter 
{
 public:
  IteratorWithDepth() {};
  void Reset(TERM* const t,unsigned d) 
  {
   //TERM::Iterator::Reset(t);
   Iterator::Reset(t);
   BracketCounter<HellConst::MaxTermDepth>::Reset();
   DepthCounter::Reset(d);
  };
  TERM Symbol() const 
  { 
   return Iterator::Symbol();   
   //return TERM::Iterator::Symbol(); 
  };
  bool CurrentSymIsVar() const { return Iterator::CurrentSymIsVar(); };
  bool Next()
  {
   if (Symbol().arity())
    {
     BracketCounter<HellConst::MaxTermDepth>::Open(Symbol().arity());
     DepthCounter::Down();
    }
   else DepthCounter::Up(BracketCounter<HellConst::MaxTermDepth>::Close());
   return Iterator::Next();
  };
  unsigned Depth() const { return DepthCounter::Depth(); }; 
 }; // class TERM::IteratorWithDepth
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TERM
 #define DEBUG_NAMESPACE "SmartDepthCounter<ulong MaxDepth>"
#endif
#include "debugMacros.h"
//=================================================
template <ulong MaxDepth>
class SmartDepthCounter 
 : private BracketCounter<MaxDepth>, private DepthCounter
{
 public:  
  SmartDepthCounter() : BracketCounter<MaxDepth>(), DepthCounter() {};
  void Reset(const unsigned d)
  {
   BracketCounter<MaxDepth>::Reset();
   DepthCounter::Reset(d);
  };
  void NonconstantFunctor(int arity) 
   {
    CALL("NonconstantFunctor(int arity)");
    ASSERT(arity > 0);
    ASSERT(DepthCounter::Depth() < HellConst::MaxTermDepth);
    BracketCounter<MaxDepth>::Open(arity);
    DepthCounter::Down();
   };
  void VarOrConst() { DepthCounter::Up(BracketCounter<MaxDepth>::Close()); };
  void NextSym(const TERM& sym)
  {
   if (sym.arity()) { NonconstantFunctor(sym.arity()); }
   else VarOrConst();       
  }; 
  unsigned Depth() { return DepthCounter::Depth(); };
}; // class SmartDepthCounter<ulong MaxDepth>

//=================================================================
#undef DEBUG_NAMESPACE   
#ifdef DEBUG_TERM
 #define DEBUG_NAMESPACE "TERM::OpenCompact"
#endif 
#include "debugMacros.h"
//=================================================

class TERM::OpenCompact
{
  // Don't forget to give it enough memory!
  // HellConst::MaxTermSize may not be enough because every complex subterm
  // takes one additional word for the pointer to it. 
  // 2*HellConst::MaxTermSize will suffice.
 public:
  OpenCompact() 
   : _holes("_holes"), 
     _numOfHoles("_numOfHoles")
     #ifdef DEBUG_NAMESPACE
      , _debugDepthCounter()
     #endif
  {
  };
  OpenCompact(TERM* mem) 
   : _holes("_holes"), 
     _numOfHoles("_numOfHoles")
     #ifdef DEBUG_NAMESPACE
      , _debugDepthCounter()
     #endif
  {   
   _currHole = &_term;
   _currNumOfHoles = 1;
   _freeMem = mem;
   DOP(_debugDepthCounter.Reset(0));
   DOP(_debugSizeCounter = 1);
  };
  void reset(TERM* mem)
  {
   _holes.reset();
   _numOfHoles.reset();
   _currHole = &_term;
   _currNumOfHoles = 1;
   _freeMem = mem;
   DOP(_debugDepthCounter.Reset(0));
   DOP(_debugSizeCounter = 1);
  };
  void pushVar(ulong var) 
  {
   CALL("pushVar(ulong var)");
   DOP(_debugDepthCounter.VarOrConst());
   ASSERT((_debugDepthCounter.Depth() >= 0) && (_debugDepthCounter.Depth() <= HellConst::MaxTermDepth));
   ASSERT(_currHole);
   _currHole->MakeVar(var);
   _currNumOfHoles--;
   if (_currNumOfHoles) 
    {
     _currHole = _currHole->Next();
    }
   else
    if (_holes.nonempty())
     {
      _currHole = _holes.pop();  
      _currNumOfHoles = _numOfHoles.pop();
      ASSERT(_currNumOfHoles);
     }
    else
     {
      DOP(_currHole = (TERM*)0);     
     };
  };
  void pushVar(const TERM& var)
  {
   CALL("pushVar(const TERM& var)");
   ASSERT(var.isVariable());
   DOP(_debugDepthCounter.VarOrConst());
   ASSERT((_debugDepthCounter.Depth() >= 0) && (_debugDepthCounter.Depth() <= HellConst::MaxTermDepth));
   ASSERT(_currHole);
   *_currHole = var;
   _currNumOfHoles--;
   if (_currNumOfHoles) 
    {
     _currHole = _currHole->Next();
    }
   else
    if (_holes.nonempty())
     {
      _currHole = _holes.pop();  
      _currNumOfHoles = _numOfHoles.pop();
      ASSERT(_currNumOfHoles);
     }
    else
     {
      DOP(_currHole = (TERM*)0);
     };
  };
  void pushConst(const TERM& c) 
  {
   CALL("pushConst(const TERM& c)");
   ASSERT(c.isComplex());
   ASSERT(!c.arity());
   DOP(_debugDepthCounter.VarOrConst());
   ASSERT((_debugDepthCounter.Depth() >= 0) && (_debugDepthCounter.Depth() <= HellConst::MaxTermDepth));
   ASSERT(_currHole);
   _currHole->SetFirst(_freeMem);   
   _currNumOfHoles--;
   *_freeMem = c;
   _freeMem++;
   if (_currNumOfHoles) 
    {
     _currHole = _currHole->Next();
    }
   else
    if (_holes.nonempty())
     {
      _currHole = _holes.pop();  
      _currNumOfHoles = _numOfHoles.pop();
      ASSERT(_currNumOfHoles);
     }
    else
     {
      DOP(_currHole = (TERM*)0);
     };    
  };
  void pushNonconstFun(const TERM& f)
  {
   CALL("pushNonconstFun(const TERM& f)");
   ASSERT(f.isComplex());
   ASSERT(f.arity());
   DOP(_debugDepthCounter.NonconstantFunctor(f.arity()));
   ASSERT((_debugDepthCounter.Depth() >= 0) && (_debugDepthCounter.Depth() <= HellConst::MaxTermDepth));
   DOP(_debugSizeCounter += f.arity());
   ASSERT(_debugSizeCounter <= HellConst::MaxTermSize);
   ASSERT(_currHole);
   _currHole->SetFirst(_freeMem);
   _currNumOfHoles--;
   if (_currNumOfHoles) 
    {
     _holes.push(_currHole->Next());
     _numOfHoles.push(_currNumOfHoles);
    };
   *_freeMem = f;
   _currHole = _freeMem->Args();
   _freeMem += (f.arity() + 1);
   //_currHole = _freeMem;
   _currNumOfHoles = f.arity();
  };
  void push(const TERM& sym)
  {
   CALL("push(const TERM& sym)");
   if (sym.isVariable()) { pushVar(sym); }
   else 
    if (sym.arity()) { pushNonconstFun(sym); } 
    else pushConst(sym);
  };
  TERM* freeMemory() const { return _freeMem; };
 private:
  Stack<TERM*,HellConst::MaxTermDepth> _holes;
  Stack<ulong,HellConst::MaxTermDepth> _numOfHoles;
  TERM* _currHole;
  ulong _currNumOfHoles;
  TERM _term;
  TERM* _freeMem;
  #ifdef DEBUG_NAMESPACE
   SmartDepthCounter<HellConst::MaxTermDepth + 1> _debugDepthCounter;
   ulong _debugSizeCounter;
  #endif
}; // class TERM::OpenCompact



//=================================================
#undef DEBUG_NAMESPACE   
#ifdef DEBUG_TERM
 #define DEBUG_NAMESPACE "TERM::OpenUnshared"
#endif 
#include "debugMacros.h"
//=================================================

class TERM::OpenUnshared
{
 public:
  OpenUnshared() : _depthCounter(), _holes(), _freeHoles(), _allocated()
  { 
   CALL("constructor OpenUnshared()");
   reset(); 
  };
  ~OpenUnshared() {};
  void reset() 
  { 
   _currHole = &_res; 
   _numOfFreeHoles = 1UL;
   _size = 1UL; 
   _depthCounter.Reset(0);
   _holes.reset();
   _freeHoles.reset(); 
   _allocatedLit = (TERM*)0;
   _allocated.reset();
  };
  bool litHeader(const TERM& hd)
  {
   CALL("litHeader(const TERM& hd)"); 
   ulong arity = hd.arity();

   _currHole->SetFirst(TERM::AllocLit(hd));
   _allocatedLit = _currHole->First();
   if (arity) 
    { 
     _depthCounter.NonconstantFunctor(arity);
     _numOfFreeHoles = arity;         
     _currHole = _currHole->First()->Args();
    };
   return true;
  };
  bool functor(const TERM& f)
  {
   CALL("functor(const TERM& f)");
   ASSERT(_numOfFreeHoles);
   ASSERT(_numOfFreeHoles <= HellConst::MaxNumOfTermArguments);
   ulong arity = f.arity();
   _size += arity;
   if (_size > HellConst::MaxTermSize) 
    {
     recycle(); 
     return false; 
    };
   _numOfFreeHoles--;
   if (arity) 
    {
     _currHole->SetFirst(new (arity) TERM(f)); 
     _allocated.push(_currHole->First()); 
     if (_depthCounter.Depth() >= HellConst::MaxTermDepth) 
      {
       recycle(); 
       return false; 
      };
     _depthCounter.NonconstantFunctor(arity);        
     if (_numOfFreeHoles)
      { 
       _holes.push(_currHole->Next());
       _freeHoles.push(_numOfFreeHoles);    
      };
     _currHole = _currHole->First()->Args();
     _numOfFreeHoles = arity; 
    }
   else // constant
    {
     _currHole->SetFirst(sharedConst(f.functor()));
     _depthCounter.VarOrConst(); 
     if (_numOfFreeHoles) { _currHole = _currHole->Next(); }
     else // backtrack
      if (_holes)
       {
	_currHole = _holes.pop();
        _numOfFreeHoles = _freeHoles.pop();
        ASSERT(_numOfFreeHoles);
        ASSERT(_numOfFreeHoles <= HellConst::MaxNumOfTermArguments);
       };     
    };
   return true;
  }; // bool functor(const TERM& f)

  void variable(const TERM& v)
  {
   CALL("variable(const TERM& v)"); 
   ASSERT(_numOfFreeHoles);
   ASSERT(_numOfFreeHoles <= HellConst::MaxNumOfTermArguments);
   _depthCounter.VarOrConst(); 
   _numOfFreeHoles--;
   *_currHole = v;
   if (_numOfFreeHoles) { _currHole = _currHole->Next(); }
   else // backtrack   
    if (_holes)
     {
      _currHole = _holes.pop();
      _numOfFreeHoles = _freeHoles.pop();
      ASSERT(_numOfFreeHoles);
      ASSERT(_numOfFreeHoles <= HellConst::MaxNumOfTermArguments);
     };  
  }; // void variable(const TERM& v)

  bool symbol(const TERM& s)
  { 
   CALL("symbol(const TERM& s)");
   // can not be applied to a predicate header!
   if (s.isVariable()) { variable(s); return true; } else return functor(s);
  };  

  const TERM& result() const { return _res; };
  void recycle()
  {
   CALL("recycle()");
   if (_allocatedLit) { _allocatedLit->DeleteLit(); _allocatedLit = (TERM*)0; };
   while (_allocated) delete _allocated.pop();    
  }; 

 private: 
  TERM _res;
  TERM* _currHole;
  ulong _size;
  ulong _numOfFreeHoles;
  SmartDepthCounter<HellConst::MaxTermDepth> _depthCounter;
  Stack<TERM*,HellConst::MaxTermDepth> _holes;
  Stack<ulong,HellConst::MaxTermDepth> _freeHoles;
  TERM* _allocatedLit;
  Stack<TERM*,HellConst::MaxTermSize> _allocated;
}; // class TERM::OpenUnshared


//=================================================
#undef DEBUG_NAMESPACE   
#ifdef DEBUG_TERM
 #define DEBUG_NAMESPACE "TERM::FastOpenUnshared"
#endif 
#include "debugMacros.h"
//=================================================

class TERM::FastOpenUnshared
{
 public:
  FastOpenUnshared() : _holes(), _freeHoles()
  { 
   reset(); 
  };
  ~FastOpenUnshared() {};
  void reset() 
  { 
   _currHole = &_res; 
   _numOfFreeHoles = 1UL;
   _holes.reset();
   _freeHoles.reset(); 
  };

  void litHeader(const TERM& hd)
  {
   CALL("litHeader(const TERM& hd)"); 
   ulong arity = hd.arity();

   _currHole->SetFirst(TERM::AllocLit(hd));
   if (arity) 
    { 
     _numOfFreeHoles = arity;         
     _currHole = _currHole->First()->Args();
    };
  };

  void functor(const TERM& f)
  {
   CALL("functor(const TERM& f)");
   ASSERT(f.isComplex());
   ASSERT(_numOfFreeHoles);
   ASSERT(_numOfFreeHoles <= HellConst::MaxNumOfTermArguments);
   ulong arity = f.arity();
   _numOfFreeHoles--;
   if (arity) 
    {       
     _currHole->SetFirst(new (arity) TERM(f)); 
     if (_numOfFreeHoles)
      { 
       _holes.push(_currHole->Next());
       _freeHoles.push(_numOfFreeHoles);    
      };
     _currHole = _currHole->First()->Args();
     _numOfFreeHoles = arity; 
    }
   else // constant
    {
     _currHole->SetFirst(sharedConst(f.functor()));
     if (_numOfFreeHoles) { _currHole = _currHole->Next(); }
     else // backtrack
      if (_holes)
       {
	_currHole = _holes.pop();
        _numOfFreeHoles = _freeHoles.pop();
        ASSERT(_numOfFreeHoles);
        ASSERT(_numOfFreeHoles <= HellConst::MaxNumOfTermArguments);
       };     
    };
  }; // void functor(const TERM& f)

  void variable(const TERM& v)
  {
   CALL("variable(const TERM& v)"); 
   ASSERT(v.isVariable());
   ASSERT(_numOfFreeHoles);
   ASSERT(_numOfFreeHoles <= HellConst::MaxNumOfTermArguments);
   _numOfFreeHoles--;
   *_currHole = v;
   if (_numOfFreeHoles) { _currHole = _currHole->Next(); }
   else // backtrack   
    if (_holes)
     {
      _currHole = _holes.pop();
      _numOfFreeHoles = _freeHoles.pop();
      ASSERT(_numOfFreeHoles);
      ASSERT(_numOfFreeHoles <= HellConst::MaxNumOfTermArguments);
     };  
  }; // void variable(const TERM& v)


  void symbol(const TERM& s)
  { 
   CALL("symbol(const TERM& s)");
   // can not be applied to a predicate header!
   if (s.isVariable()) { variable(s);  } else functor(s);
  };  

  const TERM& result() const { return _res; };

 private: 
  TERM _res;
  TERM* _currHole;
  ulong _numOfFreeHoles;
  Stack<TERM*,HellConst::MaxTermDepth> _holes;
  Stack<ulong,HellConst::MaxTermDepth> _freeHoles;
}; // class TERM::FastOpenUnshared


//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TERM
 #define DEBUG_NAMESPACE "TERM"
#endif
#include "debugMacros.h"
//=================================================
inline bool TERM::Plain() const 
{
 for (const TERM* arg = Args(); !arg->IsNull(); arg = arg->Next())
  { 
   if (arg->isVariable())
    {
     for (const TERM* prev_var = Args(); prev_var != arg; prev_var = prev_var->Next())
      if ((*prev_var) == (*arg)) return false;
    }
   else return false;
  };   
 return true;
}; // bool TERM::Plain()

inline bool TERM::EqualComplexTerms(const TERM* t1,const TERM* t2)
 {
  if (*t1 == *t2)
   {
    Iterator iter1;
    Iterator iter2;
    iter1.Reset(t1);
    iter2.Reset(t2);
    while (iter1.Next())
     {
      iter2.Next();
      if (iter1.Symbol() != iter2.Symbol()) { return false; };
     };
    return true;
   }
  else { return false; };
 }; // bool TERM::EqualComplexTerms(const TERM* t1,const TERM* t2)

inline Comparison TERM::CompareLex(const TERM* t) const
{
 CALL("CompareLex(const TERM* t) const");
 ASSERT(isComplex());
 ASSERT(t->isComplex());
 Iterator iter1(this);
 Iterator iter2(t);
 do
  {
   if (iter1.symbolRef() != iter2.symbolRef())
    { 
     if (iter1.CurrentSymIsVar() || iter2.CurrentSymIsVar()) return Incomparable;    
     return iter1.symbolRef().CompareFunctor(iter2.symbolRef());
    };  
   iter2.Next();
  }
 while (iter1.Next());
 return Equal;
}; // Comparison CompareLex(const TERM* t) const


inline ostream& operator<<(ostream& str,const TERM& t)
{
 return TERM::output(str,t);
}; 

inline ostream& operator<<(ostream& str,const TERM* t)
{
 return TERM::output(str,t);
}; 

#ifndef NO_DEBUG_VIS
 inline ostream& operator<<(ostream& str,const TERM::OccList* occ) 
 {
  str << '{';
  if (!occ) return str << '}';
  
  for (const TERM::OccList::Node* node = occ->first(); node; node = node->next())
   {
    str << '[' << node->key() << ':';
    node->value().output(str) << "] ";
   };   
  return str << '}';  
 };
#endif


static TERM::InitStatic termInitStatic;

//================================================
#undef DEBUG_NAMESPACE   
//==================================================
#endif
