//
// File:         variables.h
// Description:  Various classes for dealing with variables and substitutions.
// Created:      Feb 15, 2000, 19:40
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#ifndef VARIABLES_H
//===================================================================
#define VARIABLES_H
#ifndef NO_DEBUG_VIS
 #include <iostream.h>
#endif
#include "jargon.h"
#include "HellDebugFlags.h"
#include "HellConst.h"
#include "term.h"
#include "MultisetOfVariables.h"
#include "Stack.h"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VARIABLES
 #define DEBUG_NAMESPACE "VAR"
#endif
#include "debugMacros.h"
//=================================================
class PREFIX_SYM;
class VAR
 {
  private: // static data
   static VAR var[HellConst::MaxNumOfIndices*HellConst::MaxNumOfVariables];
   static unsigned offset[HellConst::MaxNumOfIndices];
   static const void* end_of_vars;
  private: // structure
   void* binding;
  public:
   VAR() : binding(0) {};
  public:
   static void Reset() 
    {       
     for (ulong i = 0; i<HellConst::MaxNumOfIndices; i++)
      {
       offset[i] = i*HellConst::MaxNumOfVariables;
      };
    };
   void Bind(void* bind) { binding = bind; };
   void Unbind() { binding = 0; };
   void* Binding() const { return binding; };
   static VAR* Base(ulong v) { return var+v; };
   static unsigned Offset(ulong index) { return offset[index]; }; 
   static VAR* Var(ulong v,ulong index) 
    { return var + v + offset[index]; };
   static VAR* First() { return var; };
   static VAR* First(ulong index) { return var + offset[index]; };
   static VAR* Last(ulong index) { return (var + offset[index] + HellConst::MaxNumOfVariables - 1); };
   VAR* Next() { return this+1; };
   VAR* Previous() { return this-1; };
   static ulong Index(unsigned var_offset) 
    {
     return (var_offset/HellConst::MaxNumOfVariables);
    };
   ulong Index() const { return (this - var)/HellConst::MaxNumOfVariables; }; 
   void NumAndIndex(ulong& num,ulong& index) const
    {
     unsigned pos = this - var;
     index = pos/HellConst::MaxNumOfVariables;
     num = pos%HellConst::MaxNumOfVariables;
    };
   

   static bool IsVar(void* ptr) 
    { return ((ptr >= ((void*)var)) && (ptr<((void*)end_of_vars))); };

   
   /************** Can be more efficient variant:
   static bool IsVar(void* ptr) 
    { 
     return ((ptr<((void*)end_of_vars)) && (ptr >= ((void*)var))); 
    }; 
   *****************************************************/ 
    
   void* Unref() 
    {
     void* res = Binding();
     if (res)
      {
       check_binding:
        if (IsVar(res))
         {
          void* tmp = ((VAR*)res)->Binding();
          if (tmp)
           {
            res = tmp;
            goto check_binding;
           }
          else
           {
            return res; // free variable
           };
         }
        else // complex term 
         {
          return res;
         };
      }
     else
      {
       return this;
      };
    }; 
   void* Unref(bool& free_var) const
    {
     void* res = Binding();
     if (res)
      {
       check_binding:
        if (IsVar(res))
         {
          void* tmp = ((VAR*)res)->Binding();
          if (tmp)
           {
            res = tmp;
            goto check_binding;
           }
          else
           {
            free_var = true;
            return res; // free variable
           };
         }
        else // complex term 
         {
          free_var = false;
          return res;
         };
      }
     else
      {
       free_var = true;
       return (void*)this;
      };
    };
  public: // methods for debugging
   #ifndef NO_DEBUG_VIS
    ostream& output(ostream& str) const
    {
     ulong num;
     ulong index;
     NumAndIndex(num,index);    
     return str << 'X' << num << '{' << index << '}';
    };   
    
    ostream& outputBinding(ostream& str,ostream& (*outputTerm)(ostream& str,const PREFIX_SYM* t)) const
    { 
     output(str) << " := ";
     if (!Binding()) return str << "NULL"; 
     if (VAR::IsVar(Binding()))
      return ((VAR*)Binding())->output(str);
     return outputTerm(str,(const PREFIX_SYM*)Binding());
    };
   #endif
 }; // class VAR

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VARIABLES
 #define DEBUG_NAMESPACE "SUBSTITUTION"
#endif
#include "debugMacros.h"
//=================================================
class SUBSTITUTION
 {
  private: 
   VAR* bound[HellConst::MaxNumOfVariables*HellConst::MaxNumOfIndices];
   VAR** next;
   VAR** section[HellConst::MaxNumOfVariables*HellConst::MaxNumOfIndices];
   VAR*** next_section;
  public:
   SUBSTITUTION() 
    {
     next = bound;
     next_section = section;
    };
   ~SUBSTITUTION() { CALL("destructor ~SUBSTITUTION()"); };
   #ifndef NO_DEBUG
    bool clean() const
    {
     return ((next == bound) && (next_section == section));
    };
   #endif
   void reset() { UnbindAll(); };
   void UnbindAll() 
    {
     while (next != bound)
      {
       next--;
       (*next)->Unbind();
      };
     next_section = section;
    };
    
   #ifndef NO_DEBUG
    bool NoBoundVariables() const 
    {
     return (next == bound) && (next_section == section);
    };
   #endif   

   void Bind(VAR* v,void* val) 
    {
     CALL("Bind(VAR* v,void* val)");
     ASSERT(v != ((VAR*)val));
     v->Bind(val);
     *next = v;
     next++;
    };
   
   void BindAlone(VAR* v,void* val)
    {
     Bind(v,val);
    }; 
   
   void BindInSection(VAR* v,void* val)
    {
     Bind(v,val);
    };
    
   VAR* LastBound() const { return *(next-1); };
   void UnbindLast()
    {
     CALL("UnbindLast()");
     next--;
     (*next)->Unbind();
     
     ASSERT(SectionsOK());
    };
   void MarkSection() 
    {
     CALL("MarkSection()");
     *next_section = next; 
     next_section++;
     ASSERT(SectionsOK());
    };
   VAR*** LastMark() const
    {
     CALL("LastMark()");
     ASSERT(next_section > section);
     return next_section-1; 
    };
   
   void UnbindDownto(VAR*** mark)
    {
     CALL("UnbindDownto(VAR*** mark)");
     VAR** m = *mark;
     next_section = mark;
     while (next != m)
      {
       next--;
       (*next)->Unbind();
      };
     ASSERT(SectionsOK());
    };

   void UnbindLastSection()
    {
     CALL("UnbindLastSection()");
     ASSERT(next_section > section);
     next_section--;
     VAR** section_mark = *next_section;
     
     ASSERT(section_mark >= bound);
     ASSERT(section_mark <= next);
     
     while (next != section_mark)
      {
       next--;
       (*next)->Unbind();
      };
     ASSERT(SectionsOK());
    };
   VAR*** MarkState()
    {
     MarkSection();
     return LastMark();
    };
 
  public:
   bool VariablesAffected(VAR* first,VAR* last)
    {
     for (VAR** bv = bound; bv < next; bv++)
      if (((*bv) >= first) && ((*bv) <= last)) return true;    
     return false;
    };
   bool VariablesAffected(ulong index)
    { 
     return VariablesAffected(VAR::First(index),VAR::Last(index));
    };

  public: // methods for debugging
   #ifndef NO_DEBUG
    bool SectionsOK() const
     {
      VAR*** sect = next_section - 1;
      if (sect >= ((VAR***)section))
       {
        if ((*sect) > next)
         {
         
          //GlobalSubst.Output(OutputPrefix); // debug
         
          return false;
         };
       };
      return true;     
     };
    unsigned NumOfBindings() const { return (next - (VAR**)bound); };
   #endif

   #ifndef NO_DEBUG_VIS
    ostream& output(ostream& str,ostream& (*outputTerm)(ostream& str,const PREFIX_SYM* t)) const
    { 
     for(const VAR* const * p = bound; p<next; p++)
      (*p)->outputBinding(str,outputTerm) << '\n'; 
     return str;        
    };
   #endif
 }; // class SUBSTITUTION

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VARIABLES
 #define DEBUG_NAMESPACE "LARGE_VAR_TABLE"
#endif
#include "debugMacros.h"
//=================================================

class LARGE_VAR_TABLE
 {
  private:
   void* table[HellConst::MaxNumOfIndices*HellConst::MaxNumOfVariables];
   unsigned offset; 
  public:
   LARGE_VAR_TABLE() { offset = table - (void**)(VAR::First()); };
   void Initialize(void* val)
    {
     for (ulong i = 0; i< HellConst::MaxNumOfIndices*HellConst::MaxNumOfVariables; i++)
      {
       table[i] = val;
      };
    };
   void** Entry(const VAR* const v) const { return ((void**)v) + offset; };
 }; // class LARGE_VAR_TABLE

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VARIABLES
 #define DEBUG_NAMESPACE "SMALL_VAR_TABLE"
#endif
#include "debugMacros.h"
//=================================================
 
class SMALL_VAR_TABLE
 {
  private:
   void* table[HellConst::MaxNumOfVariables];
   long offset; 
  public:
   SMALL_VAR_TABLE() { offset = table - (void**)(VAR::First()); };
   void SetIndex(ulong ind)
    {
     offset = (table - (void**)(VAR::First())) - VAR::Offset(ind);
    };
   void Initialize(void* val)
    {
     for (ulong i = 0; i< HellConst::MaxNumOfVariables; i++)
      {
       table[i] = val;
      };
    };
   void** Entry(const VAR* const v) const { return ((void**)v) + offset; };
 }; // class SMALL_VAR_TABLE 

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VARIABLES
 #define DEBUG_NAMESPACE "LARGE_VAR_COUNTER"
#endif
#include "debugMacros.h"
//=================================================
 
class LARGE_VAR_COUNTER
 {
  private:
   LARGE_VAR_TABLE counter;
   VAR* registered[HellConst::MaxNumOfVariables*HellConst::MaxNumOfIndices]; // stack
   VAR** next;
   VAR** iter;
  public:
   LARGE_VAR_COUNTER() : counter() 
    { 
     counter.Initialize((void*)(0L));
     next = registered; 
    };
   void Reset()
    {
     while (next != registered)
      {
       next--; 
       *(counter.Entry(*next)) = (void*)(0L);
      };
    };
  private:  
   void Unregister(const VAR* v)
    {
     CALL("Unregister(const VAR* v)");
     ASSERT(!Score(v));  
     for (VAR** ptr = registered; ptr < next; ptr++)  
      {
       if (v == (*ptr)) { next--; *ptr = *next; return; };
      };
    };

  public:
   long Register(const VAR* const v) 
    // returns previous value of the counter for the variable
    {
     long res = (long)(*(counter.Entry(v)));
     *(counter.Entry(v)) = (void*)(res + 1);
     if (!res) { *next = (VAR*)v; next++; }
     else 
      if (!Score(v)) 
       {
        // variable with zero score must be ungeristered
        Unregister(v);
       };
     return res;      
    };

   long Register(long coefficient,const VAR* const v) 
    // returns previous value of the counter for the variable
    {
     long res = (long)(*(counter.Entry(v)));
     *(counter.Entry(v)) = (void*)(res + coefficient);
     if (!res) { *next = (VAR*)v; next++; }
     else 
      if (!Score(v)) 
       {
        // variable with zero score must be ungeristered
        Unregister(v);
       };
     return res;      
    };

   void MultiplyBy(long coefficient)
    {
     if (coefficient)
      {
       for (VAR** ptr = registered; ptr < next; ptr++)  
	{
         *(counter.Entry(*ptr)) = (void*)(coefficient*Score(*ptr)); 
        };
      }
     else Reset();
    };
 
   long Score(const VAR* const v) const
    {
     return (long)(*(counter.Entry(v)));
    };

   ulong NumOfRegistered() const { return (next - registered); };

   void ResetIteration()
    {
     iter = registered;
    };
   VAR* NextRegistered()
    {
     if (iter == next) { return (VAR*)0; };
     VAR* res = *iter;
     iter++;
     return res;
    };
 }; // class LARGE_VAR_COUNTER

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VARIABLES
 #define DEBUG_NAMESPACE "SMALL_VAR_COUNTER"
#endif
#include "debugMacros.h"
//=================================================

class SMALL_VAR_COUNTER
 {
  private:
   SMALL_VAR_TABLE counter;
   VAR* registered[HellConst::MaxNumOfVariables]; // stack
   VAR** next;
   VAR** iter;
  public:
   SMALL_VAR_COUNTER() : counter() 
    { 
     counter.Initialize((void*)((unsigned)0));
     next = registered; 
    };
   void SetIndex(ulong ind) { counter.SetIndex(ind); }; 
   void Reset()
    {
     while (next != registered)
      {
       next--; 
       *(counter.Entry(*next)) = (void*)((unsigned)0);
      };
    };
   unsigned Register(const VAR* const v) 
    // returns previous value of the counter for the variable
    {
     unsigned res = (unsigned)(*(counter.Entry(v)));
     *(counter.Entry(v)) = (void*)(res + 1);
     if (!res) { *next = (VAR*)v; next++; };
     return res;      
    };
   
   unsigned Score(const VAR* const v) const
    {
     return (unsigned)(*(counter.Entry(v)));
    };
    
   void ResetIteration()
    {
     iter = registered;
    };
   VAR* NextRegistered()
    {
     if (iter == next) { return (VAR*)0; };
     VAR* res = *iter;
     iter++;
     return res;
    };
 }; // class SMALL_VAR_COUNTER

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VARIABLES
 #define DEBUG_NAMESPACE "VAR_TO_INT_RENAMING"
#endif
#include "debugMacros.h"
//=================================================
class VAR_TO_INT_RENAMING
 {
  private:
   LARGE_VAR_TABLE renaming;
   ulong next_var;
   VAR* renamed[HellConst::MaxNumOfVariables*HellConst::MaxNumOfIndices]; // stack
   VAR** next;
  public: 
   VAR_TO_INT_RENAMING() : renaming()
    {
     renaming.Initialize((void*)((ulong)0));
     next_var = 1;
     next = renamed;
    };
   void Reset()
    {
     next_var = 1;
     while (next != renamed)
      {
       next--; 
       *(renaming.Entry(*next)) = (void*)((ulong)0);
      };
    };
   ulong Rename(VAR* var)
    {
     ulong res = (ulong)(*(renaming.Entry(var)));
     if (!res) 
      {
       res = next_var;
       *(renaming.Entry(var)) = (void*)res;
       next_var++;
       *next = var;
       next++;
      };
     return res;
    };
     
  public: // for debugging
   #ifndef NO_DEBUG_VIS
    ostream& output(ostream& str) const
    {
     str << "VAR_TO_INT_RENAMING:\n";
     for (const VAR* const * r = renamed; r < next; r++)
      (*r)->output(str) << " -> " << (ulong)(*(renaming.Entry(*r))) << '\n';
     return str << "END OF VAR_TO_INT_RENAMING.\n";
    };  
   #endif
 }; // class VAR_TO_INT_RENAMING

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VARIABLES
 #define DEBUG_NAMESPACE "INT_VAR_RENAMING"
#endif
#include "debugMacros.h"
//=================================================

class INT_VAR_RENAMING
 {
  private:
   ulong renaming[HellConst::MaxNumOfVariables];
   ulong next_var;
   ulong renamed[HellConst::MaxNumOfVariables]; // stack
   ulong* next;
  public: 
   INT_VAR_RENAMING() 
    {
     for (ulong v = 0; v < HellConst::MaxNumOfVariables; v++)
      {
       renaming[v] = 0;
      };
     next_var = 1;
     next = renamed;
    };
   void Reset()
    {
     next_var = 1;
     while (next != renamed)
      {
       next--; 
       renaming[*next] = 0;
      };
    };
   ulong Rename(ulong var)
    {
     ulong res = renaming[var];
     if (!res) 
      {
       res = next_var;
       renaming[var] = res;
       next_var++;
       *next = var;
       next++;
      };
     return res - 1;
    };
  public: // for debugging
   #ifndef NO_DEBUG_VIS
    ostream& output(ostream& str) const
    {
     str << "VARIABLE RENAMING:\n";
     for (const ulong* r = renamed; r < next; r++)
      str << 'X' << *r << " -> " << 'X' << renaming[*r] << '\n';
     return str << "END OF VARIABLE RENAMING.\n";
    };
   #endif
 }; // class INT_VAR_RENAMING


//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VARIABLES
 #define DEBUG_NAMESPACE "VAR_WEIGHT"
#endif
#include "debugMacros.h"
//================================================= 
class VAR_WEIGHT
 {
  private: // structure
   ulong num_of_occurences;
   ulong different_terms; 
  public:
   VAR_WEIGHT() : num_of_occurences(0), different_terms(0) {};
   void Reset() { num_of_occurences = 0; different_terms = 0; };
   bool Nonnull() { return (num_of_occurences || different_terms); };
   void MoreOccurences(const ulong n) { num_of_occurences += n; };
   void AnotherTerm() { different_terms++; };
   Comparison Compare(const VAR_WEIGHT& w) const
    {
     if (num_of_occurences == w.num_of_occurences)
      {
       if (different_terms == w.different_terms) { return Equal; }
       else
        { 
         if (different_terms > w.different_terms) { return Greater; }
         else return Less; 
        };
      }
     else
      {
       if (num_of_occurences > w.num_of_occurences) { return Greater; }
       else return Less; 
      }; 
    };
 }; // class VAR_WEIGHT

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VARIABLES
 #define DEBUG_NAMESPACE "VAR_WEIGHTING"
#endif
#include "debugMacros.h"
//=================================================  
class VAR_WEIGHTING
 {
  private: 
   VAR_WEIGHT weight[HellConst::MaxNumOfVariables];
   Stack<ulong,HellConst::MaxNumOfVariables> registered;
  public:
   VAR_WEIGHTING() : registered("HellConst::MaxNumOfVariables in VAR_WEIGHTING::registered")
    {
     for (ulong v = 0; v < HellConst::MaxNumOfVariables; v++) weight[v].Reset(); 
    };
   void Reset()
    {
     while (!(registered.empty())) { weight[registered.pop()].Reset(); };
    };
   void MoreOccurences(const ulong var,const ulong occ)
    {
     if (!(weight[var].Nonnull())) { registered.push(var); };
     weight[var].MoreOccurences(occ);
    };
   void AnotherTerm(const ulong var)
    {
     if (!(weight[var].Nonnull())) { registered.push(var); };
     weight[var].AnotherTerm();
    };
   Comparison Compare(const ulong var1,const ulong var2) const
    {
     return weight[var1].Compare(weight[var2]);
    };
   void AnotherTermVariables(MultisetOfVariables<HellConst::MaxNumOfVariables>& vars)
    {
     vars.ResetIterator();
     const ulong* next_var = vars.NextVar();
     while (next_var)
      {
       AnotherTerm(*next_var);
       MoreOccurences(*next_var,vars.Occurences(*next_var));
       next_var = vars.NextVar();
      };
    };
 }; // class VAR_WEIGHTING  
//================================================
#undef DEBUG_NAMESPACE 
//===================================================================
#endif 





