#ifndef WEIGHTING_H
//=================================================
#define WEIGHTING_H
#include <iostream.h>
#include <limits.h>
#include "jargon.h"
#include "HellDebugFlags.h"
#include "HellConst.h"
#include "Tuple.h"
#include "GSortTree.h"
#include "term.h"
#include "Tabulation.h"
#include "Signature.h"
extern Signature TheSignature;
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_HEADER_WEIGHTING
 #define DEBUG_NAMESPACE "HEADER_WEIGHTING"
#endif 
#include "debugMacros.h"
//=================================================

class HEADER_WEIGHT_LIST;

class HEADER_WEIGHTING
 {
  public:
   class ArityFuncPair : Tuple2<ulong,ulong> 
   {
    public: 
     ArityFuncPair(const ulong& ar,const ulong& fn) : Tuple2<ulong,ulong>(ar,fn) {};
     const ulong& arity() const { return Tuple2<ulong,ulong>::el1; };
     ulong& arity() { return Tuple2<ulong,ulong>::el1; };
     const ulong& func() const { return Tuple2<ulong,ulong>::el2; };
     ulong& func() { return Tuple2<ulong,ulong>::el2; };
     bool operator<(const ArityFuncPair& afp) const 
     {
      return (arity() < afp.arity()) || ((arity() == afp.arity()) && (func() < afp.func())); 
     };
     bool operator>(const ArityFuncPair& afp) const 
     {
      return (arity() > afp.arity()) || ((arity() == afp.arity()) && (func() > afp.func())); 
     };
     bool operator==(const ArityFuncPair& afp) const 
     {
      return (arity() == afp.arity()) && (func() == afp.func()); 
     };
     bool operator!=(const ArityFuncPair& afp) const 
     {
      return (arity() != afp.arity()) || (func() != afp.func()); 
     }; 
   };
  private: 
   typedef GSortTree<GlobAlloc,HEADER_WEIGHTING::ArityFuncPair,HEADER_WEIGHTING> SortTree;
  private:
   long weight[HellConst::MaxNumOfDifferentHeaders];
   long min_weight;
   long max_weight;
  private:
   long ranking_by_arity;
   bool ranking_kinky; 
  private:
   long& WeightRef(const TERM& header)
    {
     CALL("WeightRef(const TERM& header)");
     ASSERT((header.functor()*2)+header.polarity() < (ulong)HellConst::MaxNumOfDifferentHeaders);
     return weight[(header.functor()*2)+header.polarity()];
    };
   const long& WeightRef(const TERM& header) const
    {
     CALL("WeightRef(const TERM& header) const");
     ASSERT((header.functor()*2)+header.polarity() < (ulong)HellConst::MaxNumOfDifferentHeaders);
     return weight[(header.functor()*2)+header.polarity()];
    };
  public:
   HEADER_WEIGHTING() : ranking_by_arity(0), ranking_kinky(false) { Reset(); };
   ~HEADER_WEIGHTING() { CALL("destructor ~HEADER_WEIGHTING()"); };
   void Reset() 
    {
     min_weight = LONG_MAX;
     max_weight = LONG_MIN;
     for (ulong i = 0; i < HellConst::MaxNumOfDifferentHeaders; i++)
      {
       weight[i] = 0L;
      };
    };
   
   void SetRankingByArity(long rba) { ranking_by_arity = rba; };
   void SetKinkyRanking(bool fl) { ranking_kinky = fl; };
   
   void AssertHeaderWeighting(const HEADER_WEIGHT_LIST* lst,bool inversePolarity);
  
   long Weight(const TERM& header) const { return WeightRef(header); };
   void SetWeight(const TERM& header,const long w) 
    {
     WeightRef(header) = w;
     if (w > max_weight) { max_weight = w; };
     if (w < min_weight) { min_weight = w; };
    };
   long MinWeight() const { return min_weight; };
   long MaxWeight() const { return max_weight; };
   void MakeMinimal(const TERM& header)
    {     
     long new_weight = MinWeight() - 1;
     if (!new_weight) { new_weight = -1; };
     SetWeight(header,new_weight);
    };
   void MakeMaximal(const TERM& header)
    {
     long new_weight = MaxWeight() + 1;
     if (!new_weight) { new_weight = 1; };
     SetWeight(header,new_weight);
    };
   
   Comparison Compare(const TERM& header1,const TERM& header2)
    {
     CALL("Compare(const TERM& header1,const TERM& header2)");
     if (Weight(header1) > Weight(header2)) { return Greater; }
     else
      {
       if (Weight(header1) < Weight(header2)) { return Less; }
       else 
        { 
         ASSERT((header1 == header2) || (header1.IsEquality() && header2.IsEquality() && (header1.polarity() == header2.polarity()))); 
         return Equal; 
        };
      };
    };
  public:
   void Tabulate(Tabulation* tab) 
   {
    for (ulong pred = 0; pred < TheSignature.nextFunctorNumber(); pred++)
     {
      if (TheSignature.isPredicate(pred) 
          //&& (!HellConst::isEquality(pred))
          && (!HellConst::isReservedPred(pred)))
       {   
        TERM h_pos((TERM::Arity)(TheSignature.arity(pred)),(TERM::Functor)pred,TERM::PositivePolarity); 
        TERM h_neg((TERM::Arity)(TheSignature.arity(pred)),(TERM::Functor)pred,TERM::NegativePolarity); 
        tab->headerRank(h_pos,Weight(h_pos));      
        tab->headerRank(h_neg,Weight(h_neg)); 
       }; 
     };
   }; // void Tabulate(Tabulation* tab)  
   ostream& output(ostream& str) const 
   {
    for (ulong pred = 0; pred < TheSignature.nextFunctorNumber(); pred++)
     {
      if (TheSignature.isPredicate(pred) 
	  //&& (!HellConst::isEquality(pred)) 
          && (!HellConst::isReservedPred(pred)))
       {   
        TERM h_pos((TERM::Arity)0,(TERM::Functor)pred,TERM::PositivePolarity); 
        TERM h_neg((TERM::Arity)0,(TERM::Functor)pred,TERM::NegativePolarity); 
        str << "weight(" <<  h_pos << '/' << TheSignature.arity(pred) << ") = ";
        str << Weight(h_pos) << "   ";
        str << "weight(" << h_neg << '/' << TheSignature.arity(pred) << ") = ";
        str << Weight(h_neg) << '\n';
       }; 
     };
    return str;
   }; // ostream& output(ostream&) const 
 }; // class HEADER_WEIGHTING


inline ostream& operator<<(ostream& str,const HEADER_WEIGHTING& hw)
{
 return hw.output(str);
}; // ostream& operator<<(ostream& str,const HEADER_WEIGHTING& hw)

inline ostream& operator<<(ostream& str,const HEADER_WEIGHTING* hw)
{
 return hw->output(str);
}; // ostream& operator<<(ostream& str,const HEADER_WEIGHTING* hw)

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_HEADER_WEIGHTING
 #define DEBUG_NAMESPACE "HEADER_WEIGHT"
#endif 
#include "debugMacros.h"
//=================================================
class HEADER_WEIGHT
 {
  public:
   enum EXTREMUM
    {
     MIN,
     MAX
    };
  private:
   TERM header;
   long weight;
  public:
   HEADER_WEIGHT(const TERM& h,const long& w) 
    : header(h),
      weight(w)
    {};
   HEADER_WEIGHT(const TERM& h,const EXTREMUM& e) 
    : header(h)
    {
     if (e == MIN)
      {
       weight = LONG_MIN;
      }
     else
      {
       weight = LONG_MAX;
      };
    };
   const TERM& Header() const { return header; };
   ulong functor() const { return header.functor(); };
   long Weight() const { return weight; };
   bool Min() const { return (weight == LONG_MIN); };
   bool Max() const { return (weight == LONG_MAX); };
   bool Extremal() const { return (Min() || Max()); };
  public:
   ostream& output(ostream& str) const
   {
    str << "weight(" << header << ") = ";
    if (Min()) { str << "min"; }
    else 
     if (Max()) { str << "max"; }
     else str << weight;
    return str;
   }; // ostream& output(ostream& str) const  
 }; // class HEADER_WEIGHT

inline ostream& operator<<(ostream& str,const HEADER_WEIGHT& hw)
{
 return hw.output(str);
}; // ostream& operator<<(ostream& str,const HEADER_WEIGHT& hw)

inline ostream& operator<<(ostream& str,const HEADER_WEIGHT* hw)
{
 return hw->output(str);
}; // ostream& operator<<(ostream& str,const HEADER_WEIGHT* hw)


//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_HEADER_WEIGHTING
 #define DEBUG_NAMESPACE "HEADER_WEIGHT_LIST"
#endif 
#include "debugMacros.h"
//=================================================
class HEADER_WEIGHT_LIST
 {
  private:
   const HEADER_WEIGHT* h_weight;
   HEADER_WEIGHT_LIST* next;
  public:
   HEADER_WEIGHT_LIST(const HEADER_WEIGHT* const h_w,HEADER_WEIGHT_LIST* const n)
    : h_weight(h_w),
      next(n)
     {};
   const HEADER_WEIGHT* HWeight() const { return h_weight; };
   HEADER_WEIGHT_LIST* Next() const { return next; };
   static void Add(const HEADER_WEIGHT* h_w,HEADER_WEIGHT_LIST*& lst)
    {
     lst = new HEADER_WEIGHT_LIST(h_w,lst);
    }; 
   bool Contains(const ulong& pred) const
   {
    if (!this) return false;
    if (h_weight->functor() == pred) return true;
    return next->Contains(pred);
   };
 
 }; // class HEADER_WEIGHT_LIST


inline ostream& operator<<(ostream& str,const HEADER_WEIGHT_LIST* hwl)
{
 while (hwl) 
  {
   str << hwl->HWeight() << '\n';
   hwl = hwl->Next();
  };
 return str;
};  
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_HEADER_WEIGHTING
 #define DEBUG_NAMESPACE "HEADER_WEIGHTING"
#endif 
#include "debugMacros.h"
//=================================================
inline void HEADER_WEIGHTING::AssertHeaderWeighting(const HEADER_WEIGHT_LIST* lst,bool inversePolarity)
 {
  CALL("AssertHeaderWeighting(const HEADER_WEIGHT_LIST* const lst,bool inversePolarity)");

  Reset();
  /********* I don't remember why is this so: **********************/
  SetWeight(TERM((TERM::Arity)0,(TERM::Functor)(HellConst::MaxNumOfPredicates-1),TERM::PositivePolarity),1);
  /*****************************************************************/   

  const HEADER_WEIGHT_LIST* tmp_lst = lst;
  const HEADER_WEIGHT* h_w;
  while (tmp_lst)
   {
    h_w = tmp_lst->HWeight();
    if (!(h_w->Extremal())) 
     {
      SetWeight(h_w->Header(),h_w->Weight());
     };
    tmp_lst = tmp_lst->Next();
   };

  if (ranking_by_arity)
   {
    // order predicates by arity   
    SortTree sortTree;
  
    for (ulong pred = 0; pred < TheSignature.nextFunctorNumber(); pred++)
     if (TheSignature.isPredicate(pred))    
      sortTree.insert(HEADER_WEIGHTING::ArityFuncPair(TheSignature.arity(pred),pred));
    
    if (ranking_by_arity > 0)  
     {
      // greater arity => greater rank
      SortTree::RightLeftIter sortTreeIter(sortTree);
      ulong pred;
      while (sortTreeIter) 
       {
	pred = sortTreeIter.value().func();
        if ((!lst->Contains(pred)) && (!HellConst::isEquality(pred)))
	 {
          TERM h_pos((TERM::Arity)0,(TERM::Functor)pred,TERM::PositivePolarity);
          TERM h_neg((TERM::Arity)0,(TERM::Functor)pred,TERM::NegativePolarity);
          if (inversePolarity) 
	   {
	    h_pos.Inversepolarity();
            h_neg.Inversepolarity();
           };
          if (ranking_kinky) { MakeMaximal(h_neg); }
          else MakeMinimal(h_neg);           
          MakeMinimal(h_pos);
         };
	sortTreeIter.next();
       };

     }
    else // ranking_by_arity < 0, less arity => greater rank
     {
      SortTree::LeftRightIter sortTreeIter(sortTree);
      ulong pred;
      while (sortTreeIter) 
       {
	pred = sortTreeIter.value().func();
        if ((!lst->Contains(pred)) && (!HellConst::isEquality(pred)))
	 {
          TERM h_pos((TERM::Arity)0,(TERM::Functor)pred,TERM::PositivePolarity);
          TERM h_neg((TERM::Arity)0,(TERM::Functor)pred,TERM::NegativePolarity);
          if (inversePolarity) 
	   {
	    h_pos.Inversepolarity();
            h_neg.Inversepolarity();
           };
          if (ranking_kinky) { MakeMaximal(h_neg); }
          else MakeMinimal(h_neg);
          MakeMinimal(h_pos);
         };
	sortTreeIter.next();
       };
     };    

   }
  else // ranking is done in the order of appearance   
   {
    for (ulong pred = 0; pred < TheSignature.nextFunctorNumber(); pred++)
     {
      if (TheSignature.isPredicate(pred) && (!HellConst::isEquality(pred))) 
       {
        TERM h_pos((TERM::Arity)0,(TERM::Functor)pred,TERM::PositivePolarity);
        TERM h_neg((TERM::Arity)0,(TERM::Functor)pred,TERM::NegativePolarity);
        if (inversePolarity) 
	 {
	  h_pos.Inversepolarity();
          h_neg.Inversepolarity();
         };
        if (!(Weight(h_neg))) 
         {
	  if (ranking_kinky) { MakeMaximal(h_neg); }
          else MakeMinimal(h_neg);
         };        
        if (!(Weight(h_pos))) 
         {
          MakeMinimal(h_pos);
         };
       };
     };  
   };
 
  tmp_lst = lst; 
  while (tmp_lst)
   {
    h_w = tmp_lst->HWeight();
    if (h_w->Min()) 
     {
      MakeMinimal(h_w->Header());
     }
    else
     {
      if (h_w->Max())
       {
        MakeMaximal(h_w->Header());
       }; 
     };
    tmp_lst = tmp_lst->Next();
   };

  // set ranks to the equality headers
  if (!Weight(TERM::negativeUnordEqHeader())) MakeMinimal(TERM::negativeUnordEqHeader());
  SetWeight(TERM::negativeOrdEqHeader(),Weight(TERM::negativeUnordEqHeader()));
  if (!Weight(TERM::positiveUnordEqHeader())) MakeMinimal(TERM::positiveUnordEqHeader());
  SetWeight(TERM::positiveOrdEqHeader(),Weight(TERM::positiveUnordEqHeader()));
 }; // void HEADER_WEIGHTING::AssertHeaderWeighting(const HEADER_WEIGHT_LIST* lst,bool inversePolarity)              

//================================================
#undef DEBUG_NAMESPACE            
//================================================
#endif
