//
// File:         InferenceDispatcher.h
// Description:  Interface to several inference engines.
// Created:      Mar 5, 2000, 20:40
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#ifndef INFERENCE_DISPATCHER_H
//====================================================
#define INFERENCE_DISPATCHER_H
#include "jargon.h"
#include "MasterDebugFlags.h"
#include "clause.h"
#include "or_index.h"
#include "binary_resolution.h"
#include "hyperresolution.h"
#include "forward_superposition.h"
#include "eq_factoring.h"
#include "backward_superposition.h"
#include "NewClause.h"
//====================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_INFERENCE_DISPATCHER
 #define DEBUG_NAMESPACE "InferenceDispatcher"
#endif
#include "debugMacros.h"
//====================================================
class InferenceDispatcher
{
 public:
  InferenceDispatcher(NewClause* freshClause)
    : binResFlag(false),
      hyperResFlag(false),
      _paramodulationFlag(false),
      _noInfBetweenRulesFlag(false),  
      _ordResIndex(),
      _binRes(freshClause,&_ordResIndex), 
      _hyperresolution(freshClause,&_ordResIndex),
      _forwardSuperposition(freshClause,&_ordResIndex),
      _equalityFactoring(freshClause), 
      _backwardSuperposition(freshClause,&_ordResIndex),
      _query((Clause*)0),
      _freshClause(freshClause)
  {
   CALL("constructor InferenceDispatcher(NewClause* freshClause)");
   HYPERRESOLUTION::SetDinamicLHSOrdering(true);
  };
  void destroy() 
  {
   _ordResIndex.Destroy();
   _hyperresolution.Destroy();
  };

  #ifndef NO_DEBUG  
   bool clean() const
   {
    return true
           && _ordResIndex.clean()
           && _hyperresolution.clean()          
           ;  
   };
  #endif

  void setParamodulationFlag(bool fl) 
  {
   _paramodulationFlag = fl;
   _ordResIndex.SetParamodulation(fl); 
  };
  void setSimBackSup(long degree)
  {
   _backwardSuperposition.setSimultaneous(degree);
  };
  void orientHyperresolution(bool positive) { _hyperresolution.Orient(positive); };
  void setUnitConflict(bool fl) { _unitConflictFlag = fl; _hyperresolution.SetUnitConfict(fl); };
  void setNoInfBetweenRulesFlag(bool fl)
  {
   _noInfBetweenRulesFlag = fl;
   _binRes.setNoInfBetweenRulesFlag(fl);
   _forwardSuperposition.setNoInfBetweenRulesFlag(fl);
   _backwardSuperposition.setNoInfBetweenRulesFlag(fl);
  };
  void newQuery(const Clause* cl) 
  { 
   _query = cl; 
   _continuation = (hyperResFlag) ? FirstHyperRes : FirstBinRes; 
  };
  bool nextInference() { return nextInf(); };
  bool hyperRule(const Clause* cl) const;
  void integrateHyperRule(Clause* cl);
  void integrateDefinition(Clause* cl,const TERM& definedHeader);
  void removeDefinition(Clause* cl);
  void makeActive(Clause* cl);
  void remove(Clause* cl);
  bool unitConflict(Clause* cl);
 public:
  bool binResFlag;
  bool hyperResFlag;
 private:
  enum Continuation 
  { 
   FirstHyperRes,
   NextHyperRes,
   FirstBinRes, 
   NextBinRes, 
   FirstForwSup, 
   NextForwSup,
   FirstEqFactor,
   NextEqFactor,
   FirstBackSup,
   NextBackSup 
  };
 private:
  bool nextInf();
 private:
  bool _paramodulationFlag;
  bool _unitConflictFlag;
  bool _noInfBetweenRulesFlag;
  OR_INDEX _ordResIndex;
  BIN_ORD_RES _binRes;
  HYPERRESOLUTION _hyperresolution;
  FORWARD_SUPERPOSITION _forwardSuperposition;
  EQ_FACTORING _equalityFactoring;
  BACKWARD_SUPERPOSITION _backwardSuperposition;
  const Clause* _query;
  NewClause* _freshClause;
  Continuation _continuation;
}; // class InferenceDispatcher

//====================================================
#endif
