//
// File:         Main.cpp
// Description:  Reading command line, parsing input and starting Vampire.
// Created:      Mar 6, 2000, 15:30
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#include "jargon.h"
#include <time.h>
#if ((TARGET_OS_TYPE != OS_TYPE_WIN32) || (COMPILER == COMPILER_GCC))
 #include <unistd.h>
#else 
 #include <process.h>
 typedef int pid_t; 
#endif
#include <limits.h>
#include <iostream.h>
#include <fstream.h> 
#include "BitWord.h"           // to initialise the static part
#include "CommandLine.h"
#include "Exit.h"
#include "Timer.h"
#include "MasterDebugFlags.h"
#include "Signature.h"
#include "weighting.h"
#include "Tabulation.h"
#include "TPTPParser.h"
#include "ClauseQueue.h"
#include "Master.h"
#include "Vampire.h"
#include "RuntimeError.h"
#if ((defined INTERCEPT_FORWARD_MATCHING_INDEXING) || (defined INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING))
 #include "Logging.h"
 #include "HellInterceptMacros.h"
#endif

//====================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_MAIN
 #define DEBUG_NAMESPACE "Main"
#endif    
#include "debugMacros.h"
//====================================================
HEADER_WEIGHT_LIST* HWList = (HEADER_WEIGHT_LIST*)0;
Signature TheSignature;
HEADER_WEIGHTING GlobalHeaderWeighting;
SUBSTITUTION GlobalSubst; // temporary, must be made local in Master 
 
const float VampireVersionNum = 2.00;
const char* const VampireVersionDesc = " (CASC 2001) "; 

ofstream tab_stream;
ofstream* usage_pace_stream = (ofstream*)0;   
ofstream* weight_limit_changes_stream = (ofstream*)0;   

void reportInternalError1(const char* msg)
{
 cout << "% Internal error: " << msg << "\n";
 cout << "% Trying to abort smoothly...\n";  
};


void reportInternalError2(const char* msg)
{
 cout << "% Abort complete.\n";
 cout << "% Internal error: " << msg << "\n";  
};
const char* basename(const char* fileName)
{
 int c = strlen(fileName) - 1; 
 while ((c >= 0) && (fileName[c] != '/')) c--;
 c++;
 char* res = new char[strlen(fileName) - c + 1];
 strcpy(res,fileName + c); 
 return res;
};

const char* computeJobId(tm* gmt,pid_t& jobProcessId)
{
 int jobDateDD = gmt->tm_mday;
 int jobDateMM = gmt->tm_mon;    
 int jobDateYY = 1900 + gmt->tm_year;
 int jobTimeHH = gmt->tm_hour;
 int jobTimeMM = gmt->tm_min;
 int jobTimeSS = gmt->tm_sec; 
 ostrstream str;
 const char* hhh = (jobTimeHH < 10) ? "0" : "";
 const char* mmm = (jobTimeMM < 10) ? "0" : "";
 const char* sss = (jobTimeSS < 10) ? "0" : "";
 str << jobDateDD << jobDateMM  << jobDateYY;
 str << hhh << jobTimeHH << mmm << jobTimeMM << sss << jobTimeSS;
 str << jobProcessId << '\0';
 char* res = new char[strlen(str.str()) + 1];
 strcpy(res,str.str());
 return res; 
}; // const char* computeJobId(tm* gmt,pid_t& jobProcessId)

void cancelInterruptHandling();

void handleInterrupt(int sigNum)
{
 const char* msg = "External Interrupt by an unknown signal"; 
 switch (sigNum)
  {
   case SIGTERM: 
	msg = "External Interrupt by SIGTERM signal"; 
	break;
   
   #if (COMPILER != COMPILER_MS_VISUAL_C_PLUS_PLUS)
    case SIGQUIT: 
	 msg = "External Interrupt by SIGQUIT signal";
	 break;
    case SIGHUP:  
	 msg = "External Interrupt by SIGHUP signal"; 
	 break;
    case SIGXCPU: msg = "External Interrupt by SIGXCPU signal"; break;
   #else
    case SIGINT: msg = "External Interrupt by SIGINT signal"; break;
    case SIGBREAK: msg = "External Interrupt by SIGBREAK signal"; break;
   #endif
  };

 cancelInterruptHandling();
 Vampire::terminate(msg);
 Exit::exit(0);
}; // void handleInterrupt(int sigNum)

void initInterruptHandler()
{
 signal(SIGTERM,handleInterrupt);

 #if (COMPILER != COMPILER_MS_VISUAL_C_PLUS_PLUS)
  signal(SIGQUIT,handleInterrupt); 
  signal(SIGHUP,handleInterrupt); 
  signal(SIGXCPU,handleInterrupt);
 #else 
  signal(SIGINT,handleInterrupt);
  signal(SIGBREAK,handleInterrupt);
 #endif
};

void cancelInterruptHandling()
{
 signal(SIGTERM,SIG_DFL);

 #if (COMPILER != COMPILER_MS_VISUAL_C_PLUS_PLUS)
  signal(SIGQUIT,SIG_DFL); 
  signal(SIGHUP,SIG_DFL); 
  signal(SIGXCPU,SIG_DFL);
 #else 
  signal(SIGINT,SIG_DFL);
  signal(SIGBREAK,SIG_DFL);
 #endif
};

void handleErrorSig(int sigNum)
{
 const char* msg = "an unknown error signal received";
 switch (sigNum)
  {
   #ifdef TARGET_OS_TYPE
    #if (TARGET_OS_TYPE == OS_TYPE_UNIX)
     case SIGFPE: msg = "error signal SIGFPE received"; break;
     case SIGILL: msg = "error signal SIGILL received"; break;
     case SIGSEGV: msg = "error signal SIGSEGV received"; break;
     case SIGBUS: msg = "error signal SIGBUS received"; break;  
     case SIGTRAP: msg = "error signal SIGTRAP received"; break;
     case SIGABRT: msg = "error signal SIGABRT received"; break;
    #elif (TARGET_OS_TYPE == OS_TYPE_WIN32)
     case SIGFPE: msg = "error signal SIGFPE received"; break;
     case SIGSEGV: msg = "error signal SIGSEGV received"; break;
     case SIGILL: msg = "error signal SIGILL received"; break; 
     #if (COMPILER == COMPILER_MS_VISUAL_C_PLUS_PLUS)
      case SIGABRT: msg = "error signal SIGABRT received"; break;
     #endif
    #else
     #error "Unsupported target OS type, check TARGET_OS_TYPE"
    #endif
   #else
    #error "Undefined TARGET_OS_TYPE"
   #endif
  };
 
 reportInternalError1(msg);
 Vampire::abort();  
 reportInternalError2(msg);
 Exit::exit(1);
}; // void handleErrorSig(int sigNum)

void initErrorSigHandler()
{
 #ifdef TARGET_OS_TYPE 	 
  #if (TARGET_OS_TYPE == OS_TYPE_UNIX)
   signal(SIGFPE,handleErrorSig);
   signal(SIGILL,handleErrorSig);
   signal(SIGSEGV,handleErrorSig);
   signal(SIGBUS,handleErrorSig);  
   signal(SIGTRAP,handleErrorSig);
   signal(SIGABRT,handleErrorSig);
  #elif (TARGET_OS_TYPE == OS_TYPE_WIN32)
   signal(SIGFPE,handleErrorSig);
   signal(SIGILL,handleErrorSig);
   signal(SIGSEGV,handleErrorSig); 
   signal(SIGABRT,handleErrorSig);
  #else
   #error "Unsupported target OS type, check TARGET_OS_TYPE"
  #endif
 #else
  #error "Undefined TARGET_OS_TYPE"
 #endif
};

void cancelErrorSigHandling()
{
 #ifdef TARGET_OS_TYPE 	 
  #if (TARGET_OS_TYPE == OS_TYPE_UNIX)
   signal(SIGFPE,SIG_DFL);
   signal(SIGILL,SIG_DFL);
   signal(SIGSEGV,SIG_DFL);
   signal(SIGBUS,SIG_DFL);  
   signal(SIGTRAP,SIG_DFL);
   signal(SIGABRT,SIG_DFL);
  #elif (TARGET_OS_TYPE == OS_TYPE_WIN32)
   signal(SIGFPE,SIG_DFL);
   signal(SIGILL,SIG_DFL);
   signal(SIGSEGV,SIG_DFL); 
   signal(SIGABRT,SIG_DFL);
  #else
   #error "Unsupported target OS type, check TARGET_OS_TYPE"
  #endif
 #else
  #error "Undefined TARGET_OS_TYPE"
 #endif
};


bool releaseMemory(ulong size) { return Vampire::releaseMemory(size); }; 

void reportAllocationFailure(ulong size) 
{ 
 cout << "% ------------- ALLOCATION FAILURE: " << size << " bytes -----------------\n"; 
 Vampire::reportAllocationFailure(size); 
 cancelInterruptHandling(); 
 cancelErrorSigHandling();
 if (tab_stream) tab_stream.close();
 if (usage_pace_stream) usage_pace_stream->close();
 if (weight_limit_changes_stream) weight_limit_changes_stream->close();
 Exit::exit(0); 
};



void describeOptions(CommandLine& CL)
{
 CL.option("h","help on command line syntax");
 CL.option("help","same as -h");
 CL.option("v","version information");
 CL.option("I",CommandLine::String,"<dir>","specifies the directory for making includes, \
                                           overrides the value obtained from the environment \
                                           variable TPTP_DIR");
 CL.option("include",CommandLine::String,"<dir>","same as -I <dir>");
 CL.option("T",CommandLine::String,"<test id>","test identifier, \
                                                 normaly used to identify \
                                                 series of Vampire runs");
 CL.option("test",CommandLine::String,"<test id>","same as -T <test id>");

  // preprocessing-only mode
 CL.option("prepro_only","switches preprocessing-only mode");

  // automatic mode
 CL.option("a","switches automatic selection of strategy");
 CL.option("auto","same as -a");
 
  // main algorithm 

 CL.option("main_alg",CommandLine::Long,"<num>"," sets the saturation algorithm, \
                                                 <num> = 0 => OTTER algorithm (default), \
                                                 <num> = 1 => Waldmeister algorithm, passive clauses are shared,  \
                                                 <num> = 2 => Waldmeister algorithm, passive clauses are not shared,  \
                                                 <num> = 3 => Waldmeister algorithm,  special memory management for storing passive clauses ");

  // inference rules
 CL.option("hyper","switches hyperresolution");
 CL.option("no_inferences_between_rules","when on, blocks resolution \
                                          (except simplifications) \
                                          between clauses containing \
                                          both positive and negative literals");
 CL.option("static_splitting","switches static splitting");
 CL.option("dynamic_splitting","switches dynamic splitting");
 CL.option("splitting_style",CommandLine::Long,"<num>","<num> <= 0 or <num> > 2 : our old pragmatic version (CASC-17) of splitting is used (default), \
                                                        <num> = 1 approximation of the standard splitting by nonblocking splitting, \
                                                        <num> = 2 approximation of the standard splitting by blocking splitting ");

 CL.option("splitting_with_naming","switches naming in splitting");

 CL.option("static_neg_eq_splitting",CommandLine::Long,"<weight>","ground arguments of negative equalities \
                                                                   whose weight exceeds <weight> will be \
                                                                   renamed, <weight> = 0 (default) switches \
                                                                   neg. eq. splitting off"); 
 CL.option("paramodulation","switches paramodulation");
 CL.option("sim_back_sup",CommandLine::Long,"<degree>","describes how many redexes can be rewritten simultaneously when backward superposition is applied, \
                                                        <degree> = 0 => redexes are rewritten only in one literal (default), \
                                                        <degree> > 0 => redexes in the whole clause are rewritten simultaneously ");
 CL.option("hyper_positive","orients hyperresolution"); 
 CL.option("selection",CommandLine::Long,"<num>"," : choses literal selection function, \
       <num> = 1 => only maximal literals are selected; \
       <num> = 2 => positive hyperresolution selection (negative \
                    literals are always selected before any positive), \
                    from several negative literals the one with the \
                    maximal weight is selected, maximal literals are \
                    selected in positive clauses; \
       <num> = 3 => positive hyperresolution selection, \
                    a heuristic intended to minimise the expected number of \
                    possible inferences is used to chose from several negative literals, \
                    maximal literals are selected in positive clauses; \
       <num> = 4 => maximal literals are selected, then we try to \
                    improve selection by selecting one negative literal with maximal weight; \
       <num> = 5 => maximal literals are selected, then we try to \
                    improve selection by selecting one negative \
                    literal minimising the expected number of possible inferences; \
       <num> = 6 => positive hyperresolution selection (negative \
                    literals are always selected before any positive). \
                    From several negative literals the maximally nonground \
                    is selected. From such literals we select those with greater size. \
                    Maximal literals are selected in positive clauses; \
       <num> = 7 => maximal literals are selected, then we try to \
                    improve selection by selecting one negative literal with maximal \
                    number of didderent variables, smaller depths of variable occurences \
                    bigger size; \
                    Literal selections with <num> >= 1000 are arbitrary, \
                    in maximally cases they make the procedure incomplete \
                    (ond exception - Horn problems). \
       <num> = 1002 => If the clause contains some inherently selected \
                      literals, only they are selected. \
                      Otherwise, if the clause contains a literal \
                      which is not a positive equality, \
                      we never select a positive equality. \
                      Among the literals available for selection \
                      we always select a ""maximally nonground"" literal. \
                      ""Maximally nonground"" means having maximal total number \
                      of variable occurences among the literals \
                      having maximal number of different variables. \
                      Among such literals we select one with the maximal size. \
       <num> = 1003 => If the clause contains some inherently selected \
                      literals, only they are selected. \
                      Otherwise, if the clause contains a literal \
                      which is not a positive equality, \
                      we never select a positive equality. \
                      Among the literals available for selection \
                      we always select a literal of the maximal size. \
                      Among several such literals we select by the function 1002. \
       <num> = 1004 => If the clause contains some inherently selected \
                      literals, only they are selected. \
                      Otherwise, if the clause contains a literal \
                      which is not a positive equality, \
                      we never select a positive equality. \
                      Among the literals available for selection \
                      we always select a ""maximally ground"" literal. \
                      ""Maximally ground"" means having minimal total number \
                      of variable occurences \
                      among literals having minimal number of different \
                      variables. \
                      Among such literals we select one with the maximal size. \
       <num> = 1005 => If the clause contains some inherently selected \
                      literals, only they are selected. \
                      Otherwise, if the clause contains a literal \
                      which is not a positive equality, \
                      we never select a positive equality. \
                      Among the literals available for selection \
                      we always select \
                      either a ""maximally ground"" negative literal, \
                      or a ""maximally nonground"" positive literal. \
                      Among such literals we select one with the maximal size. \
      <num> = 1006 =>  Bidirectional selection.  In a rule, i.e. a clause \
                      containing both positive and negative literals we  \
                      always select one positive and one negative literal. \
                      If we have several literals of the same polarity, we \
                      select those that are maximally nonground. \
                      In a fact, i.e. a clause with all literals \
                      of the same polarity, we select one maximally \
                      nonground literal. \
                      Switches off --inherited_negative_selection. \
      <num> = 1007 =>  Bidirectional selection. In a rule, i.e. a clause \
                      containing both positive and negative literals we  \
                      always select one positive and one negative literal. \
                      If we have several literals of the same polarity, we select \
                      those with maximal size. In a fact, i.e. a clause with \
                      all literals of the same polarity, we select one maximally \
                      nonground literal. \
                      Switches off --inherited_negative_selection. \
      <num> = 1008 =>  Bidirectional selection.  In a rule, i.e. a clause \
                      containing both positive and negative literals we  \
                      always select one positive and one negative literal. \
                      If we have several literals of the same polarity, we \
                      select those that are maximally ground. In a fact, i.e. \
                      a clause with all literals of the same polarity, we \
                      select one maximally nonground literal. \
                      Switches off --inherited_negative_selection.\
       <num> = 1009 => Bidirectional selection.  In a rule, i.e. a clause \
                      containing both positive and negative literals we  \
                      always select one positive and one negative literal. \
                      If we have several positive literals, we select \
                      those that are maximally nonground. If we have several \
                      negative literals, we select those that are maximally ground. \
                      In a positive fact, i.e. a clause with only positive literals, \
                      we select one maximally nonground literal. \
                      In a negative fact, i.e. a clause with only negative literals, \
                      we select one maximally ground literal. \
                     Switches off --inherited_negative_selection.\
       <num> = -2 => same as <num> = 2 but positive selection is used instead of negative \
                     selection for nonequality literals: positive nonequality literals \
                     are treated as negative and vice versa; \
       <num> = -3 => same as <num> = 3 but positive selection is used instead of negative \
                     selection for nonequality literals: positive nonequality literals \
                     are treated as negative and vice versa; \
       <num> = -4 => same as <num> = 4 but positive selection is used instead of negative \
                     selection for nonequality literals: positive nonequality literals \
                     are treated as negative and vice versa; \
       <num> = -6 => same as <num> = 6 but positive selection is used instead of negative \
                     selection for nonequality literals: positive nonequality literals \
                     are treated as negative and vice versa; \
       <num> = -7 => same as <num> = 7 but positive selection is used instead of negative \
                     selection for nonequality literals: positive nonequality literals \
                     are treated as negative and vice versa; \
       <num> = -1005 => same as <num> = 1005 but positive nonequality literals \
                     are treated as negative and vice versa; \
       <num> = -1009 => same as <num> = 1009 but positive nonequality literals \
                     are treated as negative and vice versa");       

 CL.option("literal_comparison_mode",CommandLine::Long,"<num>"," adjust comparison\
           of literals in those literal selection function that use the notion \
           of maximal literal. \
           <num> = 0 => we simply compare the atoms with the reduction ordering. \
           There is an exception: when at least one of the literals is equality, \
           we first compare precedences of their headers (predicate + polarity), \
           and only if this check is inconclusive (both are equation \
           of the same polarity) we compare the arguments. \
           <num> = 1 => we always select literals whose headers (predicate + polarity) \
           have greater precedence. Only literals with maximal headers are compared by \
           the reduction ordering. This makes comparison faster but at the cost \
           of selection quality; \
           <num> = 2 => same as <num> = 1 but complementary headers are considered \
           to have the same precedence. "); 

 CL.option("inherited_negative_selection","switches inheriting of selection of negative literals");

  // strategy 
 CL.option("forward_subsumption","switches forward subsumption");
 CL.option("fs_set_mode","switches set mode in forward subsumption, more efficient than the default multiset mode\
                          but in rare circumstances may cause incompleteness");
CL.option("fs_optimised_multiset_mode","relaxes multiset mode for forward subsumption while preserving 
                                        completeness, can be overriden by --fs_set_mode");
 CL.option("fs_use_sig_filters","switches use of signature filters in forward subsumption");
 CL.option("fs_old_style","switches use of the nonadaptive version of code trees in forward subsumption");

 CL.option("simplify_by_fs",CommandLine::Long,"<min.length>","<min.length> = 0 switches forward \
                                                              subsumption resolution off, when <min.length> >= 1 \
                                                              forward subsumption resolution is applied to clauses \
                                                              whose length >= <min.length>");

 CL.option("forward_demodulation",CommandLine::Long,"<degree>","forward simplification by unit equalities, \
                                                                <degree> <= 0 switches it off, \
                                                                <degree> = 1 only preordered rules are used, \
                                                                <degree> = 2 nearly preordered rules can be used, \
                                                                <degree> = 3 volatile rules can be used, \
                                                                <degree> >= 4 any rules can be used"); 

 CL.option("fd_on_splitting_branches","switches forward demodulation with conditions that are marks of splitting branches");
 CL.option("normalize_by_commutativity","switches normalization by commutativity laws");
   
 CL.option("backward_demodulation",CommandLine::Long,"<degree>","backward simplification by unit equalities, \
                                                                <degree> <= 0 switches it off, \
                                                                <degree> = 1 only preordered rules are used, \
                                                                <degree> = 2 nearly preordered rules can be used, \
                                                                <degree> = 3 volatile rules can be used, \
                                                                <degree> >= 4 any rules can be used"); 
 CL.option("bd_on_splitting_branches","switches backward demodulation with conditions that are marks of splitting branches");
 CL.option("simplify_by_eq_res","switches simplification by equality resolution");
 CL.option("backward_subsumption","switches backward subsumption");
 CL.option("bs_set_mode","switches set mode in backward subsumption, more efficient than the default multiset mode\
                          but in rare circumstances may cause incompleteness");
 CL.option("bs_optimised_multiset_mode","relaxes multiset mode for backward subsumption while preserving 
                                         completeness, can be overriden by --bs_set_mode");
 CL.option("orphan_murder",CommandLine::Long,"<num>","sets handling orphans, \
                                                      <num> = 0 => orphans are used");

 CL.option("unit_conflict","switches unit conflict");
 
 CL.option("lrs",CommandLine::Long,"<num>"," : switches the Limited Resource Strategy, \
                                           <num> = 0 => the LRS is off, \
                                           <num> = 1 => the standard LRS is on, \
                                           <num> = 2 or <num> = 3 => experimental implementations of LRS ");
  
 CL.option("plot_usage_pace",CommandLine::String,"<file>","undocumented feature");
 CL.option("plot_weight_limit_changes",CommandLine::String,"<file>","undocumented feature");

 CL.option("elim_def",CommandLine::Long,"<num>", "sets limit on the number of definition elimination iterations");

 CL.option("header_rank_by_arity",CommandLine::Long,"<num>","<num> = 0 - ranking depends only \
                                                             on the order of appearance, \
                                                             <num> > 0 - headers with greater arity receive \
                                                                         greater ranks, \
                                                             <num> < 0 - headers with smaller arity receive \
                                                                         greater ranks; "); 

 CL.option("header_rank_kinky","switches the kinky ranking of literal headers, \
                                for details contact the authors; ");

 CL.option("header_rank_check_positive_selection"," if it is on and positive \
              literal selection is used, then header ranking uses inversed \
              polarity too (otherwise the strategy may be incomplete); ");
  // numeric options 
 
 CL.option("t",CommandLine::Long,"<seconds>", "sets time limit to <seconds>, where <seconds> is a long integer");
 CL.option("time_limit",CommandLine::Long,"<seconds>","same as -t <seconds>");

 CL.option("m",CommandLine::Long,"<Kb>", "sets memory limit to <Kb>, where <Kb> is a long integer");
 CL.option("memory_limit",CommandLine::Long,"<Kb>","same as -m <Kb>");
 CL.option("allocation_buffer_size",CommandLine::Long,"<num>","as a percentage of the memory limit, default value is 0");
 CL.option("allocation_buffer_price",CommandLine::Double,"<coeff>","if <def> is the allocation buffer deficit, the systems abandons any attempt to eliminate the deficit as soon as <coeff>*<def> Kb have been returned to the
 heap manager,<coeff> = 1.0 is the default value");
 CL.option("max_active",CommandLine::Long,"<number>","limit on the number of active clause"); 
 CL.option("max_passive",CommandLine::Long,"<number>","limit on the number of passive clause"); 
 CL.option("w",CommandLine::Long,"<weight>","initial value of maximal \
                                              admissible clause weight, \
                                              <weight> is a long integer");
 CL.option("max_weight",CommandLine::Long,"<weight>","same as -w <weight>");
 CL.option("opt_gr_comp_weight",CommandLine::Long,"<weight>","optimal weight of a ground \
                                                              component, used for splitting in the pragmatic mode"); 
 CL.option("age_weight_ratio",CommandLine::Long,"<number>","same as pick_given_ratio in OTTER");
 CL.option("selected_pos_eq_penalty_coefficient",CommandLine::Double,"<number>"," if a positive equality is \
            selected in a clause, the weight of the clause is multiplied by this coefficient");
 CL.option("internal_tick_frequency",CommandLine::Long,"<number>","frequency of resource checks in terms of the number of generated clauses, default 10 ");
 CL.option("first_time_check",CommandLine::Long,"<number>","the LRS is off before <number>% of the time limit is used, \
                                                            unless it is switched on by memory");
 CL.option("first_mem_check",CommandLine::Long,"<number>","the LRS is off before <number>% of the memory limit is used, \
                                                           unless it is switched on by time");

 // interface options

 CL.option("show_opt","switches output of options");  
 CL.option("show_input","switches output of input clauses");
 CL.option("show_prepro","switches output of preprocessed clauses");
 CL.option("show_gen","switches output of generated inferences");
 CL.option("show_passive","switches output of passive clauses");
 CL.option("show_active","switches output of active clauses");
 CL.option("show_hyper_rules","switches output of clauses treated as rules in hyperresolution");
 CL.option("show_definitions","switches output of clauses treated as definitions in stratified resolution");
 CL.option("p","switches output of found proofs");
 CL.option("proof","same as -p");
 CL.option("show_profile","switches output of input and preprocessed input profiles");
 CL.option("show_weight_limit_changes","switches monitoring of current weight limit");
 CL.option("tab",CommandLine::String,"<file>","specifies a file to be used for tabulation");

 #ifdef INTERCEPT_FORWARD_MATCHING_INDEXING
  CL.option("fmilog",CommandLine::String,"<file>","specifies a file to be used for logging forward \
                                                   matching index operations"); 
  CL.option("fmi_log_file_max_size",CommandLine::Long,"<Kb>","specifies maximal size (in Kbytes) of the file to be used for \
                                                                logging forward matching index operations"); 
 #endif

 #ifdef INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING
  CL.option("bsilog",CommandLine::String,"<file>","specifies a file to be used for logging backward \
                                                   subsumption index operations"); 
  CL.option("bsi_log_file_max_size",CommandLine::Long,"<Kb>","specifies maximal size (in Kbytes) of the file to be used for \
                                                                logging backward subsumption index operations"); 
 
 #endif

      // non-option arguments
 CL.nonoptMiddle("<incl.file>","files that are treated as they \
                                 were included in the problem file");
 static const char* nonoptSuffNames[] = { "<problem>" };
 static const char* nonoptSuffComments[] = { "main problem file" };
 CL.nonoptSuffix(1,nonoptSuffNames,nonoptSuffComments);

}; // void describeOptions(CommandLine& CL)

int main(int numarg,char** args)
 {
  CALL("main(int numarg,char** args)");

  //DF;  TRACE(true,true);

   // initialising static parts 
  Timer timer;
  timer.reset();
  timer.start();

  //BitWord::initStatic();
  TERM::setSignature(&TheSignature);
  //Vampire::initStatic();
  //Exit::initStatic();
  GlobAlloc::setRequestMemoryRelease(releaseMemory);
  
  GlobAlloc::setReportAllocationFailure(reportAllocationFailure);

   // getting current time, pid and host name
   
  time_t* jobTime = new time_t;
  *jobTime = time((time_t*)0);
  tm* gmt = gmtime(jobTime); 
  int jobDateDD = gmt->tm_mday;
  int jobDateMM = gmt->tm_mon;    
  int jobDateYY = 1900 + gmt->tm_year;
  int jobTimeHH = gmt->tm_hour;
  int jobTimeMM = gmt->tm_min;
  int jobTimeSS = gmt->tm_sec; 

  #if ((TARGET_OS_TYPE != OS_TYPE_WIN32) || (COMPILER == COMPILER_GCC))
   pid_t jobProcessId = getpid();  
  #else
   pid_t jobProcessId = _getpid();   
  #endif

  char hostname[256];

  #if ((TARGET_OS_TYPE != OS_TYPE_WIN32) || (COMPILER == COMPILER_GCC))
   gethostname(hostname,256);
  #else
   strcpy(hostname,"unknown_Win32_driven_machine");
  #endif

   // parsing command line 
  CommandLine CL; 
  describeOptions(CL);
  CL.arguments(numarg,args);
  if (!CL.parse())
   {
    if (CL.values("v"))
     {
      cout << "% Vampire " << 'v' << VampireVersionNum << VampireVersionDesc << '\n';
     };
    if (CL.values("h") || CL.values("help"))
     { 
      CL.printHelp("vampire");  
     };
    if (CL.values("v") || CL.values("h") || CL.values("help")) Exit::exit(1);
    CL.printError();
    CL.printUsage("vampire");
    Exit::exit(1);
   };
    
  if (CL.values("h") || CL.values("help")) CL.printHelp("vampire"); 

  const char* testId = "unspecified_test";
  (CL.values("T").next(testId) || CL.values("test").next(testId));

  const char* TPTP_DIRECTORY_ENV_VAR = "TPTP_DIR";
  const char* TPTPDir = "."; // default value
  const char* tmpTPTPdir = getenv(TPTP_DIRECTORY_ENV_VAR);
  (CL.values("I").next(tmpTPTPdir) || CL.values("include").next(tmpTPTPdir));
  if (tmpTPTPdir) TPTPDir = tmpTPTPdir;

  const char* problemFileName = (char*)(CL.nonoptVal("<problem>"));
  const char* problemFileBaseName = basename(problemFileName);
  const char* jobId = computeJobId(gmt,jobProcessId); 
 
  Tabulation* cl_tabulation = (Tabulation*)0;
  //ofstream tab_stream;
  Tabulation tabulation;

  CommandLine::Values tabFlag = CL.values("tab");

  const char* tab_file = (const char*)0;
  if (tabFlag.next(tab_file)) 
   {
    tab_stream.open(tab_file,ios::app);
    if (!tab_stream) RuntimeError::report("Can not open ",tab_file," for tabulation.\n");
    tabulation.assignStream(&tab_stream);
    tabulation.assignJob(jobId); 
    cl_tabulation = &tabulation;
   };

  #ifdef INTERCEPT_FORWARD_MATCHING_INDEXING
   const char* forwardMatchingIndexLogFile = "ForwardMatchingIndex.log";
   CL.values("fmilog").next(forwardMatchingIndexLogFile);
   FMILog.open(forwardMatchingIndexLogFile);
   maxSizeOfFMILogFile = LONG_MAX;
   if (CL.values("fmi_log_file_max_size").next(maxSizeOfFMILogFile))
    {
     maxSizeOfFMILogFile = maxSizeOfFMILogFile * 1024;
    };
  #endif

  #ifdef INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING
   const char* backwardSubsumptionIndexLogFile = "BackwardSubsumptionIndex.log";
   CL.values("bsilog").next(backwardSubsumptionIndexLogFile);

   BSILog.open(backwardSubsumptionIndexLogFile);

   maxSizeOfBSILogFile = LONG_MAX;
   if (CL.values("bsi_log_file_max_size").next(maxSizeOfBSILogFile))
    {
     maxSizeOfBSILogFile = maxSizeOfBSILogFile * 1024;
    };
  #endif

  bool cl_auto = (CL.values("a") || CL.values("auto"));
  long cl_time_limit = LONG_MAX;
  (CL.values("t").next(cl_time_limit) || CL.values("time_limit").next(cl_time_limit));
   
  #ifndef NO_DEBUG
   Debug::assignJob(problemFileBaseName);
  #endif 

   // output some information
  cout << "%=================================================================\n";
  cout << "% Vampire " << 'v' << VampireVersionNum << VampireVersionDesc << '\n';

  #ifndef NO_DEBUG 
   cout << "% WARNING: COMPILED WITHOUT -DNO_DEBUG!\n";
   cout << "%          THIS MAY AFFECT THE PERFOMANCE BADLY!\n";         
  #endif
  #ifdef INTERCEPT_FORWARD_MATCHING_INDEXING
   cout << "% WARNING: FORWARD MATCHING INDEX OPERATIONS ARE LOGGED!\n"; 
   cout << "%          THIS MAY AFFECT THE PERFOMANCE BADLY!\n";         
  #endif
  #ifdef INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING
   cout << "% WARNING: BACKWARD SUBSUMPTION INDEX OPERATIONS ARE LOGGED!\n"; 
   cout << "%          THIS MAY AFFECT THE PERFOMANCE BADLY!\n";         
  #endif
  cout << "% Started: " << asctime(localtime(jobTime));
  cout << "% Host:    " << hostname << '\n';
  cout << "% Problem: " << problemFileName << '\n';
  cout << "% Command: "; CL.printCommand(cout) << '\n';
  cout << "% Job ID: " << jobId << "  Test ID: " << testId << '\n';
  cout << "% Auto: " << ((cl_auto) ? "on" : "off") << '\n';
  cout << "%=================================================================\n";


  // tabulate some information

  if (cl_tabulation) 
   { 
    tabulation.header(testId,problemFileBaseName);   
    tabulation.problemName(problemFileBaseName);
    tabulation.testId(testId);
    tabulation.hostName((const char*)hostname);
    tabulation.jobDate(jobDateDD,jobDateMM,jobDateYY);
    tabulation.jobTime(jobTimeHH,jobTimeMM,jobTimeSS);
    tabulation.version(VampireVersionNum,VampireVersionDesc);
    tabulation.versionDate();
    tabulation.versionTime();
    tabulation.option("time_limit",cl_time_limit);  
    tabulation.option("strat_auto",cl_auto);  
   };



  // parsing TPTP files
  const char* currFile;
  CommandLine::Values incFiles = CL.nonoptMiddleValues();
  bool incompleteAPriori = false;
  while (incFiles.next(currFile)) 
   {
    if (!TPTPParser::readFile(TPTPDir,currFile)) Exit::exit(1);
    incompleteAPriori = incompleteAPriori || TPTPParser::someInputClausesRejected(); 
   };
 
  if (!TPTPParser::readFile(TPTPDir,problemFileName)) Exit::exit(1);
  incompleteAPriori = incompleteAPriori || TPTPParser::someInputClausesRejected(); 
   
  // submitting options to Vampire
  // automatic mode 

  Vampire::options.incompleteAPriori = incompleteAPriori;
 
  Vampire::options.autoStrat = cl_auto;    

  // main algorithm
  
  Vampire::options.main_alg = 0;
  CL.values("main_alg").next(Vampire::options.main_alg); 

   
  // inference rules
  Vampire::options.hyper_resolution = (bool)CL.values("hyper");
  Vampire::options.no_inferences_between_rules = (bool)CL.values("no_inferences_between_rules");

  Vampire::options.static_splitting = (bool)CL.values("static_splitting");
  Vampire::options.dynamic_splitting = (bool)CL.values("dynamic_splitting");
  Vampire::options.splitting_style = 0L; // default
  CL.values("splitting_style").next(Vampire::options.splitting_style);
  Vampire::options.splitting_with_naming = (bool)CL.values("splitting_with_naming");

  Vampire::options.static_neg_eq_splitting = 0L;
  CL.values("static_neg_eq_splitting").next(Vampire::options.static_neg_eq_splitting); 
  Vampire::options.paramodulation = (bool)CL.values("paramodulation");
  Vampire::options.sim_back_sup = 0L;
  CL.values("sim_back_sup").next(Vampire::options.sim_back_sup);  
  Vampire::options.hyper_positive = (bool)CL.values("hyper_positive");
  Vampire::options.selection = 1L;
  CL.values("selection").next(Vampire::options.selection); 
  Vampire::options.literal_comparison_mode = 1L;
  CL.values("literal_comparison_mode").next(Vampire::options.literal_comparison_mode);  
  Vampire::options.inherited_negative_selection = (bool)CL.values("inherited_negative_selection");
   
   // strategy 
  Vampire::options.forward_subsumption = (bool)CL.values("forward_subsumption");
  Vampire::options.fs_set_mode = (bool)CL.values("fs_set_mode"); 
  Vampire::options.fs_optimised_multiset_mode = (bool)CL.values("fs_optimised_multiset_mode");  
  Vampire::options.fs_use_sig_filters = (bool)CL.values("fs_use_sig_filters");
  Vampire::options.fs_old_style = (bool)CL.values("fs_old_style");

  Vampire::options.simplify_by_fs = 0L;
  CL.values("simplify_by_fs").next(Vampire::options.simplify_by_fs);

  Vampire::options.forward_demodulation = 0; // default
  CL.values("forward_demodulation").next(Vampire::options.forward_demodulation);
  Vampire::options.fd_on_splitting_branches = (bool)CL.values("fd_on_splitting_branches");

  Vampire::options.normalize_by_commutativity = (bool)CL.values("normalize_by_commutativity");

  Vampire::options.backward_demodulation = 0; // default
  CL.values("backward_demodulation").next(Vampire::options.backward_demodulation);
  Vampire::options.bd_on_splitting_branches = (bool)CL.values("bd_on_splitting_branches");
  
  

  Vampire::options.simplify_by_eq_res = (bool)CL.values("simplify_by_eq_res");
  Vampire::options.backward_subsumption = (bool)CL.values("backward_subsumption");
  Vampire::options.bs_set_mode = (bool)CL.values("bs_set_mode");  
  Vampire::options.bs_optimised_multiset_mode = (bool)CL.values("bs_optimised_multiset_mode");  
  Vampire::options.orphan_murder = 0; // default
  CL.values("orphan_murder").next(Vampire::options.orphan_murder);

  Vampire::options.unit_conflict = (bool)CL.values("unit_conflict");

  Vampire::options.lrs = 0L;
  CL.values("lrs").next(Vampire::options.lrs);  

  const char* usage_pace_file = (const char*)0; 
  if (CL.values("plot_usage_pace").next(usage_pace_file))
   { 
    usage_pace_stream = new ofstream();
    usage_pace_stream->open(usage_pace_file,ios::out);
    if (!(*usage_pace_stream)) RuntimeError::report("Can not open ",usage_pace_file,".\n"); 
   };
  Vampire::options.plot_usage_pace = usage_pace_stream;
  

  const char* weight_limit_changes_file = (const char*)0; 
  if (CL.values("plot_weight_limit_changes").next(weight_limit_changes_file))
   { 
    weight_limit_changes_stream = new ofstream();
    weight_limit_changes_stream->open(weight_limit_changes_file,ios::out);
    if (!(*weight_limit_changes_stream)) RuntimeError::report("Can not open ",weight_limit_changes_file,".\n"); 
   };
  Vampire::options.plot_weight_limit_changes = weight_limit_changes_stream;  


  Vampire::options.elim_def = 0; 
  CL.values("elim_def").next(Vampire::options.elim_def);
  Vampire::options.header_rank_by_arity = 0;  
  CL.values("header_rank_by_arity").next(Vampire::options.header_rank_by_arity);
  Vampire::options.header_rank_kinky = (bool)CL.values("header_rank_kinky");
  Vampire::options.header_rank_check_positive_selection = 
   (bool)CL.values("header_rank_check_positive_selection");
    // numeric options 
  long cl_memory_limit = LONG_MAX;
  (CL.values("m").next(cl_memory_limit) || CL.values("memory_limit").next(cl_memory_limit));
  long allocBufferSize = 0L;
  CL.values("allocation_buffer_size").next(allocBufferSize);
  Vampire::options.allocation_buffer_price = 1.0;
  CL.values("allocation_buffer_price").next(Vampire::options.allocation_buffer_price);

  CL.values("max_active").next(Vampire::options.maxActive);
  CL.values("max_passive").next(Vampire::options.maxPassive);
  (CL.values("w").next(Vampire::options.maxWeight) || CL.values("max_weight").next(Vampire::options.maxWeight));
  Vampire::options.optGrCompWeight = LONG_MAX;
  CL.values("opt_gr_comp_weight").next(Vampire::options.optGrCompWeight);
  Vampire::options.age_weight_ratio = 0;
  CL.values("age_weight_ratio").next(Vampire::options.age_weight_ratio);  
  double selPosEqPenaltyCoeff = 1.0; // default 
  CL.values("selected_pos_eq_penalty_coefficient").next(selPosEqPenaltyCoeff);
  Vampire::options.selected_pos_eq_penalty_coefficient = (float)selPosEqPenaltyCoeff;
  Vampire::options.internal_tick_frequency = 10; // default
  CL.values("internal_tick_frequency").next(Vampire::options.internal_tick_frequency);
  CL.values("first_time_check").next(Vampire::options.first_time_check);  
  CL.values("first_mem_check").next(Vampire::options.first_mem_check);   
    
    // interface options    
  Vampire::options.showOpt = (bool)CL.values("show_opt");
  Vampire::options.showInput = (bool)CL.values("show_input");
  Vampire::options.showPrepro = (bool)CL.values("show_prepro");
  Vampire::options.showGen = (bool)CL.values("show_gen");
  Vampire::options.showPassive = (bool)CL.values("show_passive");
  Vampire::options.showActive = (bool)CL.values("show_active");
  Vampire::options.showHyperRules = (bool)CL.values("show_hyper_rules");
  Vampire::options.showDefinitions = (bool)CL.values("show_definitions");
  Vampire::options.showProof = (CL.values("p") || CL.values("proof"));
  Vampire::options.showProfile = (bool)CL.values("show_profile"); 
  Vampire::options.showWeightLimitChanges = (bool)CL.values("show_weight_limit_changes"); 

  Vampire::options.tabulation = cl_tabulation;   

    // submitting input to Vampire

   for (const TPTPParser::HeaderList* lst = TPTPParser::definedHeaders(); lst; lst = lst->tl())
    Vampire::definedHeader(lst->hd());

   for (const ClauseQueue::Entry* qe = TPTPParser::inputClauses().begin(); 
         qe;
         qe = qe->next())
     { 
      if (!Vampire::inputClause(qe->value())) goto terminate;
     };                 
     
    // start proving

  Vampire::options.timeLimit = cl_time_limit - timer.elapsed();
  Vampire::options.memoryLimit = ((float)cl_memory_limit)*1024;
  GlobAlloc::setHardMemoryLimit(cl_memory_limit*1024); 
  GlobAlloc::setAllocationBufferSize(allocBufferSize);
 
  initInterruptHandler();
  Vampire::initSIGINTHandler();
  #ifndef DEBUG_ERROR_SIGNALS
   initErrorSigHandler();
  #endif
  Vampire::makeOptionsConsistent();
  if (CL.values("prepro_only"))
   {
    Vampire::preproOnly();
   }
  else Vampire::prove();

  cancelInterruptHandling(); 
  #ifndef DEBUG_ERROR_SIGNALS
   cancelErrorSigHandling();
  #endif
  terminate:
   if (tab_stream) tab_stream.close();
   if (usage_pace_stream) usage_pace_stream->close();
   if (weight_limit_changes_stream) weight_limit_changes_stream->close();
   return 0;
}; // int main(int numarg,char** args)

//====================================================




