//
// File:         Master.h
// Description:  Main loop.
// Created:      Mar 5, 2000, 15:00
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#ifndef MASTER_H
//====================================================
#define MASTER_H
#include "jargon.h"
#include "MasterDebugFlags.h"
#include "clause.h"
#include "GlobAlloc.h"
#include "Timer.h"
#include "clause_set.h"
#include "kept.h"
#include "ClauseBackground.h"
#include "FIFO.h"
#include "sharing.h"
#include "clause_index.h"
#include "NewClause.h"
#include "forward_subsumption.h"
#include "ComponentName.h"
#include "InferenceDispatcher.h"
#include "UnitConflict.h"
#include "BackwardSubsumption.h"
#include "ForwardDemodulation.h"
#include "BackDemodulation.h"
#include "ClauseNumIndex.h"
#include "SharedKBConstr.h"
#include "ClauseQueue.h"
#include "PagedPassiveClauseStorage.h"
extern SUBSTITUTION GlobalSubst; // temporary
//====================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_MASTER
 #define DEBUG_NAMESPACE "Master"
#endif
#include "debugMacros.h"
//====================================================
class Master
{
 public:
  enum TerminationReason { UnknownTR,Proof, EmptyPassive, TimeLimit, MemoryLimit, AllocationFailure, MaxPassive, MaxActive, UserInt }; 
  class Options
  {
   public:
    Options() { reset(); };
    void reset(); 
     // main algorithm
    bool incompleteAPriori;
    long main_alg;
     // inference rules
    bool binary_resolution;
    bool hyper_resolution;
    bool no_inferences_between_rules;
    bool splitting;
    long splitting_style;
    bool splitting_with_naming;
    long neg_eq_splitting;   
    bool paramodulation;
    long sim_back_sup;
    bool hyper_positive;
    long selection;
    long literal_comparison_mode;
    bool inherited_negative_selection;

     // strategy switches
    bool handle_equality;
    bool forward_subsumption;
    bool fs_set_mode;
    bool fs_optimised_multiset_mode;
    bool fs_use_sig_filters;
    bool fs_old_style;
    long simplify_by_fs;
    long forward_demodulation;
    bool fd_on_splitting_branches;
    const Clause* const * normalize_by_commutativity;
    long backward_demodulation;
    bool bd_on_splitting_branches;
    bool simplify_by_eq_res;
    bool backward_subsumption;
    bool bs_set_mode;
    bool bs_optimised_multiset_mode;
    long orphan_murder;
    bool hyper_dinamic_lhs_ordering;
    bool unit_conflict;
    long lrs;
    const bool* pred_elim_table;
    const bool* defined_header_table; // for stratified resolution 
     // numeric options
    long age_weight_ratio;
    float timeLimit;
    float memoryLimit;
    float allocation_buffer_price; 
    long maxActive;
    long maxPassive;
    long maxWeight;    
    long optGrCompWeight; 
    float first_time_check;
    float first_mem_check;
    long internal_tick_frequency;    
     // interface options
    bool showGen;
    void (*showPassive) (const Clause*);        
    void (*showActive) (const Clause*); 
    void (*showHyperRule) (const Clause*);
    void (*showDefinition) (const Clause*);
    void (*showWeightChange) (long,long);
    void (*showUsagePace) (float);
  }; // class Options

 public:
  Master(ulong (*nextClauseNumFunc)())
   : options(),
     timer(),
     terminated(false),
     proof((Clause*)0),
     _parameters(),
     _passive(),
     _pagedPassive(),
     _reservedPassive("Master::_reservedPassive"),
     _active("Master::_active"),
     _history("Master::_history"),
     _definitions("Master::_definitions"),
     _retainedNewClauses(),
     _queueForDisposal(),
     _queueForBackSimplification(),
    //_queueForUnitConflict(),
     _infDispatcher(&_freshClause),
     _unitConflict(),
     _forwardSubsumption(),
     _backSubsumption(&_clauseNumIndex),
     _forwardDemodulation(),
     _backDemodulation(&_freshClause,&_sharingIndex,&_clauseNumIndex),
     _componentName(),
     _sharingIndex(),
     _clauseIndex(&_sharingIndex),
     _inferenceDispatcherSemaphore(false),
     _backwardSubsumptionSemaphore(false),
     _backwardDemodulationSemaphore(false),
     _freshClause(),
     _clauseNumIndex(),
     _nextClauseNum(nextClauseNumFunc) 
  {
   CALL("constructor Master(ulong (*nextClauseNumFunc)())");
   init();        
  };
  ~Master() 
  {
   CALL("destructor ~Master()");
  };
  void init() 
  {
   _freshClause.SetClauseIndex(&_clauseIndex);
   _freshClause.SetComponentName(&_componentName);
   _freshClause.SetForwardSubsumption(&_forwardSubsumption);  
   _freshClause.SetSimplificationByUEq(&_forwardDemodulation);  
  };

  void setStrategy()
  { 
   initParameters();
   initObjects();
  };

  void destroy()
  {
   CALL("destroy()");
  
    // queue for disposal 
    // not after _history.DestroyAllClauses()! 
  
   disposeObsolete();

    // sets of persistent clauses
   switch (_parameters.main_alg)
    {
     case 0: destroyPassive0(); break;
     case 1: destroyPassive1(); break;
     case 2: destroyPassive2(); break;
     case 3: destroyPassive3(); break;
     default: destroyPassive0(); 
    }; 
   //_passive.DestroyAllClauses();
   _reservedPassive.DestroyAllClauses();
   _active.DestroyAllClauses();
   _history.DestroyAllClauses();
   _definitions.DestroyAllClauses();   

    // queues
   Clause* retainedNewClause;

   while (_retainedNewClauses.dequeue(retainedNewClause)) 
    { 
     ASSERT(retainedNewClause->mainSet() == Clause::MainSetNew);
     retainedNewClause->setMainSet(Clause::MainSetFree);
     disposeClause(retainedNewClause);  
    };

   _queueForBackSimplification.reset();
   //_queueForUnitConflict.reset();

    // indices and inference engines   

   _infDispatcher.destroy();
   _unitConflict.reset();
   _forwardSubsumption.Destroy();
   _backSubsumption.destroyIndex();
   _forwardDemodulation.Destroy();
   _backDemodulation.destroy();
   SharedKBConstr::destroySharingIndex(); // not before _infDispatcher.destroy(), _forwardDemodulation.Destroy() and _backDemodulation.destroy()!
   _sharingIndex.Destroy(); 
   _clauseIndex.Destroy();    
   _clauseNumIndex.reset();

  };

  void reset() 
  {
   CALL("reset()");
   GlobalSubst.reset();
   destroy();
   options.reset(); 
   resetStatistics();
   _backDemodulation.resetStat(); 
   ASSERT(clean());    
  }; 

  #ifndef NO_DEBUG
   bool clean();
  #endif 


  void endOfInput();
  void mainLoop() 
  { 
   switch (_parameters.main_alg)
    {
     case 0: mainLoop0(); break;
     case 1: mainLoop1(); break;
     case 2: mainLoop2(); break;
     case 3: mainLoop3(); break; 
     default: mainLoop0(); 
    };     
  };
  void eliminateDefinitions();
  void inputClause(Clause* cl) 
  { 
   switch (_parameters.main_alg)
    {
     case 0: inputClause0(cl); break;
     case 1: inputClause1(cl); break;
     case 2: inputClause2(cl); break;
     case 3: inputClause3(cl); break;
     default: inputClause0(cl);  
    };    
  };

  bool releaseMemory(ulong size)
  {
   CALL("releaseMemory(ulong size)");
   switch (_parameters.main_alg)
    {
     case 0: return false;
     case 1: return false;
     case 2: return releaseMemory2(size);
     case 3: return releaseMemory3(size);
     default: return false;  
    };        
  };

  bool releaseMemory2(ulong size);
  bool releaseMemory3(ulong size);

  void reportAllocationFailure(ulong size) { terminate(AllocationFailure); };

  void passiveResetIter() { _passive.ResetIterator(); };
  Clause* nextPassive() { return _passive.IterNext(); };
  void historyResetIter() { _history.ResetIterator(); };
  Clause* nextHistory() { return _history.Next(); };
  void terminate(TerminationReason tr) { terminated = true; terminationReason = tr; };
  ulong statGenerated() const { return _freshClause.statistics.generated; };
  ulong statFSubsumed() const { return _freshClause.statistics.forwardSubsumed; };   
  ulong statFSimplified() const 
  {
   return statFSimplifiedByFSR() + statFSimplifiedByDemod() + statFSimplifiedByEqRes();
  };
  ulong statFSimplifiedByFSR() const 
  { 
   return _freshClause.statistics.simplifiedByFSR; 
  };
  ulong statFSimplifiedByDemod() const 
  { 
   return _freshClause.statistics.simplifiedByDemod; 
  };
  ulong statFDOnSplittingBranches() const
  {
   return _freshClause.statistics.fDOnSplittingBranches;
  };
  ulong statFSimplifiedByEqRes() const 
  { 
   return _freshClause.statistics.simplifiedByEqRes; 
  }; 
  ulong statSplit() const { return _freshClause.statistics.split; };
  float statAvgComponents() const 
  { 
   return (_freshClause.statistics.split) ? 
           (((float)_freshClause.statistics.components)/_freshClause.statistics.split)
           :
           1;
  };
  ulong statDiffComponents() const { return _freshClause.statistics.differentComponents; };
  ulong statEqTautologies() const { return _freshClause.statEqTautologies(); };
  ulong statPropTautologies() const { return _freshClause.statPropTautologies(); };
  ulong statTooHeavy() const { return _freshClause.statTooHeavy(); };
  ulong statTooHeavyLiterals() const { return _freshClause.statTooHeavyLiterals(); };
  ulong statTooManyVariables() const { return _freshClause.statTooManyVariables(); };
  ulong statTooLong() const { return _freshClause.statTooLong(); };
  ulong statTooDeep() const { return _freshClause.statTooDeep(); }; 
  ulong statRetained() const 
  {
   if (_parameters.main_alg == 3)
    {
     return _pagedPassive.totalNumberOfClauses();
    };
   return _passive.Total(); 
  };
  ulong statRefusedAllocation() const { return _statClausesRefusedAllocation; };
  ulong statUsed() const { return _active.Total(); };
  ulong statBSubsumed() const { return _statBackSubsumed; };   
  ulong statBSimplified() const { return _statBackSimplified; };
  ulong statBDOnSplittingBranches() const
  {
   return _backDemodulation.statBDOnSplittingBranches();
  };
  ulong statRewrittenByDefUnfolding() const { return _statRewrittenByDefUnfolding; }; 

  ulong statOrphansMurdered() const 
  { 
   return _statPassiveOrphansMurdered + _statActiveOrphansMurdered; 
  };
  ulong statPassiveOrphansMurdered() const { return _statPassiveOrphansMurdered; };
  ulong statActiveOrphansMurdered() const { return _statActiveOrphansMurdered; };
  ulong statRecycled() const { return _statRecycled + _pagedPassive.numberOfCancelledClauses(); }; 
  ulong statCurrentlyPassive() const { return _passive.CurrentlyKept() + _pagedPassive.numberOfCurrentlyKeptClauses(); };
  ulong statReservedPassive() const { return _reservedPassive.Size(); };
  ulong statPassiveReachable() const { return _parameters.reachable; };
  ulong statCurrentlyActive() const { return _active.Size(); };

  bool completeStrategyUsed();

  /***************/
  void scanSharingIndex(ostream& str) // for debugging
  {
   cout << "Scanning the sharing index... ";
   cout.flush();
   _sharingIndex.ScanIndex(str);
   cout << "done\n";
  };

  void dumpPersistentClauses(ostream& str) // for debugging
  {
   Clause* cl;
   cout << "Dumping the active clauses... ";
   cout.flush();
   _active.ResetIterator();
   while ((cl = _active.Next()))
    { 
     str << cl << "\n";
    };
   cout << "done\n";

   cout << "Dumping the passive clauses... ";
   cout.flush();
   _passive.ResetIterator();
   while ((cl = _passive.IterNext()))
    {
     str << cl << "\n";
    };
   cout << "done\n";

   cout << "Dumping the history clauses... ";
   cout.flush();
   _history.ResetIterator();
   while ((cl = _history.Next()))
    { 
     str << cl << "\n";
    };
   cout << "done\n";

   cout << "Dumping the retained new clauses... ";
   cout.flush();
   for (ClauseQueue::Entry* qe = _retainedNewClauses.begin(); qe; qe = qe->next())
    {
     str << qe->value() << "\n";
    };
   cout << "done\n";

   cout << "Dumping the clauses from the queue for disposal... ";
   cout.flush();
   for (ClauseQueue::Entry* qe = _queueForDisposal.begin(); qe; qe = qe->next())
    {
     str << qe->value() << "\n";
    };
   cout << "done\n";
  };
  /*************/
 
 public:
  Options options;
  Timer timer;
  bool terminated;
  TerminationReason terminationReason; 
  Clause* proof;
 private:
  class Parameters
  {
   public:
    Parameters() {};
    // main algorithm
    bool incompleteAPriori;
    long main_alg;
    // inference rules
    bool binary_resolution;
    bool hyper_resolution;
    bool no_inferences_between_rules;
    bool splitting;  
    long splitting_style;
    bool splitting_with_naming;
    long neg_eq_splitting;
    bool paramodulation;
    long sim_back_sup;
    bool hyper_positive;
    long selection;
    long literal_comparison_mode;
    bool inherited_negative_selection;

    // strategy switches
    bool handle_equality;
    bool forward_subsumption;
    bool fs_set_mode;
    bool fs_optimised_multiset_mode;
    bool fs_use_sig_filters;
    bool fs_old_style;
    long simplify_by_fs;
    long forward_demodulation;
    bool fd_on_splitting_branches;
    const Clause* const * normalize_by_commutativity;
    long backward_demodulation;
    bool bd_on_splitting_branches;
    bool simplify_by_eq_res;
    bool backward_subsumption;
    bool bs_set_mode;
    bool bs_optimised_multiset_mode;
    long orphan_murder;
    bool hyper_dinamic_lhs_ordering;
    bool unit_conflict;
    long lrs; 
    const bool* pred_elim_table; 
    const bool* defined_header_table;
     // numeric parameters
    long age_weight_ratio;
    float timeLimit;
    float memoryLimit;
    float allocation_buffer_price; 
    long maxActive;
    long maxPassive; 
    long maxWeight;
    long maxMaxWeight;
    long optGrCompWeight; 
    float first_time_check;
    float first_mem_check; 
    long reachable;
    long unreachable;
    long internal_tick_frequency; 
     // interface
    bool showGen;
    void (*showPassive) (const Clause*);     
    void (*showActive) (const Clause*);
    void (*showHyperRule) (const Clause*);
    void (*showDefinition) (const Clause*);
    void (*showWeightChange) (long,long);
    void (*showUsagePace) (float); 

     // auxilliary
    bool useClauseNumIndex;  
  }; // class Parameters
 private:
  void resetStatistics() 
  {
   _statClausesRefusedAllocation = 0;
   _statBackSubsumed = 0;
   _statBackSimplified = 0;
   _statRewrittenByDefUnfolding = 0;
   _statPassiveOrphansMurdered = 0;
   _statActiveOrphansMurdered = 0; 
   _statRecycled = 0;
  };
  void initParameters();            
  void initObjects();
  void inputClause0(Clause* cl);
  void inputClause1(Clause* cl);
  void inputClause2(Clause* cl);
  void inputClause3(Clause* cl);
  void mainLoop0();
  void mainLoop1();
  void mainLoop2();
  void mainLoop3();
  Clause* reanimate1(Clause* cl);
  Clause* reanimate2(Clause* cl);
  Clause* reanimate3(PagedPassiveClauseStorage::StoredClause& cl); 
  void processCollected0();
  void processCollected1();
  void processCollected2();
  void processCollected3();  
  void processCollectedInput0();
  void processCollectedInput1();
  void processCollectedInput2(); 
  void processCollectedInput3(); 
  bool checkForUnitConflict(Clause* cl);
  Clause* select0();
  Clause* select1() 
  { 
   CALL("select1()");
   Clause* res = _passive.Next();
   if (res) 
    {
     ASSERT(res->mainSet() == Clause::MainSetPassive);
     res->setMainSet(Clause::MainSetFree); 
    };
   return res;  
  };
  Clause* select2() 
  {
   CALL("select2()");
   Clause* res = _passive.Next();
   if (res) 
    {
     ASSERT(res->mainSet() == Clause::MainSetPassive);
     res->setMainSet(Clause::MainSetFree); 
    };   
   return res;  
  };
  Clause* select3() 
  {
   CALL("select3()");
   Clause* res = _passive.Next();
   if (res) 
    {
     ASSERT(res->mainSet() == Clause::MainSetPassive);
     res->setMainSet(Clause::MainSetFree); 
    };   
   return res;  
  };
  
  void makeActive0(Clause* cl);
  void makeActive1(Clause* cl); 
  void makeActive2(Clause* cl);
  void makeActive3(Clause* cl);
  void makePassive0(Clause* cl);
  void makePassive1(Clause* cl);
  void makePassive2(Clause* cl);
  bool insertIntoPassive3(Clause* cl);
  void integratePassive(Clause* cl);
  void makeReserved(Clause* cl);
  void enqueueForDisposal(Clause* cl);

  void disposeObsolete();
  void eliminateAllocationBufferDeficit0();
  void eliminateAllocationBufferDeficit1();
  void eliminateAllocationBufferDeficit2();
  void eliminateAllocationBufferDeficit3();

  void killActiveOrphans0();
  void killActiveOrphans2();
  void killActiveOrphans3();

  void removeFromIndices(Clause* cl);
  void disposeClause(Clause* cl);
  void destroyPassive0() { _passive.DestroyAllClauses(); };
  void destroyPassive1() { _passive.DestroyAllClauses(); };
  void destroyPassive2();
  void destroyPassive3();
  void loadReservedClauses();
  bool collectBackSimplified(); // _backDemodulation -->> _freshClause 
  bool backSubsumeBy(Clause* cl);
  void backSimplify();  
  void backSimplifyBy1(Clause* simplifier);
  void backSimplifyBy2(Clause* simplifier);
  void backSimplifyBy3(Clause* simplifier);
  float usedMemory();
  void estimateResources()
  {
   switch (_parameters.lrs)
    {
     case 0L: 
      if (_parameters.showUsagePace)
       {
        float used_time = timer.elapsed();
        if (used_time < 1.0) used_time = 1.0;      
        long moved_from_passive = _passive.UsedAndRemoved();    
        _parameters.showUsagePace(moved_from_passive/used_time);
       };
      return;

     case 1L: estimateResources1(); break;
      
     case 2L: estimateResources2(); break;

     case 3L: estimateResources3(); break;

     default: estimateResources1(); 
    };
  };
  
  void estimateResources1(); 
  void estimateResources2();
  void estimateResources3();
  void anotherInference0();
  void anotherInference1();
  void anotherInference2();
  void anotherInference3(); // ?

  void internalTickEvent0();
  void internalTickEvent1();
  void internalTickEvent2();
  void internalTickEvent3();
 

  void integrateIntoForwardSubsumptionIndex(Clause* cl) 
  { 
   CALL("integrateIntoForwardSubsumptionIndex(Clause* cl)");
   ASSERT(!cl->flag(Clause::FlagInForwardSubsumptionIndex));
   _forwardSubsumption.Integrate(cl);
   cl->setFlag(Clause::FlagInForwardSubsumptionIndex);
  };
  
  void removeFromForwardSubsumptionIndex(Clause* cl) 
  { 
   CALL("removeFromForwardSubsumptionIndex(Clause* cl)");
   ASSERT(cl->flag(Clause::FlagInForwardSubsumptionIndex));  
   _forwardSubsumption.Remove(cl);
   cl->clearFlag(Clause::FlagInForwardSubsumptionIndex);
  };

  void integrateIntoBackwardSubsumptionIndex(Clause* cl)
  {
   CALL("integrateIntoBackwardSubsumptionIndex(Clause* cl)");
   ASSERT(!cl->flag(Clause::FlagInBackwardSubsumptionIndex));
   _backSubsumption.integrate(cl);
   cl->setFlag(Clause::FlagInBackwardSubsumptionIndex);
  };

  void removeFromBackwardSubsumptionIndex(Clause* cl) 
  { 
   CALL("removeFromBackwardSubsumptionIndex(Clause* cl)");
   ASSERT(cl->flag(Clause::FlagInBackwardSubsumptionIndex));  
   _backSubsumption.remove(cl);
   cl->clearFlag(Clause::FlagInBackwardSubsumptionIndex);
  };

  void integrateIntoForwardDemodulationIndex(Clause* cl) 
  { 
   CALL("integrateIntoForwardDemodulationIndex(Clause* cl)");
   ASSERT(!cl->flag(Clause::FlagInForwardDemodulationIndex));
   _forwardDemodulation.Integrate(cl);
   cl->setFlag(Clause::FlagInForwardDemodulationIndex);
  };
  
  void removeFromForwardDemodulationIndex(Clause* cl) 
  { 
   CALL("removeFromForwardDemodulationIndex(Clause* cl)");
   ASSERT(cl->flag(Clause::FlagInForwardDemodulationIndex));  
   _forwardDemodulation.Remove(cl);
   cl->clearFlag(Clause::FlagInForwardDemodulationIndex);
  };

  void integrateIntoBackDemodulationIndex(Clause* cl)
  {
   CALL("integrateIntoBackDemodulationIndex(Clause* cl)");
   ASSERT(!cl->flag(Clause::FlagInBackDemodulationIndex));
   _backDemodulation.integrate(cl);
   cl->setFlag(Clause::FlagInBackDemodulationIndex);
  };

  void removeFromBackDemodulationIndex(Clause* cl) 
  { 
   CALL("removeFromBackDemodulationIndex(Clause* cl)");
   ASSERT(cl->flag(Clause::FlagInBackDemodulationIndex));  
   _backDemodulation.remove(cl);
   cl->clearFlag(Clause::FlagInBackDemodulationIndex);
  };
   
  bool backDemodulationQuery(Clause* cl)
  { 
   CALL("backDemodulationQuery(Clause* cl)");
   _backDemodulationQuery = cl;
   if (_backDemodulation.query(cl)) { cl->Block(); return true; };
   return false;
  };

  bool nextBackDemodulated()
  {
   CALL("nextBackDemodulated()");
   if (!_backDemodulation.nextSimplified()) { _backDemodulationQuery->Unblock(); return false; };
   return true;  
  };

 private:
  Parameters _parameters;
  KEPT _passive;
  PagedPassiveClauseStorage _pagedPassive;
  CLAUSE_SET _reservedPassive;
  CLAUSE_SET _active;
  CLAUSE_SET _history;
  CLAUSE_SET _definitions;
  ClauseQueue _retainedNewClauses;
  ClauseQueue _queueForDisposal;

  ClauseQueue _queueForBackSimplification;

  //FIFO<GlobAlloc,Clause*> _queueForUnitConflict;
  InferenceDispatcher _infDispatcher;
  UnitConflict _unitConflict;
  FS _forwardSubsumption;
  BackwardSubsumption _backSubsumption;
  ForwardDemodulation _forwardDemodulation;
  BackDemodulation _backDemodulation;
  Clause* _backDemodulationQuery;
  ComponentName _componentName;  // temporary, must be part of a higher level module
  SHARING_INDEX _sharingIndex;
  CLAUSE_INDEX _clauseIndex; 
  bool _inferenceDispatcherSemaphore;
  bool _backwardSubsumptionSemaphore;
  bool _backwardDemodulationSemaphore;
  NewClause _freshClause;
  float _startMemory;
  ClauseNumIndex _clauseNumIndex; // temporary, must be part of a higher level module
  ulong (*_nextClauseNum)(); 
  long _nextTick;
  ulong _statClausesRefusedAllocation;
  ulong _statBackSubsumed;
  ulong _statBackSimplified;
  ulong _statRewrittenByDefUnfolding;   
  ulong _statPassiveOrphansMurdered;
  ulong _statActiveOrphansMurdered; 
  ulong _statRecycled;
  enum { _halfLongMax = (LONG_MAX/2) };
}; // class Master

#ifndef NO_DEBUG
 inline bool Master::clean()
 {
  CALL("clean()");
  if (!GlobalSubst.clean())
   {
    #ifdef DEBUG_NAMESPACE
     DMSG("Master::clean()> ") << "!GlobalSubst.clean()\n";
    #endif   
    return false;
   }; 
  if (!_passive.clean())
   {
    #ifdef DEBUG_NAMESPACE
     DMSG("Master::clean()> ") << "!_passive.clean()\n";
    #endif
    return false;
   }; 
  if (!_reservedPassive.clean())
   {
    #ifdef DEBUG_NAMESPACE
     DMSG("Master::clean()> ") << "!_reservedPassive.clean()\n";
    #endif
    return false;
   };  
  if (!_active.clean())
   {
    #ifdef DEBUG_NAMESPACE
     DMSG("Master::clean()> ") << "!_active.clean()\n";
    #endif
    return false;
   }; 
  if (!_history.clean())
   {
    #ifdef DEBUG_NAMESPACE
     DMSG("Master::clean()> ") << "!_history.clean()\n";
    #endif
    return false;
   };  
  if (!_definitions.clean())
   {
    #ifdef DEBUG_NAMESPACE
     DMSG("Master::clean()> ") << "!_definitions.clean()\n";
    #endif
    return false;
   };  
  
  if (_retainedNewClauses)
   {
    #ifdef DEBUG_NAMESPACE
     DMSG("Master::clean()> ") << "_retainedNewClauses nonempty\n";
    #endif
    return false;
   };  
  if (_queueForDisposal)
   {
    #ifdef DEBUG_NAMESPACE
     DMSG("Master::clean()> ") << "_queueForDisposal nonempty\n";
    #endif
    return false;
   };  
  if (_queueForBackSimplification)
   {
    #ifdef DEBUG_NAMESPACE
     DMSG("Master::clean()> ") << "_queueForBackSimplification nonempty\n";
    #endif
    return false;
   }; 
  /******  
  if (_queueForUnitConflict)
   {
    #ifdef DEBUG_NAMESPACE
     DMSG("Master::clean()> ") << "_queueForUnitConflict nonempty\n";
    #endif
    return false;
   }; 
  ********/
  if (!_infDispatcher.clean())        
   {
    #ifdef DEBUG_NAMESPACE
     DMSG("Master::clean()> ") << "!_infDispatcher.clean()\n";
    #endif
    return false;
   };  
  if (!_forwardSubsumption.clean())
   {
    #ifdef DEBUG_NAMESPACE
     DMSG("Master::clean()> ") << "!_forwardSubsumption.clean()\n";
    #endif
    return false;
   }; 
  if (!_backSubsumption.clean())
   {
    #ifdef DEBUG_NAMESPACE
     DMSG("Master::clean()> ") << "!_backSubsumption.clean()\n";
    #endif
    return false;
   }; 
  if (!_forwardDemodulation.clean())
   {
    #ifdef DEBUG_NAMESPACE
     DMSG("Master::clean()> ") << "!_forwardDemodulation.clean()\n";
    #endif
    return false;
   };  
  if (!_sharingIndex.clean())
   {
    #ifdef DEBUG_NAMESPACE
     DMSG("Master::clean()> ") << "!_sharingIndex.clean()\n";
    #endif
    return false;
   };  
  if (!_clauseIndex.clean())
   {
    #ifdef DEBUG_NAMESPACE
     DMSG("Master::clean()> ") << "!_clauseIndex.clean()\n";
    #endif
    return false;
   };  
  if (!_clauseNumIndex.clean())
   {
    #ifdef DEBUG_NAMESPACE
     DMSG("Master::clean()> ") << "!_clauseNumIndex.clean()\n";
    #endif
    return false;
   }; 
  return true; 
 }; // bool Master::clean()
#endif


inline void Master::Options::reset()
{
 incompleteAPriori = false;
 // main algorithm
 main_alg = 0;
 
 // inference rules
 binary_resolution = true;
 hyper_resolution = false;
 no_inferences_between_rules = false;
 splitting = false;  
 splitting_style = 0L;
 splitting_with_naming = false;
 neg_eq_splitting = 0L; 
 paramodulation = false;
 sim_back_sup = 0;
 hyper_positive = true;
 selection = 1;
 literal_comparison_mode = 1;
 inherited_negative_selection = false;

 // strategy switches
 handle_equality = false;
 forward_subsumption = true;
 fs_set_mode = false;
 fs_optimised_multiset_mode = false;
 fs_use_sig_filters = false;
 fs_old_style = false;
 simplify_by_fs = 0;
 forward_demodulation = 0;
 fd_on_splitting_branches = false;
 normalize_by_commutativity = (const Clause* const *)0;
 backward_demodulation = 0;
 bd_on_splitting_branches = false;
 simplify_by_eq_res = false;
 backward_subsumption = false;
 bs_set_mode = false;
 bs_optimised_multiset_mode = false;
 orphan_murder = 0;
 hyper_dinamic_lhs_ordering = true;
 unit_conflict = false;
 lrs = 0;
 pred_elim_table = (const bool*)0;
 defined_header_table = (const bool*)0;
  // numeric options
 age_weight_ratio = 0;
 timeLimit = (float)LONG_MAX;
 memoryLimit = (float)LONG_MAX;
 allocation_buffer_price = 1.0;
 maxActive = LONG_MAX;
 maxPassive = LONG_MAX;
 maxWeight = HellConst::MaxClauseWeight;
 optGrCompWeight = LONG_MAX;
 first_time_check = 1.0;
 first_mem_check = 1.0;
 internal_tick_frequency = LONG_MAX;
  // interface options
 showGen = false;
 showPassive = 0;
 showActive = 0;
 showHyperRule = 0;
 showDefinition = 0;
 showWeightChange = 0;
 showUsagePace = 0;
}; // void Master::Options::reset()

inline void Master::initParameters()
{
 CALL("Master::initParameters()");

 _parameters.incompleteAPriori = options.incompleteAPriori;
 
 // main algorithm

 _parameters.main_alg = options.main_alg;

 // inference rules
 _parameters.binary_resolution = options.binary_resolution;
 _parameters.hyper_resolution = options.hyper_resolution;
 _parameters.no_inferences_between_rules = options.no_inferences_between_rules;
 _parameters.splitting = options.splitting; 
 _parameters.splitting_style = options.splitting_style;
 _parameters.splitting_with_naming = options.splitting_with_naming;
 _parameters.neg_eq_splitting = options.neg_eq_splitting;  
 _parameters.paramodulation = options.paramodulation;
 _parameters.sim_back_sup = options.sim_back_sup;
 _parameters.hyper_positive = options.hyper_positive;
 _parameters.selection = options.selection;
 _parameters.literal_comparison_mode = options.literal_comparison_mode;
 _parameters.inherited_negative_selection = options.inherited_negative_selection;

 // strategy switches
 _parameters.handle_equality = options.handle_equality;
 _parameters.forward_subsumption = options.forward_subsumption;
 _parameters.fs_set_mode = options.fs_set_mode; 
 _parameters.fs_optimised_multiset_mode = options.fs_optimised_multiset_mode;
 _parameters.fs_use_sig_filters = options.fs_use_sig_filters; 
 _parameters.fs_old_style = options.fs_old_style;
 _parameters.simplify_by_fs = options.simplify_by_fs;
 _parameters.forward_demodulation = options.forward_demodulation;
 _parameters.fd_on_splitting_branches = options.fd_on_splitting_branches;
 _parameters.normalize_by_commutativity = options.normalize_by_commutativity;
 _parameters.backward_demodulation = options.backward_demodulation;
 _parameters.bd_on_splitting_branches = options.bd_on_splitting_branches;
 _parameters.simplify_by_eq_res = options.simplify_by_eq_res;
 _parameters.backward_subsumption = options.backward_subsumption;
 _parameters.bs_set_mode = options.bs_set_mode;
 _parameters.bs_optimised_multiset_mode = options.bs_optimised_multiset_mode;
 _parameters.orphan_murder = options.orphan_murder;
 _parameters.hyper_dinamic_lhs_ordering = options.hyper_dinamic_lhs_ordering;
 _parameters.unit_conflict = options.unit_conflict && options.hyper_resolution; 
                                                  // && options.hyper_resolution is temporary
 _parameters.lrs = options.lrs;
 _parameters.pred_elim_table = options.pred_elim_table;
 _parameters.defined_header_table = options.defined_header_table;
  // numeric parameters  
 _parameters.age_weight_ratio = options.age_weight_ratio;
 _parameters.timeLimit = options.timeLimit;
 _parameters.memoryLimit = options.memoryLimit;
 _parameters.allocation_buffer_price = options.allocation_buffer_price;
 _parameters.maxActive = options.maxActive;
 _parameters.maxPassive = options.maxPassive;
 _parameters.maxWeight = (options.maxWeight <= ((long)HellConst::MaxClauseWeight)) ? options.maxWeight : HellConst::MaxClauseWeight;
 _parameters.maxMaxWeight = _parameters.maxWeight;
 _parameters.optGrCompWeight = options.optGrCompWeight;
 _parameters.first_time_check = options.first_time_check; 
 _parameters.first_mem_check = options.first_mem_check;  
 _parameters.internal_tick_frequency = options.internal_tick_frequency;  
 _parameters.reachable = LONG_MAX;
 _parameters.unreachable = (-1)*_halfLongMax;     
  // interface parameters
 _parameters.showGen = options.showGen;
 _parameters.showPassive = options.showPassive;
 _parameters.showActive = options.showActive;
 _parameters.showHyperRule = options.showHyperRule;
 _parameters.showDefinition = options.showDefinition;
 _parameters.showWeightChange = options.showWeightChange;
 _parameters.showUsagePace = options.showUsagePace;
  // auxilliary
 _parameters.useClauseNumIndex = options.backward_demodulation || options.backward_subsumption;
}; // void Master::initParameters()

inline void Master::initObjects()
{
  // clause sets
 _passive.SetPickGivenRatio(_parameters.age_weight_ratio);
 _pagedPassive.setAgeWeightRatio(_parameters.age_weight_ratio);
    // _freshClause
 _freshClause.SetForwardSubsumption(_parameters.forward_subsumption);
 _freshClause.SetSimplificationByFS(_parameters.simplify_by_fs);
 _freshClause.SetUnitConflict(((_parameters.unit_conflict) ? (&_unitConflict) : ((UnitConflict*)0)));
 _freshClause.SetSimplificationByUEq(_parameters.forward_demodulation);
 _freshClause.SetFDOnSplittingBranchesFlag(_parameters.fd_on_splitting_branches);
 _freshClause.SetNormalizationByCommutativity(_parameters.normalize_by_commutativity);
 _freshClause.SetSimplificationByEqRes(_parameters.simplify_by_eq_res);  
 _freshClause.SetSplitting(_parameters.splitting);
 _freshClause.SetSplittingStyle(_parameters.splitting_style);
 _freshClause.SetSplittingWithNaming(_parameters.splitting_with_naming);
 _freshClause.SetNegEqSplitting(_parameters.neg_eq_splitting);      
 _freshClause.SetSelection(_parameters.selection);
 _freshClause.SetLiteralComparisonMode(_parameters.literal_comparison_mode);
 _freshClause.SetInheritedNegativeSelection(_parameters.inherited_negative_selection);
 _freshClause.SetEqualityPossible(_parameters.handle_equality);
 _freshClause.SetMaxWeight(_parameters.maxWeight);
 _freshClause.SetPredElimTable(_parameters.pred_elim_table);
 _freshClause.SetDefHeaderTable(_parameters.defined_header_table);
 _freshClause.SetOptimalGroundCompWeight((ulong)_parameters.optGrCompWeight);

  // _forwardSubsumption      
 _forwardSubsumption.SetSetMode(_parameters.fs_set_mode);
 _forwardSubsumption.SetOptimisedMultisetMode(_parameters.fs_optimised_multiset_mode);
 _forwardSubsumption.SetUseSigFilters(_parameters.fs_use_sig_filters);
 _forwardSubsumption.SetOldStyle(_parameters.fs_old_style);

  // _backSubsumption

 _backSubsumption.setSetMode(_parameters.bs_set_mode);
 _backSubsumption.setOptimisedMultisetMode(_parameters.bs_optimised_multiset_mode);

  // forward demodulation
  
 _forwardDemodulation.SetDegree(_parameters.forward_demodulation);
 _forwardDemodulation.SetFDOnSplittingBranchesFlag(_parameters.fd_on_splitting_branches);

  // _backDemodulation
 _backDemodulation.setDegree(_parameters.backward_demodulation);
 _backDemodulation.setBDOnSplittingBranchesFlag(_parameters.bd_on_splitting_branches);

  // _infDispatcher    
 _infDispatcher.binResFlag = _parameters.binary_resolution;
 _infDispatcher.hyperResFlag = _parameters.hyper_resolution;
 _infDispatcher.setNoInfBetweenRulesFlag(_parameters.no_inferences_between_rules);
 _infDispatcher.setParamodulationFlag(_parameters.paramodulation);
 _infDispatcher.setSimBackSup(_parameters.sim_back_sup);
 _infDispatcher.setUnitConflict(_parameters.unit_conflict);
 _infDispatcher.orientHyperresolution(_parameters.hyper_positive);
}; // void Master::initObjects()

inline float Master::usedMemory()
{ 
 float res = ((float)GlobAlloc::occupiedByObjects()) - _startMemory;
 if (res <= 0) return 0.0001;
 return res;
}; // float Master::usedMemory() 

inline void Master::anotherInference0()
{
 CALL("anotherInference0()");
 _nextTick--;
 if (!_nextTick) 
  { 
   _nextTick = _parameters.internal_tick_frequency; 
   internalTickEvent0();  
  };
}; // void Master::anotherInference0() 

inline void Master::anotherInference1()
{
 CALL("anotherInference1()");
 _nextTick--;
 if (!_nextTick) 
  { 
   _nextTick = _parameters.internal_tick_frequency; 
   internalTickEvent1();  
  };
}; // void Master::anotherInference1() 

inline void Master::anotherInference2()
{
 CALL("anotherInference2()");
 _nextTick--;
 if (!_nextTick) 
  { 
   _nextTick = _parameters.internal_tick_frequency; 
   internalTickEvent2();  
  }; 
}; // void Master::anotherInference2() 

inline void Master::anotherInference3()
{
 CALL("anotherInference3()");
 _nextTick--;
 if (!_nextTick) 
  { 
   _nextTick = _parameters.internal_tick_frequency; 
   internalTickEvent3();  
  }; 
}; // void Master::anotherInference3() 

inline void Master::internalTickEvent0()
{
 CALL("internalTickEvent0()");
 estimateResources();
 if (timer.elapsed() >= _parameters.timeLimit) { terminate(TimeLimit); return; };
}; // void Master::internalTickEvent0() 

inline void Master::internalTickEvent1()
{
 CALL("internalTickEvent1()");
 estimateResources();
 if (timer.elapsed() >= _parameters.timeLimit) { terminate(TimeLimit); return; };
 if (GlobAlloc::occupiedByObjects() >= _parameters.memoryLimit) { terminate(MemoryLimit); return; }; 
}; // void Master::internalTickEvent1() 

inline void Master::internalTickEvent2()
{
 CALL("internalTickEvent2()");
 estimateResources();
 if (timer.elapsed() >= _parameters.timeLimit) { terminate(TimeLimit); return; };
 // if (GlobAlloc::occupiedByObjects() >= _parameters.memoryLimit) { terminate(MemoryLimit); return; }; 
}; // void Master::internalTickEvent2()

inline void Master::internalTickEvent3()
{
 CALL("internalTickEvent3()");
 estimateResources();
 if (timer.elapsed() >= _parameters.timeLimit) { terminate(TimeLimit); return; };
 // if (GlobAlloc::occupiedByObjects() >= _parameters.memoryLimit) { terminate(MemoryLimit); return; }; 
}; // void Master::internalTickEvent3()


inline void Master::eliminateAllocationBufferDeficit0()
{
 CALL("eliminateAllocationBufferDeficit0()");
 if (GlobAlloc::activeBufferDeficit())
  {
   long disposalBound = (long)(GlobAlloc::occupiedByObjects() - (_parameters.allocation_buffer_price * GlobAlloc::activeBufferDeficit()));
   do
    {
     Clause* cl = _passive.TheWorstUnblockedClause();    
     if (!cl) return;  
     if (cl->flag(Clause::FlagInQueueForBackSimplification)) 
      {
       _queueForBackSimplification.removeFirst(cl); 
       cl->clearFlag(Clause::FlagInQueueForBackSimplification);
      };

     ASSERT(cl->mainSet() == Clause::MainSetPassive);
     _passive.Cancel(cl);
     cl->setMainSet(Clause::MainSetFree);
 
     //DF; TRACE(false,false);     
  
     enqueueForDisposal(cl);

     //DF; TRACINGOFF;

     _statRecycled++;
    }
   while (GlobAlloc::activeBufferDeficit() && (GlobAlloc::occupiedByObjects() > disposalBound));
   if (GlobAlloc::activeBufferDeficit()) GlobAlloc::annulBufferDeficit();
  };
}; // void Master::eliminateAllocationBufferDeficit0() 


inline void Master::eliminateAllocationBufferDeficit1()
{
 CALL("eliminateAllocationBufferDeficit1()");
 if (GlobAlloc::activeBufferDeficit())
  {
   long disposalBound = (long)(GlobAlloc::occupiedByObjects() - (_parameters.allocation_buffer_price * GlobAlloc::activeBufferDeficit()));
   do
    {
     Clause* cl = _passive.TheWorstUnblockedClause();    
     if (!cl) return;  
     if (cl->flag(Clause::FlagInQueueForBackSimplification)) 
      {
       _queueForBackSimplification.removeFirst(cl); 
       cl->clearFlag(Clause::FlagInQueueForBackSimplification);
      };

     ASSERT(cl->mainSet() == Clause::MainSetPassive);
     _passive.Cancel(cl);
     cl->setMainSet(Clause::MainSetFree);
   
     disposeClause(cl);
     _statRecycled++;
    }
   while (GlobAlloc::activeBufferDeficit() && (GlobAlloc::occupiedByObjects() > disposalBound));
   if (GlobAlloc::activeBufferDeficit()) GlobAlloc::annulBufferDeficit();    
  };
}; // void Master::eliminateAllocationBufferDeficit1()  

inline void Master::eliminateAllocationBufferDeficit2()
{
 CALL("eliminateAllocationBufferDeficit2()");
 if (GlobAlloc::activeBufferDeficit())
  {
   long disposalBound = (long)(GlobAlloc::occupiedByObjects() - (_parameters.allocation_buffer_price * GlobAlloc::activeBufferDeficit()));
   do
    {
     Clause* cl = _passive.TheWorstUnblockedClause();    
     if (!cl) return;  
     if (cl->flag(Clause::FlagInQueueForBackSimplification)) 
      {
       _queueForBackSimplification.removeFirst(cl); 
       cl->clearFlag(Clause::FlagInQueueForBackSimplification);
      };

     ASSERT(cl->mainSet() == Clause::MainSetPassive);
     _passive.Cancel(cl);
     cl->setMainSet(Clause::MainSetFree);
   
     disposeClause(cl);
     _statRecycled++;
    }
   while (GlobAlloc::activeBufferDeficit() && (GlobAlloc::occupiedByObjects() > disposalBound));
   if (GlobAlloc::activeBufferDeficit()) GlobAlloc::annulBufferDeficit();    
  };

}; // void Master::eliminateAllocationBufferDeficit2()  

inline void Master::eliminateAllocationBufferDeficit3()
{
 CALL("eliminateAllocationBufferDeficit3()");
 if (GlobAlloc::activeBufferDeficit())
  {
   long disposalBound = (long)(GlobAlloc::occupiedByObjects() - (_parameters.allocation_buffer_price * GlobAlloc::activeBufferDeficit()));
   do
    {
     Clause* cl = _passive.TheWorstUnblockedClause();    
     if (!cl) return;  
     if (cl->flag(Clause::FlagInQueueForBackSimplification)) 
      {
       _queueForBackSimplification.removeFirst(cl); 
       cl->clearFlag(Clause::FlagInQueueForBackSimplification);
      };

     ASSERT(cl->mainSet() == Clause::MainSetPassive);
     _passive.Cancel(cl);
     cl->setMainSet(Clause::MainSetFree);
   
     disposeClause(cl);
     _statRecycled++;
    }
   while (GlobAlloc::activeBufferDeficit() && (GlobAlloc::occupiedByObjects() > disposalBound));
   if (GlobAlloc::activeBufferDeficit()) GlobAlloc::annulBufferDeficit();    
  };

}; // void Master::eliminateAllocationBufferDeficit3()  
//====================================================
#undef DEBUG_NAMESPACE
//====================================================
#endif
