//
// File:         ProblemProfile.cpp
// Description:  Analysing simple problem characteristics.
// Created:      Apr 15, 2000, 19:00
// Author:       Alexandre Riazanov                        
// mail:         riazanov@cs.man.ac.uk
//====================================================
#include "ProblemProfile.h"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PROBLEM_PROFILE
 #define DEBUG_NAMESPACE "ProblemProfile"
#endif
#include "debugMacros.h"
//=================================================

void ProblemProfile::reset()
{
 CALL("reset()");
 _clauses.reset();
 
 _numOfClauses = 0;
 _numOfNonhorn = 0;
 _numOfUnit = 0;
 _numOfLiterals = 0;
 _numOfEqLiterals = 0;
 ulong p;
 for (p = 0; p < HellConst::MaxNumOfFunctors; p++) _predInfoTable[p].init();

 _numOfPred = 0;
 _symmetricPred = 0;
 ulong f;
 for (f = 0; f < HellConst::MaxNumOfFunctors; f++) _symmetryTable[f] = (const Clause*)0;
 
 _totalPred = 0;
 _eliminatedPred = 0;
 for (p = 0; p < HellConst::MaxNumOfFunctors; p++) _eliminationTable[p] = false; 
 
 for (f = 0; f < HellConst::MaxNumOfFunctors; f++) _funcInfoTable[f].init();
 _numOfFuncs = 0;
 _commutativeFunc = 0;
 for (f = 0; f < HellConst::MaxNumOfFunctors; f++) _commutativityTable[f] = (const Clause*)0;

}; // void ProblemProfile::reset()
 
void ProblemProfile::load(const Clause* cl) 
{
 CALL("load(const Clause* cl)"); 
 _clauses.enqueue((Clause*)cl);
 
  // general statistics
 _numOfClauses++;
 if (!cl->HornClause()) _numOfNonhorn++;
 if (cl->Unit()) _numOfUnit++;
 _numOfLiterals += cl->Length();
 _numOfEqLiterals += cl->NumOfEqLits();   

  // collecting information on predicates 

 if (cl->SymmetryLaw())
  {
   _predInfoTable[cl->LitList()->firstLit()->functor()]._symmetric = true;
   _symmetryTable[cl->LitList()->firstLit()->functor()] = cl;
  }
 else
  if (cl->TotalityLaw())
   _predInfoTable[cl->LitList()->firstLit()->functor()]._total = true;
   
 for(LiteralList::Iterator iter1(cl->LitList());iter1.notEnd();iter1.next())
  {
   const TERM* lit1 = iter1.currentLiteral();
   ulong pred1 = lit1->functor(); 
   _predInfoTable[pred1].regPred(*lit1);
   if (_predInfoTable[pred1]._status == PredInfo::Unknown)
    {
     if (lit1->arity() && (!lit1->IsEquality()))    
      {
       if (lit1->IsDefHeader())
	{
	 if (!_predInfoTable[pred1].isNondef(*lit1))
	  {
	   // might be header of a definition
    
	   for(LiteralList::Iterator iter2(cl->LitList());iter2.notEnd();iter2.next())
	    if ((iter1.currentElement() != iter2.currentElement()) && (pred1 == iter2.currentLiteral()->functor())) 
	     {
	      _predInfoTable[pred1].nondef(*lit1);
              _predInfoTable[pred1].nondef(*(iter2.currentLiteral()));
             };            
          }; 
        }
       else
	_predInfoTable[pred1].nondef(*lit1);         
      }
     else // propositional or equality
      _predInfoTable[pred1]._status = PredInfo::NotEliminated;        
    };
  };               

 // collecting information on functors
 ulong commFunc;
 if (cl->CommutativityLaw(commFunc))
  {
   _funcInfoTable[commFunc]._commutative = true;
   _commutativityTable[commFunc] = cl;
  };
 TERM::Iterator iter;
 for(LiteralList::Iterator litIter(cl->LitList());litIter.notEnd();litIter.next())
  {
   if (litIter.currentLiteral()->arity())
    {
     iter.Reset(litIter.currentLiteral());
     ALWAYS(iter.Next());
     do
      {     
       if (!iter.CurrentSymIsVar()) 
        {
         _funcInfoTable[iter.symbolRef().functor()]._registered = true; 
         _funcInfoTable[iter.symbolRef().functor()]._symbol = iter.symbolRef();
        };
      }
     while (iter.Next());
    };
  };
}; // void ProblemProfile::load(const Clause* cl) 

void ProblemProfile::summarise()
{
 CALL("summarise()");
 for (const ClauseQueue::Entry* qe = _clauses.begin(); qe; qe = qe->next())
  markEliminated(qe->value());

 for (ulong p = 0; p < HellConst::MaxNumOfFunctors; p++) 
  {                
   if ((_predInfoTable[p]._registered) && (!_predInfoTable[p]._symbol.IsEquality()))
    {
     _numOfPred++;
     if ((_eliminationTable[p] = (_predInfoTable[p]._status == PredInfo::Eliminated)))             
      _eliminatedPred++;
     if (_predInfoTable[p]._symmetric) _symmetricPred++;
     if (_predInfoTable[p]._total) _totalPred++;
    };
  };  

 for (ulong f = 0; f < HellConst::MaxNumOfFunctors; f++) 
  { 
   if (_funcInfoTable[f]._registered) 
    {
     _numOfFuncs++;
     if (_funcInfoTable[f]._commutative) _commutativeFunc++;
    };
  };

}; // void ProblemProfile::summarise()
   
void ProblemProfile::markEliminated(const Clause* cl)
{
 CALL("markEliminated(const Clause* cl)");

 for(LiteralList::Iterator iter(cl->LitList());iter.notEnd();iter.next())
  {
   const TERM* lit = iter.currentLiteral(); 
   if (_predInfoTable[lit->functor()]._status == PredInfo::Eliminated)
    {
     for(LiteralList::Iterator iter1(cl->LitList());iter1.notEnd();iter1.next())
      if (iter1.currentElement() != iter.currentElement())
       {
	ASSERT(iter1.currentLiteral()->functor() != lit->functor());
        ASSERT(_predInfoTable[iter1.currentLiteral()->functor()]._status != PredInfo::Eliminated);
        _predInfoTable[iter1.currentLiteral()->functor()]._status = PredInfo::NotEliminated;            
       };
     return;
    };
  };

 bool isDefinition = false;

 for(LiteralList::Iterator iter(cl->LitList());iter.notEnd();iter.next())
  {
   const TERM* lit = iter.currentLiteral(); 
   if (isDefinition)
    {
     ASSERT(_predInfoTable[lit->functor()]._status != PredInfo::Eliminated);
     _predInfoTable[lit->functor()]._status = PredInfo::NotEliminated;                  
    }
   else
    if (_predInfoTable[lit->functor()]._status == PredInfo::Unknown)
     {        
      _predInfoTable[lit->functor()]._status = PredInfo::Eliminated; 
      isDefinition = true;
     };      
  };
}; // void ProblemProfile::markEliminated(const Clause* cl)


//=================================================
