//
// File:         ProblemProfile.h
// Description:  Analysing simple problem characteristics.
// Created:      Apr 15, 2000, 19:00
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#ifndef PROBLEM_PROFILE_H
#include <iostream.h>      
#include "jargon.h"
#include "ClauseQueue.h"
#include "clause.h"
#include "term.h"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PROBLEM_PROFILE
 #define DEBUG_NAMESPACE "ProblemProfile"
#endif
#include "debugMacros.h"
//=================================================
class ProblemProfile
{
 public:
  enum Category { NNE, HNE, NEQ, HEQ, PEQ, UEQ };
  class PredInfo
  {
   public: 
    enum Status { Unknown, Eliminated, NotEliminated }; 
   public:
    PredInfo() : _symbol() 
    {
     CALL("constructor PredInfo()");
    };
    void init() 
    { 
     _status = Unknown;
     _registered = false;
     _positiveNondef = false;
     _negativeNondef = false; 
     _symmetric = false;
     _total = false;
    };
    const TERM& symbol() const { return _symbol; };
    const Status status() const { return _status; };
    bool registered() const { return _registered; };
    bool statusEstablished() const { return _status != Unknown; };
    bool statusEliminated() const { return _status == Eliminated; };  
    bool symmetric() const { return _symmetric; };
    bool total() const { return _total; }; 
    static const char* spell(const Status& s)
    {
     switch (s) 
      { 
       case Unknown: return "Unknown"; 
       case Eliminated: return "Eliminated";
       case NotEliminated: return "NotEliminated"; 
      };
     return "WrongValue";  
    };
    ostream& output(ostream& str) const 
    {
     str << _symbol << "/" << _symbol.arity() << " : " << spell(_status);
     if (_symmetric) str << " sym";
     if (_total) str << " tot";
     return str;
    };  
 
   private:
    void regPred(const TERM& sym) 
    {
     _symbol = sym;
     _symbol.makePositive();
     _registered = true;
    };  
 
    void positiveNondef()
    {
     _positiveNondef = true;
     if (_negativeNondef) _status = NotEliminated;
    }; 
    void negativeNondef()
    {
     _negativeNondef = true;
     if (_positiveNondef) _status = NotEliminated;
    }; 
    void nondef(const TERM& hd)
    {
     if (hd.Positive()) { positiveNondef(); } else negativeNondef();
    };

    bool isNondef(const TERM& hd) const
    {
     return ((hd.Positive()) ? _positiveNondef : _negativeNondef);
    };
 
   private:
    TERM _symbol;
    Status _status;
    bool _registered;  
    bool _positiveNondef;
    bool _negativeNondef;  
    bool _symmetric;
    bool _total;
   friend class ProblemProfile;
  }; // class ProblemProfile::PredInfo

  class FuncInfo
  {
   public:
    FuncInfo() : _symbol() 
    {
     CALL("constructor FuncInfo()");
    };
    void init() 
    { 
     _registered = false;
     _commutative = false;
    };
    const TERM& symbol() const { return _symbol; };
    bool registered() const { return _registered; };
    bool commutative() const { return _commutative; };
    ostream& output(ostream& str) const 
    {
     str << _symbol << "/" << _symbol.arity();
     if (_commutative) str << " : COM";
     return str;
    }; 
   private:
    TERM _symbol;
    bool _registered;
    bool _commutative;
   friend class ProblemProfile; 
  }; // class FuncInfo


 public:

  ProblemProfile() 
   : _clauses() 
  {
    CALL("constructor ProblemProfile()");
   reset(); 
  };
  ~ProblemProfile() { CALL("destructor ~ProblemProfile()"); };
  void reset();
  void load(const Clause* cl);
  void load(const ClauseQueue& clauses)
  {
   CALL("load(const ClauseQueue& clauses)");
   for (const ClauseQueue::Entry* qe = clauses.begin(); qe; qe = qe->next())
    load(qe->value());
   summarise();
  };                     
      
  void summarise();                        
                     
  Category category() const 
  {
   if (_numOfEqLiterals)
    if (_numOfEqLiterals == _numOfLiterals) // only equality
     {
      if (_numOfUnit == _numOfClauses) return UEQ;    
      return PEQ;
     }
    else return (_numOfNonhorn) ? NEQ : HEQ;
   // no equality
   return (_numOfNonhorn) ? NNE : HNE;
  };   

  const PredInfo* predInfoTable() const { return _predInfoTable; };
  const bool* eliminationTable() const { return _eliminationTable; };
  const Clause* const * commutativityTable() const 
  {
   return (_commutativeFunc) ? _commutativityTable : ((const Clause* const *)0);    
  };
  const Clause* const * symmetryTable() const 
  {
   return (_symmetricPred) ? _symmetryTable  : ((const Clause* const *)0);   
  };  
  static const char* spell(Category cat)
  {
   switch (cat)
    {
     case NNE: return "nne"; 
     case HNE: return "hne"; 
     case NEQ: return "neq"; 
     case HEQ: return "heq"; 
     case PEQ: return "peq"; 
     case UEQ: return "ueq";    
    };
   return "unk";
  };

  ulong numOfClauses() const { return _numOfClauses; };
  ulong numOfNonhorn() const { return _numOfNonhorn; };
  ulong numOfUnit() const { return _numOfUnit; };
  ulong numOfLiterals() const { return _numOfLiterals; };
  ulong numOfEqLiterals() const { return _numOfEqLiterals; }; 
  ulong numOfEliminated() const { return _eliminatedPred; };
  ulong numOfCommutative() const { return _commutativeFunc; };
 
  ostream& output(ostream& str) const;

 private: 
   
  void markEliminated(const Clause* cl);

 private:
  ClauseQueue _clauses;
  ulong _numOfClauses;
  ulong _numOfNonhorn;
  ulong _numOfUnit;
  ulong _numOfLiterals;
  ulong _numOfEqLiterals;
  PredInfo _predInfoTable[HellConst::MaxNumOfFunctors];   
  const Clause* _symmetryTable[HellConst::MaxNumOfFunctors];
  ulong _numOfPred;
  ulong _symmetricPred;                       
  ulong _totalPred;
  ulong _eliminatedPred;
  bool _eliminationTable[HellConst::MaxNumOfFunctors];

  
  FuncInfo _funcInfoTable[HellConst::MaxNumOfFunctors];  
  ulong _numOfFuncs;
  ulong _commutativeFunc;
  const Clause* _commutativityTable[HellConst::MaxNumOfFunctors];   
   
}; // class ProblemProfile


inline ostream& ProblemProfile::output(ostream& str) const
{
 str << "% Category: " << spell(category()) << '\n';
 str << "% Number of non-equality predicates: " << _numOfPred << '\n';
 if (_symmetricPred) str << "% Number of symmetric predicates: " << _symmetricPred  << '\n'; 
 if (_totalPred) str << "% Number of total predicates: " << _totalPred  << '\n';  
 if (_eliminatedPred) str << "% Number of eliminated predicates: " << _eliminatedPred  << '\n';        
 str << "% Predicates: {";           
 for (ulong p = 0; p < HellConst::MaxNumOfFunctors; p++)
  if ((_predInfoTable[p]._registered) && (!_predInfoTable[p]._symbol.IsEquality()))
   _predInfoTable[p].output(str << '[') << "] ";   
 str << "}\n"; 

 str << "% Number of functors: " << _numOfFuncs << '\n';
 if (_commutativeFunc) str << "% Number of commutative functors: " << _commutativeFunc << '\n';

 str << "% Functors: {";
 for (ulong f = 0; f < HellConst::MaxNumOfFunctors; f++)
  if (_funcInfoTable[f]._registered) _funcInfoTable[f].output(str << '[') << "] ";
 str << "}\n"; 
 return str;
};


//================================================
#undef DEBUG_NAMESPACE
//====================================================
#endif
                      
