//
// File:         Vampire.h
// Description:  Main module of Vampire.
// Created:      Mar 6, 2000, 14:40
// Author:       Alexandre Riazanov 
// mail:         riazanov@cs.man.ac.uk
//====================================================
#include "Vampire.h"
#if ((defined INTERCEPT_FORWARD_MATCHING_INDEXING) || (defined INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING))
 #include "Logging.h"
 #include "HellInterceptMacros.h"
#endif
//====================================================
Vampire::Options Vampire::options;
Master::Options Vampire::_manualStrat;              
Master::Options Vampire::_autoNNE;
Master::Options Vampire::_autoHNE1;
Master::Options Vampire::_autoHNE2;
Master::Options Vampire::_autoNEQ;
Master::Options Vampire::_autoHEQ;
Master::Options Vampire::_autoPEQ;
Master::Options Vampire::_autoUEQ;

Timer Vampire::_timer;
ulong Vampire::_nextClauseNum = 1;         
Master Vampire::_master(nextClauseNum); 
ClauseQueue Vampire::_inputClauses;
ClauseQueue Vampire::_preprocessedClauses;
ClauseQueue Vampire::_preprocessingHistory;
InferenceTree Vampire::_proofTree;
ClauseNumIndex Vampire::_clauseNumIndex;
ProblemProfile Vampire::_problemProfile;
bool Vampire::_definedHeader[HellConst::MaxNumOfDifferentHeaders];
ulong Vampire::_numOfDefinedHeaders = 0UL; 
bool Vampire::_solvedByPreprocessing = false;

void Vampire::initStatic()
{
 CALL("initStatic()");
 for (ulong i = 0; i < HellConst::MaxNumOfDifferentHeaders; i++) _definedHeader[i] = false;
 _numOfDefinedHeaders = 0UL;
}; // void Vampire::initStatic()

void Vampire::initPredefinedStrategies()
{
 manualStrat(_manualStrat);
 autoNNE(_autoNNE);
 autoHNE1(_autoHNE1);
 autoHNE2(_autoHNE2);
 autoNEQ(_autoNEQ);
 autoHEQ(_autoHEQ);
 autoPEQ(_autoPEQ);
 autoUEQ(_autoUEQ);    
}; // void Vampire::initPredefinedStrategies()

bool Vampire::inputClause(Clause* cl)
{
 CALL("inputClause(Clause* cl)");

 _inputClauses.enqueue(cl);
 cl->SetNumber(nextClauseNum());
 if (cl->Empty())
  {
   cout << "WARNING: EMPTY INPUT CLAUSE DETECTED!\n";
   _master.proof = cl;   
   _master.terminate(Master::Proof);
   reportTermination();
   return false;    
  };
 if (_master.terminated) { reportTermination(); return false; };
 return true;  
}; // bool Vampire::inputClause(const Clause* cl)

void Vampire::definedHeader(const TERM& hd)
{ 
 _definedHeader[hd.HeaderNum()] = true;
 _numOfDefinedHeaders++;
}; // void Vampire::definedHeader(const TERM& hd)


void Vampire::prove()
{
 CALL("prove()");

 _timer.reset();
 _timer.start();
 //initPredefinedStrategies();
 if (options.showOpt) options.output(cout);   
 if (options.showInput)
  {
   cout << "%================== Input clauses: ===================\n";
   output(_inputClauses); 
   cout << "%============== End of input clauses: ================\n";
  };
 Clause::setSelectedPosEqPenaltyCoefficient(options.selected_pos_eq_penalty_coefficient);
 _problemProfile.reset();
 COP("loading input into _problemProfile");
 _problemProfile.load(_inputClauses);
 if (options.showProfile)
  {
   cout << "%========= Profile of input:\n"; 
   _problemProfile.output(cout);
   cout << "%========= End of profile. ========\n"; 
  };
 
 if (options.autoStrat) 
  { 
   // options.header_rank_kinky is overriden 
   if (_problemProfile.numOfEqLiterals()) { options.header_rank_kinky = false; }
   else options.header_rank_kinky = true;  
  };

 setSignatureProperties();
 assertHeaderRanks(); 
 if (options.tabulation) 
  {
   options.tabulation->stat("problem_category",ProblemProfile::spell(_problemProfile.category()));
   GlobalHeaderWeighting.Tabulate(options.tabulation);
  };
 COP("PREPRO");
 _solvedByPreprocessing = true;
 preprocess(); 
 COP("AFTER PREPRO");
 if (options.tabulation)
  {
   tabulatePreprocessingStatistics();
   options.tabulation->stat("preprocessed_problem_category",ProblemProfile::spell(_problemProfile.category()));
  };

 if (_master.terminated) { reportTermination(); return; };

 _solvedByPreprocessing = false;

 TheSignature.setCommutativity(_problemProfile.commutativityTable(),_problemProfile.symmetryTable());
 TheSignature.makeCommutative(HellConst::UnordEqNum);
 TheSignature.makeCommutative(HellConst::OrdEqNum);

 //DF; TheSignature.output(cout) << "\n\n\n"; DF;

 #ifdef INTERCEPT_FORWARD_MATCHING_INDEXING
  FMILog.unblock();
  FMILog.outputSignature(TheSignature);
  FMILog.output("$\n");
 #endif

 #ifdef INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING
  BSILog.unblock();
  BSILog.outputSignature1(TheSignature);
  BSILog.output("$\n");
 #endif

 if (!_problemProfile.numOfEqLiterals())
  {
   cancelEquality();
  };

 initPredefinedStrategies();
 if (options.autoStrat) 
  {
   COP("RUN AUTO");
   runAuto(); 
  }
 else // manual strategy selection
  { 
   COP("MANUAL");
   search(availableTime(),_manualStrat); 
  };
 reportTermination();
}; // void Vampire::prove()


void Vampire::preproOnly() 
{
 CALL("preproOnly()");
 cout << "%============ Preprocessing-only mode: start preprocessing =================\n";
 _timer.reset();
 _timer.start();
 if (options.showOpt) options.output(cout);   
 if (options.showInput)
  {
   cout << "%================== Input clauses: ===================\n";
   output(_inputClauses); 
   cout << "%============== End of input clauses: ================\n";
  }; 
 _problemProfile.reset();
 _problemProfile.load(_inputClauses);
 if (options.showProfile)
  {
   cout << "%========= Profile of input:\n"; 
   _problemProfile.output(cout);
   cout << "%========= End of profile. ========\n"; 
  };
 setSignatureProperties();
 assertHeaderRanks();
 if (options.tabulation) 
  {
   options.tabulation->stat("problem_category",ProblemProfile::spell(_problemProfile.category()));
   GlobalHeaderWeighting.Tabulate(options.tabulation);
  };

 COP("PREPRO");
 preprocess(); 
 COP("AFTER PREPRO");
 if (options.tabulation)
  options.tabulation->stat("preprocessed_problem_category",ProblemProfile::spell(_problemProfile.category()));
 if (_master.terminated) { reportTermination(); };

 cout << "%============ Preprocessed clauses in the TPTP format: =====================\n";

 outputInTPTPFormat(_preprocessedClauses);
  
 cout << "%============ Preprocessing-only mode: preprocessing done  =================\n";
}; //void Vampire::preproOnly() 


void Vampire::runAuto()
{
 CALL("runAuto()"); 
 switch (_problemProfile.category())
  {
   case ProblemProfile::NNE:
    search(availableTime(),_autoNNE); break; 

     //case ProblemProfile::HNE: search(availableTime(),_autoNNE); break; // temporary

     /**************/
   case ProblemProfile::HNE: 
     // first attempt
    cout << "%================== Trying negative hyperresolution...\n"; 

    //DF; TRACE(true,false); 

    #ifdef NO_DEBUG
     search(availableTime() * 0.1,_autoHNE1);
    #else
     search(availableTime() * 0.3,_autoHNE1);
    #endif
    switch (_master.terminationReason)
     {
      case Master::UnknownTR: ICP("Unknown TR"); return;
      case Master::Proof: return;
      case Master::EmptyPassive: 
       if (_master.completeStrategyUsed()) return;
       break;
      case Master::TimeLimit: break;
      case Master::MemoryLimit: break;
      case Master::AllocationFailure: break;
      case Master::MaxPassive: break; 
      case Master::MaxActive: break;
      case Master::UserInt: return;
     };   
      

    // another attempt 
    cout << "%================== Negative resolution failed.\n";
    cout << "%================== Trying positive hyperresolution...\n"; 
    search(availableTime(),_autoHNE2); 
    break; 
 
   /*******************/  
 
   case ProblemProfile::NEQ: search(availableTime(),_autoNEQ); break;
   case ProblemProfile::HEQ: search(availableTime(),_autoHEQ); break;
   case ProblemProfile::PEQ: search(availableTime(),_autoPEQ); break;
   case ProblemProfile::UEQ: search(availableTime(),_autoUEQ); break;
  }; 
}; // void Vampire::runAuto()

ulong Vampire::nextClauseNum()
{
 _nextClauseNum++;
 return _nextClauseNum - 1;
}; // ulong Vampire::nextClauseNum()

void Vampire::preprocess()
{
 CALL("preprocess()");
 ulong iterNum = 0;
 long eliminationCycles = 0;
 for (const ClauseQueue::Entry* qe = _inputClauses.begin(); qe; qe = qe->next())
  _preprocessedClauses.enqueue(qe->value());

 simplify:
  preproSimplify(iterNum);
  _problemProfile.reset();
  _problemProfile.load(_preprocessedClauses);
  if (options.showProfile)
   {
     cout << "%========= Profile of preprocessed (" << iterNum << "):\n"; 
    _problemProfile.output(cout);
    cout << "%========= End of profile for preprocessed (" << iterNum << ") ========\n";
   }; 
  if (_master.terminated) return;
  if (_problemProfile.numOfEliminated() && (eliminationCycles < options.elim_def))
   { 
    do 
     {
      iterNum++;
      eliminationCycles++;
      preproEliminateDef(iterNum);
      _problemProfile.reset();
      _problemProfile.load(_preprocessedClauses);
      if (options.showProfile)
       {
        cout << "%========= Profile of preprocessed (" << iterNum << "):\n"; 
        _problemProfile.output(cout);
        cout << "%========= End of profile for preprocessed (" << iterNum << ") ========\n";
       }; 
      if (_master.terminated) return;
     }
    while (_problemProfile.numOfEliminated() && (eliminationCycles < options.elim_def));
    goto simplify;
   };
  // end of story
  cout << "%======== Preprocessing has been done in " << iterNum + 1 << " stages.\n";
  
}; // void Vampire::preprocess()


void Vampire::preproSimplify(ulong iterNum)
{

  //DF; TRACE(true,true);

 CALL("preproSimplify(ulong iterNum)");
 cout << "%=============== Preprocessing by simplification " << iterNum << " ============\n";

 prepareMasterForPreproSimplification();  

 load(_preprocessedClauses); 

 //if (_master.terminated) return;


 preprocessedToHistory();
 extractPreprocessed(); 

 if (options.showPrepro) 
  {
   cout << "%=============== Preprocessing  by simplification " << iterNum << " statistics: ============\n";
   outputStatistics();
   cout << "%=============== Clauses preprocessed by simplification " << iterNum << " ================\n";
   output(_preprocessedClauses);
   cout << "%============== End of clauses preprocessed by " << iterNum << "  ==========\n";
  };
 if (_master.terminated) return;
}; // void Vampire::preproSimplify(ulong iterNum)



void Vampire::preproEliminateDef(ulong iterNum)
{
 CALL("preproEliminateDef(ulong iterNum)");
 cout << "%=============== Preprocessing by def. elimination " << iterNum << " ============\n";
 prepareMasterForPreproDefElimination();  
 load(_preprocessedClauses); 
 if (_master.terminated) return;
 preprocessedToHistory();
 _master.eliminateDefinitions();
 //if (_master.terminated) return;
 extractPreprocessed(); 

 if (options.showPrepro) 
  {
   cout << "%=============== Preprocessing  by def. elimination " << iterNum << " statistics: ============\n";
   outputStatistics();
   cout << "%=============== Clauses preprocessed by def. elimination " << iterNum << " ================\n";
   output(_preprocessedClauses);
   cout << "%============== End of clauses preprocessed by " << iterNum << " ==========\n";
  };
 if (_master.terminated) return;
}; // void Vampire::preproEliminateDef(ulong iterNum)


void Vampire::prepareMasterForPreproSimplification()
{
 CALL("prepareMasterForPreproSimplification()");
 _master.reset();
 _master.options.incompleteAPriori = options.incompleteAPriori;
 _master.options.main_alg = 0;
 _master.options.binary_resolution = false; 
 _master.options.hyper_resolution = false;
 _master.options.no_inferences_between_rules = false;
 _master.options.splitting = options.static_splitting;
 _master.options.splitting_style = options.splitting_style;
 _master.options.splitting_with_naming = options.splitting_with_naming;  
 _master.options.neg_eq_splitting = options.static_neg_eq_splitting;  
 _master.options.paramodulation = false;
 _master.options.sim_back_sup = 0;
 _master.options.hyper_positive = true;
 _master.options.selection = options.selection;
 _master.options.literal_comparison_mode = options.literal_comparison_mode;
 _master.options.inherited_negative_selection = options.inherited_negative_selection;

  // strategy switches        
 _master.options.handle_equality = true;
 _master.options.forward_subsumption = true;
 _master.options.fs_set_mode = true; // not to worry about completeness here  
 _master.options.fs_optimised_multiset_mode = false;
 _master.options.fs_use_sig_filters = false;
 _master.options.fs_old_style = false;
 _master.options.simplify_by_fs = options.simplify_by_fs;
 _master.options.forward_demodulation = 0; 
 _master.options.fd_on_splitting_branches = false;
 if (options.normalize_by_commutativity)
  {
   _master.options.normalize_by_commutativity = _problemProfile.commutativityTable();
  }
 else 
  _master.options.normalize_by_commutativity = (const Clause* const *)0;
 _master.options.backward_demodulation = 0;
 _master.options.bd_on_splitting_branches = false;
 _master.options.simplify_by_eq_res = true;
 _master.options.backward_subsumption = true;
 _master.options.bs_set_mode = true; // not to worry about completeness here  
 _master.options.bs_optimised_multiset_mode = false;
 _master.options.orphan_murder = 0;
 _master.options.hyper_dinamic_lhs_ordering = true;
 _master.options.unit_conflict = true;
 _master.options.lrs = 0;
 _master.options.pred_elim_table = (const bool*)0;
  // numeric options
 _master.options.age_weight_ratio = 0;
 _master.options.timeLimit = (float)LONG_MAX;
 _master.options.memoryLimit = (float)LONG_MAX;
 _master.options.allocation_buffer_price = (float)options.allocation_buffer_price;
 _master.options.maxActive = LONG_MAX;
 _master.options.maxPassive = LONG_MAX;
 _master.options.maxWeight = LONG_MAX;
 _master.options.optGrCompWeight = options.optGrCompWeight;
 _master.options.first_time_check = 1.0; 
 _master.options.first_mem_check = 1.0;
 _master.options.internal_tick_frequency = LONG_MAX;
   // interface options
 _master.options.showGen = options.showGen;
 _master.options.showPassive = (options.showPassive) ? outputClause : 0;
 _master.options.showActive = (options.showActive) ? outputClause : 0;
 _master.options.showHyperRule = (options.showHyperRules) ? outputHyperRule : 0;
 _master.options.showDefinition = (options.showDefinitions) ? outputDefinition : 0;
 _master.setStrategy();
}; // void Vampire::prepareMasterForPreproSimplification()

void Vampire::prepareMasterForPreproDefElimination()
{
 CALL("prepareMasterForPreproDefElimination()");
 _master.reset();

 _master.options.incompleteAPriori = options.incompleteAPriori;
 _master.options.main_alg = 0;
 _master.options.binary_resolution = true; // for elimination of definitions
 _master.options.hyper_resolution = false;
 _master.options.no_inferences_between_rules = false;
 _master.options.splitting = false;   
 _master.options.splitting_style = 0L;
 _master.options.splitting_with_naming = false; 
 _master.options.neg_eq_splitting = 0L; // off  
 _master.options.paramodulation = false;
 _master.options.sim_back_sup = 0;
 _master.options.hyper_positive = true;
 _master.options.selection = options.selection;
 _master.options.literal_comparison_mode = options.literal_comparison_mode;
 _master.options.inherited_negative_selection = options.inherited_negative_selection;

  // strategy switches        
 _master.options.handle_equality = true;
 _master.options.forward_subsumption = true;
 _master.options.fs_set_mode = true; // not to worry about completeness here  
 _master.options.fs_optimised_multiset_mode = false;
 _master.options.fs_use_sig_filters = false;
 _master.options.fs_old_style = false;
 _master.options.simplify_by_fs = options.simplify_by_fs;
 _master.options.forward_demodulation = 0;
 _master.options.fd_on_splitting_branches = false;
 if (options.normalize_by_commutativity)
  {
   _master.options.normalize_by_commutativity = _problemProfile.commutativityTable();
  }
 else 
  _master.options.normalize_by_commutativity = (const Clause* const *)0;
 _master.options.backward_demodulation = 0;
 _master.options.bd_on_splitting_branches = false;
 _master.options.simplify_by_eq_res = true;
 _master.options.backward_subsumption = true;
 _master.options.bs_set_mode = true; // not to worry about completeness here  
 _master.options.bs_optimised_multiset_mode = false;
 _master.options.orphan_murder = 0;
 _master.options.hyper_dinamic_lhs_ordering = true;
 _master.options.unit_conflict = true;
 _master.options.lrs = 0;
 _master.options.pred_elim_table = _problemProfile.eliminationTable();
  // numeric options
 _master.options.age_weight_ratio = 0;
 _master.options.timeLimit = (float)LONG_MAX;
 _master.options.memoryLimit = (float)LONG_MAX;
 _master.options.allocation_buffer_price = (float)options.allocation_buffer_price;
 _master.options.maxActive = LONG_MAX;
 _master.options.maxPassive = LONG_MAX;
 _master.options.maxWeight = LONG_MAX;
 _master.options.optGrCompWeight = options.optGrCompWeight;
 _master.options.first_time_check = 1.0;
 _master.options.first_mem_check = 1.0;
 _master.options.internal_tick_frequency = LONG_MAX;
   // interface options
 _master.options.showGen = options.showGen;
 _master.options.showPassive = (options.showPassive) ? outputClause : 0;
 _master.options.showActive = (options.showActive) ? outputClause : 0;
 _master.options.showHyperRule = (options.showHyperRules) ? outputHyperRule : 0;
 _master.options.showDefinition = (options.showDefinitions) ? outputDefinition : 0;
 _master.setStrategy();
}; // void Vampire::prepareMasterForPreproDefElimination()

void Vampire::preprocessedToHistory()
{
 CALL("preprocessedToHistory()");
 Clause* cl;
 while (_preprocessedClauses)
  {
   _preprocessedClauses.dequeue(cl);
   _preprocessingHistory.enqueue(cl);
  };
}; // void Vampire::preprocessedToHistory()

void Vampire::extractPreprocessed()
{
 CALL("extractPreprocessed()");
 _master.passiveResetIter();
 _master.historyResetIter();
 Clause* cl;
 for (cl = _master.nextPassive(); cl; cl = _master.nextPassive())
  {
   Clause* copy = cl->unsharedCopy();
   _clauseNumIndex.insert(copy);
   _preprocessedClauses.enqueue(copy);
  };
 for (cl = _master.nextHistory(); cl; cl = _master.nextHistory())
  {
   Clause* copy = cl->unsharedCopy();
   _clauseNumIndex.insert(copy);
   _preprocessingHistory.enqueue(copy); 
  };
 

 ClauseBackground::ClauseList* anc;
 Clause* ancClause;
 Clause* ancCopy; 
 COP("linking1");
 const ClauseQueue::Entry* qe;
 for (qe = _preprocessedClauses.begin(); qe; qe = qe->next())
  {
   cl = qe->value();    
   for (anc = cl->Ancestors()->ancestors(); anc; anc = anc->tl())
    {
     ancClause = anc->hd();
     ancCopy = (Clause*)_clauseNumIndex.find(ancClause->Number());
     if (ancCopy)  { anc->hd() = ancCopy; }
     else 
      {
       ASSERT(!ancClause->Ancestors()->ancestors());
       ASSERT(ancClause->Ancestors()->TheRules().bit(ClauseBackground::Input));
      }; 
    };

  };
 COP("linking2");
 for (qe = _preprocessingHistory.begin(); qe; qe = qe->next())
  {
   cl = qe->value(); 
   for (anc = cl->Ancestors()->ancestors(); anc; anc = anc->tl())
    {
     ancClause = anc->hd();
     ancCopy = (Clause*)_clauseNumIndex.find(ancClause->Number());
     if (ancCopy)  { anc->hd() = ancCopy; }
     else 
      {
       ASSERT(!ancClause->Ancestors()->ancestors());
       ASSERT(ancClause->Ancestors()->TheRules().bit(ClauseBackground::Input));
      }; 
    };

  };
}; // void Vampire::extractPreprocessed()



void Vampire::load(const ClauseQueue& clauses)
{
 CALL("load(const ClauseQueue& clauses)");
 for (const ClauseQueue::Entry* qe = clauses.begin(); qe; qe = qe->next())
  {
   _master.inputClause(qe->value());
   if (_master.terminated)  return;
  };
 _master.endOfInput();
}; // void Vampire::load(const ClauseQueue& clauses)

void Vampire::search(float timeLimit,const Master::Options& strat)
{
 CALL("search(float timeLimit,const Master::Options& strat)");
 _master.reset();
 _master.options = strat;
 _master.options.timeLimit = timeLimit;
 if (_numOfDefinedHeaders) { _master.options.defined_header_table = _definedHeader; }
 else _master.options.defined_header_table = (const bool*)0; 

 _master.setStrategy();
 

 load(_preprocessedClauses);
 if (_master.terminated) return;
 _master.mainLoop();
}; // void Vampire::search(float timeLimit,const Master::Options& strat)

void Vampire::outputClause(const Clause* cl)
{
 cout << "% " << cl << '\n';

 //DF; options.tabulation->proofClause(cl);
 
}; // void Vampire::outputClause(const Clause* cl)

void Vampire::outputHyperRule(const Clause* cl)
{
 cout << "Hyper. rule: " << cl << '\n'; 
}; // void Vampire::outputHyperRule(const Clause* cl) 

void Vampire::outputDefinition(const Clause* cl)
{
 cout << "Predicate definition: " << cl << '\n'; 
}; // void Vampire::outputDefinition(const Clause* cl) 

void Vampire::output(const ClauseQueue& clauses)
{
 CALL("output(const ClauseQueue& clauses)");
 for (const ClauseQueue::Entry* qe = clauses.begin();qe;qe = qe->next())
  cout << "% " << qe->value() << '\n';  
}; // void Vampire::output(const ClauseQueue& clauses)


void Vampire::outputInTPTPFormat(const ClauseQueue& clauses)
{
 CALL("outputInTPTPFormat(const ClauseQueue& clauses)");
 for (const ClauseQueue::Entry* qe = clauses.begin();qe;qe = qe->next())
   qe->value()->outputInTPTPFormat(cout) << "\n\n"; 
}; // void Vampire::outputInTPTPFormat(const ClauseQueue& clauses)


void Vampire::showWeightChange(long prevMW,long newMW)
{
 if (options.showWeightLimitChanges) 
  cout << "Weight Limit changed: " << prevMW << " --->> " << newMW << '\n';
 if (options.plot_weight_limit_changes) 
  (*options.plot_weight_limit_changes) << _timer.elapsed() << "   " << newMW << '\n';
}; // void Vampire::showWeightChange(long prevMW,long newMW)

void Vampire::showUsagePace(float usagePace)
{
 CALL("showUsagePace(float usagePace)");
 ASSERT(options.plot_usage_pace);
 (*options.plot_usage_pace) << _timer.elapsed() << "   " << usagePace << '\n';
}; // void Vampire::showUsagePace(float usagePace)

 
void Vampire::outputStatistics()
{

  //DF; MEMMAP(true); 

 cout << "%================= Statistics: =======================\n";
 cout << "% time " << _timer.elapsed() << " (" << _master.timer.elapsed() << " current loop)\n";
 cout << "% memory " << (GlobAlloc::occupiedByObjects()/1024) 
      << "Kb (" << (GlobAlloc::occupiedByChunks()/1024) << "Kb in chunks), " 
      << "total buffer deficit " << (GlobAlloc::totalBufferDeficit()/1024) << "Kb\n"; 
 cout << "%=== Generated clauses:\n";
 cout << "%      total " << _master.statGenerated() << '\n';
 cout << "%      forward subsumed " << _master.statFSubsumed() << '\n';
 cout << "%      eq.tautologies   " << _master.statEqTautologies() << '\n';
 cout << "%      prop.tautologies " << _master.statPropTautologies() << '\n';
 cout << "%      simplified " << _master.statFSimplified() << '\n';
 cout << "%         by FSR      " << _master.statFSimplifiedByFSR() << '\n';
 cout << "%         by demod.   " << _master.statFSimplifiedByDemod() << '\n';
 cout << "%              on splitting branches " << _master.statFDOnSplittingBranches() << '\n';
 cout << "%         by eq. res. " << _master.statFSimplifiedByEqRes() << '\n';
 cout << "%      split " << _master.statSplit() 
      << "  avg.components " << _master.statAvgComponents() 
      << "  diff.components " << _master.statDiffComponents() << '\n';
 cout << "%      useless:\n";
 cout << "%        too heavy " << _master.statTooHeavy() << '\n'; 
 cout << "%        too heavy literals " << _master.statTooHeavyLiterals() << '\n'; 
 cout << "%        too long " << _master.statTooLong() << '\n';  
 cout << "%        too deep " << _master.statTooDeep() << '\n';  
 cout << "%        too many variables " << _master.statTooManyVariables() << '\n';  
 cout << "%        refused allocation " << _master.statRefusedAllocation() << '\n';

 cout << "%=== Retained clauses:\n";
 cout << "%      total " << _master.statRetained() << '\n';
 cout << "%      used " << _master.statUsed() << '\n';
 cout << "%      back subsumed " << _master.statBSubsumed() << '\n';
 cout << "%      back simplified " << _master.statBSimplified() << '\n';
 cout << "%        on splitting branches " << _master.statBDOnSplittingBranches() << '\n';
 cout << "%      rewritten by def. unfolding " << _master.statRewrittenByDefUnfolding() << '\n';
 cout << "%      orphans murdered " << _master.statOrphansMurdered() << '\n';
 cout << "%               passive " << _master.statPassiveOrphansMurdered() << '\n';
 cout << "%               active  " << _master.statActiveOrphansMurdered() << '\n';
 cout << "%      recycled " << _master.statRecycled() << '\n';
 cout << "%      currently passive " << _master.statCurrentlyPassive()        
      << " (" << _master.statPassiveReachable() << " reachable)\n";
 cout << "%      reserved passive " << _master.statReservedPassive() << '\n';
 cout << "%      currently active " << _master.statCurrentlyActive() << '\n'; 

 cout << "%========== End of statistics. ======================\n";
 cout.flush();

 
 /**********
 ofstream dumpedSharingIndex;
 dumpedSharingIndex.open("dumpedSharingIndex",ios::out); 
 ofstream dumpedPersistentClauses;
 dumpedPersistentClauses.open("dumpedPersistentClauses",ios::out);

 _master.scanSharingIndex(dumpedSharingIndex);
 _master.dumpPersistentClauses(dumpedPersistentClauses);
 *************/

}; // void Vampire::outputStatistics()

void Vampire::tabulateStatistics()
{
 options.tabulation->stat("time",_timer.elapsed());
 options.tabulation->stat("memory",GlobAlloc::occupiedByObjects()/1024);
 options.tabulation->stat("generated",_master.statGenerated());
 options.tabulation->stat("eq_tautologies",_master.statEqTautologies());
 options.tabulation->stat("prop_tautologies",_master.statPropTautologies());
 options.tabulation->stat("forward_subsumed",_master.statFSubsumed());
 options.tabulation->stat("forward_simplified",_master.statFSimplified());
 options.tabulation->stat("forward_demodulated_on_splitting_branches",_master.statFDOnSplittingBranches());
 options.tabulation->stat("too_heavy_literals",_master.statTooHeavyLiterals());
 options.tabulation->stat("too_deep",_master.statTooDeep());
 options.tabulation->stat("too_heavy_clauses",_master.statTooHeavy());
 options.tabulation->stat("too_long_clauses",_master.statTooLong());
 options.tabulation->stat("too_many_variables",_master.statTooManyVariables());
 options.tabulation->stat("backward_subsumed",_master.statBSubsumed());
 options.tabulation->stat("backward_simplified",_master.statBSimplified());
 options.tabulation->stat("backward_demodulated_on_splitting_branches",_master.statBDOnSplittingBranches());
 options.tabulation->stat("rewritten_by_def_unfolding",_master.statRewrittenByDefUnfolding());
 options.tabulation->stat("passive_orphans_murdered",_master.statPassiveOrphansMurdered()); 
 options.tabulation->stat("active_orphans_murdered",_master.statActiveOrphansMurdered()); 
 options.tabulation->stat("used",_master.statUsed());
 options.tabulation->stat("kept_total",_master.statRetained());
 options.tabulation->stat("solved_by_preprocessing",_solvedByPreprocessing);
 options.tabulation->stat("split",_master.statSplit()); 
 options.tabulation->stat("different_components",_master.statDiffComponents());
 options.tabulation->stat("average_components",_master.statAvgComponents());
}; // void Vampire::tabulateStatistics()

void Vampire::tabulatePreprocessingStatistics()
{
 options.tabulation->stat("prepro_split",_master.statSplit()); 
 options.tabulation->stat("prepro_different_components",_master.statDiffComponents());
 options.tabulation->stat("prepro_average_components",_master.statAvgComponents());
}; // void Vampire::tabulatePreprocessingStatistics() 

void Vampire::tabulateOptions()
{
 options.tabulation->option("main_alg",_master.options.main_alg);
 options.tabulation->option("hyper_resolution",_master.options.hyper_resolution);
 options.tabulation->option("no_inferences_between_rules",_master.options.no_inferences_between_rules);
 options.tabulation->option("splitting",_master.options.splitting);
 options.tabulation->option("splitting_style",_master.options.splitting_style);
 options.tabulation->option("splitting_with_naming",_master.options.splitting_with_naming);
 options.tabulation->option("neg_eq_splitting",_master.options.neg_eq_splitting);
 options.tabulation->option("paramodulation",_master.options.paramodulation);
 options.tabulation->option("sim_back_sup",_master.options.sim_back_sup);
 options.tabulation->option("hyper_positive",_master.options.hyper_positive);
 options.tabulation->option("selection",_master.options.selection);
 options.tabulation->option("literal_comparison_mode",_master.options.literal_comparison_mode);
 options.tabulation->option("inherited_negative_selection",_master.options.inherited_negative_selection);
 options.tabulation->option("forward_subsumption",_master.options.forward_subsumption);
 options.tabulation->option("fs_set_mode",_master.options.fs_set_mode);
 options.tabulation->option("fs_optimised_multiset_mode",_master.options.fs_optimised_multiset_mode);
 options.tabulation->option("fs_use_sig_filters",_master.options.fs_use_sig_filters);
 options.tabulation->option("fs_old_style",_master.options.fs_old_style);
 options.tabulation->option("simplify_by_fs",_master.options.simplify_by_fs);
 options.tabulation->option("forward_demodulation",_master.options.forward_demodulation);
 options.tabulation->option("fd_on_splitting_branches",_master.options.fd_on_splitting_branches);
 options.tabulation->option("normalize_by_commutativity",(bool)_master.options.normalize_by_commutativity);
 options.tabulation->option("backward_demodulation",_master.options.backward_demodulation);
 options.tabulation->option("bd_on_splitting_branches",_master.options.bd_on_splitting_branches);
 options.tabulation->option("simplify_by_eq_res",_master.options.simplify_by_eq_res);
 options.tabulation->option("backward_subsumption",_master.options.backward_subsumption);
 options.tabulation->option("bs_set_mode",_master.options.bs_set_mode);
 options.tabulation->option("bs_optimised_multiset_mode",_master.options.bs_optimised_multiset_mode);
 options.tabulation->option("orphan_murder",_master.options.orphan_murder); 
 options.tabulation->option("unit_conflict",_master.options.unit_conflict); 
 options.tabulation->option("lrs",_master.options.lrs);  
 options.tabulation->option("header_rank_by_arity",options.header_rank_by_arity);
 options.tabulation->option("header_rank_kinky",options.header_rank_kinky);
 options.tabulation->option("header_rank_check_positive_selection",options.header_rank_check_positive_selection);
 options.tabulation->option("pick_given_ratio",_master.options.age_weight_ratio);
 options.tabulation->option("internal_tick_frequency",_master.options.internal_tick_frequency);
 options.tabulation->option("first_time_check",_master.options.first_time_check); 
 options.tabulation->option("first_mem_check",_master.options.first_mem_check);
 

}; // void Vampire::tabulateOptions()

void Vampire::handleSIGINT(int)
{
 _timer.stop();
 outputStatistics();
 cout << "Continue? (y/n)\n";
 char a;
 cin >> a;

 if ((a == 'n') || (a == 'N'))
  {
   cout << "Bye!\n";
   terminate("terminated by the user");
   Exit::exit(0); 
  }
 else
 {
  cout << "\nContinuing proof search...\n";
  _timer.start();
  initSIGINTHandler();
 };
}; // void Vampire::handleSIGINT(int)



void Vampire::reportTermination()
{
 CALL("reportTermination()");
 outputStatistics();
 if (options.tabulation) 
  {
   tabulateOptions();
   tabulateStatistics();
  }; 
 const char* termShortExpl = "unknown   ";
 char result = '0';
 switch (_master.terminationReason)
  {
   case Master::UnknownTR: 
    ICP("UnknownTR");
    cout << "% Proof not found. Internal error: unknown termination type.\n"; 
    termShortExpl = "unknown   ";
    result = '-';
    break;  
   case Master::Proof: 
    cout << "% Proof found. Thanks to Tanya!\n"; 
    termShortExpl = "refutation";
    result = '+';  
    break; 
   case Master::EmptyPassive: 
    if (_master.completeStrategyUsed())
     {
      cout << "% Unprovable.\n";        
      termShortExpl = "unprovable";
      result = '-';
     }
    else 
     { 
      cout << "% Proof not found. Empty passive.\n"; 
      termShortExpl = "empty kept";
     };
    break;
   case Master::TimeLimit: 
    cout << "% Proof not found. Time limit.\n"; 
    termShortExpl = "time limit";
    break; 
   case Master::MemoryLimit: 
    cout << "% Proof not found. Memory limit.\n"; 
    termShortExpl = "mem. limit";
    break;
   case Master::AllocationFailure:
    cout << "% Proof not found. Allocation failure.\n";
    termShortExpl = "all. fail.";
    break;
   case Master::MaxPassive: 
    cout << "% Proof not found. Max. num. of passive reached. \n"; 
    termShortExpl = "max_kept  ";
    break; 
   case Master::MaxActive: 
    cout << "% Proof not found. Max. num. of active reached. \n"; 
    termShortExpl = "max_used  ";
    break; 
   case Master::UserInt:
    cout << "% Proof not found. User interrupt. \n";
    termShortExpl = "user int  ";
    break;
  }; 
 if (options.tabulation) 
  {
   options.tabulation->terminationReason(termShortExpl);
   options.tabulation->result(result); 
  };

 if ((_master.terminationReason == Master::Proof) && options.showProof)
  {
   _proofTree.collect(_master.proof);    
   outputProof();
   if (options.tabulation) tabulateProof(); 

   //DF; TRACE(true,true); 

  }; 
 cout.flush();
}; // void Vampire::reportTermination()


void Vampire::reportAbortion()
{
 CALL("reportAbortion()");
 // all operations here must be safe even if an error happend in a relevant piece of code
 outputStatistics();   
}; // void Vampire::reportAbortion()


void Vampire::outputProof()
{
 cout << "%======================== Proof: =========================\n"; 
 _proofTree.resetIter();   
 for (const Clause* cl = _proofTree.getNextClause(); cl; cl = _proofTree.getNextClause())
  cout << "% " << cl << '\n';
 cout << "%==================  End of proof. ========================\n"; 
 cout.flush();
}; // void Vampire::outputProof()

void Vampire::tabulateProof()
{
 _proofTree.resetIter();   
 for (const Clause* cl = _proofTree.getNextClause(); cl; cl = _proofTree.getNextClause())
  options.tabulation->proofClause(cl);
}; // void Vampire::tabulateProof()


void Vampire::terminate(const char* msg)
{
 cout << "% Forced termination: " << msg << ".\n"; 
 _master.terminate(Master::UserInt);
 reportTermination();
}; // void Vampire::terminate(const char* msg)

void Vampire::abort()
{
 _master.terminate(Master::UserInt);
 reportAbortion(); 
};


void Vampire::cancelEquality()
{
 CALL("cancelEquality()");
 if  (options.paramodulation 
      || options.inherited_negative_selection
      || options.forward_demodulation
      || options.fd_on_splitting_branches
      || options.backward_demodulation
      || options.bd_on_splitting_branches
      || options.simplify_by_eq_res)
  {
   cout << "% WARNING: No equality. All equality related features are switched off!\n";   
   options.paramodulation = false;
   options.inherited_negative_selection = false;
   options.forward_demodulation = false;
   options.fd_on_splitting_branches = false;
   options.backward_demodulation = false;
   options.bd_on_splitting_branches = false;
   options.simplify_by_eq_res = false;
  };
}; // void Vampire::cancelEquality()

void Vampire::manualStrat(Master::Options& opt)
{
 opt.incompleteAPriori = options.incompleteAPriori;
 opt.main_alg = options.main_alg;
 opt.binary_resolution = !options.hyper_resolution;
 opt.hyper_resolution = options.hyper_resolution;
 opt.no_inferences_between_rules = options.no_inferences_between_rules;
 opt.splitting = options.dynamic_splitting;  
 opt.splitting_style = options.splitting_style;
 opt.splitting_with_naming = options.splitting_with_naming;
 opt.neg_eq_splitting = 0L; // off  
 opt.paramodulation = options.paramodulation;
 opt.sim_back_sup = options.sim_back_sup;
 opt.hyper_positive = options.hyper_positive;
 opt.selection = options.selection;
 opt.literal_comparison_mode = options.literal_comparison_mode;
 opt.inherited_negative_selection = options.inherited_negative_selection;

  // strategy switches
 opt.handle_equality = options.paramodulation;
 opt.forward_subsumption = options.forward_subsumption;
 opt.fs_set_mode = options.fs_set_mode;
 opt.fs_optimised_multiset_mode = options.fs_optimised_multiset_mode;
 opt.fs_use_sig_filters = options.fs_use_sig_filters;
 opt.fs_old_style = options.fs_old_style;
 opt.simplify_by_fs = options.simplify_by_fs;
 opt.forward_demodulation = options.forward_demodulation;
 opt.fd_on_splitting_branches = options.fd_on_splitting_branches;
 if (options.normalize_by_commutativity)
  {
   opt.normalize_by_commutativity = _problemProfile.commutativityTable();
  }
 else
  {
   opt.normalize_by_commutativity = (const Clause* const *)0;
  };
 opt.backward_demodulation = options.backward_demodulation;
 opt.bd_on_splitting_branches = options.bd_on_splitting_branches;
 opt.simplify_by_eq_res = options.simplify_by_eq_res;
 opt.backward_subsumption = options.backward_subsumption;
 opt.bs_set_mode = options.bs_set_mode;
 opt.bs_optimised_multiset_mode = options.bs_optimised_multiset_mode;
 opt.orphan_murder = options.orphan_murder;
 opt.hyper_dinamic_lhs_ordering = true;
 opt.unit_conflict = options.unit_conflict;
 opt.lrs = options.lrs;

 opt.pred_elim_table = (const bool*)0;

  // numeric options
 opt.age_weight_ratio = options.age_weight_ratio;
 opt.timeLimit = options.timeLimit;
 opt.memoryLimit = options.memoryLimit;
 opt.allocation_buffer_price = (float)options.allocation_buffer_price;
 opt.maxActive = options.maxActive;
 opt.maxPassive = options.maxPassive;
 opt.maxWeight = options.maxWeight;
 opt.optGrCompWeight = options.optGrCompWeight;
 opt.first_time_check = ((float)options.first_time_check)/100;
 opt.first_mem_check = ((float)options.first_mem_check)/100;
 opt.internal_tick_frequency = options.internal_tick_frequency;
   // interface options 
 opt.showGen = options.showGen;
 opt.showPassive = (options.showPassive) ? outputClause : 0;
 opt.showActive = (options.showActive) ? outputClause : 0;
 opt.showHyperRule = (options.showHyperRules) ? outputHyperRule : 0;
 opt.showDefinition = (options.showDefinitions) ? outputDefinition : 0;
 opt.showWeightChange = showWeightChange; 
 opt.showUsagePace = (options.plot_usage_pace) ? showUsagePace : 0;
}; // void Vampire::manualStrat(Master::Options& opt)

void Vampire::autoNNE(Master::Options& opt)
{
 opt.incompleteAPriori = options.incompleteAPriori;
 opt.main_alg = options.main_alg;
 opt.binary_resolution = true;
 opt.hyper_resolution = false;
 opt.no_inferences_between_rules = false;
 opt.splitting = options.dynamic_splitting; 
 opt.splitting_style = options.splitting_style;
 opt.splitting_with_naming = options.splitting_with_naming;
 opt.neg_eq_splitting = 0L; // off    
 opt.paramodulation = false;
 opt.sim_back_sup = 0;
 opt.hyper_positive = true;
 opt.selection = options.selection;
 opt.literal_comparison_mode = options.literal_comparison_mode; 
 opt.inherited_negative_selection = options.inherited_negative_selection;

  // strategy switches
 opt.handle_equality = false;
 opt.forward_subsumption = options.forward_subsumption;
 opt.fs_set_mode = options.fs_set_mode;
 opt.fs_optimised_multiset_mode = options.fs_optimised_multiset_mode;
 opt.fs_use_sig_filters = options.fs_use_sig_filters;
 opt.fs_old_style = options.fs_old_style;
 opt.simplify_by_fs = options.simplify_by_fs;
 opt.forward_demodulation = 0;
 opt.fd_on_splitting_branches = false;
 opt.normalize_by_commutativity = (const Clause* const *)0;
 opt.backward_demodulation = 0;
 opt.bd_on_splitting_branches = false;
 opt.simplify_by_eq_res = false;
 opt.backward_subsumption = options.backward_subsumption;
 opt.bs_optimised_multiset_mode = options.bs_optimised_multiset_mode;
 opt.bs_set_mode = options.bs_set_mode;
 opt.orphan_murder = options.orphan_murder;
 opt.hyper_dinamic_lhs_ordering = true;
 opt.unit_conflict = options.unit_conflict;
 opt.lrs = options.lrs;
 opt.pred_elim_table = (const bool*)0;
  // numeric options
 opt.age_weight_ratio = 3; // options.age_weight_ratio;
 opt.timeLimit = options.timeLimit;
 opt.memoryLimit = options.memoryLimit;
 opt.allocation_buffer_price = (float)options.allocation_buffer_price;
 opt.maxActive = options.maxActive;
 opt.maxPassive = options.maxPassive;
 opt.maxWeight = options.maxWeight;
 opt.optGrCompWeight = options.optGrCompWeight;
 opt.first_time_check = ((float)options.first_time_check)/100;
 opt.first_mem_check = ((float)options.first_mem_check)/100;
 opt.internal_tick_frequency = options.internal_tick_frequency;
   // interface options 
 opt.showGen =  options.showGen;
 opt.showPassive = (options.showPassive) ? outputClause : 0;
 opt.showActive = (options.showActive) ? outputClause : 0;
 opt.showHyperRule = (options.showHyperRules) ? outputHyperRule : 0;
 opt.showDefinition = (options.showDefinitions) ? outputDefinition : 0;
 opt.showWeightChange = showWeightChange; 
 opt.showUsagePace = (options.plot_usage_pace) ? showUsagePace : 0;
}; // void Vampire::autoNNE(Master::Options& opt)

void Vampire::autoHNE1(Master::Options& opt)
{
 opt.incompleteAPriori = options.incompleteAPriori;
 opt.main_alg = options.main_alg;
 opt.binary_resolution = false;
 opt.hyper_resolution = true;
 opt.no_inferences_between_rules = false;
 opt.splitting = options.dynamic_splitting; 
 opt.splitting_style = options.splitting_style;
 opt.splitting_with_naming = options.splitting_with_naming;
 opt.neg_eq_splitting = 0L; // off    
 opt.paramodulation = false; 
 opt.sim_back_sup = 0;
 opt.hyper_positive = false;
 opt.selection = options.selection; //0; 
 opt.literal_comparison_mode = options.literal_comparison_mode;
 opt.inherited_negative_selection = options.inherited_negative_selection;

  // strategy switches
 opt.handle_equality = false;
 opt.forward_subsumption = options.forward_subsumption;
 opt.fs_set_mode = options.fs_set_mode;
opt.fs_optimised_multiset_mode = options.fs_optimised_multiset_mode;
 opt.fs_use_sig_filters = options.fs_use_sig_filters;
 opt.fs_old_style = options.fs_old_style;
 opt.simplify_by_fs = 0;
 opt.forward_demodulation = 0;
 opt.fd_on_splitting_branches = false;
 opt.normalize_by_commutativity = (const Clause* const *)0;
 opt.backward_demodulation = 0;
 opt.bd_on_splitting_branches = false;
 opt.simplify_by_eq_res = false;
 opt.backward_subsumption = options.backward_subsumption;
 opt.bs_set_mode = options.bs_set_mode;
 opt.bs_optimised_multiset_mode = options.bs_optimised_multiset_mode;
 opt.orphan_murder = options.orphan_murder;
 opt.hyper_dinamic_lhs_ordering = true;
 opt.unit_conflict = options.unit_conflict;
 opt.lrs = options.lrs;
 opt.pred_elim_table = (const bool*)0;
  // numeric options
 opt.age_weight_ratio = 5;
 opt.timeLimit = options.timeLimit;
 opt.memoryLimit = options.memoryLimit;
 opt.allocation_buffer_price = (float)options.allocation_buffer_price;
 opt.maxActive = options.maxActive;
 opt.maxPassive = options.maxPassive;
 opt.maxWeight = options.maxWeight;
 opt.optGrCompWeight = options.optGrCompWeight;
 opt.first_time_check = ((float)options.first_time_check)/100;
 opt.first_mem_check = ((float)options.first_mem_check)/100;
 opt.internal_tick_frequency = options.internal_tick_frequency;
   // interface options 
 opt.showGen =  options.showGen;
 opt.showPassive = (options.showPassive) ? outputClause : 0;
 opt.showActive = (options.showActive) ? outputClause : 0;
 opt.showHyperRule = (options.showHyperRules) ? outputHyperRule : 0;
 opt.showDefinition = (options.showDefinitions) ? outputDefinition : 0;
 opt.showWeightChange = showWeightChange; 
 opt.showUsagePace = (options.plot_usage_pace) ? showUsagePace : 0;
}; // void Vampire::autoHNE1(Master::Options& opt)

void Vampire::autoHNE2(Master::Options& opt)
{
 opt.incompleteAPriori = options.incompleteAPriori;
 opt.main_alg = options.main_alg;
 opt.binary_resolution = false;
 opt.hyper_resolution = true;
 opt.no_inferences_between_rules = false;
 opt.splitting = options.dynamic_splitting; 
 opt.splitting_style = options.splitting_style; 
 opt.splitting_with_naming = options.splitting_with_naming;
 opt.neg_eq_splitting = 0L; // off    
 opt.paramodulation = false;
 opt.sim_back_sup = 0;
 opt.hyper_positive = true;
 opt.selection = options.selection; // 0;
 opt.literal_comparison_mode = options.literal_comparison_mode;
 opt.inherited_negative_selection = options.inherited_negative_selection;

  // strategy switches
 opt.handle_equality = false;
 opt.forward_subsumption = options.forward_subsumption;
 opt.fs_set_mode = options.fs_set_mode;
 opt.fs_optimised_multiset_mode = options.fs_optimised_multiset_mode; 
 opt.fs_use_sig_filters = options.fs_use_sig_filters;
 opt.fs_old_style = options.fs_old_style;
 opt.simplify_by_fs = 0; 
 opt.forward_demodulation = 0;
 opt.fd_on_splitting_branches = false;
 opt.normalize_by_commutativity = (const Clause* const *)0;
 opt.backward_demodulation = 0;
 opt.bd_on_splitting_branches = false;
 opt.simplify_by_eq_res = false;
 opt.backward_subsumption = options.backward_subsumption;
 opt.bs_set_mode = options.bs_set_mode;
 opt.bs_optimised_multiset_mode = options.bs_optimised_multiset_mode;
 opt.orphan_murder = options.orphan_murder; 
 opt.hyper_dinamic_lhs_ordering = true;
 opt.unit_conflict = options.unit_conflict;
 opt.lrs = options.lrs;
 opt.pred_elim_table = (const bool*)0;
  // numeric options
 opt.age_weight_ratio = 5;
 opt.timeLimit = options.timeLimit;
 opt.memoryLimit = options.memoryLimit;
 opt.allocation_buffer_price = (float)options.allocation_buffer_price;
 opt.maxActive = options.maxActive;
 opt.maxPassive = options.maxPassive;
 opt.maxWeight = options.maxWeight;
 opt.optGrCompWeight = options.optGrCompWeight;
 opt.first_time_check = ((float)options.first_time_check)/100;
 opt.first_mem_check = ((float)options.first_mem_check)/100;
 opt.internal_tick_frequency = options.internal_tick_frequency;
   // interface options 
 opt.showGen =  options.showGen;
 opt.showPassive = (options.showPassive) ? outputClause : 0;
 opt.showActive = (options.showActive) ? outputClause : 0;
 opt.showHyperRule = (options.showHyperRules) ? outputHyperRule : 0;
 opt.showDefinition = (options.showDefinitions) ? outputDefinition : 0;
 opt.showWeightChange = showWeightChange;  
 opt.showUsagePace = (options.plot_usage_pace) ? showUsagePace : 0; 
}; // void Vampire::autoHNE2(Master::Options& opt)

void Vampire::autoNEQ(Master::Options& opt)
{
 opt.incompleteAPriori = options.incompleteAPriori;
 opt.main_alg = options.main_alg;
 opt.binary_resolution = true;
 opt.hyper_resolution = false;
 opt.no_inferences_between_rules = false;
 opt.splitting = options.dynamic_splitting; 
 opt.splitting_style = options.splitting_style;
 opt.splitting_with_naming = options.splitting_with_naming; 
 opt.neg_eq_splitting = 0L; // off
 opt.paramodulation = true;
 opt.sim_back_sup = options.sim_back_sup;
 opt.hyper_positive = true;
 opt.selection = options.selection; //1;
 opt.literal_comparison_mode = options.literal_comparison_mode;
 opt.inherited_negative_selection = options.inherited_negative_selection; 

  // strategy switches
 opt.handle_equality = true;
 opt.forward_subsumption = options.forward_subsumption;
 opt.fs_set_mode = options.fs_set_mode;
 opt.fs_optimised_multiset_mode = options.fs_optimised_multiset_mode; 
 opt.fs_use_sig_filters = options.fs_use_sig_filters;
 opt.fs_old_style = options.fs_old_style;
 opt.simplify_by_fs = options.simplify_by_fs;
 opt.forward_demodulation = options.forward_demodulation; 
 opt.fd_on_splitting_branches = options.fd_on_splitting_branches;
 if (options.normalize_by_commutativity)
  {
   opt.normalize_by_commutativity = _problemProfile.commutativityTable(); 
  }
 else
  {
   opt.normalize_by_commutativity = (const Clause* const *)0;
  };
 opt.backward_demodulation = options.backward_demodulation;
 opt.bd_on_splitting_branches = options.bd_on_splitting_branches;
 opt.simplify_by_eq_res = true;
 opt.backward_subsumption = options.backward_subsumption;
 opt.bs_set_mode = options.bs_set_mode;
 opt.bs_optimised_multiset_mode = options.bs_optimised_multiset_mode;
 opt.orphan_murder = options.orphan_murder;
 opt.hyper_dinamic_lhs_ordering = true;
 opt.unit_conflict = options.unit_conflict;
 opt.lrs = options.lrs;
 opt.pred_elim_table = (const bool*)0;
  // numeric options
 opt.age_weight_ratio = 5; // options.age_weight_ratio;
 opt.timeLimit = options.timeLimit;
 opt.memoryLimit = options.memoryLimit;
 opt.allocation_buffer_price = (float)options.allocation_buffer_price;
 opt.maxActive = options.maxActive;
 opt.maxPassive = options.maxPassive;
 opt.maxWeight = options.maxWeight;
 opt.optGrCompWeight = options.optGrCompWeight;
 opt.first_time_check = ((float)options.first_time_check)/100;
 opt.first_mem_check = ((float)options.first_mem_check)/100;
 opt.internal_tick_frequency = options.internal_tick_frequency;
   // interface options 
 opt.showGen =  options.showGen;
 opt.showPassive = (options.showPassive) ? outputClause : 0;
 opt.showActive = (options.showActive) ? outputClause : 0;
 opt.showHyperRule = (options.showHyperRules) ? outputHyperRule : 0;
 opt.showDefinition = (options.showDefinitions) ? outputDefinition : 0;
 opt.showWeightChange = showWeightChange; 
 opt.showUsagePace = (options.plot_usage_pace) ? showUsagePace : 0;
}; // void Vampire::autoNEQ(Master::Options& opt)

void Vampire::autoHEQ(Master::Options& opt)
{
 opt.incompleteAPriori = options.incompleteAPriori;
 opt.main_alg = options.main_alg;
 opt.binary_resolution = true;
 opt.hyper_resolution = false;
 opt.no_inferences_between_rules = false;
 opt.splitting =  options.dynamic_splitting;  
 opt.splitting_style = options.splitting_style;
 opt.splitting_with_naming = options.splitting_with_naming;   
 opt.neg_eq_splitting = 0L; // off
 opt.paramodulation = true;
 opt.sim_back_sup = options.sim_back_sup;
 opt.hyper_positive = true;
 opt.selection = options.selection;
 opt.literal_comparison_mode = options.literal_comparison_mode;
 opt.inherited_negative_selection = options.inherited_negative_selection; 



  // strategy switches
 opt.handle_equality = true;
 opt.forward_subsumption = options.forward_subsumption;
 opt.fs_set_mode = options.fs_set_mode;
 opt.fs_optimised_multiset_mode = options.fs_optimised_multiset_mode; 
 opt.fs_use_sig_filters = options.fs_use_sig_filters;
 opt.fs_old_style = options.fs_old_style;
 opt.simplify_by_fs = options.simplify_by_fs;
 opt.forward_demodulation = options.forward_demodulation; 
 opt.fd_on_splitting_branches = options.fd_on_splitting_branches;
 if (options.normalize_by_commutativity)
  {
   opt.normalize_by_commutativity = _problemProfile.commutativityTable(); 
  }
 else
  {
   opt.normalize_by_commutativity = (const Clause* const *)0;
  };
 opt.backward_demodulation = options.backward_demodulation;
 opt.bd_on_splitting_branches = options.bd_on_splitting_branches;
 opt.simplify_by_eq_res = true;
 opt.backward_subsumption = options.backward_subsumption;
 opt.bs_set_mode = options.bs_set_mode;
 opt.bs_optimised_multiset_mode = options.bs_optimised_multiset_mode;
 opt.orphan_murder = options.orphan_murder;
 opt.hyper_dinamic_lhs_ordering = true;
 opt.unit_conflict = options.unit_conflict;
 opt.lrs = options.lrs;
 opt.pred_elim_table = (const bool*)0;
  // numeric options
 opt.age_weight_ratio = 5; // options.age_weight_ratio;
 opt.timeLimit = options.timeLimit;
 opt.memoryLimit = options.memoryLimit;
 opt.allocation_buffer_price = (float)options.allocation_buffer_price;
 opt.maxActive = options.maxActive;
 opt.maxPassive = options.maxPassive;
 opt.maxWeight = options.maxWeight;
 opt.optGrCompWeight = options.optGrCompWeight;
 opt.first_time_check = ((float)options.first_time_check)/100;
 opt.first_mem_check = ((float)options.first_mem_check)/100;
 opt.internal_tick_frequency = options.internal_tick_frequency;
   // interface options
 opt.showGen =  options.showGen;
 opt.showPassive = (options.showPassive) ? outputClause : 0;
 opt.showActive = (options.showActive) ? outputClause : 0;
 opt.showHyperRule = (options.showHyperRules) ? outputHyperRule : 0;
 opt.showDefinition = (options.showDefinitions) ? outputDefinition : 0;
 opt.showWeightChange = showWeightChange; 
 opt.showUsagePace = (options.plot_usage_pace) ? showUsagePace : 0;
}; // void Vampire::autoHEQ(Master::Options& opt)


void Vampire::autoPEQ(Master::Options& opt)
{
 opt.incompleteAPriori = options.incompleteAPriori;
 opt.main_alg = options.main_alg;
 opt.binary_resolution = true;
 opt.hyper_resolution = false;
 opt.no_inferences_between_rules = false;
 opt.splitting = options.dynamic_splitting; 
 opt.splitting_style = options.splitting_style; 
 opt.splitting_with_naming = options.splitting_with_naming;  
 opt.neg_eq_splitting = 0L; // off
 opt.paramodulation = true;
 opt.sim_back_sup = options.sim_back_sup;
 opt.hyper_positive = true;
 opt.selection = options.selection;
 opt.literal_comparison_mode = options.literal_comparison_mode;
 opt.inherited_negative_selection = options.inherited_negative_selection; 

  // strategy switches
 opt.handle_equality = true;
 opt.forward_subsumption = options.forward_subsumption;
 opt.fs_set_mode = options.fs_set_mode;
 opt.fs_optimised_multiset_mode = options.fs_optimised_multiset_mode; 
 opt.fs_use_sig_filters = options.fs_use_sig_filters;
 opt.fs_old_style = options.fs_old_style;
 opt.simplify_by_fs = 0; 
 opt.forward_demodulation = options.forward_demodulation; 
 opt.fd_on_splitting_branches = options.fd_on_splitting_branches;
 if (options.normalize_by_commutativity)
  {
   opt.normalize_by_commutativity = _problemProfile.commutativityTable(); 
  }
 else
  {
   opt.normalize_by_commutativity = (const Clause* const *)0;
  };
 opt.backward_demodulation = options.backward_demodulation;
 opt.bd_on_splitting_branches = options.bd_on_splitting_branches;
 opt.simplify_by_eq_res = true;
 opt.backward_subsumption = options.backward_subsumption;
 opt.bs_set_mode = options.bs_set_mode;
 opt.bs_optimised_multiset_mode = options.bs_optimised_multiset_mode;
 opt.orphan_murder = options.orphan_murder;
 opt.hyper_dinamic_lhs_ordering = true;
 opt.unit_conflict = options.unit_conflict;
 opt.lrs = options.lrs;
 opt.pred_elim_table = (const bool*)0;
  // numeric options
 opt.age_weight_ratio = 5; // options.age_weight_ratio;
 opt.timeLimit = options.timeLimit;
 opt.memoryLimit = options.memoryLimit;
 opt.allocation_buffer_price = (float)options.allocation_buffer_price;
 opt.maxActive = options.maxActive;
 opt.maxPassive = options.maxPassive;
 opt.maxWeight = options.maxWeight;
 opt.optGrCompWeight = options.optGrCompWeight;
 opt.first_time_check = ((float)options.first_time_check)/100;
 opt.first_mem_check = ((float)options.first_mem_check)/100;
 opt.internal_tick_frequency = options.internal_tick_frequency;
   // interface options 
 opt.showGen =  options.showGen;
 opt.showPassive = (options.showPassive) ? outputClause : 0;
 opt.showActive = (options.showActive) ? outputClause : 0;
 opt.showHyperRule = (options.showHyperRules) ? outputHyperRule : 0;
 opt.showDefinition = (options.showDefinitions) ? outputDefinition : 0;
 opt.showWeightChange = showWeightChange;  
 opt.showUsagePace = (options.plot_usage_pace) ? showUsagePace : 0;
}; // void Vampire::autoPEQ(Master::Options& opt)

void Vampire::autoUEQ(Master::Options& opt)
{
 opt.incompleteAPriori = options.incompleteAPriori;
 opt.main_alg = options.main_alg;
 opt.binary_resolution = true;
 opt.hyper_resolution = false;
 opt.no_inferences_between_rules = false;
 opt.splitting = options.dynamic_splitting;
 opt.splitting_style = options.splitting_style;
 opt.splitting_with_naming = options.splitting_with_naming;
 opt.hyper_positive = true; 
 opt.neg_eq_splitting = 0L; // off          
 opt.paramodulation = true;
 opt.sim_back_sup = options.sim_back_sup;
 opt.hyper_positive = true;
 opt.selection = options.selection;
 opt.literal_comparison_mode = options.literal_comparison_mode;
 opt.inherited_negative_selection = options.inherited_negative_selection;   

  // strategy switches
 opt.handle_equality = true;
 opt.forward_subsumption = options.forward_subsumption;
 opt.fs_set_mode = options.fs_set_mode;
 opt.fs_optimised_multiset_mode = options.fs_optimised_multiset_mode; 
 opt.fs_use_sig_filters = options.fs_use_sig_filters;
 opt.fs_old_style = options.fs_old_style;
 opt.simplify_by_fs = options.simplify_by_fs;
 opt.forward_demodulation = options.forward_demodulation; 
 opt.fd_on_splitting_branches = options.fd_on_splitting_branches;
 if (options.normalize_by_commutativity)
  {
   opt.normalize_by_commutativity = _problemProfile.commutativityTable(); 
  }
 else
  {
   opt.normalize_by_commutativity = (const Clause* const *)0;
  };
 opt.backward_demodulation = options.backward_demodulation;
 opt.bd_on_splitting_branches = options.bd_on_splitting_branches;
 opt.simplify_by_eq_res = true;
 opt.backward_subsumption = options.backward_subsumption;
 opt.bs_set_mode = options.bs_set_mode;
 opt.bs_optimised_multiset_mode = options.bs_optimised_multiset_mode;
 opt.orphan_murder = options.orphan_murder;
 opt.hyper_dinamic_lhs_ordering = true;
 opt.unit_conflict = options.unit_conflict;
 opt.lrs = options.lrs;
 opt.pred_elim_table = (const bool*)0;
  // numeric options
 opt.age_weight_ratio = 5; // options.age_weight_ratio;
 opt.timeLimit = options.timeLimit;
 opt.memoryLimit = options.memoryLimit;
 opt.allocation_buffer_price = (float)options.allocation_buffer_price;
 opt.maxActive = options.maxActive;
 opt.maxPassive = options.maxPassive;
 opt.maxWeight = options.maxWeight;
 opt.optGrCompWeight = options.optGrCompWeight;
 opt.first_time_check = ((float)options.first_time_check)/100;
 opt.first_mem_check = ((float)options.first_mem_check)/100;
 opt.internal_tick_frequency = options.internal_tick_frequency;
   // interface options
 opt.showGen =  options.showGen;
 opt.showPassive = (options.showPassive) ? outputClause : 0;
 opt.showActive = (options.showActive) ? outputClause : 0;
 opt.showHyperRule = (options.showHyperRules) ? outputHyperRule : 0;
 opt.showDefinition = (options.showDefinitions) ? outputDefinition : 0;
 opt.showWeightChange = showWeightChange; 
 opt.showUsagePace = (options.plot_usage_pace) ? showUsagePace : 0;
}; // void Vampire::autoUEQ(Master::Options& opt)

void Vampire::assertHeaderRanks()
{
 CALL("assertHeaderRanks()");
 GlobalHeaderWeighting.SetRankingByArity(options.header_rank_by_arity);
 GlobalHeaderWeighting.SetKinkyRanking(options.header_rank_kinky);
 GlobalHeaderWeighting.AssertHeaderWeighting(HWList,
                                             ((options.selection < 0)
                                               && (options.header_rank_check_positive_selection)));
 if (options.showProfile) 
  { 
   cout << "%=============== Ranks of headers: ================\n"
        << GlobalHeaderWeighting 
        << "%==================================================\n";
  };
}; // Vampire::void assertHeaderRanks()


void Vampire::setSignatureProperties()
{
 CALL("setSignatureProperties()");
 AutoFunctorPrecedence autoFunctorPrecedence;
 autoFunctorPrecedence.assignFixed(HellConst::UnordEqNum,0L);
 autoFunctorPrecedence.assignFixed(HellConst::OrdEqNum,0L);
 for (ulong fn = 0; fn < TheSignature.nextFunctorNumber(); fn++)
  {
   autoFunctorPrecedence.assignArity(fn,TheSignature.arity(fn));
  }; 
 autoFunctorPrecedence.assignAllPrecedences();
 long headerPrecedence;
 for (ulong fn = 0; fn < TheSignature.nextFunctorNumber(); fn++)
  { 
   if (autoFunctorPrecedence.headerPrecedence(true,fn,headerPrecedence))
    {  
     TheSignature.setHeaderPrecedence(1UL,fn,headerPrecedence); // 1UL <=> negative
     ALWAYS(autoFunctorPrecedence.headerPrecedence(false,fn,headerPrecedence));
     TheSignature.setHeaderPrecedence(0UL,fn,headerPrecedence); // 0UL <=> positive
    };
  };
 
 //DF; TheSignature.output(cout) << "\n";

};


//====================================================
