//
//  file atom.cpp
//  implements class Atom
//
//  01/04/2001, Dagstuhl
//  11/04/2001, Manchester, replaceRelationByFunction changed (return 0 if not replaced)
//  09/05/2001 Manchester: occurs ()
//  15/05/2001 Manchester: destroy (), introduceInverse()
//  30/05/2001 Manchester, splitLargeArity/2
//  31/05/2001 Manchester, normalize/0, negativeEqualitySplit/1
//


#include "term.h"
#include "atom.h"


Atom* Atom::replaceRelationByFunction ( PSymbol* pred, FSymbol* fun, int argNo, int* vs )
{
  if ( functor () != pred )
    return 0;

  Term* lhs = args()->nth ( argNo );

  // create the argument list for the rhs
  Term::List* ts = Term::List::empty ();
  for ( int last = args()->length() - 2; last >= 0; last-- ) 
    Term::List::push ( args()->nth(vs[last]), ts );

  Term* rhs = new Term ( fun, ts );

  Term::List* newargs = new Term::List ( rhs );
  Term::List::push ( lhs, newargs );

  args()->destroy();
  delete this;

  return new Atom ( PSymbol::equality(), newargs );
} // Atom::replaceRelationByFunction


void Atom::destroy ()
{

  for ( Term::List* ts = args (); ! ts->isEmpty (); ts = ts->tail () )
    ts->head()->destroy ();

  args()->destroy();

  delete this;
} // Atom::destroy


Var Atom::maxVar ()
{
  Var v = firstVar - 1;

  Term::List::Iterator ts ( args() ); 
  while ( ts.more() ) {
    ts.next()->maxVar (v);
  }

  return v;
} // Atom::maxVar


Atom* Atom::copy () const
{
  return new Atom ( functor(), Term::copy(args()) );
} // Atom::copy 


void Atom::replace ( Var v, Term* t )
{
  for ( Term::List* ts = args(); ts->isNonEmpty(); ts = ts->tail() )
    ts->head()->replace ( v, t );
} // Atom::replace


bool Atom::equal ( Atom* t ) const
{
  return functor() == t->functor() &&
         Term::equal ( args(), t->args() );
} // Atom::equal


// atom has the form P (x1,...,xn) where xi are pairwise different
bool Atom::isLinearAndFlat () const
{
  return Term::sequenceOfUniqueVariables (args());
} // Atom::isLinearAndFlat


Atom* Atom::apply ( Subst& subst, VarSubst& vsubst, int& lastVar ) // copies this
{
  return new Atom ( functor(), Term::apply (args(), subst, vsubst, lastVar) );
} // Atom::apply


void Atom::apply ( Subst& sbst )
{
  for ( Term::List* ts = args(); ts->isNonEmpty(); ts = ts->tail() )
    ts->head()->apply ( sbst );
} // Atom::apply


ostream& operator << ( ostream& str, Atom* t )
{
  str << t->functor();
  if ( ! t->args()->isEmpty() ) 
    str << '(' << t->args() << ')';

  return str;
} // operator << ( ostream&, Atom* f )


void Atom::rename ( VarSubst& subst )
{
  for ( Term::List* ts = args (); ! ts->isEmpty (); ts = ts->tail () )
    ts->head()->rename( subst );
} // Atom::rename


// comparison of two atoms, needed to normalize
Comparison Atom::compare ( const Atom* a ) const 
{ 
  // equality is less than nonequality
  if ( isEquality() ) {
    if ( ! a->isEquality() )
      return LESS;
  }
  else if ( a->isEquality() ) {
      return GREATER;
  }
  // now either both are equality or both are not equality

  switch ( Int::compare (weight(), a->weight()) ) {
    case LESS:
      return LESS;
    case GREATER:
      return GREATER;
    case EQUAL:
      break;
  }

  switch ( Int::compare (arity(), a->arity()) ) {
    case LESS:
      return LESS;
    case GREATER:
      return GREATER;
    case EQUAL:
      break;
  }

  return Term::compare ( args(),a->args() );
}  // Atom::compare


int Atom::weight () const
{
  return 1 + Term::weight ( args() );
} // Atom::weight


int Atom::countOccurrences ( const FSymbol* f ) const
{
  int occ = 0;

  for ( Term::List* ts = args(); ts->isNonEmpty(); ts = ts->tail() ) 
    occ += ts->head()->countOccurrences (f);

  return occ;
} // Atom::countOccurrences


int Atom::countOccurrences ( const Var v ) const
{
  int occ = 0;

  for ( Term::List* ts = args(); ts->isNonEmpty(); ts = ts->tail() ) 
    occ += ts->head()->countOccurrences (v);

  return occ;
} // Atom::countOccurrences


bool Atom::occurs ( const FSymbol* f ) const
{
  return Term::occurs( args(), f );
} // Atom::occurs


// replace the atom
//   f(X,y,Z) = f(X,u,Z)
// by 
//   i(f(X,y,Z),X,Z) = y
void Atom::introduceInverse ( int argNo, FSymbol* inv )
{
  Term* l = args()->head();
  Term* r = args()->second();
  Term* u = Term::List::deleteNth ( r->_args, argNo );
  u->destroy();
  // the atom l=r takes the form f(X,y,Z) = f(X,Z)
  args()->head(new Term (inv, new Term::List (l, r->args())));    // set lhs to i(f(X,y,Z),X,Z)
  delete r;
  args()->tail()->head(new Term(l->args()->nth(argNo)->var())); // set rhs to y
} // Atom::introduceInverse


void Atom::splitLargeArity (const PSymbol* p, FSymbol::List* newFs )
{
  if ( functor() == p ) {
    args ( Term::splitLargeArity (args(),newFs) );
  }
} // Atom::splitLargeArity


void Atom::splitLargeArity (const FSymbol* f, FSymbol::List* newFs )
{
  Term::List::Iterator ts ( args() );
  while ( ts.more() ) {
    ts.next()->splitLargeArity (f,newFs);
  }
} // Atom::splitLargeArity


void Atom::normalize () 
{
  if ( ! isEquality() ) {
    return;
  }

  // equality
  Term::List* rs = args()->tail();
  if ( args()->head()->compare (rs->head()) == LESS ) { // second argument is shorter
    // swap the arguments
    Term* tmp = args()->head();
    args()->head (rs->head());
    rs->head (tmp);
  }
} // Atom::normalize


Atom* Atom::negativeEqualitySplit ( int thresholdWeight )
{
  if ( ! isEquality() )
    return 0;

  Term* l = args()->head();
  Term::List* rs = args()->tail();
  Term* r = rs->head();

  int lw; // weight of l;
  int rw; // weight of l;
  PSymbol* P; // will be used as a new name
  if ( l->isGround() && (lw = l->weight()) >= thresholdWeight ) {
    if ( r->isGround() && (rw = r->weight()) >= thresholdWeight && rw > lw ) { // both splittable but right is greater
      goto split_right;
    }
    // l is used for splitting
    P = PSymbol::createNew ("spl_", 1);
    // changing =(l,r) into P(r)
    _functor = P;
    args()->tail ( Term::List::empty() );
    args()->head (r);
    // create P(l) and return it as a result
    rs->head (l);
    return new Atom (P, rs);
  }
  // l is non-splittable
  else if ( r->isGround() && (rw = r->weight()) >= thresholdWeight )  { // r is splittable
   split_right:
    P = PSymbol::createNew ("spl", 1);
    // changing =(l,r) into P(l)
    _functor = P;
    args()->tail ( Term::List::empty() );
    // create P(r) and return it as a result
    return new Atom (P, rs);
  }
  else {  // none is splittable
    return 0;
  }
} // Atom::negativeEqualitySplit
