//
//  file clause.h 
//  defines class Clause
//
//  28/05/2000, Manchester
//  01/06/2000, Manchester, isDefinition added
//  03/06/2000, Manchester, class Literal moved to a separate file
//  29/03/2001, Machester, Functor changed to Functor* everywhere
//  09/05/2001 Manchester: occurs ()
//  15/05/2001 Manchester: replaceCancellation()
//  22/05/2001 Manchester: isLinearDefinition()
//  27/05/2001 Manchester: occurs (PSymbol*)
//  30/05/2001 Manchester, splitLargeArity/2, negativeEqualitySplit/1
//


#ifndef __clause__
#define __clause__


#include "typedefs.h"
#include "literal.h"
class Term;
class Atom;


class Clause {
 public:

  typedef List<Clause*> List;

  // constructor
  explicit Clause ( Literal::List* ll );

  // structure
  Literal::List* literals () const;
  Clause* copy ();
  
  // deallocation
  void destroy ();

  // output
  void prettyPrint ( int indent, ostream& str ) const;

  // statistics
  int countOccurrences ( const FSymbol* ) const;
  int countOccurrences ( const PSymbol* ) const;
  int countOccurrences ( const Var ) const;
  Var maxVar () const;
  bool isHorn () const;
  bool isUnit () const;
  bool hasEquality () const;
  bool hasNonEquality () const;
  bool hasFunctionSymbols () const;
  bool isPropositional () const;
  bool isTautology () const;
  int length () const;
  int weight () const;

  // properties
  Term* isLinearDefinition ( Term*& rhs ) const; 
  bool isUniquenessAxiom ( PSymbol*& pred, int& argNo ) const; // the clause is P(x,y) & P(x,z) -> y = z
  bool isCancellationAxiom ( FSymbol*& fun, int& argNo ) const; // the clause is f(U,x,Z) = f(U,y,Z) -> x = y
  Var* isFunctionalityAxiom ( const PSymbol* pred, FSymbol*& fun, int argNo, Atom*& ) const; // the clause if P(x,f(x))
  bool becomesUnitEq ( PSymbol*& pred, bool& sign );
  bool occurs ( const PSymbol* pred, bool& sign ) const; // true if predicate pred occurs, sign = polarity of pred's first occurrence
  bool occurs ( const FSymbol* ) const; 
  bool occurs ( const PSymbol* ) const; 
  Term* isDefinition ( Term*& rhs ) const; // clause is a definition of function p
                                           // the result is lhs of the definition

  // miscellaneous
  void replaceCancellation ( int argNo,FSymbol* inv );
  bool replaceRelationByFunction ( PSymbol* pred, FSymbol* fun, int argNo, int* vs );
  int removeVarEquality (); // remove x != t, result is the number of remove equalities
  void split ( FSymbol* tosplit, int maxArity, int iterations, int restArity, 
               FSymbol* firstFun, FSymbol* restFun );
  int maxLitWeight ();
  bool canRemovePredicateDefinition ( PSymbol* pred, bool sign );
  Clause* resolveOn ( const PSymbol* pred, Clause* );
  void normalize ();
  Comparison compare ( const Clause* ) const; // comparison of normalized clauses
  void splitLargeArity (const PSymbol* p, FSymbol::List* newfs );
  void splitLargeArity (const FSymbol* p, FSymbol::List* newfs );
  List* negativeEqualitySplit ( int thresholdWeight );
  void polarities ( const PSymbol* p, int& pol ) const;

 private:
  // structure
  Literal::List* _literals;
  int _weight;

  // auxiliary functions
  bool isEquationalTautology () const;
  bool isNonEquationalTautology () const;

 friend ostream& operator << ( ostream&, const Clause* );
}; // class Clause


inline
Clause::Clause ( Literal::List* ll )
  :
  _literals ( ll ),
  _weight ( -1 ) // to signal that the clause is not normalized
{
  normalize ();
} // Clause::Clause


inline
Literal::List* Clause::literals () const
{
  return _literals;
}


#endif
