/********************************************
 *                                          *
 * Library file command_line.cpp            *
 *                                          *
 * last change: 22/02/2000, Manchester      *
 *                                          *
 *******************************************/

/************************************************************************
 *                                                                      *
 *  class CommandLine                                                   *
 *                                                                      *
 *  constructors:                                                       *
 *                                                                      *
 *    CommandLine::CommandLine ( int argc, char** argv ) -              *
 *      the arguments must be the arguments to main ()                  *
 *                                                                      *
 *  destructors: no                                                     *
 *                                                                      *
 *    the class has no destructors wince it simply uses a reference     *
 *    to argv                                                           *
 *                                                                      *
 *  public:                                                             *
 *                                                                      *
 *    bool findString ( const char* option, char*& value )              *
 *                                                                      *
 *      finds the string following -option in the command line          *
 *      for example, if option is "i", it will find in the              *
 *      command line part "-i <string>" and assign string               *
 *      to value. The result is true if -option was found.              *
 *      Note that if -option was the last string in the command         *
 *      line, the result will still be true, but the empty              *
 *      string will be returned                                         *
 *                                                                      *
 *    bool findInteger ( const char* option, int& value )               *
 *      the same as findString, but finds integer.                      *
 *      However, if -option was the last record, returns false          *
 *                                                                      *
 *    bool findFloat ( const char* option, float& value )               *
 *      the same as findInteger, but finds float                        *
 *                                                                      *
 *  known problems and comments:                                        *
 *                                                                      *
 *    findInteger and findFloat do not check that the                   *
 *    string found really represents a number, but simply               *
 *    use functions atoi and atof to convert the string to              *
 *    a number                                                          *
 *                                                                      *
 *    findSting's value string is not copied, so one should             *
 *    not try to delete it                                              *
 *                                                                      *
 ************************************************************************/
 

#include <string.h>
#include <stdlib.h>

#include "command_line.h"


CommandLine::CommandLine ( int argc, char** argv )
  :
  _noOfOptions ( argc - 1 ),
  _options ( argv + 1 )
{
} // CommandLine::CommandLine


bool CommandLine::findString ( const char* option, char*& value )
{
  static char* emptyString = "";

  for ( int i = _noOfOptions - 1; i >= 0; i-- ) {
    char* current = _options [i];
    if ( current [0] == '-' && // options begin with '-'
         strcmp ( current + 1, option ) == 0 ) { // option found
      
      if ( i == _noOfOptions - 1 ||     // either the last option on the command line
           _options [i+1] [0] == '-' )  // or the next is also an option
        value = emptyString;
      else // has value
        value = _options [i+1];

      return true;
    }
  }
  // option not found
  return false;
} // CommandLine::findString


bool CommandLine::findInteger ( const char* option, int& value )
{
  for ( int i = _noOfOptions - 1; i >= 0; i-- ) {
    char* current = _options [i];
    if ( current [0] == '-' && // options begin with '-'
         strcmp ( current + 1, option ) == 0 ) { // option found
      
      if ( i == _noOfOptions - 1 )     // the last option on the command line
        return false;
      
      // otherwise has value
      value = atoi( _options [i+1] );
      return true;
    }
  }
  // option not found
  return false;
} // CommandLine::findInteger


bool CommandLine::findFloat ( const char* option, float& value )
{
  for ( int i = _noOfOptions - 1; i >= 0; i-- ) {
    char* current = _options [i];
    if ( current [0] == '-' && // options begin with '-'
         strcmp ( current + 1, option ) == 0 ) { // option found
      
      if ( i == _noOfOptions - 1 )     // the last option on the command line
        return false;
      
      // otherwise has value
      value = (float) atof( _options [i+1] );
      return true;
    }
  }
  // option not found
  return false;
} // CommandLine::findInteger

