//
//  file formula.cpp
//  implements class Formula
//
//  15/05/2000 Dallas
//  28/05/2000 Manchester, prettyPrint changed for sequences of AND and OR
//  29/05/2000 Manchester, AND and OR have become of arbitrary arity
//  12/04/2001 Manchester, weight ()
//  03/06/2001 Manchester, sort/1,2, split/5, normalize/0, normalize1/0, compare/1,2
//                isEqualityAxiom/0, isReflexivityAxiom/0, isSymmetryAxiom/0, isTransitivityAxiom/0
//                isFunctionReflexivityAxiom/0, isPredicateReflexivityAxiom/0, skolemize,
//                removeIff/0
//

#include <string.h>


#include "term.h"
#include "atom.h"
#include "formula.h"


const int Formula::oneLineSize = 2;


void Formula::connective ( Connective c )
{
  _con = c;
}


void Formula::atom ( Atom* t )
{
  ASSERT( debug_formula, _con = ATOM );

  _atom = t;
} // Formula::atom


void Formula::left ( Formula* l )
{
  ASSERT ( debug_formula, 
           _con == IFF ||
           _con == XOR ||
           _con == IMP );

  _left = l;
} // Formula::left


void Formula::right ( Formula* r )
{
  ASSERT ( debug_formula, 
           _con == IFF ||
           _con == XOR ||
           _con == IMP );

  _right = r;
} // Formula::right


void Formula::arg ( Formula* a )
{
  ASSERT ( debug_formula, 
           _con == FORALL ||
           _con == EXISTS ||
           _con == NOT );

  _arg = a;
} // Formula::arg


void Formula::args ( List* a )
{
  ASSERT ( debug_formula, 
           _con == AND ||
           _con == OR );

  _args = a;
} // Formula::args


void Formula::vars ( VarList* vs )
{
  ASSERT ( debug_formula, 
           _con == FORALL ||
           _con == EXISTS );

  _vars = vs;
} // Formula::vars


Formula::Formula ( Connective c, Formula* l, Formula* r ) // for binary formulas
  :
  _con ( c )
{
  switch ( c ) {
    case IFF:
    case XOR:
    case IMP:
      _left = l;
      _right = r;
      return;

    case AND:
    case OR:
      if ( r->connective() == c ) {
        if ( l->connective() == c ) { // both r and l are c-junctions
          _args = Formula::List::concat( l->args(), r->args() );
          delete l;
          delete r;
        }
        else { // only r is a c-junction
          _args = new List ( l, r->args() );
          delete r;
        }
      }
      else { // only l is a c-junction
        if ( l->connective() == c ) {
          _args = Formula::List::concat ( l->args(), new List(r) );
          delete l;
        }
        else { // neither l nor r is a c-junction
          _args = new List (l, new List(r));
        }
      }
      return;

    default:
      ASSERT( debug_formula, false );
  }
} // Formula::Formula


Formula::Formula ( Connective c, Formula* a ) // for unary formulas
  :
  _con ( c ),
  _arg ( a )
{
  ASSERT ( debug_formula, 
           c == NOT );
} // Formula::Formula


Formula::Formula ( Connective c, List* fs ) // for and/or formulas
  :
  _con ( c ),
  _args ( fs )
{
  ASSERT ( debug_formula, 
           c == AND || c == OR );
} // Formula::Formula


Formula::Formula ( Connective c, VarList* vs, Formula* a ) // for quantified formulas
  :
  _con ( c ),
  _arg ( a ),
  _vars ( vs )
{
  ASSERT ( debug_formula, 
           c == FORALL ||
           c == EXISTS );
} // Formula::Formula


Formula::Formula ( Atom* t ) // for atomic formulas
  :
  _con ( ATOM ),
  _atom ( t )
{
} // Formula::Formula


VarList* Formula::vars () const
{
  ASSERT ( debug_formula, 
           _con == FORALL ||
           _con == EXISTS );

  return _vars;
} // Formula::vars 


Formula* Formula::arg () const
{
  ASSERT ( debug_formula, 
           _con == FORALL ||
           _con == EXISTS ||
           _con == NOT );

  return _arg;
} // Formula::arg


Formula::List* Formula::args () const
{
  ASSERT ( debug_formula, 
           _con == AND ||
           _con == OR );

  return _args;
} // Formula::arg


Formula* Formula::left () const
{
  ASSERT ( debug_formula, 
           _con == IFF ||
           _con == XOR ||
           _con == IMP );

  return _left;
} // Formula::left


Formula* Formula::right () const
{
  ASSERT ( debug_formula, 
           _con == IFF ||
           _con == XOR ||
           _con == IMP );

  return _right;
} // Formula::right


Atom* Formula::atom () const
{
  ASSERT ( debug_formula, 
           _con == ATOM );

  return _atom;
} // Formula::atom


Formula::Connective Formula::connective () const
{
  return _con;
} // Formula::connective


int Formula::size () const
{
  switch ( connective () ) {
    case ATOM: 
      return 1;

    case AND:
    case OR: {
      int sz = -1;
      
      for ( List* fs = args(); fs->isNonEmpty (); fs = fs->tail() )
        sz += fs->head()->size() + 1;
         
      return sz;
    }
        
    case IMP:
    case IFF:
    case XOR:
      return left()->size() + right()->size () + 1;

    case NOT:
      return arg()->size () + 1;

    case FORALL:
    case EXISTS:
      return vars()->length() + arg()->size () + 1;

    default:
      ASSERT ( debug_formula, false );
      return 0;
  }
} // Formula::size


int Formula::weight () const
{
  switch ( connective () ) {
    case ATOM: 
      return atom()->weight();

    case AND:
    case OR: {
      int sz = -1;
      
      for ( List* fs = args(); fs->isNonEmpty (); fs = fs->tail() )
        sz += fs->head()->size() + 1;
         
      return sz;
    }
        
    case IMP:
    case IFF:
    case XOR:
      return left()->size() + right()->size () + 1;

    case NOT:
      return arg()->size () + 1;

    case FORALL:
    case EXISTS:
      return vars()->length() + arg()->size () + 1;

    default:
      ASSERT ( debug_formula, false );
      return 0;
  }
} // Formula::weight


char* Formula::printName ( Connective c )
{
  switch ( c ) {
    case AND:
      return "&";
    case OR:
      return "|";
    case IMP:
      return "=>";
    case NOT:
      return "~";
    case IFF:
      return "<=>";
    case XOR:
      return "<~>";
    case FORALL:
      return "!";
    case EXISTS:
      return "?";
    default:
      return "";
  }
} // Formula::printName


ostream& operator << ( ostream& str, VarList* vs )
{
  ASSERT( debug_formula,
          ! vs->isEmpty () );

  VarList* tail = vs ->tail ();
  
  while ( ! tail->isEmpty () ) {
    str << 'X' << vs->head () << ',';
    vs = tail;
    tail = tail->tail ();
  }

  str << Term::varPrefix << vs->head ();

  return str;
} // operator << ( ostream& str, VarList* vs )


ostream& operator << ( ostream& str, Formula* f )
{
  f->print ( str );

  return str;
} // operator << ( ostream& str, Formula* f )


void Formula::print ( ostream& str )
{
  print ( ATOM, str );
}


void Formula::prettyPrint ( int indent, ostream& str )
{
  for ( int i = indent; i > 0; i-- )
    str << ' ';

  prettyPrint ( ATOM, indent, str );
} // Formula::prettyPrint


void Formula::print ( Connective out, ostream& str )
{
  bool par = false; // parentheses required

  switch ( connective () ) {
    case ATOM: 
      
      str << atom ();
      return;

    case AND:
      switch ( out ) {
        case OR:
        case NOT:
          par = true;
        default:
          break;
      }
      goto and_or;

    case OR: {
      switch ( out ) {
        case AND:
        case NOT:
          par = true;
        default:
          break;
      }

      and_or:
      if ( par ) str << '(';

      List* fs = args ();
      List* next = fs->tail ();

      while ( next->isNonEmpty () ) {
        fs->head()->print ( connective(), str );
        str << ' ' << printName ( connective () );
        fs = next;
        next = next->tail ();
      }

      str << ' ';
      fs->head()->print ( connective(), str );

      if ( par ) str << ')';
      return;
    }

    case IMP:
      switch ( out ) {
        case AND:
        case OR:
        case IMP:
        case NOT:
          par = true;
        default:
          break;
      }

      binary:
      if ( par ) str << '(';
      left()->print ( connective(), str );
      str << ' ' << printName ( connective () ) << ' ';
      right()->print ( connective(), str );
      if ( par ) str << ')';
      return;

    case NOT:

      str << printName ( NOT ) << ' ';
      arg()->print ( NOT, str );
      return;

    case IFF:
    case XOR:

      if ( out != ATOM ) 
        par = true;
      goto binary;

    case FORALL:
    case EXISTS:

      str << printName ( connective () ) << " [" << vars() << "] : ";
      arg()->print ( NOT, str );
      return;

    default:
      ASSERT ( debug_formula, false );
  }
} // Formula::print ( Connective out, ostream& str )


void Formula::prettyPrint ( Connective out, int indent, ostream& str )
{
  // short formulas are printed on one line
  // for VERY long formulas the call of length () may be inefficient
  if ( size () <= Formula::oneLineSize ) {
    print ( out, str );
    return;
  }

  bool par = false; // parentheses required

  int i; // indentation counter

  switch ( connective () ) {
    case AND:
      switch ( out ) {
        case OR:
        case NOT:
          par = true;
        default:
          break;
      }

      andor: {
        if ( par ) {
          str << "( ";
          List* fs = args();
          fs->head()->prettyPrint (connective(), indent + 2, str );
          fs = fs->tail();
        
          while ( fs->isNonEmpty() ) {
            str << '\n';
            for ( int i = indent - strlen (printName(connective())); i > 0; i-- )
              str << ' ';

            str << printName ( connective() ) << ' ';
            fs->head()->prettyPrint ( connective(), indent + 2, str );
            fs = fs->tail();
          }
          str << " )";
        }
        else {
          List* fs = args();
          fs->head()->prettyPrint ( connective(), indent, str );
          fs = fs->tail();
        
          while ( fs->isNonEmpty() ) {
            str << '\n';
            for ( int i = indent - 1; i > 0; i-- )
              str << ' ';

            str << printName ( connective() ) << ' ';
            fs->head()->prettyPrint ( connective(), indent + strlen (printName(connective())) + 1, str );
            fs = fs->tail();
          }
        }
      }
      return;

    case OR:
      switch ( out ) {
        case AND:
        case NOT:
          par = true;
        default:
          goto andor;
      }

    case IMP:
      switch ( out ) {
        case AND:
        case OR:
        case IMP:
        case NOT:
          par = true;
        default:
          break;
      }

      binary:
      // first, print left subformula
      if ( par ) {
        str << " (";
        left()->prettyPrint ( connective(), indent+2, str );
      }
      else {
        str << " ";
        left()->prettyPrint ( connective(), indent+1, str );
      }
      str << '\n';

      // then print the indentation and the connective
      for ( i = indent; i > 0; i-- )
        str << ' ';
      str << printName ( connective () ) << ' ';
      if ( out == connective () &&
           (out == AND || out == OR) )
        right()->prettyPrint ( connective(), indent, str );
      else 
        right()->prettyPrint ( connective(), indent + 2 + strlen (printName (connective())), str );
      if ( par ) str << ')';
      return;

    case NOT:

      str << printName ( NOT ) << ' ';
      arg()->prettyPrint ( NOT, indent + strlen (printName(NOT)) + 1, str );
      return;

    case IFF:
    case XOR:

      if ( out != ATOM ) 
        par = true;
      goto binary;

    case FORALL:
    case EXISTS:

      str << printName ( connective () ) << " [" << vars() << "] :\n";
      for ( i = indent; i >= 0; i-- )
        str << ' ';
      arg()->prettyPrint ( NOT, indent + 1, str );
      return;

    default:
      ASSERT ( debug_formula, false );
  }
} // Formula::print ( Connective out, ostream& str )


void Formula::rename ()
{
  VarSubst sbst;
  Var last = firstVar;

  rename ( sbst, last );
} // Formula::rename


void Formula::rename ( VarSubst& sbst, Var& last )
{
  switch ( connective () ) {
    case ATOM:

      atom()->rename ( sbst );
      return;

    case AND:
    case OR: {
      for ( List* fs = args(); fs->isNonEmpty(); fs = fs->tail() ) 
        fs->head()->rename ( sbst, last );
      return;
    }

    case IMP:
    case IFF:
    case XOR:

      left()->rename ( sbst, last );
      right()->rename ( sbst, last );
      return;

    case NOT:

      arg()->rename ( sbst, last );
      return;

    case FORALL:
    case EXISTS: {
      VarList* vs = vars ();
      VarSubst::Mark mark = sbst.mark ();

      while ( ! vs->isEmpty () ) {
        last++;
        // add binding of last to the head of the list
        sbst.bind ( vs->head (), last );
        // change the head to last
        vs->head ( last );
        vs = vs->tail ();
      }

      // rename the rest
      arg()->rename ( sbst, last );
      // and restore the substitution
      sbst.restore ( mark );
      return;
    }

    default:
      ASSERT( debug_formula, false );
  }
} // Formula::rename


bool Formula::occurs ( const PSymbol* f ) const
{
  switch ( connective () ) {
    case ATOM: 
      return f == atom()->functor ();

    case AND:
    case OR: {
      for ( List* fs = args(); fs->isNonEmpty(); fs = fs->tail() ) {
        if ( fs->head()->occurs(f) )
          return true;
      }
      return false;
    }

    case IMP:
    case IFF:
    case XOR:
      return left()->occurs ( f )
          || right()->occurs ( f );

    case NOT:
    case FORALL:
    case EXISTS:
      return arg()->occurs ( f );

    default:
      ASSERT ( debug_formula, false );
      return false;
  }
} // Formula::predOccurs


Atom* Formula::isPredicateDefinition ( Formula*& rhs )
{
  Formula* f = this;

  // skip all universal quantifiers in front of the formula
  while ( f->connective() == FORALL ) {
    f = f->arg ();
  }

  if ( f->connective() != IFF ) 
    return 0;

  Formula* lft = f->left ();
  Formula* rht = f->right ();

  if ( lft->isLinearAndFlat () ) {
    Atom* a = lft->atom ();
    if ( a->isEquality() || rht->occurs (a->functor()) ) {
      return 0;
    }
    rhs = rht;
    // cout << "Definition of " << a->functor () << " found\n";
    return a;
  }
  if ( rht->isLinearAndFlat () ) {
    Atom* a = rht->atom ();
    if ( a->isEquality() || lft->occurs (a->functor()) ) {
      return 0;
    }
    rhs = lft;
    // cout << "Definition of " << a->functor () << " found\n";
    return a;
  }

  return 0;
} // Formula::isPredicateDefinition


bool Formula::isLinearAndFlat ()
{
  return ( connective () == ATOM ) 
      && atom()->isLinearAndFlat ();      
} // Formula::isLinearAndFlat



void Formula::destroy ()
{
  switch ( connective () ) {
    case ATOM:
      atom()->destroy ();
      break;

    case AND:
    case OR: {
      for ( List* fs = args(); fs->isNonEmpty(); fs = fs->tail() ) 
        fs->head()->destroy();

      args()->destroy();
      break;
    }

    case IMP:
    case IFF:
    case XOR:
      left()->destroy();
      right()->destroy();
      break;

    case NOT:
      arg()->destroy ();
      break;

    case FORALL:
    case EXISTS:
      vars()->destroy ();
      arg()->destroy ();
      break;
  }; 
  
  delete this;
} // Formula::destroy


//
// flattens the formula 
//  flattening removes multiple applications of 
//  quantifiers, and, or and double negations
void Formula::flatten ()  
{
  switch ( connective() ) {
    case ATOM:
      return;

    case AND:
    case OR: 
      args( flatten(args(), connective()) );
      return;

    case IMP:
    case IFF:
    case XOR:
      left()->flatten();
      right()->flatten();
      return;

    case NOT:
      arg()->flatten();
      if ( arg()->connective() == NOT ) {
        // transformation not not A -> A
        Formula* nf = arg ();
        copyFrom ( nf->arg() );
        delete nf->arg();
        delete nf;
      }
      // arg does not begin with NOT
      return;

    case FORALL:
    case EXISTS:
      arg()->flatten();
      if ( arg()->connective() == connective() ) {
        // transformation (forall x)(forall y)A -> (forall xy)A
        Formula* ar = arg();
        vars ( VarList::concat (vars(), ar->vars()) );
        copyFrom (ar);
        delete ar;
      }
      return;

    default:
      ASSERT( debug_formula, false );
      return;
  }
} // Formula::flatten


Formula::List* Formula::flatten ( List* fs, Connective c )
{
  if ( fs->isEmpty () ) 
    return fs;
  
  // first, flatten the tail
  fs->tail ( flatten(fs->tail(), c) );
  // then flatten the head
  fs->head()->flatten();

  if ( fs->head()->connective() == c ) {  
    // apply (A1 & ... & An ) & B1 & ... & Bm -> A1 & ... & An & B1 & ... Bm
    Formula* hd = fs->head ();
    List* result = Formula::List::concat ( fs->tail(), hd->args() );
    delete hd;
    delete fs;
    return result;
  }
  
  return fs;
} // Formula::flatten ( List*, Connective)


Formula* Formula::copy ()
{
  switch ( connective() ) {
    case ATOM:
      return new Formula ( atom()->copy() );
      
    case AND:
    case OR:
      return new Formula ( connective(), copy(args ()) );

    case IMP:
    case IFF:
    case XOR:
      return new Formula ( connective(), left()->copy(), right()->copy() );

    case NOT:
      return new Formula ( NOT, arg()->copy() );

    case FORALL:
    case EXISTS:
      return new Formula ( connective(), vars()->copy(), arg()->copy() );
    
    default:
      ASSERT( debug_formula, false );
  }
} // Formula::copy


Formula::List* Formula::copy ( List* fs )
{
  if ( fs->isEmpty() ) 
    return Formula::List::empty ();
  
  return new List ( fs->head()->copy(), copy(fs->tail()) );
} // Formula::copy ( List* fs )


int Formula::countOccurrences ( const FSymbol* f ) const
{
  switch ( connective() ) {
    case ATOM:
      return atom()->countOccurrences(f);

    case AND:
    case OR: {
      int occ = 0;

      for ( List* fs = args(); fs->isNonEmpty(); fs = fs->tail() )
        occ += fs->head()->countOccurrences ( f );

      return occ;
    }

    case IMP:
    case IFF:
    case XOR:
      return left()->countOccurrences(f) + 
             right()->countOccurrences(f);

    case NOT:
    case FORALL:
    case EXISTS:
      return arg()->countOccurrences(f);

    default:
      ASSERT ( debug_formula, false );
      return 0;
  }
}


int Formula::countOccurrences ( const PSymbol* p ) const
{
  switch ( connective() ) {
    case ATOM:
      return atom()->functor() == p ? 1 : 0;

    case AND:
    case OR: {
      int occ = 0;

      for ( List* fs = args(); fs->isNonEmpty(); fs = fs->tail() )
        occ += fs->head()->countOccurrences ( p );

      return occ;
    }

    case IMP:
    case IFF:
    case XOR:
      return left()->countOccurrences(p) + 
             right()->countOccurrences(p);

    case NOT:
    case FORALL:
    case EXISTS:
      return arg()->countOccurrences(p);

    default:
      ASSERT ( debug_formula, false );
      return 0;
  }
}

// changed on 11/06/2001 to make more compact normal forms
Clause::List* Formula::skolemize ( bool sign )
{
  removeIff ( sign ? AND : OR );
  rename ();
  flatten ();

  nnf (true);
  flatten ();

  Clause::List* result = Clause::List::empty ();
  Subst sbst;
  Formula::List* pf = Formula::List::empty ();
  Formula::List* nf = Formula::List::empty ();

  Formula::List::push ( this, sign ? pf : nf );

  skolemize ( Atom::List::empty(), Atom::List::empty(), pf, nf, VarList::empty (), sbst, result );
  pf->destroy ();
  nf->destroy ();
  return result;

} // Formula::skolemize


void Formula::skolemize ( Atom::List* pa, Atom::List* na, List* pf, List* nf, VarList* univ,
                          Subst& sbst, Clause::List*& result ) 
{
  if ( ! pf->isEmpty() ) {
    Formula* f = pf->head ();
    pf = pf->tail ();
    switch ( f->connective() ) {
      case ATOM: {
        Atom::List newpa (f->atom(),pa);
        skolemize (&newpa, na, pf, nf, univ, sbst, result);
        return;
      }

      case AND: {
        Formula::List::Iterator fs (f->args());
        while ( fs.more() ) {
          Formula::List newpf (fs.next(), pf);
          skolemize (pa, na, &newpf, nf, univ, sbst, result);
        }
        return;
      }

      case OR: {
        Formula::List* newpf = f->args()->append(pf);
        skolemize (pa, na, newpf, nf, univ, sbst, result);
        // delete newpf down to pf
        while ( newpf != pf ) {
          List* tmp = newpf->tail ();
          delete newpf;
          newpf = tmp;
        }
        return;
      }

      case IMP: {
        Formula::List newpf (f->right(), pf);
        Formula::List newnf (f->left(), nf);
        skolemize (pa, na, &newpf, &newnf, univ, sbst, result);
        return;
      }

      case IFF: 
      case XOR: 
        ASSERT ( debug_formula, false );
        return;

      case NOT: {
        // f is ~A, adding A to nf
        Formula::List newnf (f->arg(), nf);
        skolemize (pa, na, pf, &newnf, univ, sbst, result);
        return;
      }

      case FORALL: {
        VarList* newuniv = f->vars()->append(univ);
        Formula::List newpf (f->arg(), pf);
        skolemize (pa, na, &newpf, nf, newuniv, sbst, result);
        // delete newuniv down to univ
        while ( newuniv != univ ) {
          VarList* tmp = newuniv->tail ();
          delete newuniv;
          newuniv = tmp;
        }
        return;          
      }

      case EXISTS: {
        // create new sequence of skolem terms, add them to subst
        // and remember in sk
        Term::List* sk = Term::List::empty ();
        VarList::Iterator vs ( f->vars() );
        while ( vs.more() ) {
          Var v = vs.next ();
          Term* s = Term::createSkolemTerm (univ);
          Term::List::push (s,sk);
          sbst.bind (v,s);
        }

        // now skolemize with the new substitution
        Formula::List newpf (f->arg(), pf);
        skolemize (pa, na, &newpf, nf, univ, sbst, result);

        // destroy all Skolem terms and the list sk
        Term::List::Iterator ts ( sk );
        while ( ts.more() ) {
          ts.next()->destroy();
        }
        sk->destroy();
        return;
      }
    }
  }

  // pf is empty
  if ( ! nf->isEmpty() ) {
    Formula* f = nf->head ();
    nf = nf->tail ();
    switch ( f->connective() ) {
      case ATOM: {
        Atom::List newna (f->atom(),na);
        skolemize (pa, &newna, pf, nf, univ, sbst, result);
        return;
      }

      case AND: {
        Formula::List* newnf = f->args()->append(nf);
        skolemize (pa, na, pf, newnf, univ, sbst, result);
        // delete newnf down to nf
        while ( newnf != nf ) {
          List* tmp = newnf->tail ();
          delete newnf;
          newnf = tmp;
        }
        return;
      }

      case OR: {
        Formula::List::Iterator fs (f->args());
        while ( fs.more() ) {
          Formula::List newnf (fs.next(), nf);
          skolemize (pa, na, pf, &newnf, univ, sbst, result);
        }
        return;
      }

      case IMP: {
        Formula::List newnf (f->right(), nf);
        skolemize (pa, na, pf, &newnf, univ, sbst, result);
        Formula::List newpf (f->left(), pf);
        skolemize (pa, na, &newpf, nf, univ, sbst, result);
        return;
      }

      case IFF: 
      case XOR: 
        ASSERT ( debug_formula, false );
        return;

      case NOT: {
        // f is ~A, adding A to pf
        Formula::List newpf (f->arg(), pf);
        skolemize (pa, na, &newpf, nf, univ, sbst, result);
        return;
      }

      case FORALL: {
        // create new sequence of skolem terms, add them to subst
        // and remember in sk
        Term::List* sk = Term::List::empty ();
        VarList::Iterator vs ( f->vars() );
        while ( vs.more() ) {
          Term* s = Term::createSkolemTerm (univ);
          Term::List::push (s,sk);
          Var v = vs.next ();
          sbst.bind (v,s);
        }

        // now skolemize with the new substitution
        Formula::List newnf (f->arg(), nf);
        skolemize (pa, na, pf, &newnf, univ, sbst, result);

        // destroy all Skolem terms and the list sk
        Term::List::Iterator ts ( sk );
        while ( ts.more() ) {
          ts.next()->destroy();
        }
        sk->destroy();
        return;
      }

      case EXISTS: {
        VarList* newuniv = f->vars()->append(univ);
        Formula::List newnf (f->arg(), nf);
        skolemize (pa, na, pf, &newnf, newuniv, sbst, result);
        // delete newuniv down to univ
        while ( newuniv != univ ) {
          VarList* tmp = newuniv->tail ();
          delete newuniv;
          newuniv = tmp;
        }
        return;          
      }
    }
  }

  // pf and nf are both empty
  // generate a clause
  Literal::List* clause = Literal::List::empty ();
  // push in clause the collected negative atoms as negative literals
  Atom::List::Iterator nls (na);
  while ( nls.more() ) {
    Atom* a = nls.next()->copy ();
    a->apply (sbst);
    Literal::List::push ( new Literal (false,a), clause );
  }
  // push in clause the collected positive atoms as positive literals
  Atom::List::Iterator pls (pa);
  while ( pls.more() ) {
    Atom* a = pls.next()->copy ();
    a->apply (sbst);
    Literal::List::push ( new Literal (true,a), clause );
  }

  Clause::List::push ( new Clause(clause), result );
} // Formula::skolemize


void Formula::copyFrom ( const Formula* from )
{
  _con = from->_con;
  _left = from->_left;
  _right = from->_right;
} // Formula::copyFrom


void Formula::normalize ()
{
  rename ();  // to make flattening safe (no double vars in the prefix)
  flatten ();

  // reorder subformulas using compare
  normalize1 ();
  // rename ();
}


// comparison of normalized formulas
Comparison Formula::compare ( const Formula* f ) const 
{
  if ( connective () < f->connective () ) 
    return LESS;
  if ( connective () > f->connective () ) 
    return GREATER;

  // now connectives are the same
  switch ( connective() ) {
    case ATOM:
      return atom()->compare (f->atom());

    case AND:
    case OR:
      return compare (args(),f->args());

    case IMP:
    case IFF:
    case XOR:
      // first compare left subformulas and then right
      switch ( left()->compare(f->left()) ) {
        case LESS:
          return LESS;
        case GREATER:
          return GREATER;
        case EQUAL:
          return right()->compare(f->right());
      }

    case NOT:
      return arg()->compare(f->arg());

    case FORALL:
    case EXISTS:
      // first compare the length of the variable prefix, 
      //  and then the immediate subformulas
      switch ( Int::compare (vars()->length(),f->vars()->length()) ) {
        case LESS:
          return LESS;
        case GREATER:
          return GREATER;
        case EQUAL:
          return arg()->compare(f->arg());
      }
    default:
      ASSERT (debug_formula,false);
      return LESS;
  }
} // Formula::compare


// comparison of arguments of normalized formulas
Comparison Formula::compare ( const List* fs1, const List* fs2 ) 
{
  Formula::List::Iterator f1 (const_cast<List*>(fs1));
  Formula::List::Iterator f2 (const_cast<List*>(fs2));

  while ( f1.more() ) {
    if ( f2.more() ) {
      switch ( f1.next()->compare (f2.next()) ) {
        case GREATER:
          return GREATER;
        case LESS:
          return LESS;
        case EQUAL:
          break;
      }
    }
    else { 
      return GREATER;
    }
  }
  // f1 is empty
  if ( f2.more() )
    return LESS;

  return EQUAL;
} // Formula::compare

// change the order of subformulas using compare
void Formula::normalize1 ()
{
  switch ( connective() ) {
    case ATOM:
      atom()->normalize();
      return;

    case AND:
    case OR: {
      // normalize immediate subformulas first
      Formula::List::Iterator fs ( args() );
      while ( fs.more() ) {
        fs.next()->normalize1 ();;
      }
      // then reorder them
      args ( sort(args()) );
      return;
    }
      

    case IMP:
      left()->normalize1 (); 
      right()->normalize1 (); 
      return;

    case IFF:
    case XOR:
      left()->normalize1 (); 
      right()->normalize1 (); 
      if ( left()->compare(right()) == GREATER ) {
        // swap subformulas
        Formula* tmp = left();
        left (right());
        right (tmp);
      }
      return;

    case NOT:
    case FORALL:
    case EXISTS:
      arg()->normalize1 ();
      return;

    default:
      ASSERT (debug_formula,false);
      return;
  }
} // Formula::normalize1


// 02/06/2001, copied from Clause
Formula::List* Formula::sort ( List* toSort )
{
  return sort ( toSort, Formula::List::empty() );
} // Formula::sort


// 02/06/2001, copied from Clause
Formula::List* Formula::sort ( List* toSort, List* appendTo )
{
  if ( toSort->isEmpty() ) 
    return appendTo;

  if ( toSort->tail()->isEmpty() ) {
    toSort->tail ( appendTo );
    return toSort;
  }

  List* less;
  List* equal;
  List* greater;

  split ( toSort->head(), toSort->tail(), less, equal, greater );
  // sort greater and append to appendTo
  greater = sort ( greater, appendTo );
  // concatenate equal with it
  equal = Formula::List::concat ( equal, greater );
  // push head of list
  toSort->tail ( equal );
  // sort less and append to it
  return sort ( less, toSort );
} // Formula::sort


// 02/06/2001, copied from Clause
void Formula::split ( Formula* l, List* spl, 
                      List*& less, List*& equal, List*& greater )
{
  ASSERT ( debug_clause, spl->isNonEmpty() );

  less = Formula::List::empty();
  equal = Formula::List::empty();
  greater = Formula::List::empty();

  for (;;) {
    List* tail = spl->tail ();

    switch ( spl->head()->compare(l) ) {
      case GREATER:
        spl->tail ( greater );
        greater = spl;
        break;
      case EQUAL:
        spl->tail ( equal );
        equal = spl;
        break;
      case LESS:
        spl->tail ( less );
        less = spl;
        break;
    }

    if ( tail->isEmpty() )
      return;
    spl = tail;
  }
} // Formula::split


// 03/06/2001
bool Formula::isReflexivityAxiom () const
{
  if ( connective() != FORALL ) {
    return false;
  }

  // f must be x = x
  Formula* f = arg ();
  if ( f->connective() != ATOM ) {
    return false;
  }

  // a must be x = x
  Atom* a = f->atom ();
  if ( ! a->isEquality() )
    return false;

  Term* t1 = a->args()->head();
  // t1 must be x
  if ( ! t1->isVar() )
    return false;

  Term* t2 = a->args()->second();
  // t2 must be x
  if ( ! t2->isVar() )
    return false;

  return t1->var() == t2->var();
} // Formula::isReflexivityAxiom


// 03/06/2001
bool Formula::isSymmetryAxiom () const
{
  if ( connective() != FORALL ) {
    return false;
  }

  // f must be x = y -> y = x
  Formula* f = arg ();
  if ( f->connective() != IMP ) {
    return false;
  }

  // f1 must be x = y 
  Formula* f1 = f->left ();
  if ( f1->connective() != ATOM ) {
    return false;
  }
  // f2 must be y = x
  Formula* f2 = f->right ();
  if ( f2->connective() != ATOM ) {
    return false;
  }

  // a1,a2 must be x = y
  Atom* a1 = f1->atom ();
  if ( ! a1->isEquality() )
    return false;
  Atom* a2 = f2->atom ();
  if ( ! a2->isEquality() )
    return false;

  // l1,r1 must be x,y
  Term* l1 = a1->args()->head();
  if ( ! l1->isVar() )
    return false;
  Term* r1 = a1->args()->second();
  if ( ! r1->isVar() )
    return false;
  // l1, r1 must be different
  if ( l1->var() == r1->var() )
    return false;

  // l2,r2 must be y,x
  Term* l2 = a2->args()->head();
  if ( ! l2->isVar() || l2->var() != r1->var() )
    return false;
  Term* r2 = a2->args()->second();
  return r2->isVar() && r2->var() == l1->var();
} // Formula::isSymmetryAxiom


// 03/06/2001
bool Formula::isTransitivityAxiom () const
{
  if ( connective() != FORALL ) {
    return false;
  }

  // f must be x = y & y = z -> y = x
  Formula* f = arg ();
  if ( f->connective() != IMP ) {
    return false;
  }

  // c must be x = y & y = z
  Formula* c = f->left ();
  if ( c->connective() != AND || c->args()->length() != 2 ) {
    return false;
  }

  // c1,c2 must be x = y and y = z
  Formula* c1 = c->args()->head ();
  if ( c1->connective() != ATOM ) {
    return false;
  }
  Formula* c2 = c->args()->second ();
  if ( c2->connective() != ATOM ) {
    return false;
  }
  // c3 must be x = z
  Formula* c3 = f->right();
  if ( c3->connective() != ATOM ) {
    return false;
  }

  // xy,yz, xz must be x = y, y = z, and x = z
  Atom* xy = c1->atom ();
  if ( ! xy->isEquality() )
    return false;
  Atom* yz = c2->atom ();
  if ( ! yz->isEquality() )
    return false;
  Atom* xz = c3->atom ();
  if ( ! xz->isEquality() )
    return false;

  // x, y, and z must be x,y, and z
  Term* x = xy->args()->head();
  if ( ! x->isVar() )
    return false;
  Term* y = xy->args()->second();
  if ( ! y->isVar() )
    return false;
  // vx, vy, and vz are variables x,y,z
  Var vx = x->var ();
  Var vy = y->var ();
  if ( vx == vy )
    return false;
  Term* z = yz->args()->second();
  if ( ! z->isVar() )
    return false;
  Var vz = z->var();
  if ( vx == vz || vy == vz )
    return false;
  
  // check the rest of yz
  Term* y1 = yz->args()->head();
  if ( ! y1->isVar() || y1->var() != vy )
    return false;
  
  // check xz
  Term* x1 = xz->args()->head();
  if ( ! x1->isVar() || x1->var() != vx )
    return false;
  Term* z1 = xz->args()->second();
  
  return z1->isVar() && z1->var() == vz;
} // Formula::isTransitivityAxiom


// 03/06/2001
bool Formula::isFunctionReflexivityAxiom () const
{
  if ( connective() != FORALL ) {
    return false;
  }

  // f must be x = y -> F(x)=F(y)
  Formula* f = arg ();
  if ( f->connective() != IMP ) {
    return false;
  }

  // both subformulas must be atoms
  if ( f->left()->connective() != ATOM || 
       f->right()->connective() != ATOM ) {
    return false;
  }
  Atom* xy = f->left()->atom(); // x=y
  if ( ! xy->isEquality() )
    return false;
  Term* x = xy->args()->head();
  Term* y = xy->args()->second();
  if ( ! x->isVar() || ! y->isVar() )
    return false;

  Var vx = x->var();
  Var vy = y->var();
  // variables must be different
  if ( vx == vy ) 
    return false;

  Atom* FxFy = f->right()->atom(); // F(x) = F(y)
  if ( ! FxFy->isEquality() )
    return false;
  Term* Fx = FxFy->args()->head();
  Term* Fy = FxFy->args()->second();

  if ( Fx->functor() != Fy->functor() )
    return false;
  
  return Term::equalUpTo (Fx->args(),Fy->args(),vx,vy);
} // Formula::isFunctionReflexivityAxiom


// 03/06/2001
bool Formula::isPredicateReflexivityAxiom () const
{
  if ( connective() != FORALL ) {
    return false;
  }

  // f must be x = y & P(x) -> P(y)
  Formula* f = arg ();
  if ( f->connective() != IMP ) {
    return false;
  }

  // c must be x = y & P(x) (but in unclear order)
  Formula* c = f->left ();
  if ( c->connective() != AND || c->args()->length() != 2 ) {
    return false;
  }
  Formula* c1 = c->args()->head();
  if ( c1->connective() != ATOM ) {
    return false;
  }
  Formula* c2 = c->args()->second();
  if ( c2->connective() != ATOM ) {
    return false;
  }

  Formula* fPy = f->right (); // P(y)
  if ( fPy->connective() != ATOM ) {
    return false;
  }
  Atom* Py = fPy->atom();
  if ( Py->isEquality() )
    return false;

  Atom* xy; // x = y
  Atom* Px; // P(x)
  if ( c1->atom()->isEquality() ) {
    if ( c2->atom()->isEquality() ) {
      return false;
    }
    xy = c1->atom ();
    Px = c2->atom ();
  }
  // c1 is not equality
  else if ( c2->atom()->isEquality() ) {
    xy = c2->atom ();
    Px = c1->atom ();
  }
  else {  // c1,c2 are both not equalities
    return false;
  }

  Term* x = xy->args()->head();
  Term* y = xy->args()->second();
  if ( ! x->isVar() || ! y->isVar() )
    return false;

  Var vx = x->var();
  Var vy = y->var();
  // variables must be different
  if ( vx == vy ) 
    return false;

  if ( Px->functor() != Py->functor() )
    return false;
  
  return Term::equalUpTo (Px->args(),Py->args(),vx,vy);
} // Formula::isPrediacateReflexivityAxiom


// 03/06/2001
bool Formula::isEqualityAxiom () const
{
  return isReflexivityAxiom () || 
         isSymmetryAxiom () ||
         isTransitivityAxiom () ||
         isFunctionReflexivityAxiom () ||
         isPredicateReflexivityAxiom ();
} // Formula::isEqualityAxiom


// 03/06/2001
void Formula::removeIff () 
{
  switch ( connective() ) {
    case ATOM:
      return;

    case AND:
    case OR: {
      Formula::List::Iterator fs (args());
      while ( fs.more() ) {
        fs.next()->removeIff ();
      }
      return;
    }

    case IMP:
      left()->removeIff();
      right()->removeIff();
      return;

    case IFF: {
      left()->removeIff();
      right()->removeIff();
      // using (A <-> B) <=> (A -> B) & (B -> A)
      List* newarg = new List (new Formula (IMP, left(), right()),
                     new List (new Formula (IMP, right()->copy(), left()->copy())) );
      connective (AND);
      args (newarg);
      return;
    }
    case XOR: {
      left()->removeIff();
      right()->removeIff();
      // using (A xor B) <=> ~(A <-> B)
      Formula* newarg = new Formula ( AND,
                                      new List (new Formula (IMP, left(), right()),
                                      new List (new Formula (IMP, right()->copy(), left()->copy())) ) );
      connective (NOT);
      arg (newarg);
      return;
    }

    case NOT:
    case FORALL:
    case EXISTS:
      arg()->removeIff();
      return;
  }
} // Formula::removeIff


// 11/06/2001
// remove IFF and XOR, c is the external connective (AND or OR)
void Formula::removeIff ( Connective c ) 
{
  switch ( connective() ) {
    case ATOM:
      return;

    case AND:
    case OR: {
      Formula::List::Iterator fs (args());
      while ( fs.more() ) {
        fs.next()->removeIff (connective());
      }
      return;
    }

    case IMP:
      left()->removeIff(AND);
      right()->removeIff(OR);
      return;

    case IFF: 
      left()->removeIff(c);
      right()->removeIff(c);
      if ( c == AND ) {
        // using (A <-> B) <=> (A -> B) & (B -> A)
        List* newarg = new List (new Formula (IMP, left(), right()),
                       new List (new Formula (IMP, right()->copy(), left()->copy())) );
        connective (AND);
        args (newarg);
      }
      else { // C == OR
        // using (A <-> B) <=> (A & B) \/ (~B & ~A)
        List* newarg = new List (new Formula (AND, left(), right()),
                       new List (new Formula (AND, 
                                              new Formula (NOT, left()->copy()), 
                                              new Formula (NOT, right()->copy())) ) );
        connective (OR);
        args (newarg);
      }
      return;

    case XOR: {
      left()->removeIff();
      right()->removeIff();
      Formula* newarg;
      // using (A xor B) <=> ~(A <-> B)
      if ( c == OR ) {
        newarg = new Formula ( AND,
                               new List (new Formula (IMP, left(), right()),
                               new List (new Formula (IMP, right()->copy(), left()->copy())) ) );
      }
      else {  // c == AND
        newarg = new Formula ( OR, new List (new Formula (AND, left(), right()),
                                   new List (new Formula (AND, 
                                                          new Formula (NOT, left()->copy()), 
                                                          new Formula (NOT, right()->copy())) ) ) );
      }
      connective (NOT);
      arg (newarg);
      return;
    }

    case NOT:
      arg()->removeIff( c == AND ? OR : AND );
      return;

    case FORALL:
    case EXISTS:
      arg()->removeIff(c);
      return;
  }
} // Formula::removeIff


// 04/06/2001
void Formula::polarities ( const PSymbol* p, bool sign, int& pol ) const
{
  switch ( connective() ) {
    case ATOM:
      if ( atom()->functor() == p ) {
        if ( sign ) {
          pol |= 1;
        }
        else { // negative occurrence
          pol |= 2;
        }
      }
      return;

    case AND:
    case OR: {
      Formula::List::Iterator fs (args());
      while ( fs.more() ) {
        fs.next()->polarities (p,sign,pol);
        if ( pol == 3 )
          return;
      }
      return;
    }

    case IMP: 
      left()->polarities (p,! sign,pol);
      if ( pol == 3 )
        return;
      right()->polarities (p,sign,pol);
      return;

    case IFF:
    case XOR:
      if ( occurs(p) ) {
        pol = 3;
      }
      return;

    case NOT:
      arg()->polarities (p,! sign,pol);
      return;

    case FORALL:
    case EXISTS:
      arg()->polarities (p,sign,pol);
      return;
  }
} // Formula::polarities


// 04/06/2001
// the formula is a definition (P(X) <=> rhs)
// is pos, then it should be changed into (P(X)->rhs), otherwise into
// (rhs->P(X)), i.e. pos is the polarity of rsh in the new formula
void Formula::makeHalfDefinition ( bool positive, const Formula* rhs )
{
  Formula* f = this;
  if ( f->connective() == FORALL ) {
    f = f->arg();
  }

  ASSERT ( debug_formula, f->connective() == IFF );
  f->connective (IMP);
  // there are two cases for f, either (P(x) <-> rhs) or (rhs <-> P(X))
  if ( (f->right() == rhs) != positive ) {
    // two sides of f must be swapped
    Formula* tmp = f->right();
    f->right (f->left());
    f->left (tmp);
  }
} // Formula::makeHalfDefinition


// 12/06/2001
// applicable only to formulas without equivalences
void Formula::nnf (bool sign)
{
  switch ( connective() ) {
    case ATOM: 
      if ( ! sign ) {
        Formula* f = new Formula (atom());
        connective (NOT);
        arg (f);
      }
      return;

    case AND:
    case OR: {
      Formula::List::Iterator fs (args());
      while ( fs.more() ) {
        fs.next()->nnf (sign);
      }
      return;
    }

    case IMP: {
      left()->nnf (! sign);
      right()->nnf (sign);
      List* junction = new List ( left(), 
                       new List ( right() ) );
      connective ( sign ? OR : AND );
      args (junction);
      return;
    }

    case IFF:
    case XOR:
      ASSERT (debug_formula, false );

    case NOT:
      arg()->nnf (! sign);
      return;

    case FORALL:
    case EXISTS:
      arg()->nnf (sign);
      return;
  }
} // Formula::nnf

/*

// 12/06/2001
// miniscope essentially universal quantifiers
// applicable only to nnf formulas
void Formula::miniscope ( bool sign )
{
  // the quantifier does not matter
  miniscope (sign, FORALL, VarList::empty());
  flatten ();
  rename ();
} // Formula::miniscope


// 12/06/2001
// applicable only to nnf formulas
void Formula::miniscope (bool sign, Connective quantifier, VarList* vs)
{
  switch ( connective () ) {
    case ATOM: {
      // create a sublist of variables in vs occurring in the formula
      VarList* us = occur (vs);
      if ( us->isNonEmpty () ) {
        Formula* f = new Formula ( quantifier, us, this );
        copyFrom (f);
      }
      return;
    }

    case AND: 
      if ( (sign && quantifier == FORALL) || vs->isEmpty() ) {
        Formula::List::Iterator ar ( args() );
        while ( ar.more () ) {
          ar.next()->miniscope (sign, FORALL, vs);
        }
        return;
      }
      // vs is non-empty
      // either sign = false or quantifier = EXISTS
      if ( sign ) { // and hence quantifier == EXISTS
        // (Ex)A 
        // list of formulas to be moved outwards
        Formula::List* outer = Formula::List::empty ();

        // check if there exists some argument in which A does not occur
        Formula::List::DelIterator ar ( _args );
        while ( ar.more () ) {
          Formula* f = ar.next ();
          VarList* us = f->occur (vs);
          if ( us->isEmpty() ) { // none of the variables occurs in this argument
            // delete f from the arguments
            ar.del ();
            // and add it to outer
            Formula::List::push (f,outer);
          }
          else {
            vs->destroy ();
          }
        }
        // now outer contains formulas to be pushed outward
        if ( outer->isEmpty() ) {
          // simply add the quantifier in front of the formula
          VarList* us = occur (vs);
          if ( us->isNonEmpty() ) {
            Formula* f = new Formula ( AND, args() );
            connective (EXISTS);
            vars ( us );
            arg (f);
            // miniscope the rest
            f->miniscope (sign, FORALL, VarList::empty());
          }
          return;
        }
        // now outer is non-empty, change (out1 & ... & outm & rest1 & ... & restn) into
        //     (out1 & ... & outm & EX(rest1 & ... & restn), if n >= 2
        //     (out1 & ... & outm & EX(rest1), if n == 1
        //     (out1 & ... & outm), if n == 0
        // first, miniscope outer
        Formula::List::Iterator out ( outer );
        while ( out.more () ) {
          out.next()->miniscope (sign,FORALL,VarList::empty());
        }

        if ( args()->isEmpty() ) { // n == 0
          args (outer);
          // miniscope the rest
          Formula::List::Iterator ar ( args() );
          while ( ar.more () ) {
            ar.next()->miniscope (sign,VarList::empty());
          }
          return;
        }
        // args is not empty, n >= 1
        if ( args()->tail()->isEmpty() ) { // n == 1
          // miniscope rest1
          args->head()->miniscope (sign,vs);
          // miniscope outer
          Formula::List::Iterator ar ( outer );
          while ( ar.more () ) {
            ar.next()->miniscope (sign,VarList::empty());
          }
          // append outer to rest1
          args()->tail (outer);
          return;
        }
        // now n >= 2
        // miniscope rest1 ... restn
        Formula::List::Iterator ar ( args() );
        while ( ar.more () ) {
          ar.next()->miniscope (sign,VarList::empty());
        }
        if ( outer->isEmpty () ) {
          // add existential quantifier to args
        }
      }
    }
      // sign == false 


    case OR:
    case IMP:
    case IFF:
    case XOR:
      ASSERT ( debug_formula, false );
      return this;

    case NOT:
    case FORALL:
    case EXISTS:
      return 0;
  } 
} // Formula::miniscope 


// 11/06/2001
VarList* Formula::occur ( const VarList* vs ) const // sublist of those variables in vs occuring in the formula
{
  if ( vs->isEmpty() ) 
    return VarList::empty ();

  if ( occurs (vs->head()) ) {
    return new VarList ( vs->head(), occur(vs->tail()) );
  }
  else {
    return occur (vs->tail());
  }
} // VarList* Formula::occur

*/