//
//  file formula.h 
//  defines class Formula
//
//  15/05/2000, Dallas
//  29/05/2000, Manchester, AND and OR have become of arbitrary arity
//  29/03/2001, Machester, Functor changed to Functor* everywhere
//  03/06/2001 Manchester, sort/1,2, split/5, normalize/0, normalize1/0, compare/1,2
//                isEqualityAxiom/0, isReflexivityAxiom/0, isSymmetryAxiom/0, isTransitivityAxiom/0,
//                isFunctionReflexivityAxiom/0, isPredicateReflexivityAxiom/0, skolemize,
//                removeIff/0
//

#ifndef __formula__
#define __formula__


#include "typedefs.h"
#include "term.h"
#include "clause.h"


class Atom;


class Formula {
 public:

  typedef List <Formula*> List;
  
  enum Connective {
    ATOM,
    AND,
    OR,
    IMP,
    IFF,
    XOR,
    NOT,
    FORALL,
    EXISTS
  }; // enum Formula::Connective

  // constructors
  Formula ( Connective, Formula*, Formula* ); // for binary formulas
  Formula ( Connective, Formula* ); // for unary formulas
  Formula ( Connective, List* ); // for and/or formulas
  Formula ( Connective, VarList*, Formula* ); // for quantified formulas
  Formula ( Atom* ); // for atomic formulas

  // query the structure
  Connective connective () const;
  Atom* atom () const;
  Formula* left () const;
  Formula* right () const;
  Formula* arg () const;
  List* args () const; // for conjunctions and disjunctions
  VarList* vars () const;

  // deallocation
  void destroy ();

  // statistics
  bool isEqualityAxiom () const;  // is one of the equality axioms
  bool occurs ( const PSymbol* f ) const; // predicate symbol f occurs in the formula

  // miscellaneous
  int size () const; // the number of connective, quantifiers and atoms
  int weight () const; // the number of symbols
  void prettyPrint ( int indent, ostream& );
  Atom* isPredicateDefinition ( Formula* & rhs ); // return lhs of the definition or 0
  Clause::List* skolemize ( bool sign );
  int countOccurrences ( const PSymbol* p ) const;
  int countOccurrences ( const FSymbol* f ) const;
  void normalize ();
  Comparison compare ( const Formula* ) const; // comparison of normalized formulas
  void polarities ( const PSymbol* p, bool sign, int& pol ) const;
  void makeHalfDefinition ( bool positive,const Formula* rhs );

 private: 
  // structure
  Connective _con;

  union {
    Formula* _left;
    Atom* _atom;
    VarList* _vars;
  };
  union {
    Formula* _right;
    Formula* _arg;
    List* _args;
  };

  // structure change
  void connective ( Connective );
  void atom ( Atom* );
  void left ( Formula* );
  void right ( Formula* );
  void arg ( Formula* );
  void args ( List* ); 
  void vars ( VarList* );

  // auxiliary functions
  static char* printName ( Connective );
  void print ( Connective outer, ostream& );
  void prettyPrint ( Connective outer, int indent, ostream& );
  void rename ( VarSubst&, Var& last ); 
  bool isLinearAndFlat (); 
  static List* flatten ( List*, Connective );  
  static List* copy( List* );
  void copyFrom ( const Formula* );
  static const int oneLineSize; // formula with size <= oneLineLength are printed on one line
  void print ( ostream& );
  void rename (); // renames variables in the order of their occurrence
  void flatten ();  // flattens the formula (and destroys the original one)
  void removeIff ();  // remove IFF and XOR
  void removeIff ( Connective c );  // remove IFF and XOR, c is the external connective (AND or OR)
  void normalize1 ();
  Formula* copy ();
  static Comparison compare ( const List*, const List* ); // comparison of normalized formulas
  static List* sort ( List* );
  static List* sort ( List* toSort, List* appendTo );
  static void split ( Formula*, List*, // for sorting
                      List*& less, List*& equal, List*& greater );
  bool isReflexivityAxiom () const;
  bool isFunctionReflexivityAxiom () const;
  bool isPredicateReflexivityAxiom () const;
  bool isSymmetryAxiom () const;
  bool isTransitivityAxiom () const;
  void miniscope (bool sign);
  void miniscope (bool sign, Connective quantifier, VarList*);
  // void moveDown (bool sign,VarList*); // auxiliary for miniscope
  VarList* occur ( const VarList* vs ) const; // sublist of those variables in vs occuring in the formula
  void nnf (bool sign); 

  typedef Alist <Var,Term*> Subst;
  static void skolemize ( Atom::List* pa, Atom::List* na, List* pf, List* nf, VarList* univ,
                          Subst& sbst, Clause::List*& result );
  bool occurs ( Var v ) const; // variable v occurs in the formula
 friend ostream& operator << ( ostream&, Formula* );
}; // class Formula


ostream& operator << ( ostream&, Formula* );


#endif // __formula__
