//
//  File problem.cpp
//
//  implements the class Problem
//
//  14/05/2000, Dallas
//  29/05/2000, Manchester, output for InputUnitType added
//  01/06/2000, Manchester, printDefinitions changed
//              findClausalDefinition()
//  07/06/2000, Manchester, several statistical functions added, eg isHorn()
//  09/06/2000, Manchester, removeAllRelationalFunctionDefinitions(), removeVarEquality()
//                          splitLargeArity ()
//  10/06/2000, Manchester, cutTooLongClauses(), canRemovePredicateDefinition(), removePredicateDefinition()
//                          resolveOn()
//  29/03/2001, Manchester, normalize(), sort(), split(), compare() 
//  08/04/2001, Manchester, removeVarEquality - tab argument added
//  09/04/2001, Manchester, removeTautologies - tab argument added
//  11/04/2001, Manchester, Problem::removeOneRelationalFunctionDefinition changed to report the inferences
//  11/04/2001, Manchester, InputUnit::replaceRelationByFunction added
//  12/04/2001, Manchester, InputUnit renamed to Unit and moved to unit.cpp
//  12/04/2001, Manchester, replace reimplemented
//  09/05/2001, Manchester, several functions zipped out
//  09/05/2001 Manchester: occurs (), removeDummyFunctionDefinitions()
//  15/05/2001 Manchester: replaceCancellationAxioms ()
//  22/05/2001 Manchester, applyFunDef(), removeLinearFunctionDefinitions()
//  27/05/2001 Manchester: eliminateResolvable, eliminatePure, removePredicate, removeResolvable
//  30/05/2001 Manchester, splitLargeArity/1, removeAllFunctionDefinitions/1
//  31/05/2001 Manchester, normalize/0 changed, negativeEqualitySplit/2
//  03/06/2001 Manchester, removeEqualityAxioms/0, removeDummyPredicateDefinitions/1
//

#include <string.h>
#include <limits.h>
#include <stdlib.h>


#include "clause.h"
#include "unit.h"
#include "problem.h"
#include "replace_code.h"
#include "literal.h"
#include "int.h"
#include "term.h"
#include "formula.h"
#include "array.h"
#include "string.h"
#include "resolvable.h"
#include "pure.h"


Problem::Problem ()
{
} // Problem::Problem


void Problem::build ( char* name, Unit::List* lst )
{
  _units = lst;
  _name = new char [ strlen(name) + 1 ];
  strcpy ( _name, name );
} // Problem::Problem


Problem::~Problem ()
{
  delete [] _name;

  // destroy all input units in the list
  for ( Unit::List* lst = _units; ! lst->isEmpty (); lst = lst->tail () ) 
    lst->head ()->destroy ();

  // destroy the list itself
  _units->destroy ();
} // Problem::~Problem


ostream& operator << ( ostream& str, Problem& p )
{
  str << "%\n"
         "% The original problem comes from the file "
      << p.name () 
      << "\n"
      << "%\n\n"
      << p.units ();

  return str;
} // operator << ( ostream& str, Problem* p )


// finds and removes all relational definitions from the problem
int Problem::removeAllRelationalFunctionDefinitions ( ostream& tab )
{  
  int removed = 0;

  while ( removeOneRelationalFunctionDefinition ( tab ) ) 
    removed ++;

  return removed;
} // Problem::removeAllRelationalFunctionDefinitions


// finds and removes all relational definitions from the problem
bool Problem::removeOneRelationalFunctionDefinition ( ostream& tab )
{  
  Unit::List::Iterator it ( _units );
  while ( it.more() ) {
    Unit* unit = it.next ();
    PSymbol* pred;
    int argNo;
    // check that the current unit is P(X,y) & P(X,z) -> y = z
    if ( unit->isUniquenessAxiom (pred, argNo) )  {
      FSymbol* fun;
      Var* vs;
      // check that the current clause is P(X,f(X))
      Atom* atom;
      Unit* funax;
      Unit::List::Iterator ulit ( _units );
      while ( ulit.more() ) {
        funax = ulit.next();
        vs = funax->isFunctionalityAxiom (pred, fun, argNo, atom);
        if ( vs ) {
          // output inference( P(X,y) --> y = f(X), replace_P_by_f, [ppp,fff] ).
          // note that atom is P(X,f(X))
          tab << "inference(" << pred << "(";
          Term::List* args = atom->args ();
          int arg = 0;
          for (;;) {
            if ( arg == argNo ) 
              tab << "Y";
            else
              tab << args->head ();
            args = args->tail();
            if ( args->isEmpty() )
              break;
            // args is nonempty
            tab << ",";
            arg++;
          }
          tab << ") --> Y=" << atom->args()->nth(argNo) 
              << ",replace_" << pred << "_by_" << fun 
              << ",[" << unit->name() << "," << funax->name() << "]).\n";

          // remove the uniqueness axiom and the functionality axiom from the set
          _units = _units->remove ( unit );
          _units = _units->remove ( funax );

          // replace P(S,t) by t = f(S) in all units
          replaceRelationByFunction ( pred, fun, argNo, vs, tab );
          return true;
        }
      }
    }
  }

  return false;
} // Problem::removeOneRelationalFunctionDefinition


// find all cancellation axioms
//   f(X,y,Z) = f(X,u,Z) -> y = u
// and replace them by 
//   i(f(X,y,Z),X,Z) = y
int Problem::replaceCancellationAxioms ( ostream& tab )
{  
  Unit::List::DelIterator it ( _units );
  int found = 0;
  while ( it.more() ) {
    Unit* unit = it.next ();
    FSymbol* fun;
    int argNo;
    if ( unit->isCancellationAxiom (fun, argNo) )  {
      found ++;
      // create new function symbol of arity 2 * arity (f) - 1
      FSymbol* inv = FSymbol::createNew ( "inv", fun->arity() * 2 - 1);
      unit->replaceCancellation (fun,argNo,inv,tab);
    }
  }

  return found;
} // Problem::replaceCancellationAxioms


int Problem::countOccurrences ( const PSymbol* p ) const
{
  int occ = 0;

  Unit::List::Iterator us (_units);
  while ( us.more() ) {
    occ += us.next()->countOccurrences (p);
  }
    
  return occ;
} // Problem::countOccurrences


bool Problem::occurs ( const FSymbol* f ) const
{
  Unit::List::Iterator us (_units);
  while ( us.more() ) {
    if ( us.next()->occurs (f) )
      return true;
  }
    
  return false;
} // Problem::occurs


bool Problem::occurs ( const PSymbol* p ) const
{
  Unit::List::Iterator us (_units);
  while ( us.more() ) {
    if ( us.next()->occurs (p) )
      return true;
  }
    
  return false;
} // Problem::occurs


int Problem::countOccurrences ( const FSymbol* f ) const
{
  int occ = 0;

  Unit::List::Iterator us (_units);
  while ( us.more() ) {
    occ += us.next()->countOccurrences (f);
  }
    
  return occ;
} // Problem::countOccurrences


void Problem::printDefinitions ( ostream& str )
{
  Term* lhs;
  Term* rhs;
  Unit* unit;

  Unit::List::Iterator units (_units);
  while ( units.more() ) {
    unit = units.next();
    ASSERT( debug_problem, unit->unitType() == CLAUSE );

    lhs = unit->isDefinition(rhs);
    if ( lhs ) {
      str << "Unit " 
          << unit->name() 
          << " is a definition of the function "
          << lhs->functor() << '\n'
          << "  having "
          << countOccurrences ( lhs->functor() )
          << " occurrences in the problem\n";
    }
  }
} // Problem::printDefinitions


// 11/06/2001 Manchester, made from a previous version that had caused
// incompleteness
int Problem::removeDummyFunctionDefinitions ( ostream& report )
{
  // left- and right-hand sides of all definitions
  int noUnits = _units->length ();
  Array<Term*> lhs (noUnits);
  Array<Term*> rhs (noUnits);
  // definitions themselves
  Array<Unit*> defs (noUnits); 
  // units in which the corresponding function symbol occurs
  Array<Unit::List*> occ (noUnits); 

  // first, copy all definitions from _units to the arrays
  int nextDef = 0; // next index of definition in the arrays

  Unit::List::Iterator uns (_units);

  while ( uns.more() ) {
    Unit* unit = uns.next ();
    Term* r;
    if ( Term* l = unit->isDefinition(r) ) {
      if ( l ) { // is a definition
        // memorize the definition
        lhs[nextDef] = l;
        rhs[nextDef] = r;
        defs[nextDef] = unit;
        // create a new list of units and fill it with the list of units in which
        // the function symbol occurs
        FSymbol* f = l->functor ();
        occ[nextDef] = Unit::List::empty ();
        Unit::List::Iterator us (_units);
        while ( us.more() ) {
          Unit* u = us.next();
          if ( u != unit && u->occurs (f) ) {
            Unit::List::push (u, occ[nextDef]);
          }
        }

        nextDef ++;
      }
    }
  }

  int removed = 0;

  int lastDef = nextDef - 1;
  while ( lastDef >= 0 ) { // there are definitions left
    int removedThisTime = 0;

    for ( int i = lastDef; i >= 0; i-- ) {
      if ( occ[i]->isEmpty() ) {
        // write report 
        report << "prepro(removed_as_dummy_definition," << defs[i]->name() << ",[++ " 
               << lhs[i] << "=" << rhs[i] << "]).\n";

        // memorize the definition
        Unit* def = defs[i];

        removedThisTime ++;
        // exchange i with lastDef
        if ( i != lastDef ) {
          lhs[i] = lhs[lastDef];
          rhs[i] = rhs[lastDef];
          defs[i] = defs[lastDef];
          occ[i] = occ[lastDef];
        }
        lastDef--;
        // delete def from every list in occ
        for ( int i = lastDef; i >= 0; i-- ) {
          occ[i] = occ[i]->remove (def);
        }
        // delete def from _units;
        _units = _units->remove (def);
        def->destroy ();
      }
    }

    if ( removedThisTime == 0 ) {
      // destroy all lists in occ
      for ( int i = lastDef; i >= 0; i-- ) {
        occ[i]->destroy ();
      }
      return removed;
    }
    // otherwise continue
    removed += removedThisTime;
  }

  return removed;
} // Problem::removeDummyFunctionDefinitions


// 03/06/2001
// the result is the number of removed symbols
int Problem::removeDummyPredicateDefinitions ( ostream& report )
{
  // left- and right-hand sides of all definitions
  int noUnits = _units->length ();
  Array<Atom*> lhs (noUnits);
  Array<Formula*> rhs (noUnits);
  // definitions themselves
  Array<Unit*> defs (noUnits); 

  // first, move all definitions from _units to the arrays
  int nextDef = 0; // next index of definition in the arrays

  Unit::List::DelIterator uns (_units);

  while ( uns.more() ) {
    Unit* unit = uns.next ();
    Formula* r;
    if ( Atom* l = unit->isPredicateDefinition(r) ) {
      if ( l ) { // is a definition
        // delete the definition from the list
        uns.del();
        // memorize the definition
        lhs[nextDef] = l;
        rhs[nextDef] = r;
        defs[nextDef] = unit;
        nextDef ++;
      }
    }
  }

  if ( nextDef == 0 ) // no definitions found at all
    return 0;

  // if there is a double definition, move one part of it back to the problem
  for ( int i = 0; i < nextDef - 1; i++ ) {
    for ( int j = i+1; j < nextDef; ) {
      if ( lhs[i]->functor() == lhs[j]->functor() ) {
        // push the unit back
        Unit::List::push( defs[j], _units );
        // decrease nextDef, and move nextDef-1 to j
        lhs[j] = lhs[--nextDef];
        rhs[j] = rhs[nextDef];
        defs[j] = defs[nextDef];
      }
      else {
        j++;
      }
    }
  }
      
  // divide all definitions in two parts, those with > 0 occurrences in the rest
  // of the problem, and with 0 occurrences. Those with zero occurrences will be
  // at the positions fist ... nextDef-1 in the arrays
  int first = nextDef;
  {for ( int i = 0; i < first; ) {
    if ( ! occurs(lhs[i]->functor()) ) { // the symbol does not occur in the rest
      // decrease first and exchange first with i
      Atom* a = lhs[--first];
      lhs[first] = lhs[i];
      lhs[i] = a;
      Formula* f = rhs[first];
      rhs[first] = rhs[i];
      rhs[i] = f;
      Unit* u = defs[first];
      defs[first] = defs[i];
      defs[i] = u;
    }
    else { // occurs
      i++;
    }
  }}

  if ( first != nextDef ) {
    // for all definitions f(..) = t in the non-zero part and (j)
    //   for all definitions g(..) = ... in the zero part (k)
    //     if g occurs in t, move g(..) = ... in the non-zero part
    //   then move f(..) = t back to problem clauses
    for ( int j = 0; j < first; j++ ) {
      for ( int k = first; k < nextDef; k++ ) {
        if ( rhs[j]->occurs(lhs[k]->functor()) ) {
          // exchange k with first and increase first
          Atom* a = lhs[first];
          lhs[first] = lhs[k];
          lhs[k] = a;
          Formula* f = rhs[first];
          rhs[first] = rhs[k];
          rhs[k] = f;
          Unit* u = defs[first];
          defs[first] = defs[k];
          defs[k] = u;
          first++;
        }
      }
    }
  }

  // return all non-zero definitions back to _units
  {for ( int i = 0; i < first; i++ ) {
    Unit::List::push( defs[i], _units );
  }}

  // delete all zero definitions and report
  {for ( int i = first; i < nextDef; i++ ) {
    report << "prepro(removed_as_dummy_definition," << defs[i]->name() << ",[++ " 
           << lhs[i] << "=" << rhs[i] << "]).\n";

    defs[i]->destroy ();
  }}

  return nextDef - first;
} // Problem::removeDummyPredicateDefinitions


// 04/06/2001
// the result is the number of removed symbols
// this function can be opimized considerably but I have no time now
// (eg one can remember some properties of units, eg being a definition
//  and reuse them at different iterations)
int Problem::removePredicateSemiDefinitions ( ostream& report )
{
  int removed = 0;

  // there are several iterations since after elimination of a definition
  // other definitions may become eligible
  for (;;) { // one iteration
    int removedAtLastIteration = 0;
    Unit::List::DelIterator uns (_units);
    while ( uns.more() ) {
      Unit* unit = uns.next ();
      Formula* r;
      if ( Atom* l = unit->isPredicateDefinition(r) ) {
        if ( l ) { // is a definition
          // temporarily delete this definition
          uns.del ();

          switch ( polarities (l->functor()) ) {
            case 0: // no occurrences at all, something is wrong
              ASSERT ( debug_unit, false );

            case 1: // positive occurrences only
              removedAtLastIteration ++;
              // cout << "Positive only: " << l->functor() << "\n";
              unit->makeHalfDefinition (true,r);
              break;

            case 2: // negative occurrences only
              removedAtLastIteration ++;
              // cout << "Negative only: " << l->functor() << "\n";
              unit->makeHalfDefinition (false,r);
              break;

            case 3: // both positive and negative occurrences 
              // do nothing
              break;
          }

          // put the definition back
          uns.restore (unit);
        }
      }
    }
    if ( removedAtLastIteration == 0 )
      return removed;
    removed += removedAtLastIteration;
  }
} // Problem::removePredicateSemiDefinitions


// the result is the number of removed symbols
int Problem::removeLinearFunctionDefinitions ( ostream& report )
{
  int eliminated = 0;

  Unit::List::DelIterator uns (_units);

  while ( uns.more() ) {
    Unit* unit = uns.next ();
    Term* r;
    if ( Term* l = unit->isLinearDefinition(r) ) {
      if ( l ) { // is a definition
        // delete the definition from the list
        uns.del();
        eliminated ++;
        applyFunDef ( l, r, unit, report );
      }
    }
  }

  return eliminated;
} // Problem::removeLinearFunctionDefinitions


// the result is the number of removed symbols
int Problem::removeAllFunctionDefinitions ( ostream& report )
{
  int eliminated = 0;

  Unit::List::DelIterator uns (_units);

  while ( uns.more() ) {
    Unit* unit = uns.next ();
    Term* r;
    if ( Term* l = unit->isDefinition(r) ) {
      if ( l ) { // is a definition
        // delete the definition from the list
        uns.del();
        eliminated ++;
        applyFunDef ( l, r, unit, report );
      }
    }
  }

  return eliminated;
} // Problem::removeAllFunctionDefinitions


// apply rewriting lhs -> rhs
void Problem::applyFunDef ( const Term* lhs, const Term* rhs, const Unit* def, ostream& report ) 
{
  ReplaceCode code ( lhs, rhs );

  Unit::List::Iterator us ( _units );
  while ( us.more () ) {
    Unit* unit = us.next ();
    unit->apply (code,def,report);
  }
} // Problem::applyFunDef


bool Problem::hasEquality () const
{
  Unit::List::Iterator unit ( _units ); 
  while ( unit.more() ) {
    if ( unit.next()->hasEquality() )
      return true;
  }
  
  return false;
} // Problem::hasEquality


bool Problem::isHorn () const
{
  Unit::List::Iterator unit ( _units ); 
  while ( unit.more() ) {
    if ( ! unit.next()->isHorn() )
      return false;
  }
  
  return true;
} // Problem::isHorn


// 06/06/2001 made from pureEquality
bool Problem::hasNonEquality () const
{
  Unit::List::Iterator unit ( _units ); 
  while ( unit.more() ) {
    if ( unit.next()->hasNonEquality() )
      return true;
  }
  
  return false;
} // Problem::hasNonEquality


bool Problem::hasFunctionSymbols () const
{
  Unit::List::Iterator unit ( _units ); 
  while ( unit.more() ) {
    if ( unit.next()->hasFunctionSymbols() )
      return true;
  }
  
  return false;
} // Problem::hasFunctionSymbols


bool Problem::isPropositional () const
{
  Unit::List::Iterator unit ( _units ); 
  while ( unit.more() ) {
    if ( ! unit.next()->isPropositional() )
      return false;
  }
  
  return true;
} // Problem::isPropositional


int Problem::numberOfClauses () const
{
  int result = 0;

  Unit::List::Iterator unit ( _units ); 
  while ( unit.more() ) {
    if ( ! unit.next()->unitType() == CLAUSE )
      result++;
  }

  return result;
} // Problem::numberOfClauses


int Problem::numberOfLiterals () const
{
  int result = 0;

  Unit::List::Iterator unit ( _units ); 
  while ( unit.more() ) {
    result += unit.next()->numberOfLiterals();
      result++;
  }

  return result;
} // Problem::numberOfLiterals


bool Problem::isUnit () const
{
  Unit::List::Iterator unit ( _units ); 
  while ( unit.more() ) {
    if ( ! unit.next()->isUnit() )
      return false;
  }
  
  return true;
} // Problem::isUnit


// appends statistics to the file
void Problem::statistics ( ostream& str ) const
{ 
  str << "problem_info( '" 
      << name ()
      << "', [equality("
      << (isUnit() ? "unit" : ( hasEquality() ? ( hasNonEquality() ? "yes" : "pure" ) : "no" ))
      << "), horn(" 
      << (isHorn () ? "yes" : "no")
      << "), functions("
      << ( isPropositional() ? "prop" : ( hasFunctionSymbols () ? "yes" : "no") )
      << "), clauses("
      << numberOfClauses ()
      << "), literals("
      << numberOfLiterals ()
      << ")] ).\n";
} // Problem::statistics


void Problem::replaceRelationByFunction ( PSymbol* pred, FSymbol* fun, int argNo, int* vs, ostream& report )
{
  Unit::List::Iterator ii (_units);
  while ( ii.more() ) {
    Unit* unit = ii.next();
    ASSERT( debug_problem, unit->unitType() == CLAUSE );
    unit->replaceRelationByFunction (pred, fun, argNo, vs, report);
  }

  delete [] vs;
} // Problem::replaceRelationByFunction


int Problem::removeVarEquality ( ostream& tab )
{
  int removed = 0;

  Unit::List::Iterator inp ( _units );
  while ( inp.more () ) {
    if ( inp.next()->removeVarEquality (tab) ) 
      removed++;
  }

  return removed;
} // Problem::removeVarEquality


int Problem::removeTautologies ( ostream& tab )
{
  int removed = 0;

  Unit::List::DelIterator us (_units);
  while ( us.more() ) {
    if ( us.next()->isTautology (tab) ) {
      us.del ();
      removed++;
    }
  }

  return removed;
} // Problem::removeTautologies


bool Problem::splitLargeArity ( int maxArity )
{
  bool splitted = false;

  FSymbol** fsig = FSymbol::signature (); 
  for ( int j = FSymbol::noOfSymbols() - 1; j >= 0; j-- ) {
    FSymbol* f = fsig[j];
    int arity = f->arity ();
    if ( arity > maxArity ) {
      splitted = true;
      div_t qr = div ( arity, maxArity );
      int iterations = qr.quot;
      int restArity = qr.rem;
      // list storing new predicate symbols
      FSymbol::List* fs = FSymbol::List::empty ();
      if ( restArity ) {
        FSymbol::List::push ( FSymbol::createNew(f->name(),restArity), fs );
      }
      while ( iterations-- > 0 ) {
        FSymbol::List::push ( FSymbol::createNew(f->name(),maxArity), fs );
      }
      fs = fs->reverse ();

      Unit::List::Iterator us ( _units );
      while ( us.more() ) {
        us.next()->splitLargeArity (f, fs);
      }
      f->arity ( fs->length() );
      fs->destroy ();
    }
  }

  delete [] fsig;

  PSymbol** sig = PSymbol::signature (); 
  for ( int i = PSymbol::noOfSymbols() - 1; i >= 0; i-- ) {
    PSymbol* p = sig[i];
    int arity = p->arity ();
    if ( arity > maxArity ) {
      splitted = true;
      div_t qr = div ( arity, maxArity );
      int iterations = qr.quot;
      int restArity = qr.rem;
      // list storing new predicate symbols
      FSymbol::List* ps = FSymbol::List::empty ();
      if ( restArity ) {
        FSymbol::List::push ( FSymbol::createNew(p->name(),restArity), ps );
      }
      while ( iterations-- > 0 ) {
        FSymbol::List::push ( FSymbol::createNew(p->name(),maxArity), ps );
      }
      ps = ps->reverse ();

      Unit::List::Iterator us ( _units );
      while ( us.more() ) {
        us.next()->splitLargeArity (p, ps);
      }
      p->arity ( ps->length() );
      ps->destroy ();
    }
  }

  delete [] sig;

  return splitted;
} // Problem::splitLargeArity


int Problem::maxLitWeight ()
{
  int result = Int::min ();

  Unit::List::Iterator unit ( _units );
  while ( unit.more() ) {
    int mlcw = unit.next()->maxLitWeight(); 
    if ( mlcw > result )
      result = mlcw;
  }

  return result;
} // Problem::maxLitWeight


int Problem::cutTooLongClauses ( int maxLiterals )
{
  return 0;
}


void Problem::removeResolvable ( const PSymbol* pred, bool sign, ostream& report )
{
  Unit::List* pos = Unit::List::empty(); // clauses with a sign-occurrence of pred
  Unit::List* neg = Unit::List::empty(); // clauses with a not-sign-occurrence of pred

  Unit::List::DelIterator us (_units);
  while ( us.more() ) {
    Unit* u = us.next ();
    ASSERT( debug_problem, u->unitType() == CLAUSE );
    bool litSign;

    if ( u->clause()->occurs(pred,litSign) ) {
      us.del ();
      if ( sign == litSign ) { // move unit to pos
        Unit::List::push (u,pos);
      }
      else { // move it to neg
        Unit::List::push (u,neg);
      }
    }
  }

  ASSERT( debug_problem, pos->isNonEmpty() );
  ASSERT( debug_problem, neg->isNonEmpty() );
  ASSERT( debug_problem, pos->tail()->isEmpty() );

  // create all resolvents between pos and neg and move them to rest
  for ( Unit::List* p = pos; p->isNonEmpty(); p = p->tail() ) {
    for ( Unit::List* n = neg; n->isNonEmpty(); n = n->tail() ) {
      // create resolvent and add it to _units
      Unit::List::push ( p->head()->resolveOn (pred, n->head(),report), _units );
    }
  }

  Unit::List* first;
  // destroy pos and neg, and write report
  for ( first = pos; first->isNonEmpty(); first = first->tail() ) {
    Unit* u = first->head();
    report << "prepro(removed_as_resolved," << u->name() << u->clause() << "]).\n";
    u->destroy();
  }
  pos->destroy ();
  for ( first = neg; first->isNonEmpty(); first = first->tail() ) {
    Unit* u = first->head();
    report << "prepro(removed_as_resolved," << u->name() << "," << u->clause() << "]).\n";
    u->destroy();
  }
  neg->destroy ();
} // Problem::removeResolvable


void Problem::normalize ()
{
  Unit::List::Iterator us (_units);
  while ( us.more() ) {
    us.next()->normalize ();
  }

  _units = Unit::sort (_units);
} // Problem::normalize


int Problem::eliminateResolvable ( ostream& report )
{
  int symbols = PSymbol::noOfSymbols();
  if ( symbols == 0 ) {
    return 0;
  }

  int eliminated = 0;

  for (;;) {
    // starting next round 
    Resolvable res ( symbols );

    Unit::List::Iterator us ( _units );
    while ( us.more () ) {
      Unit* unit = us.next ();
      ASSERT( debug_problem, unit->unitType() == CLAUSE );

      if ( ! res.getClause(unit->clause()) ) {
        return eliminated;
      }
    }

    // round terminated
    bool sign;
    PSymbol* p = res.first (sign);
    if ( p ) {
      eliminated++;
      removeResolvable (p,sign,report);
    }
    else {
      return eliminated;
    }
  }
} // Problem::eliminateResolvable


int Problem::eliminatePure ( ostream& report )
{
  int symbols = PSymbol::noOfSymbols();
  if ( symbols == 0 ) {
    return 0;
  }

  // at least one symbol present
  int eliminated = 0;

  for (;;) {
    // one round of eliminating pure literals
    Pure pure ( symbols );

    Unit::List::Iterator us ( _units );
    while ( us.more () ) {
      Unit* unit = us.next ();
      ASSERT( debug_problem, unit->unitType() == CLAUSE );

      if ( ! pure.getClause(unit->clause()) ) {
        return eliminated;
      }
    }
    // round terminated, check if there are pure literals
    // and remove them
    int elim = 0;
    Pure::Iterator symb (pure);
    PSymbol* p;
    while ( (p = symb.next()) != 0 ) {
      elim++;
      removePredicate ( p, report );
    }
    if ( elim == 0 ) // no symbols eliminated this round
      return eliminated;
    eliminated += elim;
  } // for (;;)
} // Problem::eliminatePure


void Problem::removePredicate ( const PSymbol* p, ostream& report )
{
  Unit::List::DelIterator us ( _units );
  while ( us.more() ) {
    Unit* u = us.next ();
    if ( u->occurs (p) ) {
      u->destroy ();
      report << "prepro(removed_as_pure," << u->name() << "," 
             << p << ").\n";
      us.del ();
    }
  }
} // Problem::eliminatePure


int Problem::negativeEqualitySplit ( int thresholdWeight, ostream& tab )
{
  int changed = 0;

  Unit::List::Iterator us (_units);
  while ( us.more() ) {
    Unit::List* newUnits = us.next()->negativeEqualitySplit (thresholdWeight, tab);
    changed += newUnits->length();
    _units = Unit::List::concat (newUnits, _units);
  }

  return changed;
} // Problem::negativeEqualitySplit


// 03/06/2001
int Problem::removeEqualityAxioms ()
{
  int removed = 0;

  Unit::List::DelIterator us (_units);
  while ( us.more() ) {
    if ( us.next()->isEqualityAxiom() ) {
      us.del();
      removed++;
    }
  }

  return removed;
} // Problem::removeEqualityAxioms


// 04/06/2001
int Problem::clausify ()
{
  int clauses = 0;
  Unit::List* newUnits = Unit::List::empty ();

  Unit::List::DelIterator us (_units);
  while ( us.more() ) {
    Unit* u = us.next();
    int newClauses = u->skolemize ( newUnits );
    if ( newClauses ) { // formula was converted to clauses
      us.del ();
    }
    clauses += newClauses;
  }

  _units = Unit::List::concat ( _units, newUnits );

  return clauses;
} // Problem::clausify


// 04/06/2001
int Problem::polarities ( const PSymbol* p ) const
{
  int pls = 0;

  Unit::List::Iterator us (_units);
  while ( us.more() ) {
    us.next()->polarities (p,pls);
    if ( pls == 3 )
      return 3;
  }

  return pls;
} // Problem::polarities


// 06/06/2001
bool Problem::hasEqualityInGoal () const
{
  Unit::List::Iterator unit ( _units ); 
  while ( unit.more() ) {
    Unit* u = unit.next();
    if ( u->isGoal() && u->hasEquality() ) {
      return true;
    }
  }
  
  return false;
} // Problem::hasEquality


// 08/06/2001
int Problem::goalWeight ()
{
  int w = 0;
  Unit::List::Iterator unit ( _units ); 
  while ( unit.more() ) {
    Unit* u = unit.next();
    ASSERT( debug_problem, u->unitType() == CLAUSE );
    if ( u->isGoal() ) {
      w += u->clause ()->weight();
    }
  }
  
  return w;
} // Problem::goalWeight
