//
//  file problem.h 
//  defines class Problem
//
//  14/05/2000, Dallas
//  28/05/2000, Manchester, input of clauses added
//  01/06/2000, Manchester, printDefinitions changed
//              findClausalDefinition added
//  07/06/2000, Manchester, several statistical functions added, eg isHorn()
//  11/06/2000, Manchester, InputUnit moved to a separate file
//  29/03/2001, Machester, Functor changed to Functor* everywhere
//  12/04/2001, Manchester, InputUnit changed to Unit
//  09/05/2001 Manchester: occurs ()
//  15/05/2001 Manchester: replaceCancellationAxioms ()
//  27/05/2001 Manchester: eliminateResolvable, eliminatePure, removePredicate
//  30/05/2001 Manchester, splitLargeArity/1, removeAllFunctionDefinitions/1
//  31/05/2001 Manchester, negativeEqualitySplit/2
//  03/06/2001 Manchester, removeEqualityAxioms/0, removeDummyPredicateDefinitions/1
//


#ifndef __problem__
#define __problem__


#include "typedefs.h"
#include "unit.h"


class FSymbol;


// ****************** class Problem, definition *******************


class Problem {
 public:
  // constructors & destructors
  Problem ();
  void build ( char* name, Unit::List* );
  ~Problem ();

  // query the structure
  char* name () const;
  Unit::List* units () const;

  // miscellaneous
  void printDefinitions ( ostream& str );
  int removeDummyFunctionDefinitions ( ostream& tab ); // the result is the number of removed symbols
  int removeDummyPredicateDefinitions ( ostream& tab ); // the result is the number of removed definitions
  int removePredicateSemiDefinitions ( ostream& tab ); // the result is the number of removed parts of definitions
  int removeLinearFunctionDefinitions ( ostream& tab ); // the result is the number of removed symbols
  int removeAllFunctionDefinitions ( ostream& tab ); // the result is the number of removed symbols
  int removeAllRelationalFunctionDefinitions ( ostream& tab ); // the result is the number of removed symbols
  int removeEqualityAxioms ();
  int replaceCancellationAxioms ( ostream& tab ); 
  int eliminateResolvable ( ostream& tab);
  int eliminatePure ( ostream& tab);
  int removeVarEquality ( ostream& tab );
  int removeTautologies ( ostream& tab );
  int cutTooLongClauses ( int maxLiterals );
  bool splitLargeArity ( int maxArity );
  int negativeEqualitySplit ( int thresholdWeight, ostream& tab );
  int clausify (); 

  // statistics
  bool hasEquality () const;
  bool hasEqualityInGoal () const;
  bool isHorn () const;
  bool hasNonEquality () const;
  bool hasFunctionSymbols () const;
  bool isPropositional ()const;
  bool isUnit () const;
  int numberOfClauses () const;
  int numberOfLiterals () const;
  void statistics ( ostream& ) const; // appends statistics to the file
  int maxLitWeight ();
  void normalize ();  // normalize and sort clauses/formulas in the input
  int goalWeight ();  

 private:

  // structure
  char* _name;
  Unit::List* _units;

  // auxiliary functions
  int countOccurrences ( const FSymbol* f ) const;
  bool occurs ( const FSymbol* f ) const;
  bool occurs ( const PSymbol* p ) const;
  int countOccurrences ( const PSymbol* p ) const;
  void applyFunDef ( const Term* lhs, const Term* rhs, const Unit* def, ostream& report ); // replace definition using lhs = rhs
  bool removeOneRelationalFunctionDefinition ( ostream& ); 
  void replaceRelationByFunction ( PSymbol* pred, FSymbol* fun, int argNo, int* vs, ostream& tab );
  void removeResolvable ( const PSymbol* pred, bool sign, ostream& report );
  void removePredicate ( const PSymbol* p, ostream& report ); // remove all occurrences of p
  int polarities ( const PSymbol* p ) const;
};


ostream& operator << ( ostream&, Problem& );

 
// ****************** class Problem, implementation *******************


inline
char* Problem::name () const
{
  return _name;
} // Problem::name


inline
Unit::List* Problem::units () const
{
  return _units;
} // Problem::input


#endif // __problem__