//
//  file term.cpp
//  implements class Term
//
//  15/05/2000 Dallas
//  28/05/2000 Manchester, many changes
//  01/06/2000 Manchester, occurs(Functor), size(), unify()
//  08/06/2000 Manchester, occurs(Var), replace(Var,Term), replaceRelationByFunction()
//  09/06/2000 Manchester, equal(), hasLargerArity(), split()
//  10/06/2000 Manchester, isFlat()
//  29/02/2000 Manchester, compare ()
//  01/04/2001 Dagstuhl, several functions moved to atom
//  30/05/2001 Manchester, splitLargeArity
//  31/05/2001 Manchester isGround/0
//  03/06/2001 equalUpTo/3, varArgs/1, createSkolemTerm/1
//
#include <string.h>


#include "debug.h"
#include "term.h"


const char* Term::varPrefix = "X";


ostream& operator << ( ostream& str, Term::List* ts )
{
  ASSERT( debug_term,
          ! ts->isEmpty () );

  Term::List* tail = ts ->tail ();
  
  while ( ! tail->isEmpty () ) {
    str << ts->head () << ',';
    ts = tail;
    tail = tail->tail ();
  }

  str << ts->head ();

  return str;
} // operator << ( ostream& str, List* vs )


ostream& operator << ( ostream& str, Term* t )
{
  if ( t->isVar () )
    str << Term::varPrefix << t->var ();
  else { // t is not a variable
    str << t->functor();
    if ( ! t->args()->isEmpty() ) 
      str << '(' << t->args() << ')';
  }

  return str;
} // operator << ( ostream&, Term* f )


void Term::rename ( VarSubst& subst )
{
  if ( isVar () ) {
    Var v;
    ASSERT ( debug_term, subst.bound (var (), v) );
    subst.bound (var(), v);
    var ( v );
    return;
  }

  // not variable
  for ( List* ts = args (); ! ts->isEmpty (); ts = ts->tail () )
    ts->head()->rename( subst );
} // Term::rename


// term has the form f (x1,...,xn) where xi are pairwise different
bool Term::isLinearAndFlat () const
{
  if ( isVar () )
    return false;

  return sequenceOfUniqueVariables (args());
} // Term::isLinearAndFlat


// term has the form f (x1,...,xn) where xi are variables
bool Term::sequenceOfUniqueVariables ( const List* terms )
{
  // check that all terms are variables
  for ( const List* ts = terms; ! ts->isEmpty (); ts = ts->tail () ) {
    // check that this argument is a variables
    Term* t = ts->head ();
    if ( ! t->isVar() )
      return false;

    // check that it is not equal to any of the previos variables
    for ( const List* ns = terms; ns != ts; ns = ns->tail () ) {
      if ( ts->head()->var() == ns->head()->var() )
        return false;
    }
  }

  return true;
} // Term::sequenceOfUniqueVariables


void Term::destroy ()
{
  if ( ! isVar() ) {
    for ( List* ts = args (); ! ts->isEmpty (); ts = ts->tail () )
      ts->head()->destroy ();

    args()->destroy();
  }

  delete this;
} // Term::destroy


Term* Term::copy () const
{
  if ( isVar () )
    return new Term ( var() );
  else
    return new Term ( functor(), copy( args() ) );
} // Term::copy 


Term::List* Term::copy ( const List* ts )
{
  if ( ts->isEmpty () )
    return Term::List::empty ();
  else
    return new List ( ts->head()->copy(), copy(ts->tail()) );
} // Term::copy ( List* ts )


int Term::weight () const
{
  if ( isVar() )
    return 1;

  return 1 + weight ( args() );
} // Term::weight


int Term::weight ( List* ts )
{
  int sz = 0;

  while ( ts->isNonEmpty() ) {
    sz += ts->head()->weight();
    ts = ts->tail ();
  }

  return sz;
} // Term::weight ( const List* )


bool Term::occurs ( const FSymbol* f ) const
{
  if ( isVar() )
    return false;

  if ( functor () == f )
    return true;

  return occurs( args(), f );
} // Term::occurs


bool Term::occurs ( Var v ) const
{
  if ( isVar() )
    return var() == v;

  return occurs( args(), v );
} // Term::occurs


bool Term::occurs ( const List* ts, const FSymbol* f )
{
  while ( ts->isNonEmpty() ) {
    if ( ts->head()->occurs(f) )
      return true;

    ts = ts->tail();
  }

  return false;
} // Term::occurs ( List* ts, FSymbol f )


bool Term::occurs ( List* ts, Var v )
{
  while ( ts->isNonEmpty() ) {
    if ( ts->head()->occurs(v) )
      return true;

    ts = ts->tail();
  }

  return false;
} // Term::occurs ( List* ts, Functor f )


bool Term::unify ( IndexedTerm t1, IndexedTerm t2, Dsubst& s )
{
  IndexedTerm u;

  Term* v1 = t1.right();

  if ( v1->isVar() ) {
    IndexedVar vv1 (t1.left(),v1->var());
    if ( s.bound (vv1, u) ) 
      return unify ( u, t2, s );

    // v1 is an unbound variable 
    if ( occurs (vv1, t2, s) )
      return false;
    
    // this variable is unbound and does not occur in t
    s.bind( vv1, t2 );
    return true;
  }

  // t1 is not a variable
  Term* v2 = t2.right();

  if ( v2->isVar() ) {
    IndexedVar vv2 (t2.left(),v2->var());
    if ( s.bound (vv2, u) ) 
      return unify ( u, t1, s );

    // v2 is an unbound variable 
    if ( occurs (vv2, t1, s) )
      return false;
    
    // this variable is unbound and does not occur in t
    s.bind( vv2, t1 );
    return true;
  }

  // both terms are not variables
  if ( v1->functor() != v2->functor() )
    return false;

  List* a1 = v1->args ();
  List* a2 = v2->args ();
  for (;;) {
    if ( ! unify( IndexedTerm( t1.left(), a1->head() ),
                  IndexedTerm( t2.left(), a2->head() ), s ) )
      return false;
    
    a1 = a1->tail ();
    if ( a1->isEmpty() )
      return true;
    
    a2 = a2->tail ();
  }
} // Term::unify


bool Term::occurs ( IndexedVar u, IndexedTerm t, Dsubst& s )
{

  Term* v = t.right();
  if ( v->isVar() ) {
    IndexedVar vv (t.left(),v->var());
    IndexedTerm w;

    if ( s.bound (vv, w) ) 
      return occurs ( u, w, s );

    // this variable is unbound
    return ( u.left() == t.left() && u.right() == v->var() );
  }

  // t is not a variable
  for ( List* vs = v->args(); vs->isNonEmpty(); vs = vs->tail() ) {
    if ( occurs( u, IndexedTerm (t.left(),vs->head()), s) )
      return true;
  }
  return false;
} // Term::occurs


int Term::countOccurrences ( const FSymbol* f ) const
{
  if ( isVar () )
    return 0;

  int occ = functor() == f ? 1 : 0;

  for ( List* ts = args(); ts->isNonEmpty(); ts = ts->tail() ) 
    occ += ts->head()->countOccurrences (f);

  return occ;
} // Term::countOccurrences


int Term::countOccurrences ( const Var v ) const
{
  if ( isVar () )
    return var() == v ? 1 : 0;

  int occ = 0;

  for ( List* ts = args(); ts->isNonEmpty(); ts = ts->tail() ) 
    occ += ts->head()->countOccurrences (v);

  return occ;
} // Term::countOccurrences


// in the definition this = rhs, every variable of this occurs
// at most once in rhs
bool Term::isLinear ( Term* rhs ) 
{
  for ( List* ts = args (); ts->isNonEmpty(); ts = ts->tail () ) {
    if ( rhs->countOccurrences(ts->head()->var()) > 1 )
      return false;
  }

  return true;
} // Term::isLinear


void Term::maxVar ( Var& max )
{
  if ( isVar () ) {
    if ( var() > max )
      max = var ();
    return;
  }

  for ( List* ts = args(); ts->isNonEmpty(); ts = ts->tail() ) {
    ts->head()->maxVar (max);
  }
} // Term::maxVar


// 04/06/2001 bug fixed (after checking isVar(), continued to the non-var case)
void Term::apply ( Subst& sbst )
{
  Term* t;

  if ( isVar() ) {
    if ( sbst.bound (var(),t) ) { // variable is bound to t in the substitution
      if ( t->isVar() ) {
        var ( t->var() );
        return;
      }
      // t is not variable
      tag ( COMPOUND );
      functor ( t->functor() );
      args ( copy(t->args()) );
      return;
    }
    // s is unbound var, do nothing
    return;
  }
  // t is not a variable
  for ( List* ts = args(); ts->isNonEmpty(); ts = ts->tail() )
    ts->head()->apply ( sbst );
} // Term::apply


void Term::replace ( Var v, Term* t )
{
  if ( isVar() ) {
    if ( var() == v ) {
      if ( t->isVar() ) {
        var ( t->var() );
        return;
      }
      // t is not variable
      tag ( COMPOUND );
      functor ( t->functor() );
      args ( copy(t->args()) );
      return;
    }
    // t is a variable but not v
    return;
  }

  // t is not a variable
  for ( List* ts = args(); ts->isNonEmpty(); ts = ts->tail() )
    ts->head()->replace ( v, t );
} // Term::replace


bool Term::equal ( Term* t ) const
{
  if ( isVar() ) 
    return t->isVar() && 
           var() == t->var();

  if ( t->isVar() )
    return false;

  return functor() == t->functor() &&
         equal ( args(), t->args() );
} // Term::equal


bool Term::equal ( const List* ts1, const List* ts2 )
{
  while ( ts1->isNonEmpty () ) {
    ASSERT ( debug_term, ts2->isNonEmpty() );
    
    if ( ! ts1->head()->equal(ts2->head()) )
      return false;
    ts1 = ts1->tail ();
    ts2 = ts2->tail ();
  }
  ASSERT ( debug_term, ts2->isEmpty() );

  return true;
} // Term::equal( List* , List* )


void Term::split ( FSymbol* tosplit, int maxArity, int iterations, int restArity, 
                   FSymbol* firstFun, FSymbol* restFun )
{
  if ( isVar() )
    return;

  for ( List* ts = args(); ts->isNonEmpty(); ts = ts->tail() )
    ts->head()->split ( tosplit, maxArity, iterations, restArity, firstFun, restFun );

  if ( functor() != tosplit ) 
    return;
    
  args ( split(args(), tosplit, maxArity, iterations, restArity, firstFun, restFun) );
} // Term::split


Term::List* Term::split ( List* ts, FSymbol* tosplit, int maxArity, 
                          int iterations, int restArity, 
                          FSymbol* firstFun, FSymbol* restFun )
{
  // note that the arguments are already splitted
  if ( iterations == 0 ) {
    if ( restArity == 0 ) {
      ASSERT( debug_term, ts->isEmpty() );

      return Term::List::empty ();
    }
    // restArity != 0;
    return new List ( new Term (restFun, ts) );
  }

  // iterations != 0
  // chop from the head of ts a list of length maxArity
  List* last = ts;
  for ( int i = maxArity - 1; i > 0; i-- ) {
    last = last->tail ();
  } 
  
  // last points to the last chain before the rest
  List* rest = last->tail();
  last->tail ( Term::List::empty() );

  return new List ( new Term (firstFun,ts),
                        split ( rest, tosplit, maxArity, iterations - 1, 
                                restArity, firstFun, restFun ) );
    
} // Term::split


Term* Term::apply ( Subst& subst, VarSubst& vsubst, int& lastVar ) // copies the argument
{
  if ( ! isVar() ) 
    return new Term ( functor(), apply (args(), subst, vsubst, lastVar) );

  Term* t;
  if ( subst.bound (var(), t) ) 
    return t->copy ();

  Var v;
  if ( vsubst.bound (var(), v) ) 
    return new Term ( v );

  lastVar++;
  vsubst.bind (var(), lastVar);

  return new Term ( lastVar );
} // Term::apply


Term::List* Term::apply ( List* ts, Subst& subst, VarSubst& vsubst, int& lastVar ) // copies the argument
{
  if ( ts->isEmpty() )
    return Term::List::empty ();

  return new List ( ts->head()->apply(subst, vsubst, lastVar),
                    apply (ts->tail(), subst, vsubst, lastVar) );
} // Term::apply


//  t1 is less than t2 if
//    (1) t1 is variable and t2 is not, or else
//    (2) t1.weight < t2.weight,        or else
//    (3) t1->arity < t2.arity,         or else
//    (4) t1.args << t2.args,  where << is the lexicographhic comparison
//
Comparison Term::compare ( const Term* t ) const
{
  if ( isVar () ) {
    if ( t->isVar() ) 
      return EQUAL;
    else 
      return LESS;
  }

  // this is nonvariable
  if ( t->isVar() ) 
    return GREATER;

  // both are nonvariable
  switch ( Int::compare (weight(), t->weight()) ) {
    case LESS:
      return LESS;
    case GREATER:
      return GREATER;
    case EQUAL:
      break;
  }

  switch ( Int::compare (arity(), t->arity()) ) {
    case LESS:
      return LESS;
    case GREATER:
      return GREATER;
    case EQUAL:
      break;
  }

  return compare ( args(),t->args() );
} // Term::compare


Comparison Term::compare ( const List* t1, const List* t2 ) 
{
  while ( t1->isNonEmpty() ) {
    ASSERT( debug_term, t2->isNonEmpty() );

    switch ( t1->head()->compare (t2->head()) ) {
      case LESS:
        return LESS;
      case GREATER:
        return GREATER;
      case EQUAL:
        break;
    }    
    t1 = t1->tail();
    t2 = t2->tail();
  }

  ASSERT( debug_term, t2->isEmpty() );
  return EQUAL;
} // Term::compare


Term::List* Term::splitLargeArity ( Term::List* ts, FSymbol::List* newFs )
{
  if ( newFs->isEmpty() ) {
    ASSERT ( debug_term, ts->isEmpty() );
    return Term::List::empty ();
  }

  // 
  List* rest;
  FSymbol* f = newFs->head();
  List* first = ts->split (f->arity(), rest);
  return new Term::List ( new Term (f, first),
                          splitLargeArity (rest, newFs->tail()) );

} // Term::splitLargeArity


void Term::splitLargeArity (const FSymbol* f, FSymbol::List* newFs )
{
  if ( isVar() )
    return;

  Term::List::Iterator ts ( args() );
  while ( ts.more() ) {
    ts.next()->splitLargeArity (f,newFs);
  }

  if ( functor() == f ) {
    args ( Term::splitLargeArity (args(),newFs) );
  }
} // Term::splitLargeArity


bool Term::isGround () const
{
  if ( isVar() )
    return false;

  Term::List::Iterator ts ( args() );
  while ( ts.more() ) {
    if ( ! ts.next()->isGround() ) {
      return false;
    }
  }

  return true;
} // Term::isGround


// 03/06/2001
// ls can be obtained from rs by swapping x and y
bool Term::equalUpTo ( const List* ls, const List* rs, Var x, Var y )
{
  Term::List::Iterator ts1 (const_cast<List*>(ls));
  Term::List::Iterator ts2 (const_cast<List*>(rs));

  while ( ts1.more() ) {
    // ls and rs must be of the same length
    ASSERT ( debug_term, ts2.more() );
    Term* l = ts1.next ();
    Term* r = ts2.next ();
    if ( l->equal(r) )
      continue;

    // l != r
    if ( ! l->isVar() || ! r->isVar() )
      return false;

    if ( l->var() == x ) {
      if ( r->var() != y )
        return false;
    }
    else if ( l->var() == y ) {
      if ( r->var() != x )
        return false;
    }
    else { // l is neither x nor y
      return false;
    }
  }

  return true;
} // Term::equalUpTo


// 03/06/2001
Term::List* Term::varArgs ( const VarList* vs)
{
  if ( vs->isEmpty() )
    return Term::List::empty();

  return new List ( new Term (vs->head()), varArgs(vs->tail()) );
} // Term::varArgs


// 03/06/2001
Term* Term::createSkolemTerm ( const VarList* vs )
{
  return new Term ( FSymbol::createNew("sk",vs->length()),
                    varArgs (vs) );
} // Term::createSkolemTerm
