//
//  file term.h 
//  defines class Term
//
//  15/05/2000, Dallas
//  28/05/2000, Manchester, Functor made into a class
//  01/06/2000, Manchester, occurs(Functor), size and unify added
//  03/06/2000, Manchester, class Functor moved to separate file
//  09/06/2000, Manchester, isEquality()
//  29/03/2001, Machester, Functor changed to Functor* everywhere
//  31/03/2001, Machester, size changed to weight
//  01/04/2001, Dagstuhl, several functions moved to atom
//  30/05/2001 Manchester, splitLargeArity/2
//  31/05/2001 Manchester isGround/0
//  03/06/2001 equalUpTo/3, varArgs/1, createSkolemTerm/1
//

#ifndef __term__
#define __term__


#include "typedefs.h"
#include "int.h"
#include "fsymbol.h"


// ****************** class Term, definition *************************


class Term {
 public:

  typedef List<Term*> List;
  
  enum Tag {
    VAR,
    COMPOUND
  }; // enum Term::Tag

  // constructors
  Term ( Var ); // for variables
  Term ( FSymbol*, List* ); // for compound terms
  bool isLinearAndFlat () const; // term has the form f (x1,...,xn) where xi are pairwise different
  static bool sequenceOfUniqueVariables ( const List* ); // sequence has the form (x1,...,xn) where xi are different variables

  // query the structure
  FSymbol* functor () const;
  List* args () const;
  Var var () const;
  void var ( Var );
  bool isVar () const;
  int arity () const;

  // deallocation and copying
  void destroy ();
  Term* copy () const;
  static List* copy ( const List* );

  // properties
  bool occurs ( const FSymbol* ) const;
  static bool occurs ( const List*, const FSymbol* );
  bool occurs ( Var ) const;
  int countOccurrences ( const FSymbol* ) const;
  int countOccurrences ( Var ) const;
  bool isLinear ( Term* rhs ); // in the definition this = rhs, every variable of this occurs
                               // at most once in rhs
  int weight () const;
  static int weight ( List* );
  bool equal ( Term* ) const;
  static bool equal ( const List*, const List* );
  bool isGround () const;

  // miscellaneous
  void rename ( VarSubst& );
  static bool unify ( IndexedTerm, IndexedTerm, Dsubst& );
  void maxVar (Var& v); // computes maximal variable and puts it in v
  void apply ( Subst& );  // actually reuses the memory allocated for the term
  void replace ( Var, Term* );  // actually reuses the memory allocated for the term
  void split ( FSymbol* tosplit, int maxArity, int iterations, int restArity, 
               FSymbol* firstFun, FSymbol* restFun );
  Term* apply ( Subst&, VarSubst&, int& lastVar ); // copies the argument
  static List* apply ( List*, Subst&, VarSubst&, int& lastVar ); // copies the list
  Comparison compare ( const Term* ) const;  // comparison of two terms, needed to normalize
  static Comparison compare ( const List*, const List* );  // comparison of two termlists
  static List* splitLargeArity ( List*, FSymbol::List* newFs );
  void splitLargeArity ( const FSymbol* f, FSymbol::List* newFs );
  static bool equalUpTo ( const List* ls, const List* rs, Var x, Var y); // ls obtained from rs by swapping x and y
  // for printing
  static const char* varPrefix;
  static Term* createSkolemTerm ( const VarList* vs);

 private: 
  // structure
  Tag _tag;
  union {
    Var _var;
    List* _args;
  };
  FSymbol* _functor;

  void tag ( Tag );
  void functor ( FSymbol* );
  void args ( List* );

  static bool occurs ( List*, Var );
  static bool occurs ( IndexedVar, IndexedTerm, Dsubst& );
  static List* split ( List*, FSymbol* tosplit, int maxArity, int iterations, int restArity, 
                           FSymbol* firstFun, FSymbol* restFun );

  static List* varArgs ( const VarList* vs);
  friend class Atom;
}; // class Term


ostream& operator << ( ostream&, Term* );
ostream& operator << ( ostream&, Term::List* );


// ****************** class Term, implementation *************************


inline
Term::Term ( Var v ) // for variables
  :
  _tag ( VAR ),
  _var ( v )
{
} // Term::Term


inline
Term::Term ( FSymbol* f, List* args ) // for compound terms
  :
  _tag ( COMPOUND ),
  _functor( f ),
  _args ( args )
{
} // Term::Term


inline
void Term::tag ( Tag t )
{
  _tag = t;
} // Term::tag


inline
void Term::functor ( FSymbol* f )
{
  _functor = f;
} // Term::functor


inline
void Term::args ( List* ts )
{
  _args = ts;
} // Term::args


inline
Term::List* Term::args () const
{
  ASSERT ( debug_term, 
           _tag == COMPOUND );

  return _args;
} // Term::args


inline
Var Term::var () const
{
  ASSERT ( debug_term, 
           _tag == VAR );

  return _var;
} // Term::var


inline
void Term::var ( Var v )
{
  ASSERT ( debug_term, 
           _tag == VAR );

  _var = v;
} // Term::var


inline
FSymbol* Term::functor () const
{
  ASSERT ( debug_term, 
           _tag == COMPOUND );

  return _functor;
} // Term::functor


inline
bool Term::isVar () const
{
  return _tag == VAR;
} // Term::isVar


inline
int Term::arity () const
{
  return _functor->arity ();
} // Term::arity


#endif // __term__
