//
//  file unit.cpp
//
//  12/04/2001, Manchester, made from part of problem.cpp
//  09/05/2001 Manchester: occurs ()
//  15/05/2001 Manchester: replaceCancellation()
//  22/05/2001 Manchester: apply () reimplemented, isLinearDefinition()
//  27/05/2001 Manchester: occurs (PSymbol*)
//  30/05/2001 Manchester: resolveOn/3 instead of resolveOn/3, splitLargeArity/2
//  31/05/2001 Manchester: normalize/0, negativeEqualitySplit/1
//  03/06/2001 Manchester: isEqualityAxiom/0, isPredicateDefinition/1, occurs(PSymbol*)
//


#include <string.h>

#include "debug.h"
#include "unit.h"
#include "formula.h"
#include "clause.h"
#include "string.h"
#include "replace_code.h"


ostream& operator << ( ostream& str, InputType t )
{
  switch ( t ) {
    case AXIOM:
      str << "axiom";
      break;

    case CONJECTURE:
      str << "conjecture";
      break;

    case HYPOTHESIS:
      str << "hypothesis";
      break;
  }

  return str;
} // operator << ( ostream& str, InputType t )


ostream& operator << ( ostream& str, UnitType t )
{
  switch ( t ) {
    case CLAUSE:
      str << "input_clause";
      break;

    case FORMULA:
      str << "input_formula";
      break;
  }

  return str;
} // operator << ( ostream& str, UnitType t )


ostream& operator << ( ostream& str, Unit::List* p )
{
  while ( ! p->isEmpty () ) {
    Unit* unit = p->head ();

    str << unit->unitType () 
        << '('
        << unit->name ()
        << ','
        << unit->inputType()
        << ",\n";

    switch ( unit->unitType() )  {
      case FORMULA:
        unit->formula()->prettyPrint ( 4, str );
        break;

      case CLAUSE:
        unit->clause()->prettyPrint ( 4, str );
        break;
    }
    str << " ).\n\n";
    p = p->tail ();
  }

  return str;
} // operator << ( ostream& str, List* p )


Unit::Unit ( char* name, InputType intype, Clause* c )
  :
  _untype ( CLAUSE ),
  _intype ( intype ),
  _name( name ),
  _clause ( c )
{
} // Unit::Unit


// 04/06/2001
Unit::Unit ( const String* prefix, const char* postfix, InputType intype, Clause* c )
  :
  _untype ( CLAUSE ),
  _intype ( intype ),
  _name( prefix, postfix ),
  _clause ( c )
{
} // Unit::Unit


Unit::Unit ( char* name, InputType intype, Formula* f )
  :
  _untype ( FORMULA ),
  _intype ( intype ),
  _name( name ),
  _formula ( f )
{
} // Unit::Unit


Unit::~Unit ()
{
  switch ( _untype ) {
    case FORMULA:
      _formula->destroy ();
      return;

    case CLAUSE:
      _clause->destroy ();
      return;
  }
} // Unit::Unit


void Unit::destroy ()
{
  switch ( _untype ) {
    case FORMULA:
      _formula->destroy ();
      return;

    case CLAUSE:
      _clause->destroy ();
      return;
  }

  delete this;
} // Unit::destroy


int Unit::countOccurrences ( const PSymbol* p ) const
{
  switch ( unitType() ) {
    case CLAUSE:
      return clause()->countOccurrences (p);

    case FORMULA:
      return formula()->countOccurrences (p);

    default:
      ASSERT( debug_unit, false );
      return false;
  }
} // Unit::countOccurrences


int Unit::countOccurrences ( const FSymbol* p ) const
{
  switch ( unitType() ) {
    case CLAUSE:
      return clause()->countOccurrences (p);

    case FORMULA:
      return formula()->countOccurrences (p);

    default:
      ASSERT( debug_unit, false );
      return false;
  }
} // Unit::countOccurrences


bool Unit::occurs ( const FSymbol* p ) const
{
  ASSERT( debug_unit, unitType() == CLAUSE );
  
  return clause()->occurs (p);
} // Unit::occurs


// 03/06/2001
bool Unit::occurs ( const PSymbol* p ) const
{
  switch ( unitType() ) {
    case CLAUSE:
      return clause()->occurs (p);

    case FORMULA:
      return formula()->occurs (p);

    default:
      ASSERT( debug_unit, false );
      return false;
  }
} // Unit::occurs


Var* Unit::isFunctionalityAxiom ( const PSymbol* pred, FSymbol*& fun, int argNo, Atom*& atom ) const
{
  ASSERT( debug_unit, unitType() == CLAUSE );

  return clause()->isFunctionalityAxiom (pred,fun,argNo, atom);
} // Unit::isFunctionalityAxiom


// the unit is P(x,y) & P(x,z) -> y = z
bool Unit::isUniquenessAxiom ( PSymbol*& pred, int& argNo ) const
{
  ASSERT( debug_unit, unitType() == CLAUSE );

  return clause()->isUniquenessAxiom (pred, argNo);
} // Unit::isUniquenessAxiom


// the unit is f(X,y,Z) = f(X,u,Z) -> y = u
bool Unit::isCancellationAxiom ( FSymbol*& fun, int& argNo ) const
{
  ASSERT( debug_unit, unitType() == CLAUSE );

  return clause()->isCancellationAxiom (fun, argNo);
} // Unit::isCancellationAxiom


// replace the cancellation axiom 
//   f(X,y,Z) = f(X,u,Z) -> y = u
// by 
//   i(f(X,y,Z),X,Z) = y
void Unit::replaceCancellation ( const FSymbol* fun,int argNo,FSymbol* inv, ostream& report )
{
  ASSERT( debug_unit, unitType() == CLAUSE );

  clause()->replaceCancellation (argNo, inv);

  report << "inference(" << clause() << "," << name() 
         << "_rcbi,[" << name () << "],[rcbi]]).\n";
  _name += "_rcbi"; // replace cancellation by inverse
} // Unit::replaceCancellation


void Unit::apply ( const ReplaceCode& code, const Unit* def, ostream& report ) 
{
  ASSERT( debug_unit, unitType() == CLAUSE );

  if ( code.apply (clause()) ) {
    report << "inference(" << clause() << "," << name() 
           << "_dapp,[" << name () << "," << def->name () << "],[dapp]]).\n";
    _name += "_dapp"; // replace cancellation by inverse
  }
} // Unit::apply


int Unit::maxLitWeight ( const ReplaceCode& code ) const
{
  ASSERT( debug_unit, unitType() == CLAUSE );

  return code.maxLitWeight (clause());
} // Unit::maxLitWeight


int Unit::weight () const 
{
  switch ( unitType() ) {
    case CLAUSE:
      return clause()->weight ();

    case FORMULA:
      return formula()->weight ();

    default:
      ASSERT( debug_unit, false );
      return false;
  }
} // Unit::weight


Term* Unit::isDefinition ( Term* & rhs ) const
{
  ASSERT( debug_unit, unitType() == CLAUSE );

  return clause()->isDefinition (rhs);
} // Unit::isDefinition


// 03/06/2001
Atom* Unit::isPredicateDefinition ( Formula* & rhs ) const
{
  switch ( unitType() ) {
    case CLAUSE:
      return 0;

    case FORMULA:
      switch ( inputType() ) {
        case AXIOM:
        case HYPOTHESIS:
          return formula()->isPredicateDefinition (rhs);
        case CONJECTURE:
          return 0;
      }

    default:
      ASSERT ( debug_unit, false );
      return 0;
  }
} // Unit::isPredicateDefinition


Term* Unit::isLinearDefinition ( Term* & rhs ) const
{
  ASSERT( debug_unit, unitType() == CLAUSE );

  return clause()->isLinearDefinition (rhs);
} // Unit::isLinearDefinition


bool Unit::hasFunctionSymbols () const
{
  ASSERT( debug_unit, unitType() == CLAUSE );

  return clause()->hasFunctionSymbols ();
} // Unit::hasFunctionSymbols


bool Unit::hasEquality () const
{
  ASSERT( debug_unit, unitType() == CLAUSE );

  return clause()->hasEquality ();
} // Unit::hasEquality


bool Unit::isHorn () const
{
  ASSERT( debug_unit, unitType() == CLAUSE );

  return clause()->isHorn ();
} // Unit::isHorn


bool Unit::isUnit () const
{
  ASSERT( debug_unit, unitType() == CLAUSE );

  return clause()->isUnit ();
} // Unit::isUnit


bool Unit::isPropositional () const
{
  ASSERT( debug_unit, unitType() == CLAUSE );

  return clause()->isPropositional ();
} // Unit::isPropositional


// 06/06/2001 made from pureEquality
bool Unit::hasNonEquality () const
{
  ASSERT( debug_unit, unitType() == CLAUSE );

  return clause()->hasNonEquality ();
} // Unit::hasNonEquality


int Unit::numberOfLiterals () const
{
  ASSERT( debug_unit, unitType() == CLAUSE );

  return clause()->length ();
} // Unit::numberOfLiterals


void Unit::replaceRelationByFunction ( PSymbol* pred, FSymbol* fun, int argNo, int* vs, ostream& report )
{
  if ( clause()->replaceRelationByFunction (pred, fun, argNo, vs) ) {
    report << "inference(" << clause() << "," << name() 
           << "_rrbf,[" << name () << ",replace_" << pred << "_by_" << fun  << "],[rrbf]]).\n";
    _name += "_rrbf";

    removeVarEquality (report);
  }
} // Unit::replaceRelationByFunction


Unit* Unit::resolveOn ( const PSymbol* f, Unit* unit, ostream& report )
{
  static int res_number = 0;
  static const char* prefix = "res_";
  static const int prefixLength = strlen (prefix);

  // making name for the new unit
  char* newName = new char [prefixLength + 10];
  strcpy ( newName, prefix );
  char* cursor = newName + prefixLength;
  Int::toString ( res_number, cursor );
  res_number++;

  Clause* c = clause()->resolveOn (f, unit->clause());
  Unit* result = new Unit ( newName, inputType(), c );
  report << "inference(" << c << "," << newName << ",[" 
         << name () <<"," << unit->name () << "],[br]).\n";
  
  return result;
} // Unit::resolveOn


Comparison Unit::compare ( const Unit* u ) const
{
  // only units of the same type can be compared at this moment
  ASSERT ( debug_unit, _untype == u->_untype );

  switch ( _untype ) {
    case CLAUSE:
      return clause()->compare(u->clause());

    case FORMULA:
      return formula()->compare(u->formula());

    default:
      ASSERT ( debug_unit, false );
  }
} // Unit::compare


bool Unit::removeVarEquality ( ostream& report )
{
  // only clauses can be treated at this moment
  ASSERT ( debug_unit, _untype == CLAUSE );

  if ( clause()->removeVarEquality() ) {
    report << "inference(" << clause() << "," << name() << "_rve,[" << name () << "],[rve]).\n";
    _name += "_rve";
    return true;
  }

  return false;
} // Unit::removeVarEquality


bool Unit::isTautology ( ostream& report ) 
{
  // only clauses can be treated at this moment
  ASSERT ( debug_unit, _untype == CLAUSE );

  if ( clause()->isTautology() ) {
    report << "prepro(removed_as_tautology," << name() << "," << clause() << ").\n";
    return true;
  }

  return false;
} // Unit::isTautology


Unit::List* Unit::sort ( List* toSort, List* appendTo )
{
  if ( toSort->isEmpty() ) 
    return appendTo;

  if ( toSort->tail()->isEmpty() ) {
    toSort->tail ( appendTo );
    return toSort;
  }

  List* less;
  List* equal;
  List* greater;

  split ( toSort->head(), toSort->tail(), less, equal, greater );
  // sort greater and append to appendTo
  greater = sort ( greater, appendTo );
  // concatenate equal with it
  equal = Unit::List::concat ( equal, greater );
  // push head of list
  toSort->tail ( equal );
  // sort less and append to it
  return sort ( less, toSort );
} // Unit::sort


Unit::List* Unit::sort ( List* toSort )
{
  return sort ( toSort, Unit::List::empty() );
} // Unit::sort


void Unit::split ( Unit* l, List* spl, 
                   List*& less, List*& equal, List*& greater )
{
  ASSERT ( debug_unit, spl->isNonEmpty() );

  less = Unit::List::empty();
  equal = Unit::List::empty();
  greater = Unit::List::empty();

  for (;;) {
    List* tail = spl->tail ();

    switch ( spl->head()->compare(l) ) {
      case GREATER:
        spl->tail ( greater );
        greater = spl;
        break;
      case EQUAL:
        spl->tail ( equal );
        equal = spl;
        break;
      case LESS:
        spl->tail ( less );
        less = spl;
        break;
    }

    if ( tail->isEmpty() )
      return;
    spl = tail;
  }
} // Unit::split


int Unit::maxLitWeight ()
{
  // only clauses can be treated at this moment
  ASSERT ( debug_unit, _untype == CLAUSE );

  return clause()->maxLitWeight();
} // Unit::maxLitWeight


void Unit::splitLargeArity ( const PSymbol* p, FSymbol::List* newFs )
{
  // only clauses can be treated at this moment
  ASSERT ( debug_unit, _untype == CLAUSE );

  clause()->splitLargeArity (p, newFs);
} // Unit::splitLargeArity


void Unit::splitLargeArity ( const FSymbol* f, FSymbol::List* newFs )
{
  // only clauses can be treated at this moment
  ASSERT ( debug_unit, _untype == CLAUSE );

  clause()->splitLargeArity (f, newFs);
} // Unit::splitLargeArity


void Unit::normalize ()
{
  // only clauses can be treated at this moment
  switch ( _untype ) {
    case CLAUSE:
      clause()->normalize();
      return;

    case FORMULA:
      formula()->normalize();
      return;
  }
} // Unit::normalize


// no time to tab the split now
Unit::List* Unit::negativeEqualitySplit ( int thresholdWeight, ostream& tab )
{
  // only clauses can be treated at this moment
  ASSERT ( debug_unit, _untype == CLAUSE );

  Clause::List* cs = clause()->negativeEqualitySplit( thresholdWeight );

  List* newUnits = Unit::List::empty ();

  Clause::List::Iterator cls ( cs );
  while ( cls.more() ) {
    static int number = 0;
    static char name [] = "term_def_          ";
    Int::toString ( ++number, name+9 );
    char* nm = new char [strlen(name)+1];
    strcpy (nm, name);
    Unit::List::push ( new Unit (nm, AXIOM, cls.next()), newUnits );
  }

  cs->destroy ();

  return newUnits;
} // Unit::negativeEqualitySplit


// 03/06/2001
// is one of the equality axioms
bool Unit::isEqualityAxiom () const
{
  switch ( unitType() ) {
    case CLAUSE:  
      // should be reimplemented, no time now
      return false;

    case FORMULA:
      switch ( inputType() ) {
        case CONJECTURE:
          return false;
        case AXIOM:
        case HYPOTHESIS:
          return formula ()->isEqualityAxiom ();
      }

    default:
      ASSERT ( debug_unit, false );
      return 0;
  }
} // Unit::isEqualityAxiom


// 04/06/2001
// if the unit is a formula, it is skolemized and the resulting units
// are appended to appendTo, the result is the number of new units
int Unit::skolemize ( List* & appendTo )
{
  switch ( unitType() ) {
    case CLAUSE:  
      return 0;

    case FORMULA: {
      bool polarity = true;

      switch ( inputType() ) {

        case AXIOM:
        case HYPOTHESIS:
          break;

        case CONJECTURE:
          polarity = false;
          break;
      }
      
      // skolemize
      Clause::List* newClauses = formula ()->skolemize (polarity);
      // to generate names for the new units
      int appendix = 0; // to append to the unit name
      char app [10];  
      app[0] = '_';
      // create new units and append them to appendTo
      Clause::List::Iterator cls (newClauses);
      while ( cls.more() ) {
        // make a new name by appending the current name with appendix
        Int::toString (appendix++,app+1);
        // create a new unit and append it
        Unit::List::push ( new Unit (name(), app, inputType(), cls.next()),
                           appendTo );
      }
      int result = newClauses->length ();
      newClauses->destroy ();
      return result;
    }  

    default:
      ASSERT ( debug_unit, false );
      return 0;
  }  
} // Unit::skolemize


// 04/06/2001
void Unit::polarities ( const PSymbol* p, int& pol ) const
{
  switch ( unitType() ) {
    case CLAUSE:  
      clause()->polarities (p,pol);
      return;

    case FORMULA:
      switch ( inputType() ) {
        case AXIOM:
        case HYPOTHESIS:
          formula()->polarities (p,true,pol);
          return;
        case CONJECTURE:
          formula()->polarities (p,false,pol);
          return;
      }

    default:
      ASSERT ( debug_unit, false );
      return;
  }
} // Unit::polarities


// 04/06/2001
// the unit is a definition (P(X) <=> rhs)
// is pos, then it should be changed into (P(X)->rhs), otherwise into
// (rhs->P(X)), i.e. pos is the polarity of rsh in the new formula
void Unit::makeHalfDefinition ( bool positive, const Formula* rhs )
{
  ASSERT ( debug_unit, _untype == FORMULA );
  ASSERT ( debug_unit, _intype != HYPOTHESIS );
  
  _name += positive ? "_neg" : "_pos";
  formula()->makeHalfDefinition ( positive, rhs );
} // Unit::makeHalfDefinition


// 06/06/2001
bool Unit::isGoal () const
{
  return inputType () == CONJECTURE;
} // Unit::isGoal


