//
//  file unit.h 
//
//  11/06/2000 Manchester, moved to from problem.h
//  29/03/2001 Macnchester, Functor changed to Functor*
//  08/04/2001 Manchester, char* changed to String in _name
//  12/04/2001 Manchester, InputUnit changed to Unit
//  09/05/2001 Manchester: occurs ()
//  15/05/2001 Manchester: replaceCancellation()
//  22/05/2001 Manchester: isLinearDefinition()
//  27/05/2001 Manchester: occurs (PSymbol*)
//  30/05/2001 Manchester, splitLargeArity/2
//  31/05/2001 Manchester, normalize/0, negativeEqualitySplit/1
//  03/06/2001 Manchester: isEqualityAxiom/0, isPredicateDefinition/1
//


#ifndef __unit__
#define __unit__


#include "typedefs.h"
#include "string.h"
#include "int.h"
#include "fsymbol.h"


class ReplaceCode;
class Formula;
class Clause;
class PSymbol;
class FSymbol;
class Atom;


enum InputType {
  AXIOM,
  CONJECTURE,
  HYPOTHESIS
};


enum UnitType {
  CLAUSE,
  FORMULA
};


// ****************** class Unit, definition *******************


class Unit {
 
 public:

  typedef List<Unit*> List;

  Unit ( char* name, InputType untype, Formula* );
  Unit ( char* name, InputType untype, Clause* );
  Unit ( const String* prefix, const char* postfix, InputType intype, Clause* c );
  ~Unit ();

  // query/change structure
  const String* name () const { return &_name; }
  InputType inputType () const;
  UnitType unitType () const;
  Formula* formula () const;
  Clause* clause () const;
  void clause ( Clause* );

  // statistics
  int weight () const;
  int countOccurrences ( const FSymbol* f ) const;
  bool occurs ( const FSymbol* f ) const;
  bool occurs ( const PSymbol* p ) const;
  int countOccurrences ( const PSymbol* p ) const;
  bool hasFunctionSymbols () const;
  bool hasEquality () const;
  bool isHorn () const;
  bool isUnit () const; // contains only one literal
  bool isPropositional () const;
  bool hasNonEquality () const;
  int numberOfLiterals () const;
  int maxLitWeight ();
  bool isGoal () const;

  // destroy the unit
  void destroy ();

  // miscellaneous
  Unit* resolveOn ( const PSymbol*, Unit*, ostream& report );
  Comparison compare ( const Unit* ) const;
  bool removeVarEquality ( ostream& report );
  bool isTautology ( ostream& report ); 
  void replaceRelationByFunction ( PSymbol*, FSymbol*, int argNo, int* vs, ostream& report ); 
  Var* isFunctionalityAxiom ( const PSymbol* pred, FSymbol*& fun, int argNo, Atom* & atom ) const;
  bool isUniquenessAxiom ( PSymbol*& pred, int& argNo ) const; // the unit is P(x,y) & P(x,z) -> y = z
  bool isCancellationAxiom ( FSymbol*& pred, int& argNo ) const; // the unit is f(X,y,Z) = f(X,u,Z) -> y = u
  bool isEqualityAxiom () const;  // is one of the equality axioms
  void apply ( const ReplaceCode&, const Unit* def, ostream& report );
  int maxLitWeight ( const ReplaceCode& ) const;
  Term* isDefinition ( Term* & rhs ) const; 
  Term* isLinearDefinition ( Term* & rhs ) const; 
  Atom* isPredicateDefinition ( Formula* & rhs ) const;
  static List* sort ( List* );
  void replaceCancellation ( const FSymbol*,int argNo,FSymbol* inv, ostream& report);
  void splitLargeArity ( const PSymbol* p, FSymbol::List* newFs );
  void splitLargeArity ( const FSymbol* p, FSymbol::List* newFs );
  void normalize ();
  List* negativeEqualitySplit ( int thresholdWeight, ostream& tab );
  int skolemize ( List* & appendTo );
  void polarities ( const PSymbol* p, int& pol ) const;
  void makeHalfDefinition ( bool positive,const Formula* rhs );

 private:

  // structure
  UnitType _untype;
  InputType _intype;
  String _name;
  union {
    Clause* _clause;
    Formula* _formula;
  };

  static List* sort ( List* toSort, List* appendTo );
  static void split ( Unit*, List*, List*& less, List*& equal, List*& greater );
}; // class Unit


ostream& operator << ( ostream&, InputType );
ostream& operator << ( ostream&, UnitType );
ostream& operator << ( ostream&, Unit::List* );


// ****************** class Unit, implementation *******************


inline
InputType Unit::inputType () const
{
  return _intype;
}


inline
UnitType Unit::unitType () const
{
  return _untype;
}


inline
Formula* Unit::formula () const
{
  ASSERT ( debug_problem, unitType() == FORMULA );

  return _formula;
}


inline
Clause* Unit::clause () const
{
  ASSERT ( debug_problem, unitType() == CLAUSE );
  
  return _clause;
}

 
inline
void Unit::clause ( Clause* c )
{
  ASSERT ( debug_problem, unitType() == CLAUSE );

  _clause = c;
} // Unit::clause


#endif // __unit__
